// -*- tab-size:4 -*-

// View the specified file using Gnome 1 settings.  We use the setting
// for "view" if there is one, otherwise "open".  If there's an optional
// second argument, we use that verb instead of view or open.

// Copyright 2001, 2005 Lugaru Software, Ltd.

// Compile with:
//  cc -o eview-gnome1 eview-gnome1.c `gnome-config --cflags --libs gnomeui`

#include <gnome.h>
#include <string.h>
#include <stdio.h>

int verbose, test;

void usage()
{
	fprintf(stderr, "usage: eview-gnome1 [-vt] filename [ key ]\n");
	fprintf(stderr, "Uses Gnome 1 settings to display the file.\n");
	fprintf(stderr, "Action is a Gnome key like open or view.\n");
	fprintf(stderr, "If omitted, use view, or if undefined, open.\n");
	fprintf(stderr, "-v\tverbose\n");
	fprintf(stderr, "-t\ttest, don't actually run the program\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	const char *mime, *val;
	GList *gl;
	char cmdline[1024], *p;

	for (argc--, argv++; argc > 0 && **argv == '-'; argc--, argv++)
		switch (*++*argv) {
			case 'v': verbose = 1; break;
			case 't': test = 1; break;
			default: usage();
		}
	if (argc < 1)
		usage();

	gnomelib_init ("epsilon-viewer-interface", "0.2");
	mime = gnome_mime_type(argv[0]);
	if (verbose) {
		printf("The MIME type of %s: %s\n", argv[0], mime);
		printf("The program for %s data: %s\n",
		       mime, gnome_mime_program(mime));
		printf("The description for %s: %s\n",
		       mime, gnome_mime_description(mime));
		gl = gnome_mime_get_keys(mime);
		for(; gl; gl = g_list_next(gl)) {
			char *s = gl->data;
			printf("Key: %s\n", s);
		}
	}

	if (!mime)
		exit(1);
	if (argc > 1) {
		val = gnome_mime_get_value(mime, argv[1]);
		if (verbose)
			printf("Value for key %s: %s\n", argv[1], val);
	} else if (!(val = gnome_mime_get_value(mime, "view")))
		val = gnome_mime_get_value(mime, "open");
	if (!val)
		exit(1);
	if (verbose)
		printf("Run command: %s\n", val);
	strcpy(cmdline, val);
	p = strstr(cmdline, "%f");
	if (p) {
		strcpy(p, "\"");
		strcat(p, argv[0]);
		strcat(p, strstr(val, "%f") + 2);
		strcat(p, "\"");
	}
	if (verbose)
		printf("Full command line: %s\n", cmdline);
	if (!test)
		gnome_execute_shell(NULL, cmdline);
	exit(0);
}
