/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*	Copyright (C) 2003, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// Generate a list of settings in this version.  The main function
// build_settings_file first reads each .b file listed in the file
// extensions-list in the current directory.  Then it writes a list of
// settings to the file settings-list in the current directory and
// exits.

#include <eel.h>

// Define some stuff that's not in older versions.

#ifndef FILETYPE_BINARY
#define	FILETYPE_BINARY		0
#endif

#ifndef MAXKEYS
#define	MAXKEYS		NUMKEYS
#endif

#if EELVERSION >= 120
#define get_range_for_settings_file		get_range
#else
#define  key_t short

get_range_for_settings_file(tab, i, val)
short *tab;
{
	int cnt = 0;

	for (; i < NUMKEYS && tab[i] == val; i++)
		cnt++;
	return cnt;
}
#endif

// Define various helper functions, so we don't depend on the
// sometimes-different flavors of these functions in the old version.

list_table_for_settings_file(i)
{
	char tabname[FNAMELEN];
	int j, cnt;
	short *k;
	key_t onekey[2];

	strcpy(tabname, name_name(i));
	bprintf("%s: table\n", tabname);
	k = index_table(i);
	onekey[0] = 2;
	for (j = 0; j < MAXKEYS; j += cnt) {
		cnt = get_range_for_settings_file(k, j, k[j]);
		if (k[j] > 0) {
			bprintf("~%s ", tabname);
			onekey[1] = j;
			stuff_macro(onekey, 1);
			if (cnt > 1) {
				bprintf(" ");
				onekey[1] = j + cnt - 1;
				stuff_macro(onekey, 1);
			}
			bprintf(": %s\n", name_name(k[j]));
		}
	}
}

export_colors_one_scheme(int i)
{
	char *nm;
	int j, fg, bg;

	nm = name_name(i);
	noteput("Working...%s", nm);
	bprintf("color_scheme \"%s\" {\n", nm);
	save_var selected_color_scheme = i;
	for (j = 0; j < number_of_color_classes(); j++) {
#if EELVERSION >= 100
		int k = get_foreground_color(j, 1);
		if (k & ECOLOR_UNKNOWN)
			continue;
#endif
		bprintf("\tcolor_class %s ", name_color_class(j));
#if EELVERSION >= 100
		if (k & ECOLOR_COPY) {
			bprintf("= color_class %s;\n", name_color_class(COLOR_STRIP_ATTR(k)));
			continue;
		}
#endif
		fg = get_foreground_color(j);
		bg = get_background_color(j);
		bprintf("MAKE_RGB(%d, %d, %d)",
				GETRED(fg), GETGREEN(fg), GETBLUE(fg));
#ifdef EFONT_BOLD
		if (fg & EFONT_BOLD)
			bprintf(" | EFONT_BOLD");
		if (fg & EFONT_ITALIC)
			bprintf(" | EFONT_ITALIC");
		if (fg & EFONT_UNDERLINED)
			bprintf(" | EFONT_UNDERLINED");
#endif
		if (!j
#ifdef ETRANSPARENT
			  || !(bg & ETRANSPARENT) && bg != get_background_color(0)
#endif
		   )
			bprintf(" on MAKE_RGB(%d, %d, %d)",
					GETRED(bg), GETGREEN(bg), GETBLUE(bg));
		stuff(";\n");
	}
	stuff("};\n\n");
}

export_current_colors()
{
	char name[FNAMELEN];
	int i;

	strcpy(name, "mycolors.e");
	absolute(name);
	do_find(name, FILETYPE_AUTO);
	buf_zap(bufnum);
	stuff("// Automatically generated by import-customizations.\n\n");
	stuff("#include \"eel.h\"\n\n");
	export_colors_one_scheme(selected_color_scheme);
	point = 0;
	for (i = 0; i < 10; i++) {	// Assume class names have no more than 10 -'s.
		noteput("Working...%d", i);
		string_replace("(color_class .*)-", "#1_", REGEX);
		if (!replace_num_changed)	// Convert -'s in class names to _'s.
			break;
	}
	add_color_class_declarations();
	save_file();
}

// For each "= color_class foo;", add a declaration at the top.
add_color_class_declarations()
{
	int tmp;

	tmp = tmp_buf();
	buf_xfer(tmp, 0, size());
	save_var bufnum = tmp;
#if EELVERSION >= 70
	save_var sort_status = 0;
#endif
	point = 0;
	do_drop_matching_lines(0, "= color_class", 0);
	point = 0;
	string_replace(".* = ", "", REGEX);
	point = 0;
	do_sort_region(0, size(), 0, 0);
	point = 0;
	do_uniq(1, 1, 0);
	restore_vars();
	point = 0;
	re_search(1, "(#include .*\n)+\n*");
	to_begin_line();
	grab_buffer(tmp);
	stuff("\n");
	buf_delete(tmp);
}

// Append to buffer b a description of name table entry i.
list_one_for_settings_file(b, i)
{
	int j, last, val, type;
	char *p, *nm, *tname;

	save_var bufnum = b;
	save_var point = size();
	save_var use_default;
	nm = name_name(i);
	switch (name_type(i)) {
		case NT_COMMAND:
		case NT_AUTOLOAD:
			bprintf("%s: command\n", nm); break;
		case NT_SUBR:
		case NT_AUTOSUBR:
			bprintf("%s: subroutine\n", nm); break;
		case NT_MACRO:
			bprintf("%s: macro ", nm);
			stuff_macro(get_macro(i), 1);
			insert('\n');
			break;
		case NT_TABLE:
			list_table_for_settings_file(i);
			break;
		case NT_WINVAR:
		case NT_BUFVAR: case NT_VAR:
			use_default = 1;
			p = varptr(i);
			type = 0;
			if (!name_user(i) && (strncmp(nm, "-our-", 5)
								  || !strstr(nm, "-scheme")))
				break;
			switch (vartype(i)) {
				case TYPE_CHAR:	val = *(char *) p;
								tname = "char ";
								break;
				case TYPE_SHORT: val = *(short *) p;
								 tname = "short ";
								 break;
#ifdef TYPE_INT32
				case TYPE_INT32:
								val = *(int *) p;
								tname = "int32 ";
								break;
#endif
				case TYPE_INT:	val = *(int *) p;
								tname = "int ";
								break;
				case TYPE_CPTR:	tname = "char ptr ";
								type = 2; break;
				case TYPE_CARRAY: tname = "char array ";
								  type = 1; break;
				default:	tname = ""; type = 2; break;
			}
			bprintf("+%s %s%svariable", nm,
					(name_type(i) == NT_WINVAR)
					? "window " :
					(name_type(i) == NT_BUFVAR)
					? "buffer " : "", tname);
			if (type == 0) {
				if (val && !strncmp(nm, "-our-", 5)
					  && parse_string(-1, "^%+-our-(color|mono|gui"
									  "|unixconsole)-scheme .*"))
					bprintf(": %s\n", name_name(val));
				else
					bprintf(": %d\n", val);
			} else if (type == 2)
				stuff(": <complex>\n");
			else {
				last = ptrlen(p);
				bprintf(": %d \"", last);
				while (last > 0 && !p[last - 1])
					last--;
				for (j = 0; j < last; j++)
					switch (p[j]) {
						case '\t':	stuff("\\t"); break;
						case '\r':	stuff("\\r"); break;
						case '\n':	stuff("\\n"); break;
						case 0:		stuff("\\z"); break;
						case '\\':
						case '"':	insert('\\');
						default:	insert(p[j]); break;
					}
				stuff("\"\n");
			}
			use_default = 0;
			break;
	}
}

// Force an exit even if there's an error.  Delay until all pending
// messages have appeared, but give up after a while and discard them,
// since one might be never-expiring.

exit_for_settings_file()
{
	TIMER tm;
	time_begin(&tm, 500);
	while (unseen_msgs() && !time_done(&tm))
		if (char_avail())
			getkey();
	drop_pending_says();
	has_arg = iter = 1;
	exit();
}

do_build_settings_file(debug)
{
	char ext_file[FNAMELEN], settings_file[FNAMELEN], eel_file[FNAMELEN];
	int tmp = tmp_buf(), i, err, max;

	for (i = 1; i < argc; i++) {	// Recover directory to use.
		if (!strncmp(argv[i], "-l", 2) && strstr(argv[i], "epsout.b")) {
			strcpy(eel_file, argv[i] + 2);
			*get_tail(eel_file, 1) = 0;
			if (chdir(eel_file))
				say("Can't cd to %s, error %d.", eel_file, errno);
			if (debug) {
				getcd(eel_file);
				buffer_printf("import-debug",
							  "Using directory %s for files.\n", eel_file);
			}
		}
	}
	strcpy(ext_file, "extensions-list");
	strcpy(settings_file, "settings-list");
	save_var bufnum = tmp;
	if (err_file_read(ext_file, FILETYPE_BINARY)) {
		if (debug) {
			buffer_printf("import-debug",
						  "No extensions list file %s, ignoring.\n", ext_file);
		} else
			exit_for_settings_file();
	}
	point = 0;
	while (grab_line(bufnum, eel_file)) {
		sayput("Loading %s...", eel_file);
		load_commands(eel_file);
	}
	buf_zap(bufnum = tmp);
	sayput("Listing settings...");
	max = final_index();
	for (i = 1; i <= max; i++)
		list_one_for_settings_file(bufnum, i);
	use_default = 0;
	bufnum = tmp;
	point = size();
	buf_printf(bufnum, "+path-sep char variable: %d\n", path_sep);
	buf_printf(bufnum, "+abort-key int variable: %d\n", abort_key);
	if (debug)
		buffer_printf("import-debug",
					  "Write settings to %s.\n", settings_file);
	if (err = file_write(settings_file, FILETYPE_BINARY)) {
		file_error(err, settings_file, "write error");
		delete_file(settings_file);	// Ensure caller notices failure.
	}
	if (debug)
		say("Done, exit this instance to continue the import process.");
	else
		exit_for_settings_file();
}

build_settings_file()
{
	do_build_settings_file(0);
}

debug_build_settings_file()
{
	do_build_settings_file(1);
}
