/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1995, 2021 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// C/C++/EEL parsing and indenting

jmp_buf c_indent_giveup;	/* come here when we hit narrowed edge */
user buffer char Closeback = 1;		/* select default C indenting style */
user buffer char Topindent = 1;
user buffer char Matchdelim = 1;
#define closeback	Closeback
#define topindent	Topindent
#define matchdelim	Matchdelim
user buffer short c_indent = 0;	/* indent by this amt; 0 means use tab size */
user buffer short c_case_offset = 0; /* alter normal indentation of case statements */
user buffer short c_access_spec_offset = 0; /* ... of public:, private:, protected: */
user buffer char c_tab_always_indents = 0;	// if 1, tab fixes indentation
					// from anywhere on line
user buffer char new_c_comments = 1;	/* insert new-style c comments */
user buffer short c_label_indent = 0;	/* indent labels to this column */
user buffer short c_contin_offset = 0;	/* alter normal indentation of contin lines */
user buffer char c_align_contin_lines = 48; // line up contin lines up to this col
user buffer int c_align_extra_space = 2;	// Adjust certain contin lines.
user buffer int c_align_selectors = 48;	// Align Objective-C selectors up to this col.
user buffer int c_align_contin_max_width = 0;	// Don't align to make wider line than this.
user buffer int c_align_contin_max_offset = 0;	// Right-align such wide lines.
user buffer char indent_preprocessor_contin = 4; // Such lines get this default indent.
user buffer short c_brace_offset = 0;	// move braces this much from usual position
user int c_look_back = 100000;	/* max # of chars to examine (for speed) */
user buffer int c_biggest_declarator = 20000; // Max chars scanned for "func(par1, par2, ...) {" sequence
user buffer int c_top_braces = 0;	// indent for outer {}'s in func
user buffer int c_top_contin = 3;	// indent for contin lines (decls) outside func
user buffer int c_top_struct = 8;	// indent for body of outer struct/class
user buffer int c_top_template = 0; // indent for template decl
user buffer int c_param_decl = 0;	// indent for old-style func param decls
user buffer char c_indent_after_extern_c = 0;	// 0: no extra indenting for extern "C"
user buffer char c_indent_after_namespace = 0;	// 0: no extra indenting for namespace
user buffer char c_align_open_paren;	// A multi-line format for func defs and similar.
user buffer char c_align_inherit = 4;	// Indent C++ contin lines using inheritance.
user buffer char c_align_break_with_case;	// Align break; to match last case:
user buffer int c_fill_column = 72;	// fill comments at this column (or -1 to disable)
user buffer int c_auto_fill_mode = 1;	// 0 never auto-fill 1 not after code 2 always
user buffer char fill_c_comment_plain;
user buffer char auto_fill_comment_rules = 3;	// 1: break on <Enter>, 2: on other chars
user buffer char c_reindent_previous_line = 1;
user buffer int php_top_level_indent;	// Indent for PHP top level.
user buffer int java_indent = 4;	/* indent by this amt; 0 means use tab size */
user buffer char c_quick_scan = 1;		// OK to assume brace at left margin is at top level

user buffer int comment_repeat_indentation_lines = 2;
user char find_linked_file_ignores_angles = 0;
user char mac_framework_dirs[FNAMELEN]
   = "~/Library/Frameworks:/Library/Frameworks:/System/Library/Frameworks"
	 ":/System/Library/PrivateFrameworks";

#define INCL_PATH_UNIX	"/usr/include:/usr/include/*:/usr/*/include"
#define INCL_PATH_WIN	"\\Program Files\\*Studio*\\vc*\\include"

	/* possible return values of in_c_comment() */
#define IN_COMMENT		0x100  // general mask for any comment
#define IN_OLD_COMMENT	0x101  // in a /* */-type comment
#define IN_NEW_COMMENT	0x102  // in a //-type comment
#define IN_STRING		0x200  // general mask for any string-like thing
#define IN_STR_CONST	0x201  // inside double quotes
#define IN_CHAR_CONST	0x202  // inside single quotes

// Flags returned by find_c_func_info.
#define CF_INFO_TYPE	1				/* Found a name like "struct" or "class" */
#define CF_INFO_CLASS	2				/* Found a class name. */
#define CF_INFO_FUNC	4				/* Found a function name. */

user int reindent_after_c_yank = 10000;
	// Reindent blocks smaller than this after yanking in C-mode

	// Below variables: 0=never, 1=except when yanking, 2=always.
user buffer char reindent_c_comments = 17;		// For /* */ comments.
user buffer char reindent_one_line_c_comments = 1; // For // comments.

user buffer char c_delete_trailing_spaces = 0; // Fix spaces when saving?
user buffer char reindent_c_preprocessor_lines = 1;	// 0 means leave # lines alone.
user buffer char reindent_perl_comments = 1;
user buffer char reindent_php_comments = 17;

user int c_tab_override = -1;	// No C-specific setting for tab width.
user int eel_tab_override = 4;	// EEL setting for tab width.
keytable c_tab;			/* key table for c mode */
buffer char in_perl_buffer;
buffer char in_shell_buffer;
#define SUBLANG_PHP	3

char _c_mode_name[] = "C";
void c_indenter();
int c_move_level(int dir, int stop_on_key);
int ifdef_auto_show_delimiter();
char *c_class_separator();

user char compile_csharp_cmd[128] = "csc \"%r\"";  // run C# compiler
user char compile_typescript_cmd[128]	  = "tsc \"%r\"";  // run TypeScript compiler on .ts files

user char c_auto_show_delim_chars[20] = "{[()]}";

#define PREPROC_LINE	"^[ \t]*#[ \t]*(if|else|endif|elif|asm)"
#define PREPROC_IF		"^[ \t]*#[ \t]*(if|asm)"
#define PREPROC_ELSE	"^[ \t]*#[ \t]*(else|elif)"
#define PREPROC_ENDIF	"^[ \t]*#[ \t]*(endif|endasm)"

// Matches C++ operator declarations.
#define CPP_OPLIST	"(new|delete)? *([-+*/%^&|<>=!]*|%(%)|%[%])"
#define CPP_OPSPEC	"(operator *" CPP_OPLIST ")"

#define OBJC_TOPBLOCKS "implementation|interface|protocol|class"

char tag_pattern_c[] = "([a-zA-Z0-9_]+::)*([a-zA-Z0-9_]+|" CPP_OPSPEC ")";

#define MS_MACRO_GROUP_NAMES \
	"(ACCESSOR|ADO_BINDING|CATEGORY|COLUMN|COM|CONNECTION(_PART|_POINT)?|" \
	"DISPATCH|EVENTSINK|EXTENSION_SNAPIN_NODEINFO|INTERFACE(_PART_DERIVE)?" \
	"|MESSAGE|MSG|OBJECT|OLECMD|PARAM|PROP(SET)?|PROPERTY(_SET(_EX)?)?" \
	"PROVIDER_COLUMN|SCHEMA|SERVICE|SINK|SNAPIN(COMMAND|TOOLBARID))" \
	"(_MAP|_PART)?"

volatile user char c_block_macro_open_pat[4096]
   = "BEGIN_" MS_MACRO_GROUP_NAMES;
volatile user char c_block_macro_close_pat[4096]
   = "END_" MS_MACRO_GROUP_NAMES;
volatile user char c_block_macro_inner_pat[4096]
   = "ADO_((FIXED|VARIABLE)_LENGTH|NUMERIC)_ENTRY[2-4]?"
	 "|ALT_MSG_MAP|BLOB_ENTRY(_STATUS)?|BOOKMARK_ENTRY"
	 "|CHAIN_(MSG_MAP(_ALT(_MEMBER)?|_DYNAMIC|MEMBER)?"
	 "|PROPERTY_SET|SNAPINCOMMAND_MAP)"
	 "|COM_INTERFACE_ENTRY[A-Z0-9_]*|COMMAND_((CODE|ID|RANGE)_)?HANDLER"
	 "|CONNECTION_(IID|PART|POINT_ENTRY)"
	 "|DECLARE_((DISPATCH|INTERFACE|OLDCMD)_MAP|GET_CONTROLLING_UNKNOWN)"
	 "|DEFAULT_PARSE_COMMAND|EXT_SNAPINMENUID"
	 "|EXTENSION_SNAPIN_(DATACLASS|NODEINFO_ENTRY)"
	 "|IMPLEMENTED_CATEGORY|INIT_INTERFACE_PART(_DERIVE)?"
	 "|INTERFACE_(AGGREGATE|PART)|MESSAGE_(RANGE_)?HANDLER"
	 "|NOTIFY_((CODE|ID|RANGE)_)?HANDLER|OBJECT_ENTRY(_NON_CREATEABLE)?"
	 "|ON_DSCNOTIFY(_RANGE)?|ON_EVENT(_(RANGE|REFLECT))?"
	 "|ON_PARSE_COMMAND(_PARAMS)?"
	 "|ON_PROPNOTIFY(_(RANGE|REFLECT))?|PROP_(DATA_)?ENTRY(_EX)?|PROP_PAGE"
	 "|PROPERTY_(INFO|COLUMN)_ENTRY[A-Z0-9_]*|REQUIRED_CATEGORY"
	 "|(SCHEMA|SERVICE)_ENTRY(_CHAIN)?|SINK_ENTRY(_(INFO|EX))?"
	 "|SNAPINCOMMAND_(RANGE_)?ENTRY|SNAPINMENUID|SNAPINTOOLBARID_ENTRY"
	 "|DECLARE_(PROTECT_FINAL_CONSTRUCT"
	 "|GET_CONTROLLING_UNKNOWN|CLASSFACTORY[A-Z0-9_]*|OBJECT_DESCRIPTION"
	 "|((NOT|ONLY|POLY)_)?AGGREGATABLE)";

user char c_prototype_macro_names[FNAMELEN] = "_ANSI_ARGS_|__P|PROTO";

user int use_c_macro_rules = 2;
