/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2021 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#define EELVERSION	140					/* major version */
#define EEL_MINOR_VERSION		"14.04"
user char eel_version[20] = EEL_MINOR_VERSION;

#include <codes.h>

#define DISPCHARS	"\xC4\xCD\xB3\xBA" \
			"\xDA\xC2\xBF\xC3\xC5\xB4\xC0\xC1\xD9" \
			"\xD6\xD2\xB7\xC7\xD7\xB6\xD3\xD0\xBD" \
			"\xD5\xD1\xB8\xC6\xD8\xB5\xD4\xCF\xBE" \
			"\xC9\xCB\xBB\xCC\xCE\xB9\xC8\xCA\xBC" \
			"\xBF    ^^^M-M-^x" \
			"\x1E\x1F\x04\x12"

#define AUTOSAVENAME	"%p#%b%e.asv#"
#define SESSDEFDIR	""	/* put session in directory on path */
#define SESSROOTDIR	"NONE"	/* skip tree search by default */

#define SAMECOL		1	/* this_cmd code for up-line, down-line */
#define APPENDKILL	2	/* this_cmd code for killing commands */
#define C_INDENT	3	/* this_cmd code for c-indent command */
#define CMP_WINDOWS	4	/* this_cmd code for compare-windows */
/* the next group are used by complete.e, and set by same-named cmds */
#define INP_EXIT	5	/* <Enter> in completion */
#define INP_ADD		6	/* normal keys in menu */
#define INP_EDIT	7	/* ^O in view or menu (obsolete) */
#define INP_ABORT	8	/* ^G in view, menu, or completion */
#define INP_GET_EXIT	9	/* <Enter> in menu */
#define INP_SHOW_MATCHES 10	/* ? in menu or completion */
#define INP_TRY		11	/* <Space> in completion */
#define INP_DEFAULT	12	/* ^S or ^R in completion */
#define INP_SEEN	13	/* paged out of view mode */
#define INP_COMPLETE	14	/* inp-complete: complete, then maybe exit */
#define INP_GET		15	/* <Tab> in menu */
#define LAST_BOOKMARK	16	/* this_cmd code for jump-to-last-bookmark */
#define INP_RECALL	17	/* this_cmd code for picking prev response */
#define INP_GRAB_WORD	18	/* this_cmd code for grabbing word from buf */
#define INP_GRAB_LINE	19	/* this_cmd code for grabbing line from buf */
#define INP_TO_LIST		20	/* this_cmd code for switching to listbox */
#define CTR_WINDOW		21	/* this_cmd code for ^L */
#define INP_LAST_RECALL	22	/* this_cmd code: grab newest prev response */
#define CMD_ID_COMPL	23	/* just auto-completed an identifier */
#define CMD_SHOW_COMPL	24	/* just showed an identifier's completion */
#define CMD_INDENT_REG	25	/* indenting a region */
#define CMD_PROC_COMPL	26	/* process completion */
#define INP_GET_OTHER	27	/* get some suggested name */
#define INP_NEXT_RECALL	28	/* this_cmd code: grab newer prev response */
#define INP_PREV_RECALL	29	/* this_cmd code: grab older prev response */
#define UNDO_MOVEMENTS	30	/* this_cmd code: for undo-movements cmd */
#define UNDO_CMD		31	/* this_cmd code: for other undo/redo cmds */
#define SPELL_CORRECT	32	/* this_cmd code: for spell-correct cmd */

#define MAXKEYS     0x7fffff00	// A value higher than any possible key.
// Leave NUMKEYS undefined to detect old EEL code that loops on it, so
// it can be updated to use new functions.

#define NULL		((char *) 0)
#define UNREFERENCED_PARAMETER(p)	((p) = (p))		// Squelch EEL warning.
#define FNAMELEN	4096
#define TAGLEN		4096

	/* These macros start at FKEY(1), not 0.) */
#define FSHIFT(c)	(NUMSHIFT(FKEY(c)))
#define FCTRL(c)	(NUMCTRL(FKEY(c)))
#define FALT(c)		(NUMALT(FKEY(c)))
#define FPLAIN(c)	(NUMPLAIN(FKEY(c)))


typedef int void;		/* for now, let void be synonym for int */

buffer short *mode_keys;	/* the buffer-specific key table */
buffer char *major_mode;	/* name of the current major mode */
char *strkeep(char *s);		// Preserve mode names across loading.
char **strkeep_data;			// Keep its data here.
zeroed buffer char *mode_extra;	/* other stuff to display on mode line */
zeroed char mode_more_extra[FNAMELEN];	// More stuff, not buffer-specific.
buffer char retain_forced_mode;		// Leave buffer's mode alone.
zeroed buffer char sublanguage_code;	// For modes that support variations.
user char full_path_on_mode_line;	// display the full path always?
char display_func_name[FNAMELEN];	// name of current function, for display
zeroed int start_of_function;	// where this function starts, if known
zeroed char must_find_func_name;	// Don't stop looking when user hits a key.
user buffer short tab_size = 8;	/* distance between tab stops */
user buffer short soft_tab_size;	// Indent based on this size, not physical tabs.
buffer char *_display_class;	/* display class table */
char *_echo_display_class;	/* display class table for echo area */
buffer short position_window_on_screen_line = 50; /* percentage down window */
zeroed buffer short filevar_tab_size;	// file said to use this tab size
user buffer char over_mode = 0;	/* are we in overwrite mode? */
user buffer char fill_mode = 0;	/* are we in auto-fill mode? */
buffer int (*indenter)();	/* automatic indentation function */
zeroed int indenter_action;	// Bits for why we're indenting.
#define IACT_AUTO_INDENT		1	// Pressed <Enter>, indenting new line.
#define IACT_AUTO_FILL			2	// Breaking long line.
#define IACT_COMMENT			4	// Indenting comment as code.
#define IACT_FILL_COMMENT		8	// Explicit fill command in comment.
#define IACT_AUTO_FILL_COMMENT	16	// Breaking long line in comment.
#define IACT_REINDENT_PREV		32	// Pressed <Enter>, indenting old line.
zeroed buffer int (*mode_restrict_break)();	// Don't break line here.
user char default_reindent_previous_line = 0; // Should auto-fill reindent both lines?
user char sentence_end[70] = "([.?!]<rsquare|rparen|squote|dquote>*"
							 "[ \t\n]*([ \t]|\n))";
user char sentence_end_double_space = 0; /* fill with two spaces? */
user buffer char auto_indent = 0; /* does <Enter> do indentation? */
user buffer char indent_with_tabs = 1; /* should indenting use tab chars? */
user char resize_rectangle_on_tab;	// Does <Tab> expand rectangular region?
user buffer short goal_column = -1; // up-, down-line stays in column if >= 0
user buffer int margin_right = 70;	/* right-hand margin */
user buffer char auto_fill_indents = 1;	/* does auto-fill do indenting? */
user int misc_language_fill_mode = 0xfffffff; // auto-fill comments in various langs?
user buffer char virtual_space = 0;	// position where there are no chars?
				// if 2, even fwd- & bwd-char cmds do it
user char typing_deletes_highlight = 1;
	// do self-inserting chars delete highlighted region?
user char typing_hides_highlight = 1;	// do self-inserting chars turn off highlight?
user char shift_selects = 1;	// does shift+arrow key select text?
zeroed char shift_selecting;	// the current selection started like this
user char national_keys_not_alt = 2;	// OS-specific var.
user buffer char recognize_password_prompt = 3;	// pop up prompt in 1=telnet, 2=process buf
user char recognize_password_pattern[FNAMELEN]	// regex to notice the prompt
   = "^((Enter |type the |.*@.* |%[.*%])?pass(word|phrase)( for .*)?"
	 "|Password: *\nResponse):[ \t]*";

zeroed window short _prev_buffer;	/* most recent buffer */
#define MAXOLDBUFS	30	/* number of previous bufs to remember */
zeroed short _prev_old_buffers[MAXOLDBUFS];	/* list of previous bufs */
zeroed short _prev_num;		/* # of remembered previous buffers */
zeroed buffer int buf_accessed;	// highest number is most recently used buf
zeroed int buf_accessed_clock;	// numbers to put in buf_accessed

window char _window_flags;	/* misc data */

user char want_auto_save = 1;	/* if 0, never do autosave */
buffer int auto_save_failed;
short _auto_save_counter = 0;	/* counts cmds */
user short auto_save_count = 500; /* autosave when counter reaches this # */
user int auto_save_biggest_file = 5_000_000;
user char auto_save_name[FNAMELEN] = AUTOSAVENAME;	/* autosave filename */
user int auto_save_idle_seconds = 30;	// After this, force autosave.
TIMER auto_save_idle_timer;
user char backup_name[FNAMELEN] = "%p%b.bak";
user char backup_by_renaming = 2;
user char want_state_file_backups = 1;
user char state_file_backup_name[FNAMELEN] = "%pebackup%e";
user char ignore_file_extensions[FNAMELEN] = "|.obj|.exe|.o|.b|";
user char ignore_file_pattern[FNAMELEN * 2];
user char ignore_file_basename[FNAMELEN] = "%.";
user char include_directories[4096];
user char warn_before_overwrite = 1;	// warn before overwriting file?
user char default_delete_trailing_spaces = 0; // Fix spaces when saving?
user char default_add_final_newline = 0; // Add a newline when saving?
zeroed buffer char file_write_newfile;		// fail if file exists?
user char print_destination[FNAMELEN] = "lpt1";	/* printer name or printing cmd */
user char print_destination_unix[FNAMELEN] = "!lpr \"%f\"";
user char print_tabs = 0;	// if 0, convert tabs to spaces when printing
user char print_in_color = 1; // 0=never, 1=if it's a color printer, 2=always
user char print_color_scheme[30] = "standard-gui";
						// Name of special color scheme for printing.
user int print_line_numbers;	// Include line numbers?
user char print_doublespaced;	// Double-space the output?
user char print_long_lines_wrap = 1;		// Wrap long lines?
zeroed struct print_options print_options;   // retain page range here
zeroed char quiet_write_state;	// Write-state doesn't prompt.
user char uncompress_files = 3;	// Act on extensions like .gz, .bz2?

#define PRHEAD_FNAME	(1)	/* display the file name */
#define PRHEAD_PAGECNT	(2)	/* display the page number and total */
#define PRHEAD_DATE		(4)	/* display the current date */
#define PRHEAD_PATH		(8)	/* display the path if file name displayed */
user char print_heading = (PRHEAD_FNAME | PRHEAD_PAGECNT | PRHEAD_DATE
						   | PRHEAD_PATH);

zeroed buffer char already_made_backup;
user buffer char want_backups = 0;	// create .bak file on each save?
user char save_all_without_asking = 0;	// Don't ask for names of fileless bufs
zeroed buffer char save_without_prompt;	// Save this buffer without asking.
user buffer char case_fold = 1;	    // should searches consider A and a equal?
user buffer char sort_case_fold = 2;	// should sorts?  2 means see case_fold
user buffer char delete_hacking_tabs = 0;
				/* should deleting change tabs to spaces? */
user char delete_options;	// Flags for deleting: 1=never append to kill.
user buffer int undo_size = 500000;	/* max # of chars to save for undo */
buffer char undo_keeps_narrowing;	// Undo may not remove narrowing.
buffer int _buffer_saved_here;	// Val of undo_count(1) where the buffer was saved.

// Values for undo_flag:
#define UNDO_READ_FILE		5	/* Read this file from disk. */
#define UNDO_WRITE_FILE		6	/* Wrote this file to disk. */
#define UNDO_UNMODIFIED		7	/* Marked this buffer unmodified. */

/* for get_window_info() and low_window_create() */
struct window_info {
	short left, top, right, bottom;
	short textcolor, hbordcolor, vbordcolor, titlecolor;
	short borders, other, bufnum;
	int point, dpoint;	/* primitives fill in up to & incl this line */
	int dcolumn;	/* preserve more window data */
	short prevbuf;
	short colscheme;
};

/* a screen worth of window info */
struct screen_info {
	struct window_info *wins;	/* window information structs */
	int lines, cols;	/* size of screen */
	int num;		/* # of tiled window on screen */
	int cur_window;		/* # of current window */
	int echo_line;		// line for echo area
	int avoid_top;		// # of lines at top with no tiled windows
	int avoid_bottom;	// # of lines at bottom with no tiled windows
	int permanent_menu;	// want perm menu?
} zoom_info, push_info;

zeroed char _recursion_level;	/* depth of recursive editing */
user short find_lines_visible = 8; // move dialog if we can't see this many

#define NUM_BOOKMARKS	36
spot _bookmark_spots[NUM_BOOKMARKS];
zeroed int _bookmark_saved_pos[NUM_BOOKMARKS];	// Set the spot from this later.
int _bookmark_slot;	/* # of last-set bookmark in ring, 0...9 */
int _last_bookmark;	/* last bookmark we jumped to */
int _bookmark_next;	/* next bookmark for repeated jump_to_last_bookmark */
int (*_blist_cmd)();		/* call this with selected bookmark */
int do_jump_to_bookmark();	/* possible values of above */
int say_set_bookmark();
user int set_bookmark_automatically;

keytable reg_tab, cx_tab;	/* the standard key tables */
keytable cc_tab;		// The Ctrl-C prefix key.
keytable buf_tab;		/* the default buffer-specific key table */
keytable dired_tab;	/* key table for dired mode */
keytable bufed_tab;	/* key table for bufed mode */
keytable blist_tab;	/* key table for bookmark list mode */

#define MAX_MACRO	500
int def_text[MAX_MACRO];	/* keys in macro being defined */
zeroed short _len_def_mac;	/* size of macro, incl. len, or 0 if none */
zeroed char recording_suspended; // stopped recording in middle of macro?
zeroed char final_macro_pause;	// did macro end with a pause-macro?
int key_binding[30];		// ask_key() puts key info here
user char macro_runs_immediately = 1;	// 0 means just queue the keystrokes.

user char mention_delay = 0;	// time before mention() in .1 secs
user int see_delay = 100;  // time error msgs must be made visible in .01 secs
user short key_repeat_rate = 40; // make keys rpt even faster
user short new_search_delay = 250; // this long btw keys means new text in list

user char server_raises_window = 1;	// -add under X foregrounds like MS-Windows?

// Ways to default based on current buffer's directory:
#define PBD_NONE	0	/* don't do anything special */
#define PBD_TEMPCD	1	/* while prompting for a file, cd there */
#define PBD_INSERT	2	/* insert buffer's dir at prompt, strip // ala Gnu */
#define PBD_JUSTINS	3	/* insert buffer's dir at prompt, don't strip // */
user char prompt_with_buffer_directory = PBD_INSERT;
user char grep_prompt_with_buffer_directory = PBD_TEMPCD;	// For grep/file-repl.

// Windows common file dialog variables.
user char want_common_file_dialog = 1;	// Use if run from menu or toolbar.
user char common_open_use_directory;	// Which directory to use with File/Open.
zeroed void (*use_alternate_dialog)();	// Call this func to create dialog.
zeroed char *use_alternate_dialog_name;	// Use this dialog name instead.
zeroed char *use_alternate_dialog_tag;  // Use this dialog tag name.
zeroed buffer char insert_newlines_in_dialog; // <Enter> doesn't exit dialog

user short beep_duration = 5;	// duration of beep in .01 seconds
user short beep_frequency = 370; // freq of beep in hertz, or 0 for system beep

user short kill_buffers = 10;		/* number of kill buffers */
user char clipboard_access = 1;		// Use the clipboard?
user short clipboard_format = CF_DEFAULT;
user int clipboard_convert_unicode = 4;
user int clipboard_convert_mac_lines = 2;
int clipboard_translation_type;

int _preserve_column;	/* used to maintain the current column */
#define MAXRECENT	60
int _recent_keys[MAXRECENT];
short _recent_place;
char _abort_load_on_error = 1;	// 1: yes, 0: mark up file, 2: during init
char _errors_in_cmd_file = 0;	/* # of errors load-buffer found */
zeroed char now_parsing_cmds;	// Currently parsing a cmd file?
zeroed char load_fail_ok;	// give error if we can't autoload?
user char einit_file_name[FNAMELEN] = "einit.ecm";	// Init from this file.

user char scroll_at_end = 1;	// should up- & down-line scroll if offscreen?
user char search_in_menu = 0;	/* does typing with menu up auto-search? */
user char response_tracks_list = 0; // does listbox copy its line to prompt?
user char insert_default_response = 1;	// prompts begin with default?
user char insert_file_remembers_file;	// Does this command assume same file?
user short window_overlap = 2; /* #lines or -percent overlap when paging */
user char paging_centers_window = 1;	// put cursor in middle after page?
char _display_characters[66] = DISPCHARS; /* chrs to use for borders, contin */
user char mode_format[60] = " %b%E%M %l,%c %d%p %s%f";
user char mode_line_at_top = 0;	// want mode line at top of screen?
#define MODE_EDGE	(mode_line_at_top ? BTOP : BBOTTOM)
user short mode_line_position = TITLELEFT(2);	// put mode line at left
user char display_definition = 1;	// 1=gui: in title bar, 2=non-gui mode line
user char display_c_options = 1;	// Include namespace in display-defintion?
user int copy_include_file_name_options;
user char want_display_host_name = 1;	// put this computer's name in title bar?
user char window_caption_file[128] = "%s - Epsilon";
user char window_caption_buffer[128] = "(%s) - Epsilon";
user char window_caption[128] = "Epsilon";
user char want_window_borders = 1;	// put lines between windows?
zeroed short last_window_color_scheme;	// for borderless windows
user buffer char show_spaces = 0;	// make spaces & tabs visible
zeroed buffer char last_show_spaces;	// are spaces & tabs already visible?
char *_show_disp_class; // version of _display_class for show-spaces
char *_show_disp_chars; // version of _display_characters for show-spaces

user char show_when_idle[120];		// put text in echo area when not busy
user short show_when_idle_column = 48;	// starting at this column

user char border_left = 0;	// put border on left edge of screen?
user char border_top = 0;	// put border on top edge of screen?
user char border_right = 0;	// put border on right edge of screen?
user char border_bottom = 0;	// put border on bottom edge of screen?
user char border_inside = 1;	// put border between side-by-side windows?
user char tiled_border = ALL_BORD(BSINGLE);	// type of lines for border

zeroed short _help_on_cmd;	// if zero, give general help, else help on cur cmd
char *_show_help_on;	// but override with name if nonzero
user char preserve_filename_case;	// Bits control how to treat X vs x in filenames.
user char file_pattern_wildcards = FPAT_ALL;	// Turn off some wildcard chars?
user char file_pattern_rules;	// Ignore some symlinks in wildcarding?
user char file_pattern_ignore_directories[FNAMELEN];
user char file_pattern_unc_domains[FNAMELEN];	// Restrict matching to these domains.
user int cygwin_filenames = 0;	// Rules for translating Cygwin file syntax.
user char cygwin_root[FNAMELEN];	// Root of Cygwin system.
user int drive_match_types = FSYS_LOCAL | FSYS_CDROM;
zeroed char was_quoted;	// Did the last file name typed have "" around it?
user int file_interpret_env_vars = 65535;
char *filename_subst_var(char *var);
user window int display_column = 0;	// 1st col to show, or -1 to wrap

#define WRAP_NEVER		(0)
#define WRAP_ALWAYS		(1)
#define WRAP_PRESERVE	(2)		/* Keep window's existing wrap setting. */
#define WRAP_SPECIAL	(3)		/* Special logic. */
user int wrap_info_mode			= WRAP_ALWAYS;
user int wrap_dired_live_link	= WRAP_ALWAYS;
user int wrap_split_vertically	= WRAP_NEVER;
user int wrap_view_error_lines	= WRAP_PRESERVE;
user int wrap_grep				= WRAP_SPECIAL;

zeroed window char _highlight_control;	// want highlight in this window?
zeroed window char is_current_window;	// is this window the current one?
zeroed char disable_recording_last_keys;	// Prompting for password?
user char restore_color_on_exit = 1; // dos: set after-exiting color at start?
user char default_word[60] = "<word>+";
user char console_ansi_font = 0;	// Can Windows console version use non-OEM font?
user int default_translation_type = FILETYPE_AUTO;
user int force_remote_translation_type = FILETYPE_AUTO;
buffer int fallback_translation_type = FILETYPE_AUTO;
user int fallback_remote_translation_type = FILETYPE_AUTO;
user int new_buffer_translation_type = FILETYPE_AUTO;
user char default_write_encoding[30] = "UTF-8-No-BOM";
user char default_read_encoding[60];
user char detect_encodings = 127;		// Detect everything we can.
user int file_read_kibitz = 1;	// Show message reading binary files?
user char locate_path_unix[FNAMELEN] = "/{*bin*,etc,home*,lib,opt,root,usr*,var*}/**/";
user buffer char *word_pattern;
user char forward_word_to_start;	// Does Alt-f go to start of word?
user buffer char indents_separate_paragraphs = 0;
	/* otherwise, only blank lines and ^L's (form-feeds) do */
user buffer char tex_paragraphs = 0;	// exclude \ @ . lines at parag edges?
user char pull_word_from_tags = 2;		// 1=only if tags are loaded, 2=always

user buffer int comment_column = 40;
user char indent_comment_as_code = 1;
user buffer char comment_start[100] = ";[ \t]*";	/* pattern to find them */
user buffer char comment_begin[100] = "; ";	/* text to insert to make */
user buffer char comment_end[100] = "";		/* text to insert after */
user buffer char comment_pattern[100] = ";.*$";	/* match whole comment */
user char comment_region_options = 0;
	// 1=make comments indented, 2=remove them too, 4=skip blank lines

// Access above variables via these functions, to allow for embedding.
char *get_str_by_mode();
#define get_comment_start()		(get_str_by_mode("comment_start"))
#define get_comment_begin()		(get_str_by_mode("comment_begin"))
#define get_comment_end()		(get_str_by_mode("comment_end"))
#define get_comment_pattern()	(get_str_by_mode("comment_pattern"))

user short near_pause = 50; // delay in .01 secs when delimiter is on screen
user short far_pause = 100; // delay when it's not visible without new window
buffer int (*mode_move_level)();	/* mode-specific brace matcher */
user char fundamental_auto_show_delim_chars[20] = "";
zeroed char inside_show_matching_delimiter;
buffer int (*mute_unmatched_delimiter)();

/* The export-colors command scans the following definitions.  It
expects that all color names will use the format below, that the
first will be named "black", and that END_COLOR_MARKER will follow
the last color name. */

#define black		MAKE_RGB(0, 0, 0)
#define dark_red	MAKE_RGB(128, 0, 0)
#define dark_green	MAKE_RGB(0, 128, 0)
#define brown		MAKE_RGB(128, 128, 0)
#define dark_blue	MAKE_RGB(0, 0, 128)
#define dark_magenta	MAKE_RGB(128, 0, 128)
#define bluish_purple	MAKE_RGB(140, 0, 255)
#define medium_cyan	MAKE_RGB(0, 144, 180)
#define dark_cyan	MAKE_RGB(0, 128, 128)
#define light_grey	MAKE_RGB(192, 192, 192)
#define money_green	MAKE_RGB(192, 220, 192)
#define sky_blue	MAKE_RGB(166, 202, 240)
#define cream		MAKE_RGB(255, 251, 240)
#define medium_red	MAKE_RGB(192, 0, 0)
#define medium_blue	MAKE_RGB(0, 0, 192)
#define grey		MAKE_RGB(160, 160, 164)
#define dark_grey	MAKE_RGB(128, 128, 128)
#define red			MAKE_RGB(255, 0, 0)
#define green		MAKE_RGB(0, 255, 0)
#define yellow		MAKE_RGB(255, 255, 0)
#define blue		MAKE_RGB(0, 0, 255)
#define magenta		MAKE_RGB(255, 0, 255)
#define cyan		MAKE_RGB(0, 255, 255)
#define light_red	MAKE_RGB(255, 192, 192)
#define light_green	MAKE_RGB(192, 255, 192)
#define light_blue	MAKE_RGB(192, 192, 255)
#define light_yellow	MAKE_RGB(255, 255, 224)
#define light_olive	MAKE_RGB(230, 214, 0)
#define ocher		MAKE_RGB(200, 100, 0)
#define bisque		MAKE_RGB(238, 213, 183)
#define white		MAKE_RGB(255, 255, 255)
#define light_pink	MAKE_RGB(255, 235, 235)
#define dark_brown	MAKE_RGB(96, 48, 0)
#define END_COLOR_MARKER	/* Export-colors needs this here. */

/* the next 16 must be the first color classes, in this order */
color_class text;
color_class highlight;
color_class echo_area;
color_class mode_line;
color_class horiz_border;
color_class vert_border;
color_class popup_title;
color_class popup_border;
color_class after_exiting;
color_class debug_text;
color_class debug_highlight;
color_class default;
color_class overlap_even;
color_class overlap_odd;
color_class screen_border;
color_class screen_decoration;

color_class viewed_text;
color_class viewed_border;
color_class viewed_link;		// make rest this color
color_class selected_link;		// make 1st one this color
color_class viewed_gui_text;
color_class dialog_gui_text;
color_class menu;
color_class menu_highlight;
color_class menu_bar;
color_class menu_bar_highlight;
color_class submenu;
color_class submenu_highlight;

color_class man_underlined;
color_class man_bold;
color_class pull_highlight;

short _our_mono_scheme;	/* name table index of scheme to use for mono mode */
short _our_color_scheme; /* name table index of scheme to use for color mode */
short _our_gui_scheme; /* name table index of scheme to use in gui version */
short _our_unixconsole_scheme; /*  scheme to use on unix console */
user int selectable_colors = 128; // # of colors selectable by set-color

/* default display classes for all characters */
char _std_disp_class[65536] = {
	BC, BC, BC, BC, BC, BC, BC, BC,
	BC, BTAB, BNEWLINE, BC, BC, BC, BC, BC,
	BC, BC, BC, BC, BC, BC, BC, BC,
	BC, BC, BC, BC, BC, BC, BC, BC,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, BC,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, BMC,
};

/* echo area display class */
char _echo_disp_class[65536] = {
	BC, BC, BC, BC, BC, BC, BC, BC,
	BC, BC, BC, BC, BC, BC, BC, BC,
	BC, BC, BC, BC, BC, BC, BC, BC,
	BC, BC, BC, BC, BC, BC, BC, BC,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, BC,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, BMC,
};

// Cursor shapes contain the starting scan line and the ending scan line,
// expressed as a percentage of the character height.  Values close to
// 0 or 100 are absolute scan line counts, so 98 to 99 is exactly one
// scan line high.
user int normal_cursor = CURSOR_SHAPE(98, 99);	// shape of insert-mode cursor
user int overwrite_cursor = CURSOR_SHAPE(0, 99); // shape of overwrite cursor
   // use different cursor shapes in virtual space inside characters:
user int virtual_insert_cursor = CURSOR_SHAPE(93, 99);
user int virtual_overwrite_cursor = CURSOR_SHAPE(0, 5);

// Gui cursor shapes contain the height, width, and vertical offset of
// the cursor, expressed as a percentage of the character dimensions.
// Values close to 0 or 100 are absolute pixel counts, so a width of 98
// is two pixels smaller than a character.  A width of exactly zero
// means use the default width.  All measurements are from the top left
// corner of the character.
user int normal_gui_cursor = GUI_CURSOR_SHAPE(100, 2, 0);	// insert-mode
user int overwrite_gui_cursor = GUI_CURSOR_SHAPE(100, 100, 0); // overwrite
   // use different cursor shapes in virtual space inside characters:
user int virtual_insert_gui_cursor = GUI_CURSOR_SHAPE(50, 2, 0);
user int virtual_overwrite_gui_cursor = GUI_CURSOR_SHAPE(50, 100, 0);

user char want_bell = 1;		/* allow beeping at all? */
user char bell_on_abort = 0;		/* beep when ^G'ing */
user char bell_on_search = 1;		/* beep when search fails */
user char bell_on_date_warning = 1;	/* when file's changed on disk */
user char bell_on_bad_key = 1;		/* type illegal char at prompt */
user char bell_on_completion = 1;	/* no matches/multiple matches */
user char bell_on_read_error = 1;	/* can't read file */
user char bell_on_write_error = 1;	/* can't write file */
user char bell_on_autosave_error = 1;	/* autosave file can't be written */

keytable comp_tab, menu_tab, view_tab, grep_tab;
char *_view_title;		/* title for view_buffer to use */
window short *_view_prev_mode;	/* mode of buffer before viewing */
window int _view_prev_window;	/* window before viewing */
zeroed char _view_prev_input;		/* _doing_input before viewing */
zeroed char paging_retains_view;	// Can <Space> or Ctrl-V exit viewing?
short _view_left = 2;	/* space at sides of viewed popup */
short _view_top = 2;
short _view_right = 2;
short _view_bottom = 6;
short _view_border = LR_BORD(BSINGLE) + TB_BORD(BDOUBLE);
user char completion_pops_up = 1;	/* do ? when completion would ring */
zeroed char _doing_input;	/* are we doing input (2=recursed, 1=norm)? */
zeroed char _doing_completion;	/* should ?, <Space>, <Esc> do stuff? */
zeroed int prompt_orig_buf;	// buffer before we displayed the prompt
zeroed int prompt_orig_win;	// window before we displayed the prompt
#define _MATCH_BUF	"-match-buf"	// Put possible completion values in this buffer.
char _this_match[TAGLEN];	/* value returned by a matcher during completion */
zeroed int (*completion_literal_char)();	// Insert char literally?
zeroed char completion_column_marker;	// This character isn't part of the match.
zeroed char resize_menu_list;	// Widen menu list of choices?
buffer int (*go_to_link)();	// pass link name to this to go there
int list_matches();		/* std routine for showing all matches */
int list_recalls();		// lists user's prev responses to prompt
char *general_matcher(char *s, int flags);		// Match from a list of items in buffer.
int *(*completion_lister)();	/* alternate to list_matches() */
char *(*list_finder)();		/* alternate finder for list_matches() */
char *only_file_extensions;	// completion only gives such files
char cmd_result;	// last completion special function
int _grab_word_loc;		// in input, grab chars from here in orig buf
int _grab_word_start;	// Already started grabbing.
char *recall_id;		// recall prompts via this key, not cmd name
#define OLDRESP	"-oldresp"
user int recall_maximum_size = 500_000;	// max chars of stored old responses
user int recall_longest_response = 10_000; // don't save responses bigger than this
user int recall_maximum_session = 500_000; // max chars saved in session file
user char recall_prior_response_options = 0;	// at prompts, display old in-place?
user char recall_response_selected;		// highlight recalled responses?

int w_left, w_top, w_right, w_bottom;	// coords for this window
window int (*mouse_handler)();	// do special mouse stuff for this win
zeroed char run_by_mouse;	// is this cmd being run via mouse menu?
zeroed char show_mouse_choices;	// put up menu of options?
zeroed window char _dragging_control; // 1 in window we're dragging text in
user char mouse_selection_copies = 0;	// does selecting copy text? 2=only in Unix
user char mouse_center_yanks = 0;	// 0=never, 1=always, 2=only in Unix
user int mouse_update_during_scroll = 10_000_000;
user int scroll_init_delay = 35;	/* hundredths of a second */
user int scroll_rate = 45;		/* lines per second */
char tiled_scroll_bar = 0; // want constant scroll bar on tiled wins?
user char scroll_bar_type = 1;	// Flavor of scroll bar desired.
user int cursor_blink_period = 100;	// In hundredths of a second.
user int draw_focus_rectangle = 0;	// Draw a box around current line?
user char draw_column_markers[FNAMELEN];	// Draw lines at certain columns?
user buffer char draw_line_numbers;
user buffer char line_number_width = 6;
user int follow_mode_overlap;	// Windows should overlap by how many lines?
buffer char follow_mode_on;
user int message_history_size = 1_000_000;	// Keep this much of old status msgs in #messages# buffer.

user int wheel_click_lines = -1; // # of lines to scroll per click of mouse
		// wheel, 0 to scroll by pages, or -1 to use control panel setting.

typedef short MOUSE_CURSOR;
	// mouse cursors used at various times, and what each should display
MOUSE_CURSOR std_pointer			= CURSOR_IBEAM;
MOUSE_CURSOR left_pointer			= CURSOR_ARROW;
MOUSE_CURSOR drag_and_drop_pointer	= CURSOR_ARROW;
MOUSE_CURSOR right_pointer			= CURSOR_RIGHTARR;
MOUSE_CURSOR horiz_arrows			= CURSOR_SIZEWE;
MOUSE_CURSOR vert_arrows			= CURSOR_SIZENS;
MOUSE_CURSOR move_arrows			= CURSOR_ARROW;
MOUSE_CURSOR pan_neutral			= CURSOR_PAN;
MOUSE_CURSOR pan_up					= CURSOR_PAN_UP;
MOUSE_CURSOR pan_down				= CURSOR_PAN_DOWN;

buffer char mouse_dbl_selects;	// is double-clicking like <Enter>?
char permanent_menu;		// want a permanent menu bar?
char want_gui_menu = 1;		// want a menu bar under Windows?
user char want_gui_prompts = 1;	// Use dialogs for prompting?
user char want_gui_printing = 1;	// Use Windows for printing?
user char winhelp_display_contents = 0;	// Should help show contents page?
user char mshelp2_collection[FNAMELEN];	// Name of ms-help2 collection to use.
user char alt_invokes_menu = 0;	// in Windows, does Alt key alone go to menu?
user char alt_numpad_keys = 0;	// Treat Alt-<Numpad> as separate keys, not ascii codes?
#define HELPBUF		"-help"		/* name of help buffer */

user int epsilon_manual_port = 7567;

// Settings for epsilon-help-format- vars
#define HELP_FORMAT_POPUP	1	/* Internal help in popup window */
#define HELP_FORMAT_HTML	2	/* Display in a browser */
#define HELP_FORMAT_WINHELP	3	/* Use WinHelp in Windows */
user int epsilon_help_format_win_console;
user int epsilon_help_format_win_gui;
user int epsilon_help_format_unix_gui;

char _region_file[FNAMELEN];	/* last file inserted, or region written to */
char _docname[FNAMELEN];		/* name of doc file */

short *ask_key();		/* function to ask for key */
command normal_character();
command show_matching_delimiter();
command maybe_show_matching_delimiter();
command case_indirect();
command yank();
command help();
char *temp_buf();		/* make temporary buffer */
command alt_prefix();
command ctrl_prefix();
command backward_delete_character();
command down_line();
command up_line();
command next_page();
command previous_page();
command center_window();
command scroll_up();
command scroll_down();
command beginning_of_window();
command end_of_window();
command goto_beginning();
command goto_end();
command forward_word();
command backward_word();
command forward_sentence();
command forward_level();
command backward_level();
command delete_character();
int dired_examine_in_window();
int prev_indenter();
int do_lowercase_region();
int do_uppercase_region();
int do_capitalize_region();
int allow_mouse_switching();
int inp_recall();
int inp_grab_word();
int inp_grab_line();
int inp_get_other();
int pull_word();
int pull_word_fwd();
int load_this_bytecode_file();
int compile_via_dll();
int list_file_matches();
int c_maybe_break_line();
int generic_maybe_break_line();
int literal_if_quoted();
int literal_always();
command edit_variables();
command fill_comment();
command prefix_fill_paragraph();
short *get_color_scheme_variable();
char *help_on_search();
int help_on_mode();
command abort();
char *nt_match();
char *give_buffer_name();
char *get_var_val();
char *get_int_buf();
char *f_match();
char *do_file_match();
char *b_match();
char *var_match();
char *tag_match();
char *variable_type_name();
char *get_cmdline();
char *default_search_string();
char **default_replace_string();
char *get_mode_string_variable();
char *preserve_buffer_type(char *buf);
char *strpbrk(char *s, char *charset);
char *strpbrk_cnt(char *s, char *charset, int skip);

struct file_info {		/* returned by check_file() */
	int fsize;			/* file size in bytes */
	int opsysattr;		/* opsys-dependent attribute */
	int raw_file_date;		/* opsys-dependent file date */
	short year, month, day;		/* file date: 1980-2099, 1-12, 1-31 */
	short hour, minute, second;	/* 0-23, 0-59, 0-59 */
	byte attr;			/* epsilon standardized attribute */
	byte check_type;		/* file/directory/device code */
};

buffer struct file_info file_info;	/* info when file was read */
zeroed buffer int unique_file_identifier[3];
user buffer char want_warn = 1;	/* warn about files updated on disk? */
buffer char ignoring_file_change;	// don't warn in this buffer for a while
user int file_date_tolerance = 2;	// Set to zero to warn about any timestamp change.
user char file_date_skip_drives = 0;	// Check for modified files on drive types?
user char date_format[FNAMELEN] = "%D-%t-%y %h:%N:%E %am";
user buffer char auto_read_changed_file = 0;
				// read a file that changed on disk w/o asking?
user char use_file_variables = 7;
user int file_variables_search_limit = 3000;

zeroed char _force_auto_read_changed_file; // Don't gripe if user said read.
buffer char _buf_readonly;	/* is buffer read-only? */
/* 0 no, 1 yes, 2 auto-set from file, 3 in pager mode, 4 set by all-read-only */
buffer char _read_aborted;	/* did we only read part of this buf? */
user char readonly_pages = 1;	// Does paging work in any r/o buffer?
user char all_read_only;	// Set to make all buffers read-only by default.

// Bits for readonly_warning:
#define ROWARN_MSG		1	/* when a file's read-only, show msg */
#define ROWARN_BUF_RO	2	/* when a file's read-only, set buffer read-only */
#define ROWARN_BELL		4	/* when a file's read-only, ring bell */
#define ROWARN_GREP		8	/* inside a grep, postpone the above */
#define ROWARN_REFRESH	16	/* inside a refresh-files, like grep */
user char readonly_warning = ROWARN_MSG | ROWARN_BUF_RO;

zeroed buffer char force_save_as;
#define FSA_NEWFILE		1	/* buffer made by File/New command */
#define FSA_READONLY	2	/* user checked read-only box */

zeroed buffer char buffer_not_saveable;	// Don't try to save, with bits:
#define NOSAVE_ALL		1	/* buffer_not_saveable: disable all tasks below */
#define NOSAVE_SESSION	2	/* don't save in session file */
#define NOSAVE_AUTOSAVE	4	/* don't autosave */
#define NOSAVE_SAVEALL	8	/* don't include in save-all-buffers */
#define NOSAVE_WARN		16	/* don't warn on exit if file's unsaved */
#define NOSAVE_BUFED	32	/* don't list in bufed if bufname starts with - */

zeroed char synchronous_file_io;		// Do internet funcs before returning.
zeroed buffer (*buffer_on_modify)();	// Call when modifying this buffer.
zeroed buffer (*buffer_maybe_break_line)();	// How to auto-fill in this buffer.
user char new_file_mode[30] = "c-mode";	// File/New on menu runs this cmd
user char new_file_ext[30] = ".c";	// File/New on menu uses this extension

buffer char *(*file_io_converter)();	// Change char set on r/w?
// Bits for func parameter above.
#define FILE_CONVERT_READ	1
#define FILE_CONVERT_WRITE	2
#define FILE_CONVERT_ASK	4
#define FILE_CONVERT_QUIET	8
char *oem_file_converter();	// A converter to r/w DOS/OEM files (Win only).
user char default_character_set = 0;

user char expand_wildcards = 0;	/* should "epsilon *.c" read all or dired? */
user char max_initial_windows = 3; // how many windows from epsilon a b c...?
char dired_sorts_files = 'n';	// how should dired sort the directory?
char dired_groups_dirs = 1;	// put all directories before files?

user char grep_keeps_files = 0;	/* when grep needs file, use temp? */
user char grep_empties_buffer = 0; /* should grep clear grep buffer first? */
char _last_was_grep;		/* which ran last: grep or proc? */
char _grep_files[FNAMELEN];	/* last file pattern we searched in */
buffer int _line_highlight;	/* region for highlighting current line */

struct disk_info {		/* returned by disk_space() */
	int sects_per_cluster, bytes_per_sector;
	int avail_clusters, tot_clusters;
};

struct time_info {		/* returned by time_and_day() */
	short year, month, day;			/* 1980-2099, 1-12, 1-31 */
	short hour, minute, second, hundredth;	/* 0-23, 0-59, 0-59, 0-99 */
	short day_of_week;	/* 0=Sunday .. 6=Saturday */
};

#define point_spot	get_spot(0)
#define mark_spot	get_spot(1)

#define to_begin_line()		(nl_reverse() ? point++ : 0)
#define to_end_line()		(nl_forward() ? point-- : 0)

/* macro for replacing a function */
/* The replacement function new_hello_start_up() can call the old
start_up() function as hello_start_up(), and is installed with
REPLACE_FUNC("hello", "start_up")
*/
#define REPLACE_FUNC(ext, func)					\
	when_loading() {					\
		if (!find_index(ext "_" func))			\
			replace_name(func, ext "_" func);	\
		else						\
			drop_name(func);			\
		replace_name("new_" ext "_" func, func);	\
	}

/* bits for full_getkey() */
#define ALTIFY_KEY	(1)	/* after alt-prefix */
#define CTRLIFY_KEY	(2)	/* after ctrl-prefix */

	/* bits for comp_read() */
#define CAUTIOUS	(0x100)	/* matches may not start with pattern */
#define COMP_FOLD	(0x200)	/* ignore case in comparison */
#define MUST_MATCH	(0x400)	/* don't allow non-matches... */
#define NONE_OK		(0x800)	/* ...but empty string is ok */
#define POP_UP_PROMPT (0x1000) /* prompt in a pop-up window */
#define COMP_FILE_BIT (0x2000)	/* completing on a file name */
#define COMP_FILE	(COMP_FILE_BIT | CAUTIOUS | SPACE_VALID)
						/* use special rules for files */
#define PASSWORD_PROMPT	(0x4000)	/* don't echo the prompt */
#define SPACE_VALID	(0x8000)	/* <Space> can be in the response. */
// Plus some codes for file_match() or finder function defined in codes.h.

user short menu_width = 35;	/* width of completion's menu of names */
user short bufed_width = 50;	/* width of bufed window */
zeroed int _nt_filter;	/* bits for name table entries to match in complete */
zeroed char *inp_get_other_name;
zeroed char show_all_variables;	// should get_var return only user vars?
zeroed char got_bad_number;	// was last numtoi() or get_number() valid?
user char record_customizations = 0;
user char load_customizations = 1;
zeroed char noinit_flag; // Was the -noinit flag used to disable customizations?

#define BIT(x)		(1 << (x))

#define FOLDFILE	(get_foldfile())
#define TERMINAL	((opsys == OS_UNIX) ? "/dev/tty" : "con")

/* Be compatible with the old names for things */
#define display_class		_display_class
#define display_characters	_display_characters
#define len_def_mac			_len_def_mac
#define get_buffer_point	get_buf_point
#define f_all_match			do_file_match
#define sort_status			show_status
#define is_win32			is_mswindows

/* Understand ANSI name for these. */
#define strchr				index
#define strrchr				rindex
#define stricmp				strfcmp

#define key_t				int

user buffer int want_debug_coloring;
#if defined(DEBUG_COLORING)
#undef set_character_color
#define set_character_color(f, t, c)	debug_set_character_color(f, t, c)
#endif

#include <search.h>
