/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2020 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

char _css_mode_name[] = "CSS";
keytable html_tab;			/* key table for html mode */
int color_css_range(from, to);
int color_html_range(from, to);

#define HTML_NAME_CHAR	"word|dot|hyphen|:"
#define HTML_NAME		"<" HTML_NAME_CHAR ">+"
#define HTML_VALUE		"('[^']*'|\"[^\"]*\"|[^\"' <>\t\n]+)"
#define HTML_EQUALS		"[ \t\n]*=[ \t\n]*"
#define HTML_ATTRIBUTE	HTML_NAME "(" HTML_EQUALS HTML_VALUE ")?"
#define XML_SORT_MARKER	 "-XML-SORT-MARKER-"
#define XML_SORT_NL_MARKER	"-XML-SORT-NL-MARKER-"
#define XML_SORT_WSPACE	"(<wspace>|-XML-SORT-NL-MARKER-)"

user char html_auto_show_delim_chars[20] = "<>";
user char html_auto_indent = 255;
user char xml_auto_indent = 255;
user int html_indenting_rules = 0xffffff;
user buffer int html_style_rules;
user buffer html_recognize_coldfusion_comments = 1;
user int color_html_look_back = 500000;
user int html_tag_match_look_back = 10_000_000;
user char auto_show_css_delimiters = 1;
user char css_auto_show_delim_chars[20] = "{[()]}";
user buffer char css_indent = 8;
user char compile_html_cmd[FNAMELEN];
user char compile_xml_cmd[FNAMELEN];
user char compile_php_cmd[FNAMELEN];

// Codes for embedded languages (or 0 for plain coloring).
#define HT_JSCRIPT	1
#define HT_VBSCRIPT	2
#define HT_PHP		3
#define HT_PYTHON	4
#define HT_CSS		5
#define HT_IGNORE	10	// Not the start of an embedded block.

user char html_asp_coloring = HT_JSCRIPT;		// <% %> sequences
user char html_php_coloring = HT_PHP;		// <? ?> sequences
user char xml_asp_coloring = HT_IGNORE;		// <% %> sequences in xml
user char xml_php_coloring = HT_IGNORE;		// <? ?> sequences in xml
user char html_other_coloring = HT_JSCRIPT;	// <Script> for unknown language.

user int html_prevent_coloring = 0;
#define HTML_SPECIFIC_COLORING(bit, var)	\
	((html_prevent_coloring & (bit)) ? HT_IGNORE : (var))

#define SUBLANG_XML	4
#define SUBLANG_CSS	5

user char php_comment_style = 1;

user char php_closeback = 1;		/* select default PHP indenting style */
user char php_topindent = 1;
user buffer short php_indent = 0;
user short php_label_indent = 0;	/* indent labels to this column */
user short php_contin_offset = 0;	/* alter normal indentation of contin lines */
user char php_align_contin_lines = 48; // line up contin lines up to this col
user short php_brace_offset = 0;	// move braces this much from usual position
user int php_top_braces = 0;	// indent for outer {}'s in func
user int php_top_contin = 3;	// indent for contin lines (decls) outside func
user int php_top_struct = 8;	// indent for body of outer struct/class

user char html_auto_fill_mode = 31;	// 1=auto-fill toggle, 2=text, 4=tag, 
user char xml_auto_fill_mode = 25;	// ... 8=html comments, 16=scripting comments
user int html_auto_fill_combine = 10; // Don't break after short element names.
user int xml_auto_fill_combine = 10;
user char html_display_definition = 15;	// 1=in func, 2=outside, 4=xml, 8=html
user int html_display_nesting_width = 40;
user buffer int html_indent = 3;
user buffer int xml_indent = 3;
user char html_no_indent_elements[500] = "|html|body|";
user char html_empty_elements[500]
   = "|hr|isindex|col|img|basefont|br|param|area|base|link|meta|frame|input|";
user char coldfusion_empty_elements[1000]
   = "|abort|applet|application|argument|associate|break|cache|chartdata|col"
	 "|collection|content|cookie|directory|dump|else|elseif|error"
	 "|exit|file|flush|ftp|gridcolumn|gridrow|gridupdate|header|htmlhead"
	 "|httpparam|import|include|index|input|insert|invokeargument|ldap|location"
	 "|log|loginuser|logout|mailparam|module|object|objectcache|param|pop"
	 "|procparam|procresult|property|queryparam|registry|rethrow|return|schedule"
	 "|search|set|setting|slider|textinput|throw|treeitem|update|wddx|";
user buffer char html_paragraph_is_container = 1;

user char html_paragraph_elements[1024]
   = "p|html|head|body|frameset|style|title|address|blockquote|center|del|div|h[1-6]"
	 "|hr|ins|noscript|pre|dir|dl|dt|dd|li|menu|ol|ul|table|caption|colgroup|col|"
	 "thead|tfoot|tbody|tr|td|th|form|button|fieldset|legend|input|label|select|optgroup"
	 "|option|textarea|frameset|frame|noframes";
user char xml_paragraph_elements[1024] = "<alphanum|.|_|hyphen>+";	// Any tag breaks paragraphs.

user char html_reindent_previous_line = 1;
user char xml_reindent_previous_line = 1;

buffer spot show_mismatch_region[2];

color_class html_text;
color_class html_tag;
color_class html_attribute;
color_class html_value;
color_class html_comment;
color_class html_altcomment;
color_class html_namespace;
color_class html_entity;
color_class html_punctuation;
color_class html_embedded;
color_class html_mismatch;

color_class css_string;
color_class css_comment;
color_class css_selector;
color_class css_class;
