/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

/* Searching flags for do_searching() and other search funcs. */
/* Some of these are defined in codes.h, with SEARCH_ names. */
#define FOLD		SEARCH_FOLD	/* case folding on? */
#define MODFOLD		2	/* use above bit? (if not, use case_fold buf var) */
#define REGEX		SEARCH_REGEX	/* regular expression mode? */
#define WORD		SEARCH_WORD	/* match only words? */
#define FAILING		16	/* current pattern not found? */
#define BAD			32	/* (if regex) current pattern illegal? */
#define QUERY		64	/* (if replace) ask at each match? */
#define REVERSE		SEARCH_REVERSE	/* look backwards for matches? */
#define INCR		256	/* search after each char? */
#define SONCE		512	/* search just once now */
#define UNKNOWN		1024	/* has search type been specified? */
#define REPLACE		2048	/* in replace, not just search? */
#define FILEGREP	4096	/* doing multi-file search/replace? */
#define BUFGREP		8192	/* doing multi-buffer search/replace? */
#define SEARCH_WRAPPED	16384	/* search hit end of buffer; try from other end */
#define SEARCH_OVERWRAPPED	32768	/* repeating prior hits due to wrapping */
#define SEARCH_EMPTY		65536	/* user entered no pattern */
#define SEARCH_CONTIN		131072	/* inter-record searching */
#define CASE_REPLACE		262144	/* smart case-based replace? */
#define SEARCH_IGNORE_COLOR	524288	/* ignore <c: > in regex? */
#define SEARCH_REGEX_HIDDEN	0x100000	// Regex behind the scenes, for nulls.
#define GREP		(FILEGREP | BUFGREP)	/* multi-buffer/file search? */
#define SMODES		(MODFOLD | FOLD | WORD | REGEX)

// Codes for a search_continuation function.
#define SCON_RECORD		-1		/* remember this location */
#define SCON_RESTORE	-2		/* return to remembered location */
#define SCON_COMPARE	-3		/* is this location past remembered one? */
								/* 1: yes, -1: no, 0: equal */

/* return codes from do_searching(), or 1=found, 0=not found */
#define DSABORT		2	/* user aborted search */
#define DSBAD		3	/* illegal regular expression */

	/* s_code: where to start searching from, for i-search */
#define FHERE	1	/* from right here */
#define FORIG	2	/* from starting position */
#define FMATCH	3	/* from last match */
#define FMADD	4	/* same, but indicates we added a char this time */
#define FMREGEX	5	/* same, but don't assume too much about pattern */

typedef struct { // An old position, for searching through multiple items.
	int pos;	// We use just this normally.
	char item[FNAMELEN];	// For use in multi-item searches.
} old_place;

char *_default_search, *_default_regex_search;
char *_default_replace, *_default_regex_replace;
int _default_flags = FOLD;	/* 2 sets of defaults */
int _default_regex_flags = FOLD;
char _last_was_regex;		/* was last type of search a regex? */
user key_t fwd_search_key = -1;	/* accept these inside search too */
user key_t rev_search_key = -1;
user char regex_first_end = 0;	// find first match to end, or first to start?
user char regex_shortest = 0;	// find shortest possible match, or longest?
user char search_wraps = 1;	// jump to other end of buffer on failure?
user char search_positions_at_start; // string searching positions to start of match?
user char search_in_region;		// search within if there's a highlighted region
user char replace_in_region;	// replace within if there's a highlighted region
user char search_defaults_from = 1; // 1: highlighted region, 2: last response
user int search_delete_match = 1;	// Pressing Del deletes match
user char replace_by_case = 1;	// 2: permit fancy case-based replace
int replace_num_found;		// in last replace, # of matches found
int replace_num_changed;	// in last replace, # of matches changed
buffer int (*search_continuation)();
char *search_other_name;	// enable alt-g in search prompt

struct restrict_search {
	spot start, end;		// Limit searches to this region.
	int colstart, colend;	// Columns to include, if rectangular region.
} restrict_search;
zeroed char restrict_search_type;	// Type of search-limiting region, or 0 if none.

