/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2006 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "asm.h"

color_asm_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	while (from > 0 && get_tagged_region("color-as-unit", from, &s) != -1
		   && s <= from)
		from = s - 1;
	c_init_color(from, to);
	point = from;
	to_begin_line();
	if (talk = (to - from > 50000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_.@#%?$]+|[\"';]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case ';':			// found comment
				nl_forward();
				set_character_color(t, point, color_class c_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class c_string);
				if (get_character_color(point, (int *) 0, &s) == 
					color_class c_string && s > to)  // fix up after
					c_init_color(point, to = s); // quoted "'s
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class c_charconst);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, asm_keyword_color(t));
				if (asm_title_line()) {
					t = point;
					point++;
					nl_forward();
					set_character_color(t, point, color_class c_string);
				}
				asm_comment_line();
				break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring ASM program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

// If we found a comment keyword, use the next non-space character as a
// delimiter, and find its match.
asm_comment_line()
{
	char pat[10];
	int p, s;

	if (!parse_string(-1, "</word>comment"))
		return;
	point += parse_string(1, "[ \t]*");
	if (!curchar() || isspace(curchar()))
		return;
	sprintf(pat, "%c", curchar());
	s = give_begin_line();
	p = ++point;
	search(1, pat);
	nl_forward();
	set_tagged_region("color-as-unit", s, point, 1);
	set_character_color(p, point, color_class c_comment);
}

// We're after an identifier.  If it's title or similar, return nonzero
// so we'll color the rest of the line as a string.  But ignore such
// keywords if we're inside a macro.

int asm_title_line()
{
	if (!parse_string(-1, "</word>(error|title|name|subttl)"))
		return 0;
	save_var point;
	while (re_search(-1, "</word>(macro|rept|endm)</word>")) {
		if (get_character_color(point) == color_class asm_preprocessor)
			return tolower(character(matchstart)) == 'e';
	}
	return 1;
}

// Return color for "identifier" from here to point
// (something with alpha or digits).
asm_keyword_color(from)
{
	char buf[500];

	if (point - from > sizeof(buf) / sizeof(char) - 10)
		save_var point = from + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	strcpy(buf + point - from + 1, "|");
	if (is_asm_preproc(buf))
		return color_class asm_preprocessor;
	if (is_asm_keyword(buf))
		return color_class asm_keyword;
	if (is_asm_register(buf))
		return color_class asm_register;
	// Check for a number last, since some keywords like .386 look like
	// numbers.
	if (index("0123456789-", buf[1]) || buf[1] == '.' && isdigit(buf[2]))
		return c_number_color(buf + 1);
	return color_class asm_identifier;
}


is_asm_preproc(p)	// is text in p (must be surrounded by |'s) a preproc kwd?
char *p;
{
	if (strstr(
			   "|%out|.186|.286|.286c|.286p|.287|.386|.386c|.386p|.387|.486|.486p|.8086"
			   "|.8087|.alpha|.break|.code|.const|.continue|.cref|.data|.data?|.dosseg"
			   "|.else|.elseif|.endif|.endw|.err|.err1|.err2|.errb|.errdef|.errdif|.errdifi"
			   "|.erre|.erridn|.erridni|.errnb|.errndef|.errnz|.exit|.fardata|.fardata?"
			   "|.if|.lall|.lfcond|.list|.listall|.listif|.listmacro|.listmacroall|.model"
			   "|.msfloat|.no87|.nocref|.nolist|.nolistif|.nolistmacro|.radix|.repeat|.sall"
			   "|.seq|.sfcond|.stack|.startup|.tfcond|.type|.until|.untilcxz|.while|.xall"
			   "|.xcref|.xlist|alias|align|assume|catstr|comm|comment|db|dd|df|dosseg|dq"
			   "|dt|dup|dw|echo|else|elseif|elseif1|elseif2|elseifb|elseifdef|elseifdif"
			   "|elseifdifi|elseife|elseifidn|elseifidni|elseifnb|elseifndef|end|endif"
			   "|endm|endp|ends|eq|equ|even|exitm|extern|externdef|extrn|for|forc|ge|goto"
			   "|group|gt|high|highword|if|if1|if2|ifb|ifdef|ifdif|ifdifi|ife|ifidn|ifidni"
			   "|ifnb|ifndef|include|includelib|instr|invoke|irp|irpc|label|le|length|lengthof"
			   "|local|low|lowword|lroffset|lt|macro|mask|mod|name|ne|offset|opattr|option"
			   "|org|page|popcontext|proc|proto|ptr|public|purge|pushcontext|record|repeat"
			   "|rept|seg|segment|short|size|sizeof|sizestr|struc|struct|substr|subtitle"
			   "|subttl|textequ|this|title|type|typedef|union|while|width|.586|.586p|.686"
			   "|.686p|.k3d|.mmx|.xmm|%depth|%incl|%linum|%noincl|%nomacs|%notrunc|.386r"
			   "|.common|.prot|.twocase|.xlisti|ideal|locals|model|$|?|@b|@f|abs|addr|all"
			   "|assumes|at|basic|byte|c|carry?|casemap|common|compact|cpu|dotname|dword"
			   "|emulator|epilogue|error|export|expr16|expr32|far|far16|far32|farstack"
			   "|flat|forceframe|fortran|fword|huge|language|large|listing|ljmp|loadds"
			   "|m510|medium|memory|near|near16|near32|nearstack|nodotname|noemulator|nokeyword"
			   "|noljmp|nom510|none|nonunique|nooldmacros|nooldstructs|noreadonly|noscoped"
			   "|nosignextend|nothing|notpublic|oldmacros|oldstructs|os_dos|overflow?|para"
			   "|parity?|pascal|private|prologue|qword|radix|readonly|real10|real4|real8"
			   "|req|sbyte|scoped|sdword|setif2|sign?|small|stack|stdcall|sword|syscall"
			   "|tbyte|tiny|use16|use32|uses|vararg|word|zero?|@@:|@CatStr|@code|@CodeSize"
			   "|@Cpu|@CurSeg|@data|@DataSize|@Date|@Environ|@fardata|@fardata?|@FileCur"
			   "|@FileName|@InStr|@Interface|@Line|@Model|@SizeStr|@stack|@SubStr|@Time"
			   "|@Version|@WordSize|err|masm51|nowarn|pword|quirks|show|warn|"
			   , p))
			return 1;
	return 0;
}

is_asm_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr(
			   "|aaa|aad|aam|aas|adc|add|and|arpl|bound|bsf|bsr|bswap|bt|btc|btr|bts|call"
			   "|cbw|cdq|clc|cld|cli|clts|cmc|cmp|cmps|cmpsb|cmpsd|cmpsw|cmpxchg|cwd|cwde"
			   "|daa|das|dec|div|enter|esc|f2xm1|fabs|fadd|faddp|fbld|fbstp|fchs|fclex"
			   "|fcom|fcomp|fcompp|fcos|fdecstp|fdisi|fdiv|fdivp|fdivr|fdivrp|feni|ffree"
			   "|fiadd|ficom|ficomp|fidiv|fidivr|fild|fimul|fincstp|finit|fist|fistp|fisub"
			   "|fisubr|fld|fld1|fldcw|fldenv|fldenvd|fldenvw|fldl2e|fldl2t|fldlg2|fldln2"
			   "|fldpi|fldz|fmul|fmulp|fnclex|fndisi|fneni|fninit|fnop|fnsave|fnsaved|fnsavew"
			   "|fnstcw|fnstenv|fnstenvd|fnstenvw|fnstsw|fpatan|fprem|fprem1|fptan|frndint"
			   "|frstor|frstord|frstorw|fsave|fsaved|fsavew|fscale|fsetpm|fsin|fsincos"
			   "|fsqrt|fst|fstcw|fstenv|fstenvd|fstenvw|fstp|fstsw|fsub|fsubp|fsubr|fsubrp"
			   "|ftst|fucom|fucomp|fucompp|fwait|fxam|fxch|fxtract|fyl2x|fyl2xp1|hlt|idiv"
			   "|imul|in|inc|ins|insb|insd|insw|int|into|invd|invlpg|iret|iretd|iretdf"
			   "|iretf|ja|jae|jb|jbe|jc|jcxz|je|jecxz|jg|jge|jl|jle|jmp|jna|jnae|jnb|jnbe"
			   "|jnc|jne|jng|jnge|jnl|jnle|jno|jnp|jns|jnz|jo|jp|jpe|jpo|js|jz|lahf|lar"
			   "|lds|lea|leave|les|lfs|lgdt|lgs|lidt|lldt|lmsw|lock|lods|lodsb|lodsd|lodsw"
			   "|loop|loopd|loope|looped|loopew|loopne|loopned|loopnew|loopnz|loopnzd|loopnzw"
			   "|loopw|loopz|loopzd|loopzw|lsl|lss|ltr|mov|movs|movsb|movsd|movsw|movsx"
			   "|movzx|mul|neg|nop|not|or|out|outs|outsb|outsd|outsw|pop|popa|popad|popf"
			   "|popfd|push|pusha|pushad|pushd|pushf|pushfd|pushw|rcl|rcr|rep|repe|repne"
			   "|repnz|repz|ret|retf|retn|rol|ror|sahf|sal|sar|sbb|scas|scasb|scasd|scasw"
			   "|seta|setae|setb|setbe|setc|sete|setg|setge|setl|setle|setna|setnae|setnb"
			   "|setnbe|setnc|setne|setng|setnge|setnl|setnle|setno|setnp|setns|setnz|seto"
			   "|setp|setpe|setpo|sets|setz|sgdt|shl|shld|shr|shrd|sidt|sldt|smsw|stc|std"
			   "|sti|stos|stosb|stosd|stosw|str|sub|test|verr|verw|wait|wbinvd|xadd|xchg"
			   "|xlat|xlatb|xor|addpd|addps|addsd|addss|addsubpd|addsubps|andnpd|andnps"
			   "|andpd|andps|clflush|cmova|cmovae|cmovb|cmovbe|cmovc|cmove|cmovg|cmovge"
			   "|cmovl|cmovle|cmovna|cmovnae|cmovnb|cmovnbe|cmovnc|cmovne|cmovng|cmovnge"
			   "|cmovnl|cmovnle|cmovno|cmovnp|cmovns|cmovnz|cmovo|cmovp|cmovpe|cmovpo|cmovs"
			   "|cmovz|cmppd|cmpps|cmpss|cmpxchg8b|comisd|comiss|cpuid|cvtdq2pd|cvtdq2ps"
			   "|cvtpd2dq|cvtpd2pi|cvtpd2ps|cvtpi2pd|cvtpi2ps|cvtps2dq|cvtps2pd|cvtps2pi"
			   "|cvtsd2si|cvtsd2ss|cvtsi2sd|cvtsi2ss|cvtss2sd|cvtss2si|cvttpd2dq|cvttpd2pi"
			   "|cvttps2dq|cvttps2pi|cvttsd2si|cvttss2si|divpd|divps|divsd|divss|emms|fcmovb"
			   "|fcmovbe|fcmove|fcmovnb|fcmovnbe|fcmovne|fcmovnu|fcmovu|fcomi|fcomip|fisttp"
			   "|fxrstor|fxsave|haddpd|haddps|hsubpd|hsubps|lddqu|ldmxcsr|lfence|loopde"
			   "|loopdne|loopwe|loopwne|maskmovdqu|maskmovq|maxpd|maxps|maxsd|maxss|mfence"
			   "|minpd|minps|minsd|minss|monitor|movapd|movaps|movd|movddup|movdq2q|movdqa"
			   "|movdqu|movhlps|movhpd|movhps|movlhps|movlpd|movlps|movmskpd|movmskps|movntdq"
			   "|movnti|movntpd|movntps|movntq|movq|movq2dq|movshdup|movsldup|movss|movupd"
			   "|movups|mulpd|mulps|mulsd|mulss|mwait|orpd|orps|packssdw|packsswb|packuswb"
			   "|paddb|paddd|paddq|paddsb|paddsw|paddusb|paddusw|paddw|pand|pandn|pause"
			   "|pavgb|pavgw|pcmpeqb|pcmpeqd|pcmpeqw|pcmpgtb|pcmpgtd|pcmpgtw|pextrw|pinsrw"
			   "|pmaddwd|pmaxsw|pmaxub|pminsw|pminub|pmovmskb|pmulhuw|pmulhw|pmullw|pmuludq"
			   "|por|prefetchh|psadbw|pshufd|pshufhw|pshuflw|pshufw|pslld|pslldq|psllq"
			   "|psllw|psrad|psraw|psrld|psrldq|psrlq|psrlw|psubb|psubd|psubq|psubsb|psubsw"
			   "|psubusb|psubusw|psubw|punpckhbw|punpckhdq|punpckhqdq|punpckhwd|punpcklbw"
			   "|punpckldq|punpcklqdq|punpcklwd|pxor|rcpps|rcpss|rdmsr|rdpmc|rdtsc|rsm"
			   "|rsqrtps|rsqrtss|sfence|shufpd|shufps|sqrtpd|sqrtps|sqrtsd|sqrtss|stmxcsr"
			   "|subpd|subps|subsd|subss|sysenter|sysexit|ucomisd|ucomiss|ud2|unpckhpd"
			   "|unpckhps|unpcklpd|unpcklps|wrmsr|xorpd|xorps|lgdte|lidte|sidte|"
			   , p))
			return 1;
	return 0;
}

is_asm_register(p)	// is text in p (must be surrounded by |'s) a reg name?
char *p;
{
	if (strstr(
			   "|ah|al|ax|bh|bl|bp|bx|ch|cl|cr0|cr2|cr3|cs|cx|dh|di|dl|dr0|dr1|dr2|dr3"
			   "|dr6|dr7|ds|dx|eax|ebp|ebx|ecx|edi|edx|es|esi|esp|fs|gs|si|sp|ss|st|tr3"
			   "|tr4|tr5|tr6|tr7|cr4|mm|mm0|mm1|mm2|mm3|mm4|mm5|mm6|mm7|st0|st1|st2|st3"
			   "|st4|st5|st6|st7|xmm|xmm0|xmm1|xmm2|xmm3|xmm4|xmm5|xmm6|xmm7|"
			   , p))
		return 1;
	return 0;
}

command asm_mode()
{
	mode_default_settings();
	mode_keys = asm_tab;		/* Use these keys. */
	major_mode = _asm_mode_name;
	compile_buffer_cmd = compile_asm_cmd;	// can compile this
	strcpy(comment_start, ";[ \t]*");
	strcpy(comment_pattern, ";.*$");
	strcpy(comment_begin, "; ");
	strcpy(comment_end, "");
	recolor_range = color_asm_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("asm-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

suffix_asm()
{
	asm_mode();
}

suffix_mac()
{
	asm_mode();
}

suffix_asi()
{
	asm_mode();
}

// We also handle .inc unless gams-files is nonzero.  See gams.e.

suffix_ah()
{
	asm_mode();
}
