/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*		Copyright (C) 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "diff.h"
#include "colcode.h"
#include "dired.h"

command compare_files()
{
	char pat1[FNAMELEN], pat2[FNAMELEN], def[FNAMELEN], dest[FNAMELEN];
	int ask = has_arg;

	iter = 1;
	strcpy(dest, COMPFILE_DIR);
	save_var _help_on_cmd = last_index;
	get_last_recall(f_match, def);
	if (url_is_a_directory(def))
		add_final_slash(def);
	get_file(pat1, "Compare file pattern/directory", def);
	if (!*pat1)
		getcd(pat1);
	save_var recall_id = "compare-files-2nd";
	get_last_recall(f_match, def);
	if (url_is_a_directory(def))
		add_final_slash(def);
	if (url_is_a_directory(pat1))
		get_file(pat2, "With directory", def);
	else
		get_file(pat2, "With file pattern/directory", def);
	recall_id = 0;
	if (!*pat2)
		getcd(pat2);
	absolute(pat1);
	absolute(pat2);
	if (!strcmp(pat1, pat2))
		error("Specify two different patterns/directories, not %s twice.",
			  pat1);
	if (ask) {
		if (!is_remote_file(pat1) && !is_remote_file(pat2))
			switch (ask_3way("Choose comparison method",
							 "Compare text after translation and decoding (default),\n"
							 "compare raw bytes (faster),\nor compare only sizes & times (fastest)?",
							 "Compare Text (default), Bytes (faster), Sizes & times (fastest) [TBS]",
							 "Text", "Bytes", "Sizes", 1)) {
				case 1: save_var compare_files_method = 2; break;
				case 2: save_var compare_files_method = 1; break;
				case 3: save_var compare_files_method = 0; break;
				default: return;
			}
		if (compare_files_method != 3)
			switch (ask_3way("Choose whether to ignore spaces",
							 "Make all runs of spaces and tabs match,\n"
							 "make all runs of spaces, tabs, and newlines match,\n"
							 "or require text to match exactly?",
							 "Runs of Spaces & tabs match, Newlines too, Exact matches only [SNE]",
							 "Spaces", "Newlines", "Exact", 2)) {
				case 1: save_var compare_files_ignore_spaces = 2; break;
				case 2: save_var compare_files_ignore_spaces = 1; break;
				case 3: save_var compare_files_ignore_spaces = 0; break;
				default: return;
			}
		get_buf(dest, "Put result in buffer",
				(mode_keys == comp_files_tab) ? bufname : COMPFILE_DIR);
	}
	noteput("Working...");
	do_compare_files(pat1, pat2, dest);
	drop_pending_says();
	compare_files_mode();
	discardable_buffer = 1;
	say("");
}

// Make a sorted list of matching files in a temp buffer and return it,
// leaving point at its start. If hier, match the hierarchy under the
// directory using the pattern and put the names relative to that dir.
int make_file_list(char *p, char *hier)
{
	char pat[FNAMELEN], *s, dir[FNAMELEN], sep = path_sep;
	int t, cnt = 0;

	strcpy(dir, p);
	strcpy(pat, p);
	if (hier) {
		if (is_remote_file(pat)) {
			sep = '/';
			if (!strncmp(hier, "**", 2))
				hier = "*";
		}
		sprintf(pat + strlen(pat), "%c%s", sep, hier);
	}
	s = do_file_match(pat, STARTMATCH | EXACTONLY | FM_NO_DIRS);
	if (!s)
		return 0;
	p = strstr(dir, "**");
	if (!hier && p) {
		*p = 0;
		hier = dir;
	}
	t = tmp_buf();
	do {
		if (hier) {
			char rel[FNAMELEN];
			relative(s, rel, dir);
			buf_printf(t, "%s\n", rel);
		} else
			buf_printf(t, "%s\n", get_tail(s, 0));
		if (++cnt % 100 == 0)
			noteput("Found %d files in %s...", cnt, dir);
	} while (s = do_file_match(pat, EXACTONLY | FM_NO_DIRS));
	save_var bufnum = t;
	case_fold = (FOLDFILE != 0);
	do_sort_region(0, size(), 0, 0);
	point = 0;
	return t;
}

// Copy src to dest, but if it contains **, delete it and following.
get_dir_part_of_pattern(char *dest, char *src)
{
	strcpy(dest, src);
	char *p = strstr(dest, "**");
	if (p)
		*p = 0;
}

// Find all matches of the pattern pat1. (If it's a directory, search
// the hierarchy.) Likewise for pat2, except if it's a directory and
// pat1 had an explicit ** pattern, use that. Then compare the two lists
// of files. Put the result in buffer dest.

do_compare_files(char *pat1, char *pat2, char *dest)
{
	int is_dir1, is_dir2, t1, t2, res;
	char *o1, *o2, *both, *pat, dir1[FNAMELEN], dir2[FNAMELEN];

	get_dir_part_of_pattern(dir1, pat1);
	get_dir_part_of_pattern(dir2, pat2);
	is_dir1 = url_is_a_directory(pat1);
	is_dir2 = url_is_a_directory(pat2);
	noteput("Listing files...");
	t1 = make_file_list(pat1, is_dir1 ? "**" : 0);
	if (!t1)
		error("No matches for %s", pat1);
	on_exit buf_delete(t1);
	pat = is_dir2 ? "**" : 0;
	if (is_dir2 && strstr(pat1, "**"))
		pat = strstr(pat1, "**");
	t2 = make_file_list(pat2, pat);
	if (!t2)
		error("No matches for %s", pat2);
	on_exit buf_delete(t2);
	o1 = temp_buf();
	o2 = temp_buf();
	both = temp_buf();
	on_exit delete_buffer(o1);
	on_exit delete_buffer(o2);
	on_exit delete_buffer(both);
	noteput("Comparing...");
	do_compare_sorted(t1, t2, o1, o2, both);
	res = zap(dest);
	buf_printf(res, "Compare: %s\n", pat1);
	buf_printf(res, "   With: %s\n", pat2);
	noteput("Building difference list...");
	build_difference_list(name_to_bufnum(both), dir1, dir2, res);
	if (is_dir2 || !is_dir1)
		make_only_in(o1, pat1, res);
	// When we compare a pattern like *.c to a dir, don't list other files
	// in the directory.
	if (is_dir1 || !is_dir2)
		make_only_in(o2, pat2, res);
	to_buffer_num(res);
	point = 0;
	nl_forward();		// Exclude the header lines from sorting.
	nl_forward();
	noteput("Finishing up...");
	do_sort_region(point, size(), 0, 0);
	point = 0;
}

// The buffer b contains a list of files only in directory dir (which
// might contain a pattern after it). Write a message for each to the
// res buffer.

make_only_in(char *b, char *dir, int res)
{
	char f[FNAMELEN], pretty[FNAMELEN];

	strcpy(pretty, dir);
	if (is_pattern(pretty))
		*get_tail(pretty, 0) = 0;
	drop_final_slash(pretty);
	save_var bufname = b;
	point = 0;
	while (point < size()) {
		grab(point, give_end_line(), f);
		buf_printf(res, "Only in %s: %s\n", pretty, f);
		nl_forward();
	}
}

// Copy this directory name or file pattern, maybe altering it so it
// can be combined with a basename by copying the basename to the
// returned pointer.
char *make_dir_part(char *to, char *from)
{
	strcpy(to, from);
	if (url_is_a_directory(to)) {
		add_final_slash(to);
		return to + strlen(to);
	} else if (strstr(to, "**"))
		return strstr(to, "**");
	else
		return get_tail(to, 1);
}

// Write to the buffer res info on the similarity of the files listed
// in the buffer b.  Each appears in both the directories named.
build_difference_list(int b, char *dir1, char *dir2, int res)
{
	char file[FNAMELEN], f1[FNAMELEN], f2[FNAMELEN], *p1, *p2;
	struct file_info i1, i2;
	int cnt = 0, total = count_lines_in_buf(b, 1);

	save_var readonly_warning |= ROWARN_GREP, no_popup_errors = 1;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	save_var file_pattern_wildcards = 0;
	p1 = make_dir_part(f1, dir1);
	p2 = make_dir_part(f2, dir2);
	set_buf_point(b, 0);
	while (grab_line(b, file)) {
		strcpy(p1, file);
		strcpy(p2, file);
		if (is_remote_file(f1) || is_remote_file(f2))
			save_var compare_files_method = 2;		// Must always compare contents.
		if (!is_remote_file(f1) && check_file(f1, &i1) != CHECK_FILE)
			buf_printf(res, "Not a file: %s\n", f1);
		else if (!is_remote_file(f2) && check_file(f2, &i2) != CHECK_FILE)
			buf_printf(res, "Not a file: %s\n", f2);
		else
			switch (files_match(f1, &i1, f2, &i2)) {
				case 2:
					buf_printf(res, "Identical (except whitespace): %s\n", file);
					break;
				case 1:
					buf_printf(res, "Identical: %s\n", file);
					break;
				case 0:
					buf_printf(res, "Different: %s\n", file);
					break;
			}
		if (++cnt % 10 == 0 || is_remote_file(f1) || is_remote_file(f2)) {
			noteput("Comparing files: %d%%...", muldiv(cnt, 100, total));
			check_abort();
		}
	}
}

// Return 1 if these files match according to compare-files rules., 2 if
// they match except for whitespace, or 0 if they don't match at all.
int files_match(char *f1, struct file_info *i1,
				char *f2, struct file_info *i2)
{
	int t1, t2, transl = FILETYPE_AUTO, orig = bufnum;

	if (!compare_files_method)
		return !compare_dates(i1, i2);
	if (compare_files_method == 1) {
		transl = FILETYPE_BINARY;
		if (i1->fsize != i2->fsize)
			return 0;
	}
	save_var abort_file_io = ABORT_JUMP;
	t1 = tmp_buf();
	t2 = tmp_buf();
	on_exit buf_delete(t1);
	on_exit buf_delete(t2);
	save_var bufnum = t1;
	quiet_file_read(f1, transl);
	save_var narrow_start = narrow_end = 0;
	bufnum = t2;
	quiet_file_read(f2, transl);
	save_var narrow_start = narrow_end = 0;
	bufnum = orig;
	if (compare_files_ignore_spaces) {
		char *pat = (compare_files_ignore_spaces == 1) ? "[ \t]+" : "[ \t\n]+";
		if (buffers_match_ignoring(t1, t2, pat))
			return buffers_identical(t1, t2) ? 1 : 2;
		else
			return 0;
	}
	return buffers_identical(t1, t2);
}

// Compare buffers a and b and return 1 if they match, 0 if they don't.
// But where they differ, if both match pat, ignore the runs that
// match.
int buffers_match_ignoring(int bufa, int bufb, char *pat)
{
	int matcha, matchb;
	int same;

	save_var bufnum;
	bufnum = bufa;
	int posa = narrow_start, enda = size() - narrow_end;
	bufnum = bufb;
	int posb = narrow_start, endb = size() - narrow_end;
	for (;;) {
		same = compare_buffer_text(bufa, posa, bufb, posb, 0);
		posa += same;
		posb += same;
		if (posa >= enda || posb >= endb)
			break;
		bufnum = bufa;
		point = posa;
		matcha = parse_string(1, pat);
		bufnum = bufb;
		point = posb;
		matchb = parse_string(1, pat);
		if (!matcha || !matchb)
			return 0;
		posa += matcha;
		posb += matchb;
	}
	return posa == enda && posb == endb;
}


color_compare_files_range(from, to) // recolor just this section
{			// last colored region may go past to
	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = from;
	to_begin_line();
	set_character_color(point, to, -1);
	while (point < to) {
		if (parse_string(1, "(Compare|   With): "))
			set_character_color(matchend, give_end_line(),
								color_class dired_header);
		else if (parse_string(1, "Different: "))
			set_character_color(matchend, give_end_line(),
								color_class dired_copied);
		else if (parse_string(1, "Only in .*: "))
			set_character_color(matchend, give_end_line(),
								color_class dired_deleted);
		if (!nl_forward())
			break;
	}
	return point;
}

// Get the pair of file names that correspond to this line, if any, and
// return 4 if they were recorded as different, 3 if identical, 2 if
// only file B exists, 1 if only file A exists, or 0 if this line
// doesn't mention a file. If common is nonzero, write the common part
// of their names there.

int compare_files_get_info(char *file1, char *file2)
{
	char t1[FNAMELEN], t2[FNAMELEN], f[FNAMELEN], dir[FNAMELEN], *s;
	int res, len;

	*file1 = *file2 = 0;
	if (!compare_files_get_patterns(t1, t2))
		return 0;
	to_begin_line();
	if (parse_string(1, "Identical( %(except whitespace%))?: "))
		res = 3;
	else if (parse_string(1, "Different: "))
		res = 4;
	else if (parse_string(1, "Only in (.*): (.*)")) {
		grab(find_group(1, 1), find_group(1, 0), dir);
		grab(find_group(2, 1), find_group(2, 0), f);
		s = make_dir_part(file1, dir);
		*s = 0;
		len = strlen(file1);
		if (len > 1 && is_path_separator(file1[len - 1]))
			len--;
		if (!strncmp(t1, file1, len))
			res = 1;
		else if (!strncmp(t2, file1, len))
			res = 2;
		else
			error("Internal error: Can't choose between %s and %s for %s", t1, t2, file1);
		strcpy(make_dir_part(file1, t1), f);
		strcpy(make_dir_part(file2, t2), f);
		return res;
	} else 
		return 0;
	grab(matchend, give_end_line(), f);
	strcpy(make_dir_part(file1, t1), f);
	strcpy(make_dir_part(file2, t2), f);
	return res;
}

when_loading()		/* disable self-inserting chars */
{
	fix_key_table(reg_tab, (short) normal_character, comp_files_tab, -1);
	set_case_indirect(comp_files_tab);
}

compare_files_help() on comp_files_tab['h'], comp_files_tab['?']
{
	help_on_command("compare-files-mode");
}

compare_files_view(int mode)
{
	char file1[FNAMELEN], file2[FNAMELEN], *s;

	switch (compare_files_get_info(file1, file2)) {
		case 0:
			error("No file on this line.");
		case 1:
		case 2:
			s = (mode == 0) ? file1 : file2;
			if (!check_file(s))
 				error("File %s does not exist.", s);
			break;
		default:
			if (mode >= 2) {
				window_one();
				std_find_it(file1);
				window_split((mode == 3) ? VERTICAL : HORIZONTAL);
				std_find_it(file2);
				return;
			}
			s = (mode == 0) ? file1 : file2;
	}	
	if (!*s)
		error("No such file");
	ask_find_it(s);
}

compare_files_view_a() on comp_files_tab['a']
{
	compare_files_view(0);
}

compare_files_view_b() on comp_files_tab['b']
{
	compare_files_view(1);
}

compare_files_view_both() on comp_files_tab['c']
{
	compare_files_view(2);
}

compare_files_view_side_by_side() on comp_files_tab['s']
{
	compare_files_view(3);
}

compare_files_copy(int from_b)
{
	char file1[FNAMELEN], file2[FNAMELEN];

	switch (compare_files_get_info(file1, file2)) {
		case 0:
			error("No file on this line.");
		case 1:
			if (from_b)
				error("Can't copy in that direction, no file there.");
			break;
		case 2:
			if (!from_b)
				error("Can't copy in that direction, no file there.");
			break;
	}	
	if (from_b)
		diff_copy(file2, file1);
	else
		diff_copy(file1, file2);
}

compare_files_copy_a() on comp_files_tab['A']
{
	compare_files_copy(0);
}

compare_files_copy_b() on comp_files_tab['B']
{
	compare_files_copy(1);
}

when_loading()
{
	set_list_keys(comp_files_tab);
}

compare_files_view_diff() on comp_files_tab['d']
{
	char file1[FNAMELEN], file2[FNAMELEN];

	if (compare_files_get_info(file1, file2) < 3)
		error("Must have two files to diff");
	else if (compare_files_diff(file1, file2, "#diff#", LISTING_DIFF_STYLE)) {
		to_buffer("#diff#");
		point = 0;
		bprintf("Compare %s to %s\n", file1, file2);
		point = 0;
		discardable_buffer = 1;
	}
}

// Make a buffer with diffs for all different files.

compare_files_list() on comp_files_tab['l']
{
	char file1[FNAMELEN], file2[FNAMELEN];
	char *temp = temp_buf(), *orig = bufname;

	save_var point = 0;
	zap(COMPFILE_LIST);
	do {
		if (compare_files_get_info(file1, file2) == 4) {
			if (*file1)
				noteput("Working... %s", file1);
			buffer_printf(COMPFILE_LIST, "%p\n", -1);
			if (compare_files_diff(file1, file2, temp, LISTING_DIFF_STYLE)) {
				buffer_printf(COMPFILE_LIST, "Compare %s to %s\n", file1, file2);
				bufname = temp;
				buf_xfer_colors(name_to_bufnum(COMPFILE_LIST), 0, size());
				bufname = orig;
			} else
				buffer_printf(COMPFILE_LIST, "Couldn't compare %s and %s: "
							  "file not found\n", file1, file2);
		}
	} while (nl_forward());
	say("");
	to_buffer(COMPFILE_LIST);
	point = 0;
	discardable_buffer = 1;
}

compare_files_mode()
{
	mode_default_settings();
	mode_keys = comp_files_tab;		/* Use these keys. */
	major_mode = _comp_files_mode_name;
	recolor_range = color_compare_files_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	mouse_dbl_selects = 1;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	make_mode();
}

// Get patterns we're comparing and return 1, else 0.
int compare_files_get_patterns(char *pat1, char *pat2)
{
	save_var point = 0;
	if (!parse_string(1, "Compare: "))
		return 0;
	grab(matchend, give_end_line(), pat1);
	if (!nl_forward() || !parse_string(1, "   With: "))
		return 0;
	grab(matchend, give_end_line(), pat2);
	return 1;
}

// Generate a diff in out.  Return 0 if a file wasn't found.
compare_files_diff(char *file1, char *file2, char *out, int style)
{
	struct screen_info wins;
	int err1 = 0, err2 = 0, temp1 = 0, temp2 = 0, had_narrow1 = 0, had_narrow2 = 0;

	save_var synchronous_file_io = 1;
	save_var readonly_warning |= ROWARN_GREP, no_popup_errors = 1;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	wins.wins = 0;
	save_screen(&wins);
	on_exit if (temp1 && buf_exist(temp1)) buf_delete(temp1);
	on_exit if (temp2 && buf_exist(temp2)) buf_delete(temp2);
	on_exit restore_screen(&wins);
	one_window();
	if (!look_file(file2)) {
		err2 = std_find_it(file2);
		if (!err2)
			temp2 = bufnum;
	} else
		window_bufnum = bufnum;
	had_narrow2 = narrow_start > 0 || narrow_end > 0;
	save_var narrow_start = narrow_end = 0;
	window_split(HORIZONTAL);
	if (!look_file(file1)) {
		err1 = std_find_it(file1);
		if (!err1)
			temp1 = bufnum;
	} else
		window_bufnum = bufnum;
	had_narrow1 = narrow_start > 0 || narrow_end > 0;
	save_var narrow_start = narrow_end = 0;
	if (!err1 && !err2)
		do_diff(out, style);
	if (err1)
		file_error(err1, file1, "read error");
	else if (err2)
		file_error(err2, file2, "read error");
	else {
		save_var bufname = out;
		save_var point = 0;
		point += parse_string(1, "Comparing.*\nwith.*\n");
		if (had_narrow1)
			bprintf("%pIgnored narrowing for %s\n", -1, file1);
		if (had_narrow2)
			bprintf("%pIgnored narrowing for %s\n", -1, file2);
		diff_mode();
		return 1;
	}
	return 0;
}

compare_files_examine() on comp_files_tab['v'], comp_files_tab['\n'],
						   comp_files_tab['\r']
{
	char file1[FNAMELEN], file2[FNAMELEN], msg[FNAMELEN];

	switch (compare_files_get_info(file1, file2)) {
		case 2:
			ask_find_it(file2);
			return;
		case 1:
			ask_find_it(file1);
			return;
		case 0:
			error("No pair of files to compare.");
	}
	if (compare_files_diff(file1, file2, COMPFILE_ONE, VISUAL_DIFF_STYLE)) {
		to_buffer(COMPFILE_ONE);
		sprintf(msg, "%s vs %s ", file1, file2);
		set_mode_message(msg);
		visual_diff_mode();
		discardable_buffer = 1;
		point = 0;
	}
}
