/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"

// Check for program prog (if nonnull), then run the command line to
// uncompress stdin to stdout.
do_decompression(char *prog, char *cmdline)
{
	char fname[FNAMELEN], *s;

	if (GET_LINE_TRANSLATE(translation_type) != FILETYPE_BINARY)
		return;
	if (is_remote_file(filename) && is_retrieving_remote_file(bufnum))
		return;
	if (prog && *prog) {
		sprintf(fname, "%s%s", prog, is_unix ? "" : ".exe");
		if (!(s = getenv("PATH")) || !look_on_path(fname, 0, s))
			return;
	}
	save_var _buf_readonly = 0;
	force_save_as = FSA_READONLY;
	if (!pipe_text(bufname, bufname, cmdline, NULL,
				   PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0))
		error("Could not uncompress %s", filename);
	point = modified = 0;
	strcpy(fname, filename);
	if (s = get_extension(fname)) {
		*s = 0;
		call_mode(fname);
	}
}

suffix_gz()
{
	// Calling gzip not gunzip works in more cases with Cygwin utils.
	// Use .exe suffix to avoid Take Command's internal incompatible gzip.
	if (uncompress_files & 1)
		do_decompression("gzip", is_mswindows ? "gzip.exe --decompress --stdout"
										  : "gzip --decompress --stdout");
}

suffix_bz2()
{
	if (uncompress_files & 2)
		do_decompression("bzip2", "bzip2 --decompress --stdout");
}

decompress_after_reading()
{
	char *ext = get_extension(filename);
	if (!strcmp(".gz", ext))
		suffix_gz();
	else if (!strcmp(".bz2", ext))
		suffix_bz2();
}
