/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 2004, 2018 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "colcode.h"

// Follow mode is a minor mode that makes adjacent windows showing the
// same buffer scroll at the same time.

int follow_when_idle(int cnt)
{
	int did_any = 0, i;
	if (follow_mode_on == 1 && !cnt) {
		follow_mode_on++;		// Lock out if error.
		int buf = bufnum, win = window_number, start = window_start;
		while (window_number > 0 && --window_number < win
			   && bufnum == buf && is_window(window_handle) == ISTILED) {
			did_any = 1;
			int old = point;
			point = start;
			i = MAX(text_height() - follow_mode_overlap, 0);
			point = window_start = prev_screen_line(i);
			if (point < old && old < start)
				point = old;
			start = window_start;
		}
		window_number = win;
		start = window_end + 1;
		if (follow_mode_overlap) {
			save_var point = start;
			start = prev_screen_line(MIN(text_height() - 1, follow_mode_overlap));
		}
		while (++window_number > win && is_window(window_handle) == ISTILED
			   && bufnum == buf) {
			did_any = 1;
			int old = point;
			point = window_start = start;
			i = MAX(text_height()- follow_mode_overlap, 0);
			start = next_screen_line(i);
			if (window_start < old && old < start)
				point = old;
		}
		window_number = win;
		modify_region(SHOW_MATCHING_REGION, MRCONTROL, 0);
		modify_region(SHOW_EXTRA_REGIONS, MRCONTROL, 0);
		follow_mode_on--;
	}
	restore_vars();
	if (did_any)
		refresh();
	return -1;
}

command follow_mode()
{
	follow_mode_on = has_arg ? (iter != 0) : !follow_mode_on;
	add_buffer_when_idle(bufnum, follow_when_idle);
	modify_region(SHOW_MATCHING_REGION, MRCONTROL,
				  follow_when_idle ? 0 : find_index("is_current_window"));
	modify_region(SHOW_EXTRA_REGIONS, MRCONTROL,
				  follow_when_idle ? 0 : find_index("is_current_window"));
	say("Follow mode %sabled for this buffer.",
		  follow_mode_on ? "en" : "dis");
}

// Called when keys repeat.
do_when_repeating_follow()
{
	if (follow_mode_on)
		follow_when_idle(0);
}
