/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2009 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "inimode.h"

color_ini_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();
	for (; point < to; nl_forward()) {
		t = point;
		switch (curchar()) {
			case '[':
				set_character_color(t, give_end_line() + 1,
									color_class inifile_section);
				break;
			case '#':
			case ';':
				set_character_color(t, give_end_line() + 1,
									color_class inifile_comment);
				break;
			default:	// Some textual .sys files use REM for comments.
				if (parse_string(1, "[ \t]*REM[ \t\n]")) {
					set_character_color(t, give_end_line(),
										color_class inifile_comment);
				} else if (parse_string(1, "[^\n=]*="))
					set_character_color(t, matchend,
										color_class inifile_definition);
		}
	}
	return point;
}

// Set display_func_name to the name of the ini section we're editing,
// and return 1.  If not in a section, set display_func_name to "" and
// return 1.  If user pressed a key and we gave up for now, return 0.

ini_func_name_finder()
{
	int from, to;
	save_var point;
	to_end_line();	// Might be starting from a section line.
	*display_func_name = 0;
	if (re_search(-1, "^<hspace>*%[(.*)%]")) {
		from = find_group(1, 1);
		to = find_group(1, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
	}
	return 1;
}

command ini_mode()
{
	mode_default_settings();
	major_mode = _ini_mode_name;
	mode_keys = inifile_tab;		/* Use these keys. */
	strcpy(comment_start, "[;#][ \t]*");
	strcpy(comment_pattern, "[;#].*$");
	strcpy(comment_begin, "; ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_ini_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 256) != 0;
	try_calling("ini-mode-hook");
	make_mode();
}

when_loading()
{
	inifile_tab[ALT('q')] = (short) fill_comment;
	inifile_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

suffix_ini()
{
	ini_mode();
}

suffix_ecm()	// Epsilon command file.
{
	ini_mode();
}

// Is it an MS-Windows file, not Info doc file or other types using
// this extension?
suffix_inf()
{
	if (matches_at(0, 1, "([ \t]*\n)*[[;]"))
		ini_mode();
}

suffix_sys()
{
	if (GET_LINE_TRANSLATE(translation_type) != FILETYPE_BINARY)
		ini_mode();
}
