/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1992, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has functions related to the menu bar.

#include "eel.h"
#include "menu.h"
#include "lowlevel.h"

command show_standard_bitmaps()
{
	want_toolbar = 1;
	toolbar_create();
	toolbar_add_button("STD_CUT", "STD_CUT", "find-file");
	toolbar_add_button("STD_COPY", "STD_COPY", "find-file");
	toolbar_add_button("STD_PASTE", "STD_PASTE", "find-file");
	toolbar_add_button("STD_UNDO", "STD_UNDO", "find-file");
	toolbar_add_button("STD_REDOW", "STD_REDOW", "find-file");
	toolbar_add_button("STD_DELETE", "STD_DELETE", "find-file");
	toolbar_add_button("STD_FILENEW", "STD_FILENEW", "find-file");
	toolbar_add_button("STD_FILEOPEN", "STD_FILEOPEN", "find-file");
	toolbar_add_button("STD_FILESAVE", "STD_FILESAVE", "find-file");
	toolbar_add_button("STD_PRINTPRE", "STD_PRINTPRE", "find-file");
	toolbar_add_button("STD_PROPERTIES", "STD_PROPERTIES", "find-file");
	toolbar_add_button("STD_HELP", "STD_HELP", "find-file");
	toolbar_add_button("STD_FIND", "STD_FIND", "find-file");
	toolbar_add_button("STD_REPLACE", "STD_REPLACE", "find-file");
	toolbar_add_button("STD_PRINT", "STD_PRINT", "find-file");
}

command show_view_bitmaps()
{
	want_toolbar = 1;
	toolbar_create();
	toolbar_add_button(	"VIEW_LARGEICONS", "VIEW_LARGEICONS", "find-file");
	toolbar_add_button(	"VIEW_SMALLICONS", "VIEW_SMALLICONS", "find-file");
	toolbar_add_button(	"VIEW_LIST", "VIEW_LIST", "find-file");
	toolbar_add_button(	"VIEW_DETAILS", "VIEW_DETAILS", "find-file");
	toolbar_add_button(	"VIEW_SORTNAME", "VIEW_SORTNAME", "find-file");
	toolbar_add_button(	"VIEW_SORTSIZE", "VIEW_SORTSIZE", "find-file");
	toolbar_add_button(	"VIEW_SORTDATE", "VIEW_SORTDATE", "find-file");
	toolbar_add_button(	"VIEW_SORTTYPE", "VIEW_SORTTYPE", "find-file");
	toolbar_add_button(	"VIEW_PARENTFOLDER", "VIEW_PARENTFOLDER", "find-file");
	toolbar_add_button(	"VIEW_NETCONNECT", "VIEW_NETCONNECT", "find-file");
	toolbar_add_button(	"VIEW_NETDISCONNECT", "VIEW_NETDISCONNECT", "find-file");
	toolbar_add_button(	"VIEW_NEWFOLDER", "VIEW_NEWFOLDER", "find-file");
	toolbar_add_button(	"HIST_BACK", "HIST_BACK", "find-file");
	toolbar_add_button(	"HIST_FORWARD", "HIST_FORWARD", "find-file");
	toolbar_add_button(	"HIST_FAVORITES", "HIST_FAVORITES", "find-file");
	toolbar_add_button(	"HIST_ADDTOFAVORITES", "HIST_ADDTOFAVORITES", "find-file");
	toolbar_add_button(	"HIST_VIEWTREE", "HIST_VIEWTREE", "find-file");
}


command standard_toolbar()
{
	want_toolbar = 1;
	toolbar_create();
	toolbar_add_separator();
	toolbar_add_button("STD_FILENEW", "New", "new_file");
	toolbar_add_button("STD_FILEOPEN", "Open", "find_file");
	toolbar_add_button("STD_FILESAVE", "Save", "save_file");
	toolbar_add_separator();
	toolbar_add_button("STD_PRINT", "Print", "print-buffer");
	toolbar_add_separator();
	toolbar_add_button("STD_CUT", "Cut", "kill_region");
	toolbar_add_button("STD_COPY", "Copy", "copy_region");
	toolbar_add_button("STD_PASTE", "Paste", "yank");
	toolbar_add_separator();
	toolbar_add_button("STD_UNDO", "Undo", "undo-changes");
	toolbar_add_button("STD_REDOW", "Redo", "redo-changes");
	toolbar_add_separator();
	toolbar_add_button(	"STD_FIND", "Find", "dialog-search");
	toolbar_add_button(	"STD_REPLACE", "Replace", "dialog-replace");
	toolbar_add_separator();
	toolbar_add_button("VIEW_DETAILS", "List Buffers", "bufed");
	toolbar_add_button("STD_DELETE", "Delete Buffer", "kill-current-buffer");
}

void set_toolbar()
{
	if (want_toolbar == 0)
		toolbar_destroy();
	else {
		toolbar_create();
		standard_toolbar();
	}
}

command toggle_toolbar()
{
	want_toolbar = !want_toolbar;
	set_toolbar();
}

init_menu()
{
	char *s;

	if (has_feature & FEAT_OS_MENUS) {
		if (s = lookpath(gui_menu_file))
			win_load_menu(s);
		win_display_menu(want_gui_menu);
	} else if (permanent_menu) {
		make_perm_menu();
		avoid_top_lines--;
	}
}

check_rebuild_menu()
{
	if ((has_feature & FEAT_OS_MENUS) && need_rebuild_menu)
		init_menu();
	need_rebuild_menu = 0;
}

load_menu()
{
	char buf[FNAMELEN], *s;
	spot sp;

	if (!exist(MENU_BNAME)) {
		if (!(s = lookpath(menu_file)))
			error("Can't find %s", menu_file);
		strcpy(buf, s);
		absolute(buf);
		save_var bufnum = zap(MENU_BNAME);
		if (err_file_read(buf, FILETYPE_AUTO))
			quick_abort();
		tab_size = 100;		// hide second column
		menu_window = 0;
		modify_region(1, MRCOLOR, color_class menu_bar_highlight);
		menu_line_spot = sp = alloc_spot();
		add_region(sp, sp, color_class submenu_highlight, REGLINE);
		submenu_highlight(0);
		save_var abort_searching = 0;
		for (point = 0; re_search(1, "^#.*\n"); )
			delete(matchstart, point);
	}
}

command rebuild_menu()
{
	if (has_feature & FEAT_OS_MENUS) {
		need_rebuild_menu = 1;
		return;
	}
	save_var bufnum;
	if (permanent_menu)
		drop_perm_menu();
	if (exist(MENU_BNAME))
		delete_buffer(MENU_BNAME);
	if (permanent_menu)
		make_perm_menu();
}

get_menu_width() // modify menu to include bindings & return new width of menu
{		// call with point just before current width number
	int width, numfrom, numto;
	char widthb[20];

	parse_string(1, "[0-9]+", widthb);	// get width of menu
	width = numtoi(widthb);
	if (menu_bindings && curchar() != '@') { // did we do col already?
		grab(numfrom = matchstart, numto = matchend, widthb);
		nl_forward();
		width += sub_menu_binding(width);
		point = numfrom;
		delete(numfrom, numto);		// put in new width &
		bprintf("%d@", width);		// mark this column done
	}
	return width;
}

sub_menu_binding(width)		/* add bindings to a single menu column */
{
	int p, textlen, puthere, i, addwidth = 0, atname;
	char func[100];

	while (p = parse_string(1, "([^\n\t]*)\t+")) {
		puthere = find_group(1, 0);
		textlen = puthere - point;
		point += p;
		atname = point;
		to_end_line();
		grab(atname, point, func);
		point = puthere;
		i = one_menu_binding(func, width - textlen);
		addwidth = MAX(addwidth, textlen - width);
		addwidth = MAX(addwidth, i);
		nl_forward();
	}
	return addwidth;
}

one_menu_binding(func, needspace)	/* add binding to a menu item */
char *func;
{
	int i;
	char b[160];

	if (*func == '#' || needspace < 0)
		return 0;
	*b = 0;
	i = find_index(func);
	if (!i || !find_bindings(b, i, root_keys, root_keys, "", 0))
		return 0;
	for (i = 0; i < needspace; i++)
		insert(' ');
	stuff(b);
	return strlen(b);
}

command toggle_menu_bar()	// toggle whether menu bar always appears
{
	if (has_feature & FEAT_OS_MENUS)
		set_gui_menu(has_arg ? (iter != 0) : !want_gui_menu);
	else
		set_menu_bar(has_arg ? (iter != 0) : !permanent_menu);
}

set_gui_menu(newval)
{
	want_gui_menu = newval;
	if (!(has_feature & FEAT_OS_MENUS))
		return;
	init_menu();
}

set_menu_bar(newval)	// turn perm menu bar on or off
{
	if (has_feature & FEAT_OS_MENUS)
		return;
	if (permanent_menu && (!menu_window || !exist(MENU_BNAME)))
		error("Permanent menu internal error: %d", menu_window);
	if (menu_window && permanent_menu && exist(MENU_BNAME))
		drop_perm_menu();
	menu_window = 0;
	permanent_menu = newval;
	if (permanent_menu)
		make_perm_menu();
}

restore_menu(want)	// do we want permanent menu?
{			// assume avoid_top_lines is already correct
	if (!(has_feature & FEAT_OS_MENUS))
		if (!permanent_menu && want) {
			make_perm_menu();
			avoid_top_lines--;
		} else if (permanent_menu && !want) {
			avoid_top_lines++;
			drop_perm_menu();
		}
	permanent_menu = want;
}

is_menu_here()	// has mouse moved onto (possibly not displayed) menu bar?
{
	return !(has_feature & FEAT_OS_MENUS) && still_on_menu_bar()
			&& (auto_menu_bar && mouse_pixel_y <= 0
				&& y_pixels_per_char() > 1 && opsys != OS_UNIX
				|| permanent_menu);
}

still_on_menu_bar()	// is mouse still on menu bar?
{
	return mouse_y <= 0 && mouse_x < screen_cols - 1;
}

num_of_popups()
{
	return number_of_popups() - (permanent_menu != 0);
}

click_menu()		// clicked over menu bar, begin doing menu stuff
{
	do {
		show_menu();
	} while (!permanent_menu && still_on_menu_bar() && await_menu_click());
	refresh();
}

show_menu_while_over()	// show menu while mouse is over it
{
	if (await_menu_click())
		click_menu();
	else
		invisible_cmd();
	if (!permanent_menu)
		remove_window(menu_window);
}

await_menu_click()	// show menu, return 1 if user clicked on it
{
	save_var mouse_cursor = left_pointer;
	load_menu();
	prep_menu_bar();
	save_var window_handle = menu_window;
	do {
		move_to_column(mouse_x);
		maybe_refresh();
		getkey();
	} while (key == MOUSE_MOVE && still_on_menu_bar());
	remove_window(menu_window);
	if (IS_MOUSE_DOWN(key) && still_on_menu_bar())
		return 1;
	ungot_key = full_key;
	return 0;
}

make_perm_menu()
{
	int oldwin = window_handle;

	if (has_feature & FEAT_OS_MENUS)
		return;
	load_menu();
	avoid_top_lines++;
	if (echo_line == 0) {
		echo_line = 1;
		say("");
	}
	prep_menu_bar();
	window_handle = menu_window;
	system_window = 1;	// user can't switch to it
	submenu_highlight(0);
	mark = point = 0;
	if (is_window(oldwin) && oldwin != menu_window)
		window_handle = oldwin;
	else
		window_number = 0;
}

drop_perm_menu()
{
	if (has_feature & FEAT_OS_MENUS)
		return;
	remove_window(menu_window);
	avoid_top_lines--;
	if (echo_line == 1) {
		echo_line = 0;
		say("");
	}
	menu_window = 0;
}

get_movement_or_release()	/* wait until mouse moves or buttons up */
{
	int ox = mouse_x, oy = mouse_y;

	while (mouse_pressed() && ox == mouse_x && oy == mouse_y)
		if (char_avail())
			getkey();
}

get_menu_cmd(cmd, sub)
char *cmd;
{
	*cmd = 0;
	if (sub && submenu_is_highlighted()) {
		window_handle = sub;
		save_var point += parse_string(1, "[^\t\n]*\t[ \t]*");
		parse_string(1, ".*", cmd);
	}
}

in_top_menu(sub, num)	// mouse on top line, maybe switch submenus
{			// return submenu handle, or 0 if none
	int new;

	window_handle = menu_window;
	new = show_submenu(sub, num);
	window_handle = menu_window;
	if (new)
		highlight_on();		// highlight on menu bar
	else
		highlight_off();
	if (sub && sub != new)		// remove old submenu
		remove_window(sub);
	return new;
}

show_submenu(old, num)  // put up submenu, return handle, or 0 if none
{			// if num, select that submenu #, else mouse col
	char word[80];
	int before;

	before = point;		// remember where on top line we were
	if (!to_submenu_name(num))
		return 0;
	mark = point;		// highlight from here to after name
	if (!parse_string(1, " [^ \n\t]+ ", word))	// get name of submenu
		return 0;
	if (before == point && is_window(old))	// same submenu selected
		return old;
	return make_submenu(word);
}

to_submenu_name(num)	// move to just before start of num'th submenu name
{			// (if -1, to name under mouse), and return nonzero
	point = 0;	// if a name's there
	if (num < 0)
		move_to_column(mouse_x + 2);
	else {
		while (num-- >= -1 && re_search(1, " +"))
			;		// get right behavior on spaces
		re_search(-1, " *");	// between words
	}
	return curchar() != '\n' && re_search(-1, " [^ \n\t]+");
}

make_submenu(word)
char *word;
{
	char pat[100];
	int sub;

	sub = add_popup(mark, 1, 30, screen_lines, LR_BORD(BBLANK), MENU_BNUM);
	prep_menu_window(color_class submenu);
	display_scroll_bar = 0;
	nl_forward();			// starting from 2nd line,
	sprintf(pat, "^%q[ \t]*", word);
	if (!re_search(1, pat)) {	// look for matching menu
		remove_window(sub);	// can't find menu, give up
		return 0;
	}
	window_width = get_menu_width() + 2;
	nl_forward();
	window_start = point;
	search(1, "\n\n");		// go to blank line after submenu items
	window_height = lines_between(window_start, point) - 1;
	point = window_start;
	return sub;
}

in_sub_menu(sub)	// mouse in a submenu, get to right line
{
	window_handle = sub;
	point_to_pos(mouse_x, mouse_y, mouse_screen);
	to_begin_line();
}

prep_menu_bar()		// Create menu bar
{
	save_var window_handle;
	menu_window = add_popup(0, 0, screen_cols, 1, 0, MENU_BNUM);
	prep_menu_window(color_class menu_bar);
	display_scroll_bar = 0;
	point = 0;
}

prep_menu_window(color)	// Do setup for a menu window
{
	text_color = color;
	set_wattrib(window_handle, PHORIZBORDCOLOR, color);
	set_wattrib(window_handle, PVERTBORDCOLOR, color);
	display_column = 0;
}

submenu_highlight(i)	// turn off submenu highlight, or put it on cur line
{
	*menu_line_spot = i ? point : size();
}

submenu_is_highlighted() // tell if any submenu line is being highlighted
{
	return *menu_line_spot != size();
}

run_menu_selection(p, shift)	// execute cmd named in string p
char *p;
{
	save_var run_by_mouse = 1;
	if (*p && *p != '#') {
		refresh();
		if (shift & (KB_CTRL_DN | KB_SHIFT_DN))
			has_arg = iter = 1;
		do_command(check_command(p));
	}
}

flash_menu_item()
{
	int i;

	for (i = 0; i < menu_bar_flashes; i++) {
		submenu_highlight(0);
		refresh();
		delay(5, 0);
		submenu_highlight(1);
		refresh();
		delay(5, 0);
	}
}

command show_menu() on reg_tab[FALT(2)]
{
	if (has_feature & FEAT_OS_MENUS) {
		invoke_menu(0);
		return;
	}
	save_var mouse_cursor = left_pointer;
	load_menu();
	if (!permanent_menu)
		prep_menu_bar();
	do_show_menu(0);
}

do_show_menu(num)
{
	char cmd[300];
	int sub, doit;

	save_var window_handle;
	save_var abort_searching = 0;
	save_var mouse_mask |= MASK_MOVE | MASK_UP;
	if (!IS_MOUSE_KEY(key))
		save_var menu_stays_after_click = 1;
	sub = show_menu_loop(num, cmd);
	restore_vars();
	doit = !exit_menu_cancel();
	finish_menu_selection(cmd, sub, doit);
	if (!echo_line)	// if echo line at top, clear menu from it
		note("");
	if (doit)
		run_menu_selection(cmd, (shift_pressed()
						   & (KB_ALT_DN | KB_CTRL_DN | KB_SHIFT_DN)));
}

show_menu_loop(num, cmd)
char *cmd;
{
	int sub = 0, line = 0, max, isdir, oldsub, orignum;

	max = max_submenus();
	if (IS_MOUSE_KEY(key))
		ungot_key = full_key;
	do {
		oldsub = sub;
		sub = to_sub_line(sub, num, line);
		if (getkey() == MOUSE_MOVE) {
			refresh();
			while (getkey() == MOUSE_MOVE)
				;
		}
		isdir = 1;		// assume it's a direction key
		switch (get_direction()) {
			case BTOP:	line--; break;
			case BBOTTOM:	line++; break;
			case BLEFT:	num--, line = 0; break;
			case BRIGHT:	num++, line = 0; break;
			default:	isdir = 0; break;
		}
		if (isalpha(key))
			line = do_menu_search(key);
		else if (IS_MOUSE_DOWN(key)) {
			if (oldsub && mouse_y <= 0)	// maybe drop menu
				orignum = count_submenus(mouse_x) - 1;
			else
				orignum = -2;	// upclick on bar won't drop
			sub = while_mouse_pressed(sub);
			num = after_mouse_pressed(sub, num, orignum);
			line = 0;
		} else if (binding_of(key, reg_tab) == (short) help) {
			get_menu_cmd(cmd, sub);
			if (*cmd && find_index(cmd))
				help_on_command(cmd);
			isdir = 1;
		}
		num = (num + max) % max;
		line = (line + window_height) % window_height;
	} while (isdir || isalpha(key) || IS_MOUSE_KEY(key));
	if (!exit_menu_doit() && !exit_menu_cancel()) {
		ungot_key = full_key;
		key = ESC;
	}
	return sub;
}

to_sub_line(sub, num, line)	// move to submenu # & line, maybe switching
{				// submenus, show result, & return new submenu
	sub = in_top_menu(sub, num);
	window_handle = sub;
	for (point = window_start; line-- > 0; )
		nl_forward();
	submenu_highlight(1);
	maybe_refresh();
	return sub;
}

while_mouse_pressed(sub)
{
	mouse_display = 1;
	while (mouse_pressed()) {
		if (mouse_y <= 0) {
			submenu_highlight(0);
			sub = in_top_menu(sub, -1);
			window_handle = menu_window;
		} else if (window_at_coords(mouse_y, mouse_x) == sub) {
			in_sub_menu(sub);
			submenu_highlight(1);
		} else
			submenu_highlight(0);
		refresh();
		get_movement_or_release();
	}
	return sub;
}

after_mouse_pressed(sub, num, orignum)	// convert mouse release to key
{					// and return new submenu #
	int i, max = max_submenus();

	if (mouse_y <= 0) {	// exit menu if we clicked from orig title
		i = count_submenus(mouse_x) - 1;
		if (orignum == num || !menu_stays_after_click)
			key = ESC;
		num = MIN(MAX(i, 0), max - 1);
		if (num != i)	// or if we clicked in blank area of menu bar
			key = ESC;
	} else if (window_at_coords(mouse_y, mouse_x) == sub) {
		in_sub_menu(sub);	// clicked on a cmd
		key = '\r';
	} else				// clicked somewhere else, exit menu
		key = ESC;
	return num;
}

count_submenus(max)	// return number of words on menu bar within max chars
{
	int i;

	save_var window_handle = menu_window;
	save_var point = 0;
	save_var narrow_end = size() - max;
	save_var abort_searching = 0;
	for (i = 0; curchar() != '\n' && re_search(1, " +"); i++)
		;
	return i;
}

max_submenus()		// give total # of submenus
{
	return count_submenus(1000) - 1;
}

do_menu_search(letter)	// look for menu item that starts with letter
{			// return its line number in submenu
	char buf[10];
	int orig = point;

	save_var case_fold = 1;
	save_var narrow_end = size() - window_end;
	nl_forward();		// find next match
	sprintf(buf, "^%c", letter);
	if (!re_search(1, buf)) {
		point = window_start;	// try again from start
		if (!re_search(1, buf))
			point = orig;
	}
	return lines_between(window_start, point);
}

exit_menu_doit()	// does this key do current menu item?
{
	switch (make_alt(key)) {
		case ALT('\r'):
		case ALT('\n'):
		case NUMALT(GREYENTER):	return 1;
	}
	return 0;
}

exit_menu_cancel()	// does this key exit from menu?
{
	return key == ESC || user_abort;
}

finish_menu_selection(cmd, sub, doit)
char *cmd;		// fill in cmd name from menu, then clean up
{
	save_var window_handle;
	if (doit) {
		get_menu_cmd(cmd, sub);
		if (*cmd && *cmd != '#')
			flash_menu_item();
	}
	window_handle = menu_window;
	highlight_off();
	if (!permanent_menu)
		remove_window(menu_window);
	if (sub)
		remove_window(sub);
}

// Display a text-based popup menu.
text_context_menu()
{
	char cmd[FNAMELEN], widthb[30];
	save_var mouse_mask |= MASK_MOVE | MASK_UP;
	load_menu();
	save_var mouse_cursor = left_pointer, window_handle;
	int popup = add_popup(mouse_x + 4, mouse_y, 30, screen_lines, ALL_BORD(BSINGLE), MENU_BNUM);
	prep_menu_window(color_class submenu);
	on_exit remove_window(popup);
	int isdir;
	display_scroll_bar = 0;
	point = 0;
	if (!re_search(1, "^ _popup[ \t]* ")	// look for popup menu
		  || !parse_string(1, "[0-9]+", widthb))	// get width of menu
		return 1;
	window_width = numtoi(widthb) + 2;
	nl_forward();
	window_start = point;
	save_var narrow_start = point;
	search(1, "\n\n");		// go to blank line after menu items
	save_var narrow_end = size() - matchstart;
	window_height = lines_between(window_start, point) + 3;
	point = window_start;
	while ((mouse_pressed() || char_avail()) && !user_abort)
		getkey();		// Wait for release of right button
	do {
		submenu_highlight(1);
		maybe_refresh();
		getkey();
		isdir = 0;
		switch (get_direction()) {
			case BTOP:
				point--;
				isdir = 1;
				break;
			case BBOTTOM:
				nl_forward();
				isdir = 1;
				break;
		}
		if (isalpha(key)) {
			go_line(do_menu_search(key) + 1);
			isdir = 1;
		}
		to_begin_line();
		if (IS_MOUSE_KEY(key) && window_at_coords(mouse_y, mouse_x, mouse_screen) == popup) {
			to_mouse_position(mouse_x, mouse_y);
			to_begin_line();
		}
	} while (IS_MOUSE_KEY(key) && !IS_MOUSE_UP(key) || isdir);
	*cmd = 0;
	if (IS_MOUSE_UP(key) && window_at_coords(mouse_y, mouse_x, mouse_screen) == popup
		  || exit_menu_doit())
		get_menu_cmd(cmd, popup);
	restore_vars();
	if (!*cmd || *cmd == '#')
		return 2;
	flash_menu_item();
	run_menu_selection(cmd, (shift_pressed()
							 & (KB_ALT_DN | KB_CTRL_DN | KB_SHIFT_DN)));
	return 0;
}

// Take two command lines to execute, separated by "!"s.  Use the first
// if there's no highlighted region (or there's just one "!").
// Otherwise use the second, and replace instances of "!" in it with the
// highlighted region.  If program to run contains spaces in its name or
// path, it must be surrounded by " characters.

run_by_pattern(p)
char *p;
{
	char text[FNAMELEN], cmdline[FNAMELEN], *s;

	s = index(p, '!');
	save_spot point, mark;
	fix_region();
	if (is_highlight_on() && s && mark - point < sizeof(text) / sizeof(*text)) {
		grab(point, mark, text);
		highlight_off();
	} else
		*text = 0;
	if (s && *text) {
		strcpy(cmdline, s + 1);
		p = index(cmdline, '!');
		if (p) {
			strcpy(p, text);
			strcat(p, index(s + 1, '!') + 1);
		}
	} else if (s) {
		strcpy(cmdline, p);
		if (s = index(cmdline, '!'))
			*s = 0;
	}
	if (is_mswindows) {
		char prog[FNAMELEN];
		say("%s", cmdline);
		if (*cmdline == '"') {
			strcpy(prog, cmdline + 1);
			s = index(prog, '"');
			if (s)
				*s = 0;
		} else {
			strcpy(prog, cmdline);
			s = index(prog, ' ');
			if (s)
				*s = 0;
		}
		winexec(prog, cmdline, SW_SHOWNORMAL, 0);
	} else
		shell("", cmdline, "");
}

// GUI menu selections or requests from WinHelp run this subroutine.

menu_invoked() on reg_tab[WIN_MENU_SELECT], reg_tab[WIN_HELP_REQUEST]
{
	int i;

	if (_doing_input == DI_LINEINPUT && is_gui) {
		ungot_key = full_key;
		error_if_input(1);
	}
	if (*menu_command == '$') {		// help file
		char old[FNAMELEN], dir[FNAMELEN];
// MSVC 2 wants the current directory to be \msvc20\bin, when using the help
// file \msvc20\help\xxx.hlp.  The other help files we know about don't care.
// So temporarily set the current directory based on the help file name.
		getcd(old);
		strcpy(dir, menu_command + 1);
		*get_tail(dir, 0) = 0;
		sprintf(get_tail(dir, 0), "bin");
		chdir(dir);
		i = windows_help_from(menu_command + 1, winhelp_display_contents);
		chdir(old);
		if (!i)
			error("Could not display help from %s", menu_command + 1);
		return;
	} else if (*menu_command == '!') {
		run_by_pattern(menu_command + 1);
		return;
	} else if (*menu_command == '%') {
		run_on_file_template(menu_command + 1);
		return;
	}
	save_var run_by_mouse = 1;
	i = find_index(menu_command);
	switch (name_type(i)) {
		case NT_MACRO:
			run_by_mouse = 0;
		case NT_COMMAND:
		case NT_SUBR:
		case NT_AUTOLOAD:
		case NT_AUTOSUBR:
			add_key_to_macro(CMD_INDEX_KEY + i);
			if (shift_pressed() & (KB_CTRL_DN | KB_SHIFT_DN))
				has_arg = iter = 1;
			do_command(i);
			break;
		case NT_WINVAR:
		case NT_BUFVAR:
		case NT_VAR:
		case NT_BUILTVAR:
			do_set_variable(i);
			break;
		default:
			error("Couldn't call `%s' from menu.", menu_command);
	}
}

#define XWIN_XFLD_COUNT	13	/* # of fields in an X11 font logical descriptor */

// Switch to the given font, then adjust the dialog window we're showing
// so it's a good size.
void xwin_switch_to_font_and_resize(char *font)
{
	if (strcmp(font_fixed, font)) {
		font_fixed = font;
		refresh();
		window_width_to_fit();
		int h = lines_between(0, size(), 0) + 5;
		h = MIN(h, screen_lines - 3);
		window_height = h;
		window_start = 0;
	}
}

// Called when we're selecting a font name for live updates. Can handle
// both foundry+family partial names, and full names.
int xwin_font_name_when_idle(int cnt)
{
	char partial_font[FNAMELEN], font[FNAMELEN], *p;
	int hyphen_cnt;

	if (cnt)
		return -1;
	save_var abort_searching = ABORT_IGNORE;
	save_var point;
	to_begin_line();
	grab(point, give_end_line(), partial_font);
	if (strlen(partial_font) > 2) {
		p = partial_font;
		for (hyphen_cnt = 0; p = strchr(p, '-'); hyphen_cnt++)
			p++;
		if (hyphen_cnt > 10)	// Looks like a complete font name.
			strcpy(font, partial_font);
		else
			xwin_expand_partial_font(font, partial_font);
		xwin_switch_to_font_and_resize(font);
		if (strcmp(font_fixed, font))
			say("Unusable font %s", font);
	}
	return -1;
}

// Modify the current font by replacing field #offset in its name
// with the value on this line. If decimal, we'll also convert from
// the point size format we show the user ("11.5" or "Custom") to the
// XLFD equivalent ("115" or "0").
void xwin_font_when_idle_helper(int offset, int decimal)
{
	char sel[FNAMELEN], font[FNAMELEN], after[FNAMELEN], *p, *q;
	int len, i;

	save_var abort_searching = ABORT_IGNORE;
	save_var point;
	to_begin_line();
	grab(point, give_end_line(), sel);
	len = strlen(sel);
	// Modify font_fixed by inserting point size, after converting it.
	if (decimal) {
		if (len >= 3 && sel[len - 2] == '.') {
			sel[len - 2] = sel[len - 1];	// Remove . from font size like 12.0
			sel[len - 1] = 0;
		} else if (!strncmp(sel, "Custom", 6)) {
			strcpy(sel, "120");
		} else
			return;
	}
	strcpy(font, font_fixed);
	for (p = font, i = 0; p && i < offset; i++)
		p = strchr(p + 1, '-');
	if (!p)
		return;
	// Insert sel into the font-size position in font.
	strcpy(after, p + 1);
	strcpy(p + 1, sel + (*sel == ' '));	// Skip initial space.
	q = strchr(after, '-');
	if (!q)
		return;
	strcat(p, q);
	xwin_switch_to_font_and_resize(font);
}

// Called when we're selecting a font size for live updates.
int xwin_font_size_when_idle(int cnt)
{
	if (!cnt)
		xwin_font_when_idle_helper(7, 1);
	return -1;
}

// Called when we're selecting a pixel size for live updates.
int xwin_pixel_size_when_idle(int cnt)
{
	if (!cnt)
		xwin_font_when_idle_helper(6, 0);
	return -1;
}

// Construct a regex pattern for an X11 font name by combining the
// elements of the p array, replacing * with a match-any pattern.
xwin_font_array_to_pattern(char *fontpat, char (*p)[1000])
{
	int i;
	sprintf(fontpat, "^-%q", p[0]);
	for (i = 1; i < XWIN_XFLD_COUNT; i++) {
		strcat(fontpat, "-");
		if (strcmp(p[i], "*")) {
			strcat(fontpat, p[i]);
			if (strchr(p[i], '-'))
				i++;
		} else
			strcat(fontpat, "[^-\n]*");
	}
	strcat(fontpat, "$");
}

// Do any fonts in the list in buf match the font pattern in p, once
// we've defined an additional field in the pattern? We construct the
// font pattern by combining the elements of the p array, replacing * with
// a match-any pattern, but replacing whatever's at position ind with
// val. Return 0 if no matches, 1 if one, 2 if two or more. If there's
// at least one match, also add the new value val to the p array.

int xwin_fill_in_field(int buf, char (*p)[1000], int ind, char *val)
{
	char fontpat[1000];
	int cnt = 0, i;

	sprintf(fontpat, "^-%q", p[0]);
	for (i = 1; i < XWIN_XFLD_COUNT; i++) {
		strcat(fontpat, "-");
		if (i == ind)
			strcat(fontpat, val);
		else if (!strcmp(p[i], "*")	// Replace * with no-dash pattern
				 || i == 6 && !strcmp(p[i], "0"))	// point size==0 means any
			strcat(fontpat, "[^-\n]*");
		else
			strcat(fontpat, p[i]);
	}
	save_var bufnum = buf;
	point = 0;
	while (re_search(1, fontpat) && cnt < 2)
		cnt++;
	if (cnt > 0)
		strcpy(p[ind], val);
	return cnt;
}

// If partial_font is just a foundry+family name, fill in the rest to
// make a full name, providing reasonable defaults.
xwin_expand_partial_font(char *font, char *partial_font)
{
	int buf = name_to_bufnum("-font-list"), i;
	char p[XWIN_XFLD_COUNT][1000], sz[20];

	strcpy(p[0], partial_font);
	for (i = 1; i < XWIN_XFLD_COUNT; i++)
		strcpy(p[i], "*");
	if (!xwin_fill_in_field(buf, p, 1, "medium"))
		xwin_fill_in_field(buf, p, 1, "bold");
	xwin_fill_in_field(buf, p, 2, "r");
	xwin_fill_in_field(buf, p, 3, "normal");
	xwin_fill_in_field(buf, p, 11, "iso8859");
	for (i = 140; i < 200; i++) {	// Try some reasonable point sizes.
		sprintf(sz, "%d", i);
		if (xwin_fill_in_field(buf, p, 6, sz))
			break;
	}

	// Try some common res_x values. If it doesn't have those specific
	// ones, but does allow any res_x, just pick a reasonable one.
	if (!xwin_fill_in_field(buf, p, 7, "100")
		  && !xwin_fill_in_field(buf, p, 7, "75"))
		if (xwin_fill_in_field(buf, p, 7, "0"))
			strcpy(p[7], "100");
	
	strcpy(font, partial_font);
	*font = 0;
	for (i = 0; i < XWIN_XFLD_COUNT; i++) {
		strcat(font, "-");
		strcat(font, p[i]);
	}
}

// Each line in the buffer buf contains the foundry+family of a set of
// one or more fonts. Delete those lines where Epsilon can't use the
// full font name we construct from it (most likely meaning the whole
// family is unusable).
xwin_prune_to_usable_fonts(char *buf)
{
	char partial_font[FNAMELEN], font[FNAMELEN], orig[FNAMELEN];
	strcpy(orig, font_fixed);
	save_var bufname = buf;
	for (point = 0; point < size(); ) {
		grab(point, give_end_line(), partial_font);
		xwin_expand_partial_font(font, partial_font);
		font_fixed = font;
		if (strcmp(font_fixed, font))
			delete(point, give_end_line() + 1);
		else if (!nl_forward())
			break;
	}
	font_fixed = orig;
}

xwin_set_font_first_prune()
{
	point = 0;
	// Now, discard proportional fonts.
	do_drop_matching_lines(REGEX, "-p-[^-\n]*-[^-\n]*-[^-\n]*$", 1);
	// Discard font aliases too.
	do_drop_matching_lines(REGEX, "^-[^-\n]*-[^-\n]*-[^-\n]*-[^-\n]*-[^-\n]*-", 0);
	// Also drop any fonts with weird charsets.
	do_drop_matching_lines(REGEX, "-fontspecific(-[^-\n]*)?$", 1);
}

// List all the possibilities for one field (with index ind) of the
// font into buffer dest, using the list of fonts in buf. First delete
// any line that doesn't match onlypat. If ind is -1, we instead list
// all font names matching onlypat.
xwin_make_font_name_list(char *dest, int buf, int ind, char *onlypat)
{
	char pat[FNAMELEN];
	int i;
	strcpy(pat, "^-");	// Make an XFLD pattern where #1 matches field ind
	for (i = 0; i < ind; i++)
		strcat(pat, "[^-]*-");
	if (ind == 0)		// Special rule: first two are together
		strcat(pat, "([^-]*-[^-]*)-.*");
	else if (ind == 12)	// Special rule: final charset may include -
		strcat(pat, "(.*)$");
	else				// Usual case: just get the one field
		strcat(pat, "([^-]*)-.*");
	save_var bufnum = zap(dest);
	grab_buffer(buf);
	point = 0;
	if (onlypat)
		do_drop_matching_lines(REGEX, onlypat, 0);
	if (ind >= 0)
		string_replace(pat, "#1", REGEX);
	do_uniq(1, 1, 0);
	point = 0;
}

// Make the list of point sizes look nicer. Add the missing decimal
// point, and put a space before single-digit sizes so they sort first.
xwin_font_size_prettify(char *buf)
{
	save_var bufname = buf;
	string_replace("^0$", "Custom size", REGEX);
	string_replace("^(<digit>*)(<digit>)$", "#1.#2", REGEX);
	string_replace("^(<digit>%.<digit>)$", " #1", REGEX);
}

// Replace all 0 point sizes with the user's selection.
xwin_add_point_size(char *buf, char *cust_point_sz)
{
	save_var bufname = buf;

	save_var point = 0;
	while (re_search(1, "^-[^-\n]*-[^-\n]*-[^-\n]*-[^-\n]*-[^-\n]*-[^-\n]*-"
					 "[^-\n]*-(0)-")) {
		delete(find_group(1, 1), matchend);
		bprintf("%s-", cust_point_sz);
	}
}

// Undo the changes above.
xwin_font_size_unprettify(char *str)
{
	int tmp = tmp_buf();
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	stuff(str);
	point = 0;
	string_replace("^ +", "", REGEX);
	string_replace("^(<digit>*)%.(<digit>)$", "#1#2", REGEX);
	string_replace("Custom size", "0", REGEX);
	grab(0, size(), str);
}

// User typed in a point size. Put it in XFLX format.
xwin_font_size_fix_user(char *str)
{
	int tmp = tmp_buf();
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	stuff(str);
	point = 0;
	string_replace("^ +", "", REGEX);	// Strip leading spaces
	string_replace(" +$", "", REGEX);	// Strip trailing spaces
	string_replace("^(<digit>+)$", "#10", REGEX);	// Change 14 to 140
	string_replace("^(<digit>*)%.(<digit>)$", "#1#2", REGEX); // 14.0 to 140
	grab(0, size(), str);
}

// Ask user to pick from a list, put response in dest, and return 1. If
// there's only one possibility, just fill it in and return 0. Display
// prompt and dialog title, and position on suggestion sug if it's in
// the list. If idlefunc, load it into matches.
int xwin_prompt_for_field(char *dest, char *prompt, char *title,
						 char *sug, int (*idlefunc)())
{
	int mbuf = name_to_bufnum(_MATCH_BUF);
	save_var bufnum = mbuf;
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	delete_final_newlines();
	point = 0;
	restore_vars();
	if (!count_lines_in_buf(mbuf, 0)) {	// Just one line?
		grab_full_line(mbuf, dest);		// Yes, so grab it.
		return 0;
	}
	save_var search_in_menu = 2;	// Keys move to the next match.
	strcpy(dest, sug);
	if (idlefunc) {
		add_buffer_when_idle(mbuf, idlefunc);
		on_exit delete_buffer_when_idle(mbuf, idlefunc);
	}
	get_choice(mbuf, dest, title, prompt, NULL, NULL, NULL);
	return 1;
}

// For fields with no special handling. Make a list of options and ask
// user to pick one, returning 1. If there's only one possibility, add
// it to the field data without prompting, and return 0.
int xwin_query_one_field(int ind, char (*fields)[1000], char *prompt,
						char *title, char *sug)
{
	char fontpat[FNAMELEN];
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, name_to_bufnum("-font-list"),
							 ind + 1, fontpat);
	return xwin_prompt_for_field(fields[ind], prompt, title, sug,
								 (int (*)()) 0);
}

// Make a list of fonts by running xlsfonts, and have the user narrow
// down the choices until just one is left. Return 0 if we can't get the
// list.
int set_font_for_unix()
{
	char fontname[1000], orig[1000], fontpat[1000], fontfamily[1000], msg[FNAMELEN];
	char fields[XWIN_XFLD_COUNT][1000], cust_point_sz[1000];
	jmp_buf *old_level = top_level, this_level;
	int i, mbuf = zap(_MATCH_BUF);
	struct screen_info retain_windows;

	retain_windows.wins = 0;
	save_screen(&retain_windows);
	strcpy(orig, font_fixed);
	*cust_point_sz = 0;
	for (i = 0; i < XWIN_XFLD_COUNT; i++)
		strcpy(fields[i], "*");
	save_var show_status = 0;
	note("Listing fonts...");
	int font_list = zap("-font-list");

	// First, extract the foundry+family of cur font.
	buf_printf(font_list, "%s\n", font_fixed);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	grab_full_line(mbuf, fontfamily);

	buf_zap(font_list);
	pipe_text(NULL, "-font-list", "if [ -x /opt/x11/bin/xlsfonts ] ; "
			  "then /opt/x11/bin/xlsfonts ; else xlsfonts ; fi", NULL,
			  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
	save_var bufnum = font_list;	// Got a list of all available fonts.
	if (lines_between(0, size(), 0) < 5) {
		say("Couldn't run xlsfonts to list the fonts.");
		return 0;
	}
	top_level = &this_level;
	if (setjmp(top_level)) {	// On abort, come here.
		font_fixed = orig;
		restore_screen(&retain_windows);	// Restore windows, since size changes
		free(retain_windows.wins);			// might have deleted some.
		top_level = old_level;
		quick_abort();
	}
	xwin_set_font_first_prune();	// Drop fonts we know we can't use.
	note("");

	// Extract the foundry+family names from font_list.
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	xwin_prune_to_usable_fonts(_MATCH_BUF);
	strcpy(msg, "Select a font:");
	if (*fontfamily)
		sprintf(msg, "Select a font [now %s]:", fontfamily);
	xwin_prompt_for_field(fields[0], msg, " Select a font ",
						 fontfamily, xwin_font_name_when_idle);

	// Now make a list of allowed point sizes for the selected font
	// family. Zero means any size is OK.
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 7, fontpat);
	xwin_font_size_prettify(_MATCH_BUF);

	i = xwin_prompt_for_field(fields[6], "Select a point size:",
							 " Point Size ", "12",
							 xwin_font_size_when_idle);

	xwin_font_size_unprettify(fields[6]);
	// If point size is 0 (custom), prompt for it and remember it.
	if (!strcmp(fields[6], "0")) {
		if (i)
			get_string(cust_point_sz, "Point size, such as 12 or 14: ");
		else
			get_strdef(cust_point_sz, "Font lacks recommended point size. "
					   "Please pick one", "14");
		xwin_font_size_fix_user(cust_point_sz);
	}

	// Ask the user to choose among other options the font family might
	// include.
	xwin_query_one_field(11, fields, "Select a char set (iso8859-1 "
						"recommended, or iso10646-1 for Unicode):",
						" Character Set ", "iso8859-1");
	xwin_query_one_field(1, fields, "Make all characters bold?",
						" Force Bold? ", "medium");
	xwin_query_one_field(2, fields, "Make all characters italic (o) not regular (r)?",
						" Force Italic? ", "r");

	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 6, fontpat);
	xwin_prompt_for_field(fields[5], "Select a pixel size:", " Pixel size ", "",
						  xwin_pixel_size_when_idle);

// any other fields?

	save_var resize_menu_list = 1;
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, -1, fontpat);

	if (*cust_point_sz)
		xwin_add_point_size(_MATCH_BUF, cust_point_sz);
	xwin_prompt_for_field(fontname, "These fonts match. Please pick one:",
						 " Font names ", "", xwin_font_name_when_idle);

	note("");
	aside("Switch to %s if %d", fontname, !user_abort);
	font_fixed = user_abort ? orig : fontname;
	if (!user_abort && strcmp(font_fixed, fontname))
		say("Sorry, can't use that font.");

	restore_screen(&retain_windows);
	free(retain_windows.wins);
	top_level = old_level;
	return 1;
}

do_set_font(char *prompt, int type, char *varname)
{
	if (has_feature & FEAT_FONT_DIALOG)
		windows_set_font(prompt, type);
	else if (has_feature & FEAT_SET_FONT)
		do_set_variable(find_index(varname));
	else
		error("Setting %s is not supported in this environment.", varname);
	update_display_class();
}

when_setting_font_fixed()
{
	update_display_class();
}

command set_font()
{
	if (is_unix != IS_UNIX_XWIN || !set_font_for_unix())
		do_set_font("Select font", 0, "font-fixed");
}

command set_printer_font()
{
	do_set_font("Select printer font", 1, "font-printer");
}

command set_dialog_font()
{
	do_set_font("Select font for dialogs", 2, "font-dialog");
}

// Helps to select odd font sizes that set-font can't handle
// (for example, many of the fonts used in the DOS box).  Press
// arrow keys to change desired height & width of the current font.
// Press <Enter> or <Space> to exit.  Press the abort key to restore
// the old font size, then exit.  This command cannot select font
// names, or toggle bold/italic settings--use set-font for that.

// To choose a font, first use the set-font command to select the
// desired font name, and select italic or bold options if desired.
// Then run change-font-size (this command) and use the arrow keys to
// move among the different font sizes.  The font size (height & width)
// will be displayed at all times.  A width or height of 0 means
// use a suitable default.

command change_font_size()
{
	char *s, name[50], buf[50], old[50];
	int h, w, dir;

	if (is_unix)
		error("Please use the set-font command to select a font.");
	if (has_arg) {
		font_size_increment(iter);
		iter = 0;
		return;
	}
	strcpy(old, font_fixed);
	strcpy(name, font_fixed);
	s = index(name, ',');
	*s = 0;
	h = numtoi(s + 1);
	s = index(s + 1, ',');
	w = numtoi(s + 1);
	s = index(s + 1, ',') + 1;		// rest of spec
	for (;;) {
		note("Using font %s %dx%d, screen is %dx%d. "
			"(Press an arrow key to change.)",
			name, h, w, screen_cols, screen_lines);
		getkey();
		if (user_abort || key == BELL || key == ESC) {
			font_fixed = old;
			break;
		}
		dir = get_direction();
		switch (dir) {
			case BTOP:		if (h > 0) h--; break;
			case BBOTTOM:	h++; break;
			case BLEFT:		if (w > 0) w--; break;
			case BRIGHT:	w++; break;
		}
		if (dir != -1) {
			sprintf(buf, "%s,%d,%d,%s", name, h, w, s);
			font_fixed = buf;
			refresh();
		} else if (IS_ENTER_KEY(key) || key == ' ')
			break;
		else if (IS_TRUE_KEY(key))
			say("Press <Esc> to abort, <Enter> to keep these settings.");
	}
	say("");
}

// Increment and decrement the font size value in the current font name under Windows.
font_size_increment(int incr)
{
	char newfont[FNAMELEN], *p, *q;

	p = strchr(font_fixed, ',');
	if (p) {
		int size = strtoi(p + 1, 10);
		q = strchr(p + 1, ',');
		if (q) {
			strncpy(newfont, font_fixed, p - font_fixed);
			sprintf(newfont + (p - font_fixed), ",%d%s", size + incr, q);
			font_fixed = newfont;
			return;
		}
	}
}

command web_search()
{
	char url[FNAMELEN], *p;

	strcpy(url, web_search_url);
	p = url + strlen(url);
	if (is_highlight_on()) {
		save_var point, mark;
		fix_region();
		if (mark - point >= ptrlen(p) - 1)
			error("Highlighted region is too big for a search parameter.");
		grab(point, mark, p);
	} else {
		save_var point--;
		re_search(RE_FORWARD, "<ident>+");
		re_search(RE_REVERSE, "<ident>+");
		grab(matchend, matchstart, p);
	}
	show_url(url);
}

command invoke_windows_menu() on reg_tab[ALT(' ')]
{
	int k = ' ';

	if (IS_EXT_ASCII_KEY(NUMPLAIN(key)))
		k = NUMPLAIN(key) & 0xff;	/* strip meta bit */
	invoke_menu(k);
}
