/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2009 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "indent.h"
#include "c.h"
#include "html.h"
#include "tex.h"
#include "python.h"

// Add >'s to lines between point and mark.
command mail_quote_region() on cc_tab['>']
{
	save_spot point, mark;
	if (!is_highlight_on())
		mark_paragraph();
	fix_region();
	to_begin_line();
	save_var narrow_start = point, narrow_end = size() - mark;
	simple_re_replace(1, "^!<any>", mail_quote_text);
	highlight_off();
}

// Figure out the section quoted by >'s; remove its quote marks and set
// point & mark around it.  But if there's a highlighted region, unquote
// just those lines.
command mail_unquote() on cc_tab['<']
{
	char prefix[FNAMELEN], quoted[FNAMELEN];

	save_var narrow_start, narrow_end;
	if (is_highlight_on()) {
		fix_region();
		to_begin_line();
		narrow_start = point;
		narrow_end = size() - mark;
	}
	to_begin_line();
	if (prefill_have_single_line(mail_quote_pattern))
		nl_forward();
	if (!parse_string(1, mail_quote_skip, prefix)) {
		highlight_off();
		error("Can't locate quoting on this line.");
	}
	sprintf(quoted, "^%q", prefix);
	if (!is_highlight_on())
		prefix_find_fill_region(0, mail_quote_skip, mail_quote_pattern);
	simple_re_replace(1, quoted, "");
	mark = size();
	point = 0;
}

// Does this line lack the fill prefix, and is it surrounded on both
// sides by lines with the fill prefix?  (If so, it's probably a
// continuation line and should be included in filling.)
int prefill_have_single_line(char *pat)
{
	save_var point;
	to_begin_line();
	if (parse_string(1, pat))
		return 0;
	if (!nl_forward() || !parse_string(1, pat))
		return 0;
	point--;
	if (!nl_reverse())
		return 0;
	to_begin_line();
	return parse_string(1, pat) != 0;
}

// Return nonzero if this line starts with pat.  If one_parag, the line
// must additionally have no longer prefix fill sequence than that, and
// must not be empty after that either.  If !one_parag, it just needs to
// start with pat.

// Special rule for when one_parag is zero: single lines with no
// matching prefix, surrounded by lines with a prefix, are considered to
// be part of the same paragraph.

// If pat is empty and one_parag is nonzero, a line is considered
// matching if it's nonempty and doesn't start with any possible prefix.

int prefill_line_matches(char *pat, int one_parag, char *skip, char *sing_pat)
{
	int i;

	save_var point;
	to_begin_line();
	if (!*pat && one_parag)
		return !parse_string(1, skip) && !matches_at(point, 1, "[ \t]*$");
	i = parse_string(1, pat);
	if (!one_parag)
		return (i || prefill_have_single_line(sing_pat));
	return i && i == parse_string(1, skip)
			&& (!one_parag || matches_at(point + i, 1, "[^ \t\n]"));
}

// Look at the fill prefix on this line; narrow to region of lines
// around this one with a similar prefix.  If one_parag, include only a
// single paragraph (within the block of similarly prefixed text).
prefix_find_fill_region(int one_parag, char *skip, char *pat)
{
	char prefix[FNAMELEN], quoted[FNAMELEN];

	to_begin_line();
	*prefix = 0;
	parse_string(1, skip, prefix);
	sprintf(quoted, "%q", prefix);
	while (nl_reverse()) {
		if (!prefill_line_matches(quoted, one_parag, skip, pat)) {
			nl_forward();
			break;
		}
	}
	narrow_start = point;
	while (nl_forward()) {
		if (!prefill_line_matches(quoted, one_parag, skip, pat)) {
			point--;
			break;
		}
	}
	narrow_end = size() - point;
}

prefix_fill_one_paragraph(char *pat, char *skip, int force)
{
	char prefix[FNAMELEN], quoted[FNAMELEN];

	save_spot point--;
	re_search(1, "[^\n\t ]");
	if (force || matches_at(give_begin_line(), 1, pat)) {
		save_spot mark, narrow_start, narrow_end;
		save_var margin_right;
		to_begin_line();
		*prefix = 0;
		parse_string(1, skip, prefix);
		if (!*prefix && !force) {
			fill_paragraph();
			return;
		}
		margin_right -= current_column();
		if (margin_right < 1)
			margin_right = 1;
		sprintf(quoted, "^%q", prefix);
		prefix_find_fill_region(1, skip, pat);
		// We've now narrowed to one paragraph, and know the quoting
		// prefix shared by all its lines.  So strip the prefix, fill,
		// and put it back.
		if (*prefix)
			simple_re_replace(1, quoted, "");
		highlight_off();
		fill_paragraph();
		if (*prefix)
			simple_re_replace(1, "^", prefix);
	} else
		fill_paragraph();
}

// Fill a region or a paragraph, using the specified rules for
// determining a possible fill prefix.  If !force, do a plain
// fill-paragraph when the line doesn't match the prefix pattern.
prefix_fill_region(char *pat, char *skip, int force)
{
	int last = 0;
	iter = 0;
	if (has_arg)
		save_var margin_right = current_column() + 1;
	if (!is_highlight_on())
		prefix_fill_one_paragraph(pat, skip, force);
	else {
		char prefix[FNAMELEN], quoted[FNAMELEN];
		save_spot point, mark, narrow_start, narrow_end;
		fix_region();
		do {
			last = point;
			prefix_fill_one_paragraph(pat, skip, 1);
			to_begin_line();
			*prefix = 0;
			parse_string(1, skip, prefix);
			sprintf(quoted, "%q", prefix);
			while (prefill_line_matches(quoted, 1, skip, pat) && nl_forward())
				check_abort();
			while (character(parse_string(1, skip) ? matchend : point) == '\n'
				   && nl_forward())	// Move past all-prefix lines to next paragraph.
				check_abort();
		} while (point < mark && point > last);
	}
}

// Fill a paragraph, but preserve its quoting.
command mail_fill_paragraph() on cc_tab[ALT('q')]
{
	prefix_fill_region(mail_quote_pattern, mail_quote_skip, 0);
}

// Fill this paragraph (or regions in paragraph) using non-alphanumeric
// text at the start of line as a fill prefix.
command prefix_fill_paragraph() on reg_tab[ALT('Q')]	// Alt-Shift-Q
{
	prefix_fill_region("^<^alphanum|(|)|wspace>",
					   "^<^alphanum|(|)|newline>+", 1);
}

when_loading()
{
	c_tab[ALT('Q')] = (int) prefix_fill_paragraph;
	html_tab[ALT('Q')] = (int) prefix_fill_paragraph;
	tex_tab[ALT('Q')] = (int) prefix_fill_paragraph;
	python_tab[ALT('Q')] = (int) prefix_fill_paragraph;
}
