/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 2000, 2003 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "tags.h"

struct {
	char prefix[100];		// Find words that start with this.
	char found[200];		// The current candidate replacement.
	char last[200];			// Previous replacement we found.
	int len;				// Length of the current candidate replacement.

	// These apply to the destination buffer.
	int savepos;		// Buffer position of the start of the prefix.

	// The rest apply to the source buffer we're searching for matches.
	int window_src;		// Get matches from this window.
	int pullsrc;		// Region we're highlighting.
	spot start;			// Its start and end.
	spot end;

	int tag_position;	// Position in tag buffer we're searching.

	spot ignore_start;	// Ignore matches in this range.
	spot ignore_end;

	int match_current_word;	// At start, match word we're inside.

	int orig;			// Position in search buffer where we started.
} pull;

zeroed int echo_window_handle;

int pull_find_from_tags(int dir)
{
	char filter_pattern[200];
	int res;

	if (pull.tag_position == -1)
		return 0;
	save_var bufname = "-tags";
	point = pull.tag_position;
	sprintf(filter_pattern, "^([^\t\n]+(::|'))?(%s)(%s)",
			pull.prefix, word_pattern);
	for (;;) {
		res = re_search(dir, filter_pattern);
		pull.tag_position = point;
		*pull.found = 0;
		if (!res)
			return 0;
		grab(find_group(3, 1), find_group(4, 0), pull.found);
		if (!strcmp(pull.found, pull.last))
			continue;
		dismiss_echo_context();
		note("Tag found: %s", pull.found);
		*pull.end = *pull.start;
		expire_message = 1;
		return 1;
	}
}

int find_word_with_prefix(int dir)
{
	char filter_pattern[200];
	int res;

	save_var point;
	if (dir == -1)
		point = *pull.start;
	else
		point = *pull.end;

	if (pull.match_current_word)	// First match should be current word.
		point -= parse_string(-dir, word_pattern) * dir;
	pull.match_current_word = 0;

	*pull.found = 0;
	if (pull.tag_position != -1) {
		res = pull_find_from_tags(dir);
		if (res || dir < 0)
			return res;
		pull.tag_position = -1;
		point = 0;
	}
	sprintf(filter_pattern, "</word>(%s)(%s)", pull.prefix, word_pattern);
	while (re_search(dir, filter_pattern) != 0){

		if (pull.ignore_start && spot_to_buffer(pull.ignore_start) == bufnum
			  && (*pull.ignore_start <= matchstart && matchstart <= *pull.ignore_end
				  || *pull.ignore_start <= point && point <= *pull.ignore_end))
			continue;

		*pull.start = find_group(1, 1);
		*pull.end = find_group(2, 0);
		grab(*pull.start, *pull.end, pull.found);
		if (!strcmp(pull.found, pull.last))
			continue;

		return 1;
	}
	if (dir == -1 && point <= narrow_start && pull_word_from_tags
		  && !pull.window_src && pull.tag_position == -1) {
		if (pull_word_from_tags == 2)
			init_tags();
		if (exist("-tags")) {
			pull.tag_position = buffer_size("-tags");
			return pull_find_from_tags(dir);
		}
	}
	return 0;
}

restore_prefix()
{
	save_var point = pull.savepos;
	delete(point, point + pull.len);
	stuff(pull.prefix);
	pull.len = strlen(pull.prefix);
}

int pl_search(int dir)
{
	restore_prefix();

	if (pull.window_src)
		save_var window_handle = pull.window_src;
	if (find_word_with_prefix(dir) != 0){
		restore_vars();
			// Get rid of the last guess.
		point = pull.savepos;
		delete(point, point + pull.len);

			// Stuff in the new guess.
		pull.len = strlen(pull.found);
		strcpy(pull.last, pull.found);
		stuff(pull.found);

		maybe_refresh();

		if (!pull.window_src && pull.tag_position == -1)
			if (is_region_visible(*pull.start, *pull.end) == 0) {
				echo_context(*pull.start);
			} else {
				dismiss_echo_context();
				say("");
			}

		return 1;
	} else {
		say("Search failed for word beginning with: %s", pull.prefix);
		*pull.last = 0;
		return 0;
	}
}

void free_pl()
{
	remove_region(pull.pullsrc);
	free_spot(pull.start);
	free_spot(pull.end);
	if (pull.ignore_start)
		free_spot(pull.ignore_start);
	free_spot(pull.ignore_end);
	dismiss_echo_context();
	if (pull.window_src) {
		save_var window_handle = pull.window_src;
		point = pull.orig;
	}
}

void alloc_pl()
{
	/* Get the word prefix, if any, that occurs
	 * immediately before point.
	 */
	pull.len = 0;
	pull.prefix[0] = 0;
	pull.ignore_end = alloc_spot(1);
	parse_string(-1, word_pattern, pull.prefix);
	pull.ignore_start = alloc_spot(0);	// Ignore original word.

	pull.savepos = point;
	pull.len = strlen(pull.prefix);

	if (_doing_input == DI_LINEINPUT || _doing_input == DI_SEARCH
		  || _doing_input == DI_SEARCH_NOINCR) {
		pull.window_src = prompt_orig_win;
		save_var window_handle = prompt_orig_win;
		free_spot(pull.ignore_start);	// Don't restrict matches.
		pull.ignore_start = 0;
	} else
		pull.window_src = 0;
	pull.start = alloc_spot(0);
	pull.end = alloc_spot(0);
	pull.orig = point;
	pull.pullsrc = add_region(pull.start, pull.end, color_class pull_highlight, REGNORM);
	*pull.last = 0;
	pull.tag_position = -1;
	pull.match_current_word = 0;
	if (_doing_input == DI_LINEINPUT || _doing_input == DI_SEARCH
		  || _doing_input == DI_SEARCH_NOINCR)
		pull.match_current_word = 1;
}

void pl_restore()
{
	point = pull.savepos;
	delete(point, point + pull.len);
	stuff(pull.prefix);
}

command pull_word() on reg_tab[NUMCTRL(KEYUP)], reg_tab[FKEY(3)]
{
	int pl_key, old_mod = modified, orig_key = full_key;

	check_modify(bufnum);
	iter = 1;
	alloc_pl();

	for (pl_key = pl_trans(full_key); 1; pl_key = pl_getkey()){
		if (full_key == orig_key && pl_key != 1 && pl_key != -1)
			pl_key = 1;
		switch(pl_key){
			case 1:
			case -1:
				if (pl_search(pl_key) == 0){
					expire_message = 1;
					pl_restore();
					dismiss_echo_context();
					maybe_refresh();
					continue;
				}
				break;
			case 7:
				pl_restore();
				free_pl();
				if (*pull.found)
					user_abort = 0;
				else
					ungot_key = full_key;
				modified = old_mod;
				return;
			case '?':
				help_on_command("pull-word");
				break;
			default:
				free_pl();
				/* Run the key that caused us
				 * to exit.
				 */
				ungot_key = full_key;
				return;
		}
	}
}

command pull_word_fwd() on reg_tab[NUMCTRL(KEYDOWN)]
{
	pull_word();
}

int pl_trans(int k)
{
	int r = reg_tab[k];
	if (r <= 0)
		r = reg_tab[generic_key(k)];
	if (r == (short) pull_word)
		return -1;
	else if (r == (short) pull_word_fwd)
		return 1;
	else if (r == (short) help)
		return '?';
	else if (r == (short) abort)
		return 7;
	else
		return 0;
}

int pl_getkey()
{
	for (;;) {
		switch (getkey()) {
		case MOUSE_MOVE:
		case GETFOCUS:
		case LOSEFOCUS:
		case LOST_SELECTION:
		case WIN_MOUSE_LEAVES:
		case WIN_MOUSE_ENTERS:
			break;
		default:
			return pl_trans(full_key);
		}
	}
}

// Are all the characters in this buffer region visible in the current
// window?

is_region_visible(int start, int end)
{
	int t, row, col;

	if (start > end){
		t = start;
		start = end;
		end = t;
	}

	if (get_window_pos(start, &row, &col) || get_window_pos(end - 1, &row, &col))
		return 0;	// If either end isn't visible, say no.

	save_var point = start;	// Only check line endings/beginnings.
	save_var narrow_start = start, narrow_end = size() - (end - 1);
	while (nl_forward())
		if (get_window_pos(point - 1, &row, &col)
			  || get_window_pos(point, &row, &col))
			return 0;
	return 1;
}

echo_context(int pos)
{
	save_var point = pos;

	if (echo_window_handle && is_window(echo_window_handle)) {
		remove_window(echo_window_handle);
	}
	save_var window_handle;
	echo_window_handle = add_popup(0, screen_lines - 1, screen_cols, 1,
								   0, bufnum);
	system_window = 1;
}



dismiss_echo_context()
{
	if (echo_window_handle && is_window(echo_window_handle)) {
		remove_window(echo_window_handle);
		return;
	}
}
