/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14.00 to 14.01.

#include "eel.h"
#include "menu.h"
#include "proc.h"

#define XWIN_XFLD_COUNT	13	/* # of fields in an X11 font logical descriptor */
int xwin_font_name_when_idle(int cnt);
int xwin_font_size_when_idle(int cnt);
int xwin_pixel_size_when_idle(int cnt);

// Make a list of fonts by running xlsfonts, and have the user narrow
// down the choices until just one is left. Return 0 if we can't get the
// list.
int set_font_for_unix()
{
	char fontname[1000], orig[1000], fontpat[1000], fontfamily[1000], msg[FNAMELEN];
	char fields[XWIN_XFLD_COUNT][1000], cust_point_sz[1000];
	jmp_buf *old_level = top_level, this_level;
	int i, mbuf = zap(_MATCH_BUF);
	struct screen_info retain_windows;

	retain_windows.wins = 0;
	save_screen(&retain_windows);
	strcpy(orig, font_fixed);
	*cust_point_sz = 0;
	for (i = 0; i < XWIN_XFLD_COUNT; i++)
		strcpy(fields[i], "*");
	save_var show_status = 0;
	note("Listing fonts...");
	int font_list = zap("-font-list");

	// First, extract the foundry+family of cur font.
	buf_printf(font_list, "%s\n", font_fixed);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	grab_full_line(mbuf, fontfamily);

	buf_zap(font_list);
	pipe_text(NULL, "-font-list", "if [ -x /opt/x11/bin/xlsfonts ] ; "
			  "then /opt/x11/bin/xlsfonts ; else xlsfonts ; fi", NULL,
			  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
	save_var bufnum = font_list;	// Got a list of all available fonts.
	if (lines_between(0, size(), 0) < 5) {
		say("Couldn't run xlsfonts to list the fonts.");
		return 0;
	}
	top_level = &this_level;
	if (setjmp(top_level)) {	// On abort, come here.
		font_fixed = orig;
		restore_screen(&retain_windows);	// Restore windows, since size changes
		free(retain_windows.wins);			// might have deleted some.
		top_level = old_level;
		quick_abort();
	}
	xwin_set_font_first_prune();	// Drop fonts we know we can't use.
	note("");

	// Extract the foundry+family names from font_list.
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	xwin_prune_to_usable_fonts(_MATCH_BUF);
	strcpy(msg, "Select a font:");
	if (*fontfamily)
		sprintf(msg, "Select a font [now %s]:", fontfamily);
	xwin_prompt_for_field(fields[0], msg, " Select a font ",
						  fontfamily, xwin_font_name_when_idle);

	// Now make a list of allowed point sizes for the selected font
	// family. Zero means any size is OK.
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 7, fontpat);
	xwin_font_size_prettify(_MATCH_BUF);

	i = xwin_prompt_for_field(fields[6], "Select a point size:",
							  " Point Size ", "12",
							  xwin_font_size_when_idle);

	xwin_font_size_unprettify(fields[6]);
	// If point size is 0 (custom), prompt for it and remember it.
	if (!strcmp(fields[6], "0")) {
		if (i)
			get_string(cust_point_sz, "Point size, such as 12 or 14: ");
		else
			get_strdef(cust_point_sz, "Font lacks recommended point size. "
					   "Please pick one", "14");
		xwin_font_size_fix_user(cust_point_sz);
	}

	// Ask the user to choose among other options the font family might
	// include.
	xwin_query_one_field(11, fields, "Select a char set (iso8859-1 "
						 "recommended, or iso10646-1 for Unicode):",
						 " Character Set ", "iso8859-1");
	xwin_query_one_field(1, fields, "Make all characters bold?",
						 " Force Bold? ", "medium");
	xwin_query_one_field(2, fields, "Make all characters italic (o) not regular (r)?",
						 " Force Italic? ", "r");

	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 6, fontpat);
	xwin_prompt_for_field(fields[5], "Select a pixel size:", " Pixel size ", "",
						  xwin_pixel_size_when_idle);

// any other fields?

	save_var resize_menu_list = 1;
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, -1, fontpat);

	if (*cust_point_sz)
		xwin_add_point_size(_MATCH_BUF, cust_point_sz);
	xwin_prompt_for_field(fontname, "These fonts match. Please pick one:",
						  " Font names ", "", xwin_font_name_when_idle);

	note("");
	aside("Switch to %s if %d", fontname, !user_abort);
	font_fixed = user_abort ? orig : fontname;
	if (!user_abort && strcmp(font_fixed, fontname))
		say("Sorry, can't use that font.");

	restore_screen(&retain_windows);
	free(retain_windows.wins);
	top_level = old_level;
	return 1;
}

repeat_search(cnt, flags)	 // search & report errors
{
	int orig, ret = 0, wrap = (search_wraps & 2) ? 1 : 0;
	char *s;

	s = default_search_string(flags);
	while (cnt-- > 0) {
		if (wrap == 2)      // Just wrapped, no more wrapping.
			wrap = 0;
		else
			orig = point;
		ret = do_color_searching(flags, s);
		if (ret == 1)
			continue;
		if (ret == DSABORT)
			say("Search cancelled.");
		else if (wrap == 1 && ask_wrap(flags)) {
			wrap = 2;
			point = (flags & REVERSE) ? size() : 0;
			say("Search wrapped.");
			cnt++;
			continue;
		} else {
			if (ret == DSBAD)
				say("%s: bad regex pattern.", s);
			else
				say("%s: not found.", s);
			point = orig;
			if (in_macro(1))
				quick_abort();
		}
		break;
	}
	if (search_positions_at_start && ret == 1)
		point = matchstart;
}

do_push(cmdline, cap, show)
char *cmdline;	// cap nonzero means capture output, show means show it to user
{		// but show -1 means don't wait for keystroke
	char dir[FNAMELEN], old[FNAMELEN];
	int result, tbuf;

	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;
	if (is_unix == IS_UNIX_XWIN && !*cmdline) {
		cmdline = push_cmd_unix_interactive;
		if (is_unix_flavor == IS_UNIX_MACOS)
			cmdline = push_cmd_unix_macos_interactive;
	}
	iter = 1;
	before_push();
	if (start_process_in_buffer_directory) {
		getcd(old);
		get_buffer_directory(dir);
		chdir(dir);
	}
	result = shell("", cmdline, cap ? bufnum_to_name(tbuf = tmp_buf())
									: "", SHELL_COPY_OUTPUT);
	if (should_restore_process_directory())
		chdir(old);
	if (cap) {
		process_captured(tbuf, cmdline);
		buf_delete(tbuf);
	}
	after_push();
	if (cmdline[0] && cap && show) {
		locate_window(PROCBUF, "");
		point = size();
	}
	if (result == -1 && errno) {
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
	maybe_restart_concurrent();
	if (cmdline[0] && result)
		say("Process returned %d", result);
	else
		check_dates(0);
	return result;
}

when_loading()
{
	if (!strcmp(push_cmd_unix_macos_interactive,
				"(xterm || /opt/X11/bin/xterm) &"))
		strcpy(push_cmd_unix_macos_interactive,
			   "if command -v xterm >/dev/null; then xterm; else /opt/X11/bin/xterm; fi &");
}
