/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2021 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14.02 to 14.03.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "diff.h"
#include "html.h"
#include "perl.h"
#include "python.h"

// Is there some inheritance info after pos? See parse_other_class().
int have_other_class(int pos)
{
	save_var point = pos;
	return parse_other_class();
}

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN], prev[FNAMELEN];
	int class_start = point, i, prevpos, thispos;

	*last = *prev = 0;	// See if "class foo" is before this.
	prevpos = 0;
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (c_ignore_declspec())
			continue;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, (c_extra_keywords & (CSHARP_KEYWORDS
				| JAVA_KEYWORDS | JS_KEYWORDS))
							  ? "[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_:]+", this))
				return 0;
			thispos = matchstart;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")
				  || !strcmp(this, "enum") || !strcmp(this, "__interface")
				  || !strcmp(this, "interface")) {
				if (class) {
					if (*class && ((display_c_options & 1)
								   || strcmp(type, "namespace"))) {
						strcat(class, c_class_separator());
						class += strlen(class);
					}
					if (*prev && !have_other_class(prevpos))	// Found "class foo bar {", use foo
						strcpy(class, prev);			// (unless it's "class foo : bar {" etc.)
					else
						strcpy(class, last);	// Is place where we started within func?
				}
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(prev, last);
			prevpos = thispos;
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

// Is point in a comment (C++ style or block) or a string or character
// constant? Return codes for each, or 0. We cache this info in a tagged
// region, and scan for it from the start of the buffer when needed.
// Note that the opening characters of the item, such as /* or ", are
// not considered part of the item and return 0.

in_c_comment(int lookback)
{
	UNREFERENCED_PARAMETER(lookback);
	if (in_perl_buffer)
		return in_perl_comment_body(point);
	int pos = point, from, to, res;
	save_var narrow_start = 0, narrow_end = 0;
	if (modified_buffer_region(&from, &to, "c-lex")) {
		// First, discard region modified since last scan.
		from = c_lex_safe_start(from);
		set_tagged_region("c-lex", from, size(), -1);	// Set to unknown.
	}
	reset_modified_buffer_region("c-lex");

	res = get_tagged_region("c-lex", pos, &from, &to);
	if (res == -1)		// No cached data, so lex it.
		c_lex_region(from, pos);
	res = get_tagged_region("c-lex", pos, &from, &to);
	if (pos >= size())		// Special rules for EOF, as we don't cache that.
		res = c_lex_eof_rules(res, from, to);
	if (res == -1) {
		say("Internal error in in_c_comment: %s at %d", bufname, pos);
		return 0;
	}
	return res;
}

to_statement_start()
{
	int lastcont, count = 100;	/* give up after this many lines */
	int orig = point;

	to_begin_line();
	if (parse_string(1, "[ \t\f]*}")) {
		search(1, "}");
		c_rev_skip_level('{', '}');
	}
	if (!is_js_object_literal(point)) {
		do {
			to_begin_line();
			lastcont = point;
			if (parse_string(1, "[ \t\f]*},")	// initializer
				  || (c_extra_keywords & JS_KEYWORDS) && parse_string(1, "<hspace>*}")) {
				search(1, "}");
				c_rev_skip_level('{', '}');
			}
		} while (!cind_try_space() && is_contin_statement(point, 0, orig)
				 && count-- > 0);
		point = lastcont;
	}
	to_indentation();
	if (curchar() == '#')
		past_initial_preprocessor();
}

char *html_comment_begin_override()
{
	save_var narrow_start = 0, narrow_end = 0;
	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point)) {
		case HT_PHP:
			int start;
			save_var point;
			point += parse_string(1, "<hspace>*");
			if (get_character_color(point, &start) == color_class perl_comment)
				if (matches_at(start, 1, "//"))
					return "// ";
				else if (character(start) == '#')
					return "# ";
			switch (php_comment_style) {
				case 1: return "# ";
				case 2: return "// ";
			}
			return "/* ";
		case HT_JSCRIPT:
			return new_c_comments ? "// " : "/* ";
		case HT_VBSCRIPT:
			return "' ";
		case HT_PYTHON:
			return "# ";
		case HT_CSS:
			return "/* ";
	}
	return 0;
}

/* Delete whitespace, then indent.  If the indentation is exactly
 * right already, don't alter the buffer at all. */
indent_to_column(int place)
{
	point -= parse_string(-1, "[ \t]*");
	int tabs = 0, spaces = 0, i, start = current_column();
	int startcolor = isspace(curchar()) ? get_character_color(point) : -1;
	if (indent_with_tabs)	// First see how many tabs and spaces to use.
		while ((i = start + tab_size - start % tab_size) <= place) {
			tabs++;
			start = i;
		}
	spaces = place - start;
	while (tabs > 0 && curchar() == '\t')	// While buffer contents is
		tabs--, point++;		// exactly right, just skip past.
	if (!tabs)
		while (spaces > 0 && curchar() == ' ')
			spaces--, point++;
	int more_spaces = parse_string(1, "[ \t]*");
	if (!tabs && !spaces && !more_spaces)
		return;		// We didn't have to do anything.

	// If the buffer hasn't changed since its last syntax highlighting,
	// maybe ensure that this change in indenting doesn't force
	// recoloring.  Only do for indenting at start of line.  This
	// provides a big speed improvement for indent-region, in modes
	// where indenting logic looks at syntax highlighting.
	int from, to;
	int ok = (coloring_flags & COLOR_IGNORE_INDENT)
			 && !modified_buffer_region(&from, &to, "needs-color")
			 && parse_string(-1, "\n[ \t]*");

	from = point;
	delete(point, point + more_spaces);
	while (tabs-- > 0)
		insert('\t');
	while (spaces-- > 0)
		insert(' ');
	if (ok) {
		reset_modified_buffer_region("needs-color");
		// If we won't recolor normally, set color of the inserted
		// indentation suitably, since some modes use coloring to detect
		// syntax elements.
		set_character_color(from, point, startcolor);
	}

	if (current_column() != place) {	// Shouldn't be needed, but be safe.
		delete_horizontal_space();
		insert_to_column(current_column(), place);
	}
}

// Return a code if we're inside a string or comment, but don't count
// the opening character as part of it.
int in_perl_comment_body(int pos)
{
	int res = in_perl_comment(pos), from, to;
	if (res) {
		get_character_color(pos, &from, &to);
		if (pos <= from + 1)
			return 0;
	}
	return res;
}

char *tag_c_ident_pattern();

/* We've parsed "class foo" but there's no { character right after.
 * See if there's some inheritance information we can skip past to find
 * the { character.
 * Return nonzero if we found ": foo {" or similar (C++).
 * Return nonzero if we found Java's "extends foo implements bar {" or similar.
 * In either case, move before the final { character.  Otherwise return
 * zero.
 */
parse_other_class()
{
	c_fwd_space();
	if (is_java_buffer() && curchar() == '<')
		c_skip_level('<', '>');
	c_fwd_space();
	if (curchar() == ':') {
		point++;
		for (;;) {
			c_fwd_space();
			if (parse_string(1, tag_c_ident_pattern())
				  || parse_string(1, "[,:]"))
				point = matchend;
			else if (!find_template_end())
				break;
		}
		c_fwd_space();
	} else if (c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) {
		while (strchr("ei", curchar())
			   && parse_string(1, "(extends|implements)</word>")) {
			point = matchend;
			skip_identifier_list();
		}
	}
	return curchar() == '{';
}
