/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2021 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14.03 to 14.04.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "diff.h"
#include "html.h"
#include "perl.h"
#include "python.h"

c_func_name_finder()
{
	if (!strfcmp(get_extension(filename), ".json")) {	// Disable this way for now
		*display_func_name = 0;
		return 1;
	}
	return find_c_func_name(display_func_name, !must_find_func_name);
}

is_c_keyword(p)		// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr("|if|else|for|do|while|return|break|continue|switch|"
			   "case|default|goto|char|short|int|static|volatile|const|"
			   "float|double|long|void|enum|", p))
		return 1;
	if (p[1] == '_' &&
		  strstr("|_asm|_based|_cdecl|_emit|_except|_export|_far|_fastcall|"
				 "_fortran|_huge|_interrupt|_loadds|_near|_pascal|_saveregs|"
				 "_segment|_segname|_stdcall|_syscall|_try|__int64|", p))
		return 1;
	if (!(c_extra_keywords & (EEL_KEYWORDS | CPP_KEYWORDS | JAVA_KEYWORDS |
							  JS_KEYWORDS | IDL_KEYWORDS | CSHARP_KEYWORDS | OBJC_KEYWORDS))
		  && strstr("|restrict|_Alignas|_Alignof|_Atomic|_Bool|_Complex|"
					"_Generic|_Imaginary|_Noreturn|_Static_assert|_Thread_local|"
					"_Decimal128|_Decimal32|_Decimal64|", p))
		return 1;		// C language only
	if ((c_extra_keywords & EEL_KEYWORDS) &&
		  strstr("|keytable|buffer|window|command|on|color_class|"
				 "spot|save_spot|save_var|on_exit|user|zeroed|byte|int32|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("|asm|catch|class|delete|friend|inline|new|operator|"
				 "private|protected|public|template|try|this|virtual|"
				 "throw|_self|and|and_eq|bitand|bitor|bool|compl|const_cast|"
				 "dynamic_cast|explicit|false|mutable|namespace|not|not_eq|"
				 "or|or_eq|reinterpret_cast|static_cast|true|typeid|typename|"
				 "using|wchar_t|xor|xor_eq|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("|alignas|alignof|char8_t|char16_t|char32_t|concept|consteval|"
				 "constexpr|constinit|co_await|co_return|co_yield|decltype|"
				 "export|noexcept|nullptr|requires|static_assert|thread_local|", p))
		return 1;		// Later C++ keywords
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("|override|final|import|module|", p))
		return 1;		// C++ "identifiers with special meaning"
	if ((c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
		  strstr("|abstract|assert|boolean|byte|catch|class|extends|false|final|"
				 "finally|function|implements|import|in|instanceof|"
				 "interface|native|new|null|package|private|protected|"
				 "public|strictfp|super|synchronized|this|throw|throws|transient|"
				 "true|try|var|with|", p))
		return 1;
	if (!(c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
		  strstr("|auto|extern|register|unsigned|signed|sizeof|"
				 "struct|union|typedef|", p))
		return 1;		// C keywords not in Java.
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		  strstr("|appobject|auto_handle|bindable|boolean|broadcast|byte|"
				 "callback|coclass|code|comm_status|context_handle|"
				 "control|decode|defaultbind|dispinterface|displaybind|"
				 "dual|enable_allocate|encode|error_status_t|"
				 "explicit_handle|fault_status|handle|handle_t|hidden|"
				 "hyper|idempotent|ignore|immediatebind|import|in|"
				 "include|interface|library|licensed|local|maybe|module|", p))
		return 1;
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		  strstr("|ms_union|nocode|nonextensible|notify|object|odl|"
				 "oleautomation|optional|out|pipe|propget|propput|"
				 "propputref|public|ptr|readonly|ref|requestedit|"
				 "restricted|retval|small|source|string|unique|"
				 "usesgetlasterror|v1_enum|vararg|", p))
		return 1;
	if ((c_extra_keywords & CSHARP_KEYWORDS) &&
		  strstr("|abstract|as|base|bool|byte|catch|checked|class|decimal|delegate|"
				 "event|explicit|false|finally|fixed|foreach|implicit|in|"
				 "interface|internal|is|lock|namespace|new|null|object|"
				 "operator|out|override|params|private|protected|public|"
				 "readonly|ref|sbyte|sealed|sizeof|stackalloc|string|this|"
				 "throw|true|try|typeof|uint|ulong|unchecked|unsafe|ushort|"
				 "using|virtual|", p))
		return 1;
	if ((c_extra_keywords & OBJC_KEYWORDS) &&
		  strstr("|self|super|BOOL|id|Class|nil|YES|NO|NS_DURING|NS_HANDLER|"
				 "NS_ENDHANDLER|in|out|inout|bycopy|byref|oneway|@interface|"
				 "@implementation|@end|@class|@private|@protected|@public|"
				 "@encode|@protocol|@selector|@try|@catch|@finally|@throw|"
				 "@synchronized|Nil|IBAction|IBOutlet|", p))
		return 1;
	return 0;
}

int c_maybe_break_line(int mode)
{
	switch (mode) {
		case 0:
		case 1:
			c_auto_fill_mode = (mode != 0);
			return 0;
		case 2:
			return c_auto_fill_mode;
	}
	if (!c_auto_fill_mode)		// We're not in auto-fill mode.
		return 1;
	return do_maybe_break_line(1);
}

// Is the character just before point in a comment or string? Return
// its code.
int get_prev_char_comment_code(int c_like)
{
	save_var point--;
	return c_like ? in_c_comment(c_look_back) : in_comment();
}

int do_maybe_break_line(int c_like)
{
	int comm, res = 1, starts_with_comm, starts_in_comm, after_prefix;
	int comm_start = -1, break_block_comment = 0;
	int generic = strcmp(major_mode, "C");
	char fill[FNAMELEN], pat[FNAMELEN], raw_comm[FNAMELEN];
	spot after, next_ends = 0;

	delete_if_highlighted();
	if (c_like && !in_perl_buffer)
		generic = 0;

	// Various reasons we might want not want to fill right now.
	if (give_end_line() - give_begin_line() > FNAMELEN - 10)
		return 1;			// The line's too long for us.
	if (IS_ENTER_KEY(key)
		  && get_prev_char_comment_code(c_like) == IN_OLD_COMMENT
		  && (auto_fill_comment_rules & 1))
		break_block_comment = 1, res = 0;
	else if (current_column() + 1 <= margin_right)
		return 1;		// We're not past the margin.
	comm = get_prev_char_comment_code(c_like);
	if (!(comm & IN_COMMENT))
		return 1;			// We're not in a comment.

	// Collect info on the situation.
	if (key == ' ') {		// Insert <Space> now, then remember not
		insert(key);		// to later, so we can break here.
		res = 0;
	}
	after = alloc_spot(1);	// Remember where we should end up.
	to_begin_line();	// See if a comment's the first thing on this line.
	strcpy(raw_comm, get_comment_start());
	sprintf(pat, "[ \t]*%s", generic ? raw_comm : "/[*/]");
	starts_with_comm = parse_string(1, pat);	// Yes, comment's first on line.
	starts_in_comm = c_like ? in_c_comment(c_look_back) : in_comment();
	// Does this line start with a comment?
	sprintf(pat, "[ \t%s]*", generic ? raw_comm : "*/");
	parse_string(1, pat, fill);	// Repeat this prefix on new lines.
	after_prefix = point;
	if (!starts_in_comm) {		// Find where this line's comment starts.
		to_begin_line();
		sprintf(pat, "<min>.*%s",
				generic ? raw_comm : comm == IN_NEW_COMMENT ? "//" : "/%*");
		point += parse_string(1, pat);
		comm_start = point;
	}
	if (nl_forward()) {		// Remember the end of the next line, unless
		to_end_line();		// there's a comment there.
		if (c_like && !in_c_comment(c_look_back))
			next_ends = alloc_spot(0);
	}

	// Look for a breaking point.
	point = *after;
	if ((c_like && c_auto_fill_mode != 1 || starts_with_comm || starts_in_comm)
		  && (auto_fill_comment_rules & 2) && !break_block_comment
		  && pick_break(margin_right) && point > after_prefix
		  && point > comm_start)
		break_block_comment = 1;

	if (break_block_comment)
		comm_break_line_here(c_like, comm, starts_with_comm, starts_in_comm,
							 comm_start, next_ends, after, fill);

	point = *after;
	free_spot(after);
	if (next_ends)
		free_spot(next_ends);
	return res;
}

do_html_indent(int manual)
{
	int region, start;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 200);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
												: html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1))) {
		if (!(var & 0x4000))
			prev_indenter();
		return;
	}
	save_var narrow_start = start;
	if (region > 0 && region <= HT_CSS && !IS_ENTER_KEY(key)
		  && !(html_indenting_rules & (1 << (region - 1)))) {
		prev_indenter();	// Can move back to prev line's indent.
		return;
	}
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	if (region >= HT_JSCRIPT && region < HT_IGNORE) {
		if (all_comments(narrow_start, point, region)) {	// Start of script block.
			narrow_start = 0;	// 0x400 indent; 0x800 indent even if <% has none.
			int i = get_indentation(start);
			if (!start || start == point)
				i = 0;
			else if ((html_indenting_rules & 0x400)
					 && ((html_indenting_rules & 0x800) || i))
				i += give_html_indent_by_type(region);
			indent_to_column(i);
			return;
		}
		html_exclude_script_head();
	}
	switch (region) {
		case HT_JSCRIPT:
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			do_vbasic_indent();
			return;
		case HT_PHP:
			do_php_indent();
			return;
		case HT_PYTHON:
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}
