/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 B9 to 14 B10.

#include "eel.h"
#include "c.h"
#include "chelp.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "hex.h"
#include "info.h"
#include "makemode.h"
#include "misclang.h"
#include "rmode.h"
#include "perl.h"
#include "proc.h"
#include "session.h"
#include "tags.h"
#include "tex.h"

char _blist_mode_name[] = "Bookmark-List";
#define BMARK_LIST	"-bookmark-list"
int _blist_window, _blist_orig_window;

make_bookmark_list(op)
char *op;
{
	int i, orig = bufnum, list, count = 0, new;

	iter = 0;
	list = zap(BMARK_LIST);
	for (i = 10; i < NUM_BOOKMARKS; i++)
		count += list_a_bookmark(list, i);
	for (i = '0'; i <= '9'; i++)
		count += list_a_bookmark(list, key_to_bookmark(i));
	bufnum = orig;
	if (!count) {
		remove_blist();
		error("No bookmarks defined");
	}
	new = (window_handle != _blist_window);
	if (new) {
		_blist_orig_window = window_handle;
		remove_blist();
		_blist_window = add_popup(2, 2, screen_cols - 4,
								  screen_lines - 6, ALL_BORD(BSINGLE), list);
		mouse_cursor = left_pointer;
	} else
		window_bufnum = list;
	build_first = 1;
	point = 0;
	buf_delete_final_newlines(list);
	text_color = color_class text;
	display_column = 0;	/* horizontally scroll this window */
	window_title(_blist_window, BTOP, TITLECENTER, " Bookmarks ");
	display_more_msg(_blist_window);
	make_line_highlight();	/* have current line always be highlighted */
	goal_column = 0;	/* next-, prev-line stay in col 0 */
	popup_near_window(_blist_window, _blist_orig_window);
	if (is_gui && want_gui_prompts && new)
		one_window_to_dialog("Select a Bookmark", _blist_window,
							 "Go", "Cancel", "Help");
	mode_keys = blist_tab;
	major_mode = _blist_mode_name;
	mouse_dbl_selects = 1;
	make_mode();
	say("Select a bookmark to %s.", op);
	modified = 0;
}

// Copy destfile to res, but if it's a directory, append basename of src to it.
make_copy_name(char *src, char *destfile, char *res)
{
	strcpy(res, destfile);
	if (url_is_a_directory(res)) {
		destfile = res + strlen(res);
		if (*res && !is_path_separator(destfile[-1]))
			*destfile++ = is_remote_file(destfile) ? '/' : path_sep;
		strcpy(destfile, get_tail(src, 0));
	}
}

// Like is_directory() but works with remote files. Can be slow if we
// have to ask the remote system.
int url_is_a_directory(char *s)
{
	switch (remote_file_type(s)) {
		case 1: return is_directory(s);
		case 4:
		case 6:
		case 7:
		case 8:	return 1;
	}
	return 0;
}

/* Ask for a file name and save buffer to that file if one is given. */
command write_file()
{
	char tmp[FNAMELEN], tmp2[FNAMELEN];
	int flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;

	iter = 0;
	if (use_common_file_dialog()) {
		strcpy(tmp, force_save_as == FSA_NEWFILE ? "" : filename);
		if (!use_common_open_file_dlg(tmp, "Save file as", &flags, 1))
			return 2;
	} else
		get_file_dir(tmp, "Write file: ");
	if (*tmp && url_is_a_directory(tmp) && *filename) {
		make_copy_name(filename, tmp, tmp2);
		strcpy(tmp, tmp2);
	}
	if (*tmp) {
		if (!is_remote_file(tmp) && get_file_read_only(tmp) == 1)
			error("File %s is read-only.", get_tail(tmp, 0));
		switch (warn_existing_file(tmp)) {
			case 1: return 0;
			case -1: save_var file_write_newfile = 1;
		}
		if (has_arg) {
			translation_type = give_line_translate(tmp);
			make_mode();
		}
		set_buffer_filename(tmp);
		already_made_backup = 0;
		modified = 1;
		return do_save_file(1, 0, 1);
	}
	say("Buffer %s not saved, no file name.", bufname);
	return 2;
}

command copy_to_file()
{
	int flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;
	char tmp[FNAMELEN], tmp2[FNAMELEN], orig[FNAMELEN];

	iter = 0;
	strcpy(orig, filename);
	if (use_common_file_dialog()) {
		strcpy(tmp, force_save_as == FSA_NEWFILE ? "" : filename);
		if (!use_common_open_file_dlg(tmp, "Copy to file", &flags, 1))
			return;
	} else
		get_file(tmp, "Copy to file",
				 force_save_as == FSA_NEWFILE ? "" : filename);
	if (*tmp && url_is_a_directory(tmp) && *filename) {
		make_copy_name(filename, tmp, tmp2);
		strcpy(tmp, tmp2);
	}
	switch (warn_existing_file(tmp)) {
		case 1: return;
		case -1: save_var file_write_newfile = 1;
	}
	filename = tmp;
	save_var already_made_backup = 0, force_save_as = 0;
	save_var synchronous_file_io = 1;	// Don't revert filename until we've reported status.
	save_var modified;
	on_exit filename = orig;
	do_save_file(1, 0, 0);
}

// Maybe warn if file s already exists.  Return 1 to abort, 0 if user
// said to go ahead, or turned off checking, or -1 if file didn't
// exist when we checked.  (If possible, cancel writing if it appears
// later.)
int warn_existing_file(char *s)
{
	char msg[FNAMELEN];

	if (is_remote_file(s))
		inet_only();
	if (!warn_before_overwrite || warn_before_overwrite == 2 && is_remote_file(s))
		return 0;
	if (is_remote_file(s)
		  ? !remote_file_match(s, STARTMATCH | EXACTONLY)
		  : (check_file(s, NULL) != CHECK_FILE))
		return -1;
	sprintf(msg, "Overwrite %r? ", s);
	if (!ask_yn("Confirm overwrite", msg, "Overwrite", "Cancel", 1))
		return 1;
	return 0;
}

when_loading()
{
}
