/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 B10 to 14 B11.

#include "eel.h"
#include "bufed.h"
#include "c.h"
#include "chelp.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "hex.h"
#include "info.h"
#include "makemode.h"
#include "misclang.h"
#include "rmode.h"
#include "perl.h"
#include "proc.h"
#include "session.h"
#include "tags.h"
#include "tex.h"

do_switch_buffers(int orig_key)
{
	char bname[FNAMELEN];
	int other_key = -1;

	// If the key that invoked us is a function key, numpad or cursor
	// pad, or "gray key", find its shifted equivalent and look for that
	// also.  Usually Shift+orig key will give us the original key code,
	// if the original included Ctrl, but sometimes it might give us a
	// different one.
	if (!IS_EXT_ASCII_KEY(orig_key) && IS_TRUE_KEY(orig_key))
		other_key = NUMSHIFT(NUMPLAIN(orig_key));
	error_if_input(1);
	tiled_only();
	if (!(switch_buffers_options & 1))
		save_var _bufed_sort = 'a';
	if (!(switch_buffers_options & 2))
		save_var bufed_grouping = 0;
	do_bufed(BUFED_SWITCH | (has_arg ? BUFED_ALL : 0));
	move_in_switch_buffers(shift_pressed() & KB_SHIFT_DN);
	while ((shift_pressed() & KB_CTRL_DN)) {
		maybe_refresh();
		if (char_avail()) {
			getkey();
			if (key == GETFOCUS || key == LOSEFOCUS) {
				do_topkey();
				continue;
			} else if (key == orig_key || key == other_key)
				move_in_switch_buffers(shift_pressed() & KB_SHIFT_DN);
			else if (binding_of(NUMPLAIN(key), root_keys) == (short) down_line)
				move_in_switch_buffers(0);
			else if (binding_of(NUMPLAIN(key), root_keys) == (short) up_line)
				move_in_switch_buffers(1);
		} else
			delay(10, COND_KEY | COND_RETURN_ABORT);
		if (user_abort || key == abort_key || key == ESC
			  || key == NUMCTRL(GREYESC)) {
			while (char_avail())	// Discard keys so we don't get a delayed
				getkey();			// Canceled msg.
			user_abort = 0;
			remove_bufed();
			return;
		}
	}
	if (!bufed_getname(bname))
		*bname = 0;
	remove_bufed();
	if (*bname && exist(bname))
		to_buffer(bname);
}

run_topkey()		/* call do_topkey() with appropriate setup */
{
	jmp_buf this_level;

	iter = 1;
	has_arg = this_cmd = table_count = 0;
	undo_mainloop();	/* make sure undoing is on for this buffer */
	save_var top_level = &this_level;
	if (!setjmp(top_level))
		do_topkey();
	prev_cmd = this_cmd;
}

do_set_show_graphic(mode)	// set graphic rep by mode code
{				// 0: multi-char, 1: graphic, 2: hexadecimal, 3: int'l
	int i, std, stdctrl;

	if (show_spaces && _display_class == _show_disp_class)
		save_var _display_class = _std_disp_class;
	std = (mode == 0) ? BM : (mode == 2) ? BHEX : BNORMAL;
	stdctrl = (mode == 3) ? BC : std;
	for (i = 0; i < 128; i++)
		if (i < 32 || i == 127) {
			if (i != '\n' && i != '\t')
				_display_class[i] = mode ? stdctrl : BC;
			int replacement = set_character_property(i + 128, CPROP_DEF_REPL, -1);
			int dispwidth = set_character_property(replacement, CPROP_DISP_WIDTH, -1);
			if (mode == 0 || dispwidth == 0 && mode != 2)
				_display_class[i + 128] = BMC;
			else if (mode != 2 && replacement != i + 128
					 && set_character_property(i + 128, CPROP_DISP_WIDTH, -1) == 0)
				// If orig char was zero-width but repl isn't, use repl.
				_display_class[i + 128] = replacement;
			else
				_display_class[i + 128] = std;
		} else {
			_display_class[i] = BNORMAL;
			_display_class[i + 128] = std;
		}
	for (i = 256; i  < 65536; i++)
		_display_class[i] = (mode == 2) ? BHEX : BNORMAL;
	fix_display_class();
	make_show_spaces();
	full_redraw = 1;
}

update_display_class()
{
	do_set_show_graphic(get_show_graphic());
}

// Prompt for name of encoding.  Pass 0 to include auto-detect, 1 to
// skip it.
int ask_encoding(char *pr, int i)
{
	char conv[FNAMELEN], *s;

	zap(_MATCH_BUF);
	save_var recall_id = "encoding";
	for (; s = encoding_to_name(i); i++)
		buffer_printf(_MATCH_BUF, "%s\n", s);
	s = encoding_to_name(GET_ENCODING(translation_type));
	prompt_comp_read(conv, pr, general_matcher, COMP_FOLD, s ? s : "");
	if ((i = encoding_from_name(conv)) >= 0)
		return i;
	say("Unknown encoding %s", conv);
	return -1;
}

do_set_font(char *prompt, int type, char *varname)
{
	if (has_feature & FEAT_FONT_DIALOG)
		windows_set_font(prompt, type);
	else if (has_feature & FEAT_SET_FONT)
		do_set_variable(find_index(varname));
	else
		error("Setting %s is not supported in this environment.", varname);
	update_display_class();
}

when_setting_font_fixed()
{
	update_display_class();
}

tag_a_file(char *s)
{
	jmp_buf this_level;
	int ok = 1, err = 0, orig = bufnum, temp = 0, oldmod, aborted = 0;

	bufnum = _tags_buffer;	/* go to right place in tags buffer */
	point = 0;
	nl_forward();
	oldmod = modified;	/* want to see if we add new tags */
	modified = 0;
	bufnum = orig;
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		if (!look_file(s)) {
			if (!temp)
				temp = tmp_buf();
			bufnum = temp;
			save_var readonly_warning = 0;
			err = read_file(s, get_default_translation_type(s));
			filename = s;
			call_mode(filename);
			make_mode();
			check_filevars();
		}
		if (!err) {
			save_var point = 0;
			ok = find_tagger(1);
		}
	} else
		aborted = 1;
	bufnum = _tags_buffer;
	if (modified)		/* we added new tags, mark unsorted */
		mark_tags(0);
	else
		modified = oldmod;
	bufnum = orig;
	if (temp)
		buf_delete(temp);
	restore_vars();
	if (err == EREADABORT)
		aborted = 1, err = 0;
	if (err && tag_batch_mode)
		buffer_printf("-tag-errors", "Couldn't read the file %s.\n", s);
	else if (err)		/* couldn't read file */
		quick_abort();	/* already showed error msg */
	else if (!ok && tag_batch_mode)
		buffer_printf("-tag-errors", "Don't know how to tag the file %s.\n", s);
	else if (!ok)
		error("Don't know how to tag the file %s", s);
	return aborted || user_abort;
}
