/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 b1 to 14 b2.

#include "eel.h"
#include "epsnet.h"

// If there's a directory name before an env variable substitution,
// remove the directory name.
delete_file_path_before_subst()
{
	char pat[FNAMELEN], var[FNAMELEN];
	if (opsys == OS_WINDOWS)
		sprintf(pat, "((%s|[a-z]):)?"	// optional a: or ftp:
			"[\\/].*[\\/]!"			// slash, stuff, slash
			"[%%](<word|@|(|)>+)[%%]",		// env var subst
			url_services);
	else
		sprintf(pat, "((%s):)?"	// optional ftp:
			"[\\/].*[\\/]!"			// slash, stuff, slash
			"[$](<word|@|(|)>+|{<word|@|(|)>+})",		// env var subst
			url_services);
	int res = parse_string(1, pat);
	if (res) {		// If it looks like it has an env var, grab it.
		grab(find_group(3, 1), find_group(3, 0), var);
		if (var[0] == '{') {		// Found {varname}, use varname.
			var[strlen(var) - 1] = 0;
			if (filename_subst_var(var + 1))
				delete(0, res);
		} else if (filename_subst_var(var))
			delete(0, res);
	}
}

// Return nonzero if any env var names start with p. If some do, and
// flags has STARTMATCH, fill _MATCH_BUF with the matching names,
// modified to start with s through p, and with final at the end.
int construct_env_var_name_matches(int flags, char *s, char *p, char *final)
{
	char pat[FNAMELEN];
	int b = tmp_buf();
	on_exit buf_delete(b);
	save_var bufnum = b;
	get_subst_var_list();
	save_var show_status = 0, case_fold = 1, point = 0;
	sprintf(pat, "^%s", p);				// Delete any names that don't start with p
	do_drop_matching_lines(REGEX, pat, 0);
	if (!size())
		return 0;
	if (flags & STARTMATCH) {
		simple_re_replace(1, "=.*", final);	// Replace values with end of match.
		strncpy(pat, s, p - s);
		simple_re_replace(1, "^", pat);		// Insert initial text at start of match.
		zap(_MATCH_BUF);
		xfer(_MATCH_BUF, 0, size());
	}
	return 1;
}

// Given a pointer to text starting with $ or %, return the end of the
// env var name portion, or null if it's only a partial name.
char *find_env_var_end(char *s)
{
	if (*s == '%')			// Windows-style: just look for delim.
		return strchr(s + 1, '%');
	else if (s[1] == '{')	// Found ${ so look for matching }.
		return strchr(s + 1, '}');
	for (s++; *s; s++)
		if (!isident(*s) && !strchr("@", *s))
			return s;
	return 0;
}

#define URL_PATTERN	"|((ftp|http|scp|telnet|ssh|https):[/\\][/\\][^\n\t >]+)"

default_find_linked_file(char *file)
{
	char *pat;

	save_var case_fold = 1;
	if (opsys == OS_UNIX)
		pat = "(%$({<ident>+}|<ident>+)/)?[+-9A-z~@]+[/.][+-9A-z@]*[--9+A-z]" URL_PATTERN;
	else
		pat = "(%%<ident>+%%[\\/])?([a-zA-Z]:)?[+-9A-z~]+<.>[+-9A-z]*[--9+A-z]" URL_PATTERN;
	point -= parse_string(-1, "[^ \t\n!-*]*");
	if (!re_search(1, pat)) {
		to_begin_line();
		if (!re_search(1, pat))
			error("Can't find any file name on this line.");
	}
	if (strchr("\"'`<", character(matchstart)))
		matchstart++;
	if (strchr(")", character(matchend - 1))
		  && strchr("\"'`.,>", character(matchend - 2)))
		matchend--;
	if (strchr("\"'`.,>", character(matchend - 1)))
		matchend--;
	grab(matchstart, matchend, file);
}

find_linked_file_here(char *file)
{
	char subr[FNAMELEN], curdir[FNAMELEN], mode[FNAMELEN], *s;
	int i = 0;

	save_var narrow_end = size() - give_end_line();
	if (*filename) {
		s = get_extension(filename);
		sprintf(subr, "%s_find_linked_file", *s ? (s + 1) : "none");
		i = find_index(subr);		// Any special method for this file extension?
	}
	if (!i) {
		clean_mode(mode);
		sprintf(subr, "%s_find_linked_file", mode);
		i = find_index(subr);		// Any special method for this buffer?
	}
	switch (name_type(i)) {		// i has name table index of subr/var
		case NT_SUBR: case NT_COMMAND:
		case NT_AUTOSUBR: case NT_AUTOLOAD:
			(*(int (*)())i)(file);		// Cast i to a func ptr, then call it.
			break;
		default:		// Look in file for something like a file name.
			default_find_linked_file(file);
			break;
	}
	get_buffer_directory(curdir);
	if (!is_remote_file(file) && file[0] != ((opsys == OS_UNIX) ? '$' : '%'))
		absolute(file, curdir);
}

// Insert the environment and any other subst variables and their
// values into the current buffer in the form VAR=VALUE, one line each.
get_subst_var_list()
{
	char dir[FNAMELEN];
	int buf = name_to_bufnum("-subst-vars"), orig = bufnum;
	if (buf) {
		grab_buffer(buf);
		return;
	}
	save_var bufnum = zap("-subst-vars");
	discardable_buffer = 1;				// Now get all env vars and values, plus shell folders.
	get_substitution_variables((file_interpret_env_vars & 16) ? 3 : 1);
	save_var point = 0;
	while (re_search(1, "^<word|@|(|)>+<^word|@|(|)|=|newline>")) {
		delete(point - 1, point);		// Remove invalid chars from the var names.
		to_begin_line();
	}
	point = 0;
	get_customization_directory(dir);
	bprintf("@ecustom=%s\n", dir);
	save_var show_status = 0;
	if (file_interpret_env_vars & 64)		// Strip vars that don't name a directory.
		do_drop_matching_lines(REGEX, is_mswindows
							   ? "^<^=>+=(\\\\|([a-z]+:)?[\\/])" : "^<^=>+=/", 0);
	if (file_interpret_env_vars & 32) {
		point = size();
		char *s = lookpath("subst_vars");
		if (s)
			do_insert_file(s, FILETYPE_AUTO);
	}
	buf_xfer(orig, 0, size());
}

// Get the value of this subst var from the list and return it, or 0 if
// none.
char *get_other_subst_var(char *var)
{
	char pat[FNAMELEN], value[FNAMELEN];
	int b = tmp_buf();
	on_exit buf_delete(b);
	save_var bufnum = b;
	get_subst_var_list();
	save_var show_status = 0, case_fold = 1, point = 0;
	sprintf(pat, "^%q=", var);
	if (!re_search(1, pat))
		return 0;
	grab(point, give_end_line(), value);
	return strkeep(value);
}

char *filename_subst_var_work_area;

// Return the value of a subst var, but if the result looks like a list
// of directories, make into an extended file pattern and return that.
char *filename_subst_var(char *var)
{
	char *s, *p;

	s = get_other_subst_var(var);
	char sep = (opsys == OS_WINDOWS) ? ';' : ':';
	if (s && (file_interpret_env_vars & 8) && strchr(s, sep)) {
		if (!filename_subst_var_work_area && ptrlen(filename_subst_var_work_area) < strlen(s) + 10)
			filename_subst_var_work_area = realloc(filename_subst_var_work_area, strlen(s) + 10);
		sprintf(filename_subst_var_work_area, "{%s", s);
		s = filename_subst_var_work_area;
		for (p = s; p = strchr(p, sep); p++) {
			if (p > s && is_path_separator(p[-1]) == 1) {
				strcpy(p - 1, p);		// Strip trailing / or \.
				p--;
			}
			*p = ',';			// Change each sep to comma.
		}
		strcat(filename_subst_var_work_area, "}");
		return filename_subst_var_work_area;
	}
	return s;
}

int filename_subst_envvars_unix(char *src, char *dest)
{
	char *s, *t, var[FNAMELEN], *val;
	int res = 0;
	while (s = strchr(src, '$')) {
		strncpy(dest, src, s - src);
		if (s[1] == '{') {
			if (t = strchr(s + 1, '}')) {
				strncpy(var, s + 2, t - s - 2);
				if (val = filename_subst_var(var)) {
					strcat(dest, val);
					src = t + 1;
					res = 1;
				} else {
					strncpy(dest, s, t - s);
					src = t + 1;
				}
			} else {
				strcpy(dest, s);
				src += strlen(src);
			}
		} else {
			for (t = s + 1; *t && (isident(*t) || strchr("@", *t)); )
				t++;
			strncpy(var, s + 1, t - s - 1);
			if (val = filename_subst_var(var)) {
				strcat(dest, val);
				src = t;
				res = 1;
			} else {
				strncpy(dest, s, t - s);
				src = t;
			}
		}
		dest = dest + strlen(dest);
	}
	strcpy(dest, src);
	return res;
}

// Find a file variable that matches this pattern; return nonzero if we
// did.  This can be on an -*- var: value -*- line at the start, or in a
// Local Variables section at the end.  If oldpat is non-NULL, it says
// what old-style pattern to look for, where just a value is present
// between -*- markers.
int look_for_filevar_pattern(char *var, char *value, char *oldpat)
{
	char fullpat[2000];

	save_var point = 0, case_fold = 1;
	save_var narrow_end = size() - 6000;	// Might have very long lines.
	if (parse_string(1, "#!"))	// Look in first line or two only.
		nl_forward();
	narrow_end = size() - give_end_line();
	point = 0;
	if (search(1, "-*-")) {	// Quick test first.
		point = 0;
		if (oldpat) {
			sprintf(fullpat, ".*-%%*-[ \t]*%s[ \t]*-%%*-", oldpat);
			if (re_search(1, fullpat))
				return 1;
		}
		point = 0;
		sprintf(fullpat, ".*-%%*-.*</word>%s:[ \t]*%s[ \t]*(;.*)?[ \t]*-%%*-",
				var, value);
		if (re_search(1, fullpat))
			return 1;
	}
	restore_vars();
	save_var point = size() - file_variables_search_limit, case_fold = 1;
	sprintf(fullpat, "</word>%s:[ \t]*%s", var, value);
	if (search(1, "Local Variables:") && re_search(1, fullpat))
		return 1;
	return 0;
}

// Try to parse a single element, either a number or a suitable
// variable name (perhaps with - before it), whose value we retrieve
// from buffer in_buf. On success, set *val to
// its value and return 1, else return 0.
int filevar_parse_value(int *val, int in_buf)
{
	char buf[FNAMELEN];
	if (parse_string(1, "-?(0x[0-9a-f]+|0o[0-7]+|ob[01]+|[0-9]+|nil</word>])", buf)) {
		*val = numtoi(buf);
		return 1;
	}
	if (!parse_string(1, "-?[a-z][-a-z0-9]*", buf))
		return 0;
	int i = find_index(buf + (buf[0] == '-'));
	if (!i || vartype_class(i) != 1)
		return 0;
	save_var bufnum = in_buf;
	*val = (buf[0] == '-') ? -get_num_var(i) : get_num_var(i);
	return 1;
}

// Check for file vars in the current buffer, then apply to target
// buffer.
check_buffer_filevars(int target)
{
	char buf[FNAMELEN], var[FNAMELEN], when_setting[FNAMELEN], *s;
	int i, cur = bufnum, val, val2, operator;

	save_var bufnum;
	if (look_for_filevar_pattern("tab-(size|width)", "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			tab_size = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(margin-right|fill-column)",
								 "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			margin_right = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(auto-)?fill-mode",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		call_maybe_break_line(i);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(indent-with-tabs|indent-tabs-mode)",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		indent_with_tabs = numtoi(buf);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(write-line-translate)",
								 "([0-9]|dos|windows|binary|unix|mac|auto)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		set_line_translate_from_text(buf);
		bufnum = cur;
	}
	save_var point = size() - file_variables_search_limit, case_fold = 1;
	if (!search(1, "Local Variables:"))
		return;
	while (re_search(1, "</word>([a-z]<hyphen|ident>*):[ \t]*!<hyphen|word>")) {
		grab(find_group(1, 1), find_group(1, 0), var);
		i = find_index(var);
		if (!i || name_type(i) != NT_BUFVAR || !name_user(i) || vartype_class(i) != 1) {
			aside("Ignoring file variable %s.", var);
			continue;	// Must be a buffer-specific numeric user var
		}
		s = name_name(i);
		if (word_in_list(s, "|fill-mode|indent-with-tabs|", 1))
			continue;		// These already received special handling above.
		if (!filevar_parse_value(&val, target)) {
			aside("Ignoring unrecognized value for %s", var);
			continue;
		}
		if (parse_string(1, "<hspace>*([-+*/])<hspace>*")) {
			operator = character(find_group(1, 1));
			point = matchend;
			if (filevar_parse_value(&val2, target))
				switch (operator) {
					case '+':		val += val2; break;
					case '-':		val -= val2; break;
					case '*':		val *= val2; break;
					case '/':		val /= val2; break;
				}
		}
		bufnum = target;
		set_num_var(i, val);
		sprintf(when_setting, "when-setting-%s", var);
		try_calling(when_setting);
		bufnum = cur;
	}
}

// Look for an .epsilon_vars file with settings for many files and apply
// its settings to the current buffer.  Do both global and per-directory
// files.
apply_vars_files()
{
	char varsfile[FNAMELEN], global[FNAMELEN], dir[FNAMELEN];

	*varsfile = 0;
	if (!stricmp(get_tail(filename, 1), ".epsilon_vars"))
		return;
	if (get_buffer_directory(dir))
		sprintf(varsfile, "%s.epsilon_vars", dir);
	get_customization_directory(global);
	strcat(global, ".epsilon_vars");
	apply_vars_file(global);
	if (*varsfile && !is_remote_file(varsfile) && strcmp(varsfile, global)) {
		if ((use_file_variables & 16) && check_file(varsfile) != CHECK_FILE)
			if (!look_up_tree(varsfile, ".epsilon_vars", dir, NULL))
				return;
		apply_vars_file(varsfile);
	}
}

// Apply the settings from the given .epsilon_vars file (local or
// global).
apply_vars_file(char *varsfile)
{
	char bname[FNAMELEN], pat[FNAMELEN];
	char *file = filename, *mode = major_mode, *s;
	int orig = bufnum;

	if (check_file(varsfile) != CHECK_FILE)
		return;
	save_var bufnum;
	if (!look_file(varsfile)) {	// Load the .epsilon_vars file we found.
		make_bname(varsfile, bname);
		bufnum = create(bname);
		find_in_this_buf(varsfile, get_default_translation_type(varsfile));
	}
	int tmp = tmp_buf();	// Make a buffer with just the matching settings.
	buf_printf(tmp, "Local Variables:\n");
	on_exit buf_delete(tmp);
	save_var bufnum;
	save_var point = 0, case_fold = 0, abort_searching = ABORT_ERROR;
	save_var file_variables_search_limit = 1000000;	// Remove usual limit for an .epsilon_vars file.
	while (re_search(1, "^(Extensions|Modes|Filenames|Basenames): *([^ \n]*)") == 1) {
		grab(find_group(2, 1), find_group(2, 0), pat);
		switch (toupper(character(matchstart))) {
			case 'E':	s = get_extension(file); break;
			case 'M':	s = mode; break;
			case 'F':	s = file; break;
			default:	s = get_tail(file, 1); break;
		}
		if (fpatmatch(s, pat, 0, 0)) {
			int start = give_end_line();
			if (re_search(1, "^(Extensions|Modes|Filenames|Basenames): *"))
				to_begin_line();
			buf_xfer(tmp, start, point);
			buf_printf(tmp, "\n");
		}
	}
	bufnum = tmp;	// Now apply settings from tmp to orig buffer.
	point = 0;
	simple_re_replace(1, "^#.*", "");
	check_buffer_filevars(orig);
	if (!retain_forced_mode)
		check_filevar_mode(orig);
}

command import_customizations()
{
	int tmp = zap("import-customizations"), i, ver = 0, err, clean = !has_arg;
	int debug = (iter == 99);
	char buf[2000], maindir[FNAMELEN], bindir[FNAMELEN];
	char origsta[FNAMELEN], cursta[FNAMELEN], *s, *exe_ext = "", pat[FNAMELEN];
	char answer[6][2000], ranswer[6][2000], oldname[FNAMELEN];
	char origdir[FNAMELEN], *epsout = lookpath("epsout.e");

	iter = 1;
	tiled_only();
	getcd(origdir);
	on_exit chdir(origdir);
	putenv("EPSRUNS");		// Prevent error msg in old Epsilons.
	save_var bufnum;
	if (!epsout)
		error("The file epsout.e is missing.");
	for (i = EELVERSION / 10 - 1; i >= 8 ; i--) {
		if (is_mswindows) {
			sprintf(buf, "EPSPATH%02d", (i < 10) ? (i * 10) : i);
			if (!verenv(buf))
				continue;
		} else {
			sprintf(buf, "~/.epsilon/epsilon-v%d.sta", i);
			if (!check_file(buf))
				continue;
		}
		buf_printf(tmp, "%d\n", ver = i);
	}
	set_buf_point(tmp, 0);
	switch (count_lines_in_buf(tmp, 0)) {
		case 0:		ver = 0; break;
		case 1:		break;
		default:
			*buf = 0;
			buf_delete_final_newlines(tmp);
			if (!get_choice(tmp, buf, " Update from version ",
							"Update from which version? ",
							"OK", "Cancel", ""))
				return;
			ver = numtoi(buf);
			say("");
	}
	*maindir = *bindir = *origsta = *cursta = 0;
	if (is_mswindows && ver) {
		exe_ext = ".exe";
		sprintf(buf, "EPSPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (s = verenv(buf)) {
			strcpy(maindir, s);
			while (s = strchr(maindir, ';')) {
				*s = 0;
				sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
				if (check_file(origsta) == CHECK_FILE)
					break;
				strcpy(maindir, s + 1);
			}
			absolute(maindir);
			sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
		}
		sprintf(buf, "BINPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (*maindir)
			sprintf(bindir, "%s\\bin", maindir);
		s = verenv(buf);
		if (s)
			strcpy(bindir, s);
		absolute(bindir);
		if (*maindir)
			sprintf(cursta, "%s\\epsilon.sta", maindir);
		if (ver >= 13) {
			get_customization_directory(cursta);
			sprintf(cursta + strlen(cursta), "\\epsilon-v%d.sta", ver);
		}
	} else if (ver) {		// Set vars for Unix.
		sprintf(buf, "/{usr/local,opt}/epsilon%d*/bin/epsilon", ver);
		int max = 0;		// Find newest match of pattern.
		struct file_info fi;
		s = file_match(buf, STARTMATCH);
		*buf = 0;
		while (s && check_file(s, &fi)) {
			if (!max || fi.raw_file_date > max) {
				strcpy(buf, s);
				max = fi.raw_file_date;
			}
			s = file_match("", 0);
		}
		if (*buf) {
			get_tail(buf, 0)[-1] = 0;
			strcpy(bindir, buf);
			strcpy(origsta, buf);
			strcpy(get_tail(origsta, 0), "original.sta");
		}
		sprintf(cursta, "~/.epsilon/epsilon-v%d.sta", ver);
	}
	save_var window_bufnum;
	to_buffer_num(tmp);
	buf_zap(bufnum);
	sprintf(oldname, "Epsilon %d", ver);
	if (!ver)
		strcpy(oldname, "the old version");
	bprintf("Epsilon will try to import any customizations you saved in your\n");
	bprintf("state file in %s.  Please correct any of\n", oldname);
	give_binding(buf, "exit-level");
	bprintf("these settings as necessary, then press %s\n", buf);
	bprintf("to begin the import process.\n\n");

	if (!ver) {
		bprintf("Epsilon couldn't find any %s,\n", is_mswindows
				? "Epsilon registry entries"
				: "customized state file in ~/.epsilon");
		bprintf("so you'll have to fill in this information.\n\n");
	}
	bprintf("1. Your customized state file for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text,
			*cursta ? cursta : "full path for your .sta file", -1);
	if (*cursta && check_file(cursta) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	bprintf("2. The original pre-customization state file for %s\n", oldname);
	bprintf("is:\n%p%s\n%p\n",
			color_class viewed_text,
			*origsta ? origsta : "full path for original.sta", -1);
	if (*origsta)
		if (check_file(origsta) != CHECK_FILE)
			bprintf("(This file doesn't exist.)\n\n");
	bprintf("3. If you customized %s by loading EEL files,\n", oldname);
	bprintf("please list each .e file's pathname here, one per line:\n%p\n%p\n",
			color_class viewed_text, -1);
	sprintf(buf, "full path for epsilon%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%c%s%s", bindir, path_sep,
				(ver < 10) ? "e32" : "epsilon", exe_ext);
	bprintf("4. The Epsilon%s executable file for %s is:\n%p%s\n%p\n",
			is_mswindows ? " (GUI, not console version)" : "", oldname,
			color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	sprintf(buf, "full path for eel%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%ceel%s", bindir, path_sep, exe_ext);
	bprintf("5. The EEL compiler executable for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	get_customization_directory(buf);
	strcat(buf, "upd");
	mkdir(buf);
	bprintf("6. Use this working directory for intermediate files:\n%p%s\n%p\n",
			color_class viewed_text, buf, -1);
	if (buf && check_file(buf) != CHECK_DIR)
		bprintf("(Warning: Couldn't create this directory.)\n\n");
	for (;;) {
		err = point = 0;
		recursive_edit();
		if (!buf_exist(tmp))
			return;
		bufnum = tmp;
		point = 0;		// Now retrieve the answers.
		for (i = 0; i < 6; i++) {
			sprintf(pat, "^%d%%. [^:]+:\n((.|\n.)*)\n\n", i + 1);
			if (!re_search(1, pat)) {
				say("Can't interpret response %d", i + 1);
				err = 1;
				break;
			}
			grab(find_group(1, 1), find_group(1, 0), answer[i]);	// 8.3 version.
			grab(find_group(1, 1), find_group(1, 0), ranswer[i]);	// For display.
			if (i == 2) {
				err = test_valid_files(find_group(1, 1), find_group(1, 0));
				if (err == 2)
					quick_abort();
				err = err ? 3 : 0;	// Don't prompt again.
			} else if (check_file(answer[i]) != ((i == 5)
				? CHECK_DIR : CHECK_FILE)) {
				say("Response %d: Can't read %s \"%f\".",
					  i + 1, (i == 5) ? "directory" : "file", answer[i]);
				err = 1;
				break;
			} else
				convert_to_8_3_filename(answer[i]);
		}
		if (!err)
			break;
		if (err != 3
			  && !ask_yn("Try again", "Would you like to correct your response? ",
						 "Yes", "No", 0))
			quick_abort();
	}
	if (chdir(answer[5]))
		error("Couldn't set directory to %f.", answer[5]);
	buf_zap(tmp);	// Compile epsout & extensions, then save list of them.
	compile_for_import_customizations(epsout, answer[4], answer[5], 0);
	compile_for_import_customizations(answer[2], answer[4], answer[5], tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}

	// Run the old version with orig state file.
	save_var reject_client_connections = 1;
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[1], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	buf_zap(tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	// Run the old version with new state file.
	sprintf(buf, " -s%s -l%s%cepsout.b -r%sbuild_settings_file",
			answer[0], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	int cur = tmp_buf();
	on_exit buf_delete(cur);
	bufnum = cur;
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	int clist = make_difference_list(tmp, cur);
	on_exit buf_delete(clist);	// Now format the list of differences.
	if (clean)
		clean_customizations();
	int cust = prep_customizations();
	bufnum = cust;
	current_time_and_date(buf);
	save_var point;
	i = point;
	bprintf("\n# Import-customizations created these settings %s\n", buf);
	bprintf("# from %s.\n", ranswer[0]);
	int ext = tmp_buf();
	on_exit buf_delete(ext);
	buf_printf(ext, "%s\n", ranswer[2]);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	bprintf("\n");
	restore_vars();
	to_buffer_num(cust);
	point = i;
	if (clean)
		comment_old_customizations(point);
	list_undefined_customizations();
	say("Done.  Save file, then restart Epsilon, to use these new settings.");
}

#define XWIN_XFLD_COUNT	13	/* # of fields in an X11 font logical descriptor */
int xwin_font_name_when_idle(int cnt);
int xwin_font_size_when_idle(int cnt);
int xwin_pixel_size_when_idle(int cnt);

// Make a list of fonts by running xlsfonts, and have the user narrow
// down the choices until just one is left. Return 0 if we can't get the
// list.
int set_font_for_unix()
{
	char fontname[1000], orig[1000], fontpat[1000], fontfamily[1000];
	char fields[XWIN_XFLD_COUNT][1000], cust_point_sz[1000];
	jmp_buf *old_level = top_level, this_level;
	int i, mbuf = zap(_MATCH_BUF);
	struct screen_info retain_windows;

	retain_windows.wins = 0;
	save_screen(&retain_windows);
	strcpy(orig, font_fixed);
	*cust_point_sz = 0;
	for (i = 0; i < XWIN_XFLD_COUNT; i++)
		strcpy(fields[i], "*");
	save_var show_status = 0;
	note("Listing fonts...");
	int font_list = zap("-font-list");

	// First, extract the foundry+family of cur font.
	buf_printf(font_list, "%s\n", font_fixed);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	grab_full_line(mbuf, fontfamily);

	buf_zap(font_list);
	pipe_text(NULL, "-font-list", "xlsfonts", NULL,
			  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
	save_var bufnum = font_list;	// Got a list of all available fonts.
	if (lines_between(0, size(), 0) < 5) {
		say("Couldn't run xlsfonts to list the fonts.");
		return 0;
	}
	top_level = &this_level;
	if (setjmp(top_level)) {	// On abort, come here.
		font_fixed = orig;
		restore_screen(&retain_windows);	// Restore windows, since size changes
		free(retain_windows.wins);			// might have deleted some.
		top_level = old_level;
		quick_abort();
	}
	xwin_set_font_first_prune();	// Drop fonts we know we can't use.
	note("");

	// Extract the foundry+family names from font_list.
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	xwin_prune_to_usable_fonts(_MATCH_BUF);
	xwin_prompt_for_field(fields[0], "Select a font:", " Select a font ",
						 fontfamily, xwin_font_name_when_idle);

	// Now make a list of allowed point sizes for the selected font
	// family. Zero means any size is OK.
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 7, fontpat);
	xwin_font_size_prettify(_MATCH_BUF);

	i = xwin_prompt_for_field(fields[6], "Select a point size:",
							 " Point Size ", "12",
							 xwin_font_size_when_idle);

	xwin_font_size_unprettify(fields[6]);
	// If point size is 0 (custom), prompt for it and remember it.
	if (!strcmp(fields[6], "0")) {
		if (i)
			get_string(cust_point_sz, "Point size, such as 12 or 14: ");
		else
			get_strdef(cust_point_sz, "Font lacks recommended point size. "
					   "Please pick one", "14");
		xwin_font_size_fix_user(cust_point_sz);
	}

	// Ask the user to choose among other options the font family might
	// include.
	xwin_query_one_field(11, fields, "Select a char set (iso8859-1 "
						"recommended, or iso10646-1 for Unicode):",
						" Character Set ", "iso8859-1");
	xwin_query_one_field(1, fields, "Make all characters bold?",
						" Force Bold? ", "medium");
	xwin_query_one_field(2, fields, "Make all characters italic (o) not regular (r)?",
						" Force Italic? ", "r");

	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 6, fontpat);
	xwin_prompt_for_field(fields[5], "Select a pixel size:", " Pixel size ", "",
						  xwin_pixel_size_when_idle);

// any other fields?

	save_var resize_menu_list = 1;
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, -1, fontpat);

	if (*cust_point_sz)
		xwin_add_point_size(_MATCH_BUF, cust_point_sz);
	xwin_prompt_for_field(fontname, "These fonts match. Please pick one:",
						 " Font names ", "", xwin_font_name_when_idle);

	note("");
	aside("Switch to %s if %d", fontname, !user_abort);
	font_fixed = user_abort ? orig : fontname;
	if (!user_abort && strcmp(font_fixed, fontname))
		say("Sorry, can't use that font.");

	restore_screen(&retain_windows);
	free(retain_windows.wins);
	top_level = old_level;
	return 1;
}

when_loading()
{
	drop_name("build-filevar-pattern");
	file_interpret_env_vars = 65535;
}
