/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 b2 to 14 b3.

#include "eel.h"
#include "epsnet.h"
#include "tags.h"
#include "colcode.h"
#include "c.h"
#include "proc.h"
#include "menu.h"
#include "tags.h"
#include "bufed.h"
#include "vspell.h"
#include "epsnet.h"

// How many times does text appear in buffer buf?
int count_matches_in_buffer(int buf, char *text)
{
	int cnt = 0;
	save_var bufnum = buf, point = 0;
	while (search(1, text))
		cnt++;
	return cnt;
}

// Just before point there's a > character that may mark the end of a template.
// Move to just before the < that starts it and return 1.  Return 0 if it
// doesn't seem to be a template, without moving.
int find_template_start()
{
	int opos = point, start;

	if (character(point - 1) != '>')
		return 0;
	start = default_move_level(-1, ">", "<");
	if (start < 0)
		return 0;
	point = start;
	if (find_template_end() && point == opos) {
		point = start;
		return 1;
	} else {
		point = opos;
		return 0;
	}
}

int find_c_func_info(char *type, char *class, char *func, int stop_on_key)
{
	char funcname[FNAMELEN], classname[FNAMELEN], pat[FNAMELEN], pat2[FNAMELEN];
	int start, funcstart, j, got_class = 0;

	start_of_function = -1;
	*classname = *funcname = 0;
	if (func)
		*func = 0;
	if (class)
		*class = 0;
	if (type)
		*type = 0;
	save_var point, matchstart, matchend, narrow_start = 0, narrow_end = 0;
	save_var tag_declarations = 0;	// We use tagging sometimes.
	if (c_extra_keywords & OBJC_KEYWORDS)
		got_class |= find_objc_class(type, class);
	to_begin_line();
	start = point;
	if (c_quick_scan) {
		point--;		// Begin searching at the last } char on the left margin...
		while (search(-1, "\n}")) {
			point++;		// ...that's not in a comment.
			if (!in_c_comment(c_look_back))
				break;
			point--;
		}
	} else			// Required if inner braces might be at the left margin.
		point = 0;

	strcpy(pat, "{|/<*|/>|(~?[a-zA-Z0-9_]+|" CPP_OPSPEC ")[ \t\n]*%(");
	if (c_extra_keywords & OBJC_KEYWORDS)
		strcat(pat, "|\n[-+]");
	while (re_search(1, pat)) {
		point = matchstart;
		if (stop_on_key && (char_avail() || is_key_repeating()))
			return -1;		// Give up search now?
		if (parse_string(-1, "^[ \t]*#[ \t]*(define|if|pragma).*(\\\n.*)*")) {
			to_end_line();	// Ignore everything in a macro definition.
			continue;
		}
		switch (curchar()) {
			case '{':		// Could be start of class def.
				j = point;
				if (c_inside_block(point, start)) {
					got_class |= find_c_class_info(start, type, class, func);
					point = j + 1;
					if (got_class)
						continue;
				}
				point = j;
				c_skip_level('{', '}');
				break;
			case '/':
				if (character(point + 1) == '*') {		// Skip over comment.
					point += 2;
					search(1, "*/");
				} else
					to_end_line();	/* c++ comment */
				break;
			case '\n':		// Objective-C method
				point++;
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				if (find_objc_method(func) == 1)
					return CF_INFO_FUNC | got_class;
				break;
			default:		// Found an identifier; is it a function def?
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				funcstart = point;
				if (!parse_string(1, "~?[a-zA-Z0-9_]+|" CPP_OPSPEC, funcname))
					return got_class;
				sprintf(pat2, "([a-zA-Z0-9_]+)[ \t\n]+(%s)",
						c_prototype_macro_names);
				if (parse_string(-1, pat2))
					grab(find_group(1, 0), find_group(1, 1), funcname);
				if (parse_string(1, "[ \t\n]*%("))
					narrow_end = size() - (point + c_biggest_declarator);
				if (tag_c_func(0) == 2 && !find_c_func_oneliner(start)) {
					point = funcstart;	// Checks out as a function def.
					start_of_function = give_begin_line();
					*classname = 0;
					if ((c_extra_keywords & JS_KEYWORDS) && !strcmp(funcname, "function"))
						find_js_func_info(funcname);
					if (parse_string(-1, (c_extra_keywords & (CSHARP_KEYWORDS | JS_KEYWORDS))
									 ? "[ \t\n]*%.[ \t\n]*"
									 : "[ \t\n]*::[ \t\n]*")) {
						point = matchend;
						if (!parse_string(-1, "[a-zA-Z0-9_]+", classname) && parse_string(-1, "%>")) {
							int template_end = point;
							if (find_template_start() && parse_string(-1, "[a-zA-Z0-9_]+", classname))
								grab(matchstart, template_end, classname + strlen(classname));
						}
					}
					if (func)
						strcpy(func, funcname);
					if (class && *classname)
						strcpy(class, classname);
					return CF_INFO_FUNC | (*classname ? CF_INFO_CLASS : 0)
							| got_class;
				}
				*funcname = 0;
				narrow_end = 0;
				break;
		}
	}
	return got_class;
}

is_c_keyword(p)		// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr("|if|else|for|do|while|return|break|continue|switch|"
		   "case|default|goto|char|short|int|static|volatile|const|"
		   "float|double|long|void|enum|", p))
		return 1;
	if (p[1] == '_' &&
	    strstr("|_asm|_based|_cdecl|_emit|_except|_export|_far|_fastcall|"
		   "_fortran|_huge|_interrupt|_loadds|_near|_pascal|_saveregs|"
		   "_segment|_segname|_stdcall|_syscall|_try|__int64|", p))
		return 1;
	if (!(c_extra_keywords & (EEL_KEYWORDS | CPP_KEYWORDS | JAVA_KEYWORDS |
							  JS_KEYWORDS | IDL_KEYWORDS | CSHARP_KEYWORDS | OBJC_KEYWORDS))
		  && strstr("|restrict|_Alignas|_Alignof|_Atomic|_Bool|_Complex|"
					"_Generic|_Imaginary|_Noreturn|_Static_assert|_Thread_local|"
					"_Decimal128|_Decimal32|_Decimal64|", p))
		return 1;		// C language only
	if ((c_extra_keywords & EEL_KEYWORDS) &&
	    strstr("|keytable|buffer|window|command|on|color_class|"
		   "spot|save_spot|save_var|on_exit|user|zeroed|byte|int32|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
	    strstr("|asm|catch|class|delete|friend|inline|new|operator|"
			   "private|protected|public|template|try|this|virtual|"
			   "throw|_self|and|and_eq|bitand|bitor|bool|compl|const_cast|"
			   "dynamic_cast|explicit|false|mutable|namespace|not|not_eq|"
			   "or|or_eq|reinterpret_cast|static_cast|true|typeid|typename|"
			   "using|wchar_t|xor|xor_eq|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("alignas|alignof|char8_t|char16_t|char32_t|concept|consteval|"
				 "constexpr|constinit|co_await|co_return|co_yield|decltype|"
				 "export|noexcept|nullptr|requires|static_assert|thread_local", p))
		return 1;		// Later C++ keywords
	if ((c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|abstract|assert|boolean|byte|catch|class|extends|false|final|"
			   "finally|function|implements|import|in|instanceof|"
			   "interface|native|new|null|package|private|protected|"
			   "public|strictfp|super|synchronized|this|throw|throws|transient|"
			   "true|try|var|with|", p))
		return 1;
	if (!(c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|auto|extern|register|unsigned|signed|sizeof|"
			   "struct|union|typedef|", p))
		return 1;		// C keywords not in Java.
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|appobject|auto_handle|bindable|boolean|broadcast|byte|"
			   "callback|coclass|code|comm_status|context_handle|"
			   "control|decode|defaultbind|dispinterface|displaybind|"
			   "dual|enable_allocate|encode|error_status_t|"
			   "explicit_handle|fault_status|handle|handle_t|hidden|"
			   "hyper|idempotent|ignore|immediatebind|import|in|"
			   "include|interface|library|licensed|local|maybe|module|", p))
		return 1;
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|ms_union|nocode|nonextensible|notify|object|odl|"
			   "oleautomation|optional|out|pipe|propget|propput|"
			   "propputref|public|ptr|readonly|ref|requestedit|"
			   "restricted|retval|small|source|string|unique|"
			   "usesgetlasterror|v1_enum|vararg|", p))
		return 1;
	if ((c_extra_keywords & CSHARP_KEYWORDS) &&
		  strstr("|abstract|as|base|bool|byte|catch|checked|decimal|delegate|"
				 "event|explicit|false|finally|fixed|foreach|implicit|in|"
				 "interface|internal|is|lock|namespace|new|null|object|"
				 "operator|out|override|params|private|protected|public|"
				 "readonly|ref|sbyte|sealed|sizeof|stackalloc|string|this|"
				 "throw|true|try|typeof|uint|ulong|unchecked|unsafe|ushort|"
				 "using|virtual|", p))
		return 1;
	if ((c_extra_keywords & OBJC_KEYWORDS) &&
		  strstr("|self|super|BOOL|id|Class|nil|YES|NO|NS_DURING|NS_HANDLER|"
				 "NS_ENDHANDLER|in|out|inout|bycopy|byref|oneway|@interface|"
				 "@implementation|@end|@class|@private|@protected|@public|"
				 "@encode|@protocol|@selector|@try|@catch|@finally|@throw|"
				 "@synchronized|Nil|IBAction|IBOutlet|", p))
			return 1;
	return 0;
}

unix_delete_file_path_before_tilde()
{
	if (!parse_string(1, "/.*/!~")) {
		if (parse_string(1, "/~"))
			delete(point, point + 1);
		return;
	}
	int p = matchend;
	save_spot point;
	int end = point + parse_string(1, "/.*/~[^/]*");
	while (point < end) {	// From tilde, go back to each /.
		if (buffer_range_has_file_name(point, end)) {
			delete(0, point);
			return;
		} else if (!search(1, "/"))
			break;
	}
	delete(0, p);
}

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!(telnet_interpret_output & 16))
		save_var point;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			point = type_point;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 0x100)
				perform_unicode_conversion(bufnum, from, size(), CONV_TO_16, "UTF-8-No-BOM");
			if (telnet_interpret_output & 0x20) {
				point = from;
				simple_re_replace(0, "<#0-#1>", "");
			}
			if (telnet_interpret_output & 0x40)
				remove_xterm_sequences(from);
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			if (telnet_interpret_output & 0x80)
				process_backspace_sequences(from - 100);
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)
				  && point == size()) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if ((net_prompt_get_directory & 1) && point == size())
				net_prompt_get_dir();
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
	}
}

// Set this buffer's current directory by examining the prompt from the
// remote system.
void net_prompt_get_dir()
{
	struct url_parts parts;
	if (!buffer_url || !parse_url(buffer_url, &parts))
		return;				// Get the host name.
	int i = find_variable_by_host("net_prompt_get_directory", parts.host);
	int flags = get_num_var(i);
	char *pat, *s = strchr(buffer_url, ':');	// Copy :host@user portion.
	i = find_variable_by_host("net_prompt_pattern", parts.host);
	if (!(pat = get_str_var(i)) || !s || !(flags & 1))
		return;
	int mode = 0;
	if (parse_string(-1, "\x1FEpsilon:cwd:(.*)\x1F"))
		mode = 1;			// Our shell script modified the prompt.
	else if (!(flags & 2) || !parse_string(-1, pat))
		if (!(flags & 4)) {	// Fall back, no cur dir info.
			mode = 2;
			parse_string(-1, "()");	// This will make dir be empty.
		} else
			return;
	int from = find_group(1, 1), to = find_group(1, 0);
	char dir[FNAMELEN], url[FNAMELEN];
	if (to - from < FNAMELEN - strlen(buffer_url) - 5)
		grab(from, to, dir);
	if (mode == 1)
		delete(matchend, matchstart);
	sprintf(url, "%s%s/", (flags & 8) ? "scp" : "ftp", s);
	if (mode != 2) {
		i = find_variable_by_host("net_prompt_get_dir_helper", parts.host);
		(*(int (*)()) i)(mode, url, dir, &parts);
	}
	process_current_directory = url;
	if (flags & 16)
		set_character_color(give_begin_line(), point, color_class process_prompt);
	if (flags & 32) {
		save_var use_process_current_directory = 0;	// Already got it.
		do_when_process_input(bufnum, 0, point);	// Helps to color user input later.
	}
}

// Modify url by adding the path corresponding to dir. Mode: 0=got dir
// from parsing the prompt, 1=got dir from magic Epsilon sequence. This
// version is for Unix-style hosts. A per-host replacement like
// net_prompt_get_dir_helper_example_com() is allowed to completely
// replace the contents of url if needed. It may use the components of
// the buffer's URL found in *parts.

void net_prompt_get_dir_helper(int mode, char *url, char *dir, struct url_parts *parts)
{
	UNREFERENCED_PARAMETER(mode);		// A user-written replacement might use these.
	UNREFERENCED_PARAMETER(parts);
	if (strcmp(dir, "~")) {		// If dir is not home dir, add it.
		if (dir[0] == '~' && dir[1] == '/')
			strcat(url, dir + 2);	// Path is relative to home dir, so 1 /.
		else if (strcmp(dir, "/"))		// Use 2 /s to make abs path, unless root dir.
			strcat(url, dir);
		strcat(url, "/");
	}
}

// For sending chars to ssh job, act like a process, but for getting
// chars, act like telnet.
ssh_activity(int activity, int buf, int param1, int param2)
{
	save_var telnet_interpret_output = ssh_interpret_output, abort_searching = 0;
	if (activity == NET_SEND) {
		save_var bufnum = buf;
		color_process_range(0, size());
		do_when_process_input(buf, param1, param2);
	} else
		telnet_activity(activity, buf, param1, param2);
}

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	if (clear_process_buffer & 0x20)
		zap(bname);
	to_buffer(bname);
	type_point = size();
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = is_unix ? 6 : 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s", &parts, 0);
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname, SHELL_NO_CD_MAGIC))
		error("Couldn't start ssh process");
	char url[FNAMELEN];
	sprintf(url, "ssh://%s", justhost);
	if (parts.have_usr)
		sprintf(url, "ssh://%s@%s", username, justhost);
	buffer_url = strkeep(url);
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/(env )?perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(php)")) {
		sprintf(res, pat, "php");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "(sh|csh|[bd]ash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

// Display topic in a browser and return 1.  Return 0 if running a web
// browser is unsupported, or -1 if we couldn't for some other reason.
int try_html_help_string(char *topic)
{
	char url[FNAMELEN], prog[FNAMELEN], *p, *q, cmdline[100], dir[FNAMELEN];

	if (epsilon_manual_port > 0)
		sprintf(url, "http://localhost:%d/man/", epsilon_manual_port);
	else
		strcpy(url, "https://www.lugaru.com/cgi-bin/lhelp/man/");
	for (p = topic, q = url + strlen(url); *p; p++, q++)
		if (*p == ' ') {
			strcpy(q, "%20");
			q += 2;
		} else
			*q = *p;
	*q = 0;
	if (!*topic)
		strcat(url, "0");
	p = lookpath("lhelp/epsilon.lhl");
	if (p) {		// Find name of directory containing our lhelp files.
		strcpy(dir, p);
		*get_tail(dir, 0) = 0;
	} else
		get_executable_directory(dir);
	convert_to_8_3_filename(dir);
	if (opsys == OS_UNIX) {
		sayput("Running a browser...");
		get_executable_file(prog, "lhelpd", 1);
		sprintf(prog + strlen(prog), " -p %d -d \"%s\" &",
				epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			shell("", prog, "");
	} else if (is_gui || is_mswindows) {
		sayput("Starting a browser, please wait...");
		expire_message = 1;
		get_executable_directory(prog);
		strcat(prog, "lhelp.exe");
		sprintf(cmdline, "lhelp -p %d -d \"%s\"", epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			winexec(prog, cmdline, SW_SHOWMINNOACTIVE, 0);
	} else
		return 0;
	if (!try_show_url(url))
		return -1;
	note("");
	return 1;
}

windows_only_help()
{
	if (!(has_feature & FEAT_WINHELP))
		error("Sorry, WinHelp isn't available in this environment.");
}

#include "lcustom.e"

await_menu_click()	// show menu, return 1 if user clicked on it
{
	save_var mouse_cursor = left_pointer;
	load_menu();
	prep_menu_bar();
	save_var window_handle = menu_window;
	do {
		move_to_column(mouse_x);
		maybe_refresh();
		getkey();
	} while (key == MOUSE_MOVE && still_on_menu_bar());
	remove_window(menu_window);
	if (IS_MOUSE_DOWN(key) && still_on_menu_bar())
		return 1;
	ungot_key = full_key;
	return 0;
}

command show_menu() on reg_tab[FALT(2)]
{
	if (has_feature & FEAT_OS_MENUS) {
		invoke_menu(0);
		return;
	}
	save_var mouse_cursor = left_pointer;
	load_menu();
	if (!permanent_menu)
		prep_menu_bar();
	do_show_menu(0);
}

command mouse_move() on reg_tab[MOUSE_MOVE]
{			/* maybe pop up a menu or something */
	if (!permanent_menu && is_menu_here())
		show_menu_while_over();
	else {
		if (!mouse_screen) {
			int win = get_window_area();
			if (mouse_over_border(win, mouse_y, BBOTTOM) &&
				mouse_y + 1 != echo_line
				|| mouse_over_border(win, mouse_y, BTOP))
				mouse_cursor = vert_arrows;
			else if (mouse_over_border(win, mouse_x, BLEFT))
				mouse_cursor = horiz_arrows;
			else if (mouse_over_border(win, mouse_x, BRIGHT))
				mouse_cursor = !is_gui && display_scroll_bar
							   ? right_pointer : horiz_arrows;
			else if (win == -1 && permanent_menu && mouse_y < 1)
				mouse_cursor = left_pointer;
			else
				mouse_cursor = std_pointer;
		} else if (tiled_scroll_bar)
			mouse_cursor = on_scroll_bar() ? right_pointer
										   : std_pointer;
		else if (on_scroll_bar())
			show_scroll_bar();
		if (ungot_key == -1 || ungot_key == MOUSE_MOVE)
			invisible_cmd();
	}
}

suffix_bashrc()
{
	shell_mode();
}

suffix_cshrc()
{
	shell_mode();
}

suffix_tcshrc()
{
	shell_mode();
}

suffix_bash_profile()
{
	shell_mode();
}

suffix_bash_logout()
{
	shell_mode();
}

suffix_login()
{
	shell_mode();
}

suffix_logout()
{
	shell_mode();
}

// Prepare a list of env var name matches of pat in the match buffer.
int proc_complete_list_envvars(char *var)
{
	char pat[FNAMELEN];
	save_var bufnum = zap(_MATCH_BUF);
	get_substitution_variables(1);
	save_var abort_searching = 0, show_status = 0;
	point = 0;
	simple_re_replace(1, "=.*", "");	// Discard values.
	sprintf(pat, "^%q", var);
	do_drop_matching_lines(REGEX, pat, 0); // Discard non-matching lines.
	do_sort_region(0, size(), 0, 0);
	return lines_between(0, size(), 0);
}

command process_complete() on process_tab['\t'], process_tab[NUMSHIFT(GREYTAB)]
{
	char pat[FNAMELEN], pref[FNAMELEN];
	int cnt = 0, start, end, buf = create(_MATCH_BUF), envvars = 0;
	int offset = IS_SHIFT_KEY(full_key) ? -1 : 1;

	start = end = point;
	if (start < type_point) {
		if (!maybe_indent_rigidly(0))
			indent_like_tab();
		return;
	}
	*pat = 0;
	if (!(process_completion_style & 16)
		  && parse_string(-1, (opsys == OS_UNIX) ? "\${?(<ident>*)" : "%%(<ident|(|)>*)")) {
		grab(start = find_group(1, 1), find_group(1, 0), pat);
		envvars = start - matchstart;	// 2 if ${, 1 if $ or %
		point = end;
	} else if (process_line_inside_quoting() || parse_string(-1, "[^\n\t <>=\"]+", pat)) {
		start = point;
		point = end;
		if (process_line_inside_quoting())
			grab(start = matchend + 1, end, pat);
		if (start < type_point && type_point < end) {
			start = type_point;		// exclude prompt
			grab(start, end, pat);
		}
	}
	if (!*pat)
		strcpy(pat, "*");
	if (!(process_completion_style & 1) || prev_cmd != CMD_PROC_COMPL) {
		int dircmdlen = matches_at_length(start, -1, process_completion_dircmds);
		int dirsonly = dircmdlen && ((dircmdlen == start - type_point)
					 || matches_at(start - dircmdlen, -1, "[][#&;()] *"));
		int initial = all_blanks(type_point, start)
					  || matches_at(start, -1, "([&;|]|</word>(sudo)) +");
		if (envvars)
			cnt = proc_complete_list_envvars(pat);
		else
			cnt = proc_complete_do_list_matches(pat, initial,
												process_completion_style & 2,
												dirsonly ? FM_ONLY_DIRS : 0);
		check_abort();
		offset = 0;
		if (!cnt)
			error("No matches for %s.", pat);
		set_buf_point(buf, 0);
	}
	if (!(process_completion_style & 1))
		find_buffer_prefix_folding(name_to_bufnum(_MATCH_BUF), pref, !is_unix);
	else {
		buf_delete_final_newlines(buf);
		grab_line_offset(buf, pref, offset, 1);
	}
	int needs_quoting = strpbrk(pref, opsys == OS_UNIX ? " \\" : " ") && !envvars;
	int quotechar = opsys == OS_UNIX ? '\'' : '"';
	if ((process_completion_style & 1) || strlen(pref) > end - start) {
		delete(start, end);
		if (needs_quoting && !process_line_inside_quoting())
			insert(quotechar);
		stuff(pref);
		if ((process_completion_style & 1)) {
			if (is_path_separator(character(point - 1)))
				delete(point - 1, point);
			this_cmd = CMD_PROC_COMPL;
			return;
		}
		if (cnt == 1 && !is_path_separator(character(point - 1))) {
			if (needs_quoting && process_line_inside_quoting())
				insert(quotechar);
			stuff(!envvars ? " " : (opsys != OS_UNIX) ? "%" : envvars == 2 ? "}" : "");
		}
	} else if (prev_cmd == CMD_PROC_COMPL) {
		proc_complete_show_matches();
		this_cmd = CMD_PROC_COMPL;
	} else if (cnt == 1 && !is_path_separator(character(point - 1))) {
		if (needs_quoting && process_line_inside_quoting())
			insert(quotechar);
		stuff(" ");
	} else if (!(process_completion_style & 4)
			   && buffer_size(_MATCH_BUF) < screen_cols - 20 - strlen(pat)
			   && buffer_size(_MATCH_BUF) < ptrlen(pref) - 2) {
		save_var bufname = _MATCH_BUF;
		simple_re_replace(1, "\n", " ");
		grab(0, size(), pref);
		say("%d matches for %s: %s", cnt, pat, pref);
		expire_message = 1;
	} else {
		say("%d matches for %s: press again to see matches.", cnt, pat);
		expire_message = 1;
		this_cmd = CMD_PROC_COMPL;
	}
}

command eval()
{
	char expr[FNAMELEN], *src = temp_buf();

	get_string(expr, "Expression: ");
	if (!*expr)
		return;
	eval_template(expr, src, has_arg);

	// Compiles the buffer src, the status messages go into the
	// buffer named "eval-comp".
	if (eel_compile(src, 0, "-s -b", "eval-comp") != 0){
		to_buffer("eval-comp");
		discardable_buffer = 1;
		delete_buffer(src);
		error("Cannot evaluate expression.");
	}
	delete_buffer(src);
	ev_template();
	iter = 0;
}

do_execute_eel(char *expr)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	execute_eel_template(expr, src);

	// Compiles the buffer src, the status messages go into the
	// buffer named "eval-comp".
	if (eel_compile(src, 0, "-s -b", "eval-comp") != 0){
		to_buffer("eval-comp");
		discardable_buffer = 1;
		error("Could not compile EEL statement.");
	}
	ev_template();
}

command import_colors()
{
	char name[FNAMELEN];
	int tmp, i;

	if (!ask_yn("Confirm", "This command is only for updating from Epsilon 7 or older. Continue? ",
				"Yes", "No", 1))
		return;
	get_file(name, "Load changes from file", "changes");
	tmp = tmp_buf();
	to_buffer_num(tmp);
	if (read_file(name, FILETYPE_AUTO))
		quick_abort();

	strcpy(name, "mycolors.e");
	absolute(name);
	if (warn_existing_file(name) > 0)
		return;
	do_find(name, FILETYPE_AUTO);
	buf_zap(bufnum);
	grab_buffer(tmp);
	point = 0;
	do_drop_matching_lines(REGEX, "^&.* color class for ", 0);
	do_drop_matching_lines(0, "[0x4000000 on 0x0]", 1);	// Unknown color.
	while (re_search(1, ": *([0-9a-fx]+)$")) {
		grab(find_group(1, 1), find_group(1, 0), name);
		point = find_group(1, 1);		// Change old-style color numbers
		delete(point, find_group(1, 0));	// to names.
		i = numtoi(name);
		give_color_combo(name, i);
		stuff(name);
	}
	point = 0;
	for (i = 0; i < 10; i++) {	// Assume class names have no more than 10 -'s.
		noteput("Working... %d%%", 20 + i * 5);
		string_replace("(color class for .*)-", "#1_", REGEX);	// Conv to _'s.
		if (!replace_num_changed)
			break;
	}
	for (i = 0; i < 10; i++) {
		string_replace("(=<word>*)-", "#1_", REGEX);	// Conv to _'s.
		if (!replace_num_changed)
			break;
	}

	noteput("Working... 70%%");
	// Make into valid EEL definitions.
	string_replace("^&(.*) color class for (.*): (.*)",
				   "color_scheme \"#1\" color_class #2 #3;", REGEX);
	noteput("Working... 80%%");
	// Convert pre-v8 color names to EEL names.
	string_replace(" (blue|green|cyan|red|magenta)", " dark_#1", REGEX);
	string_replace(" white", " grey", REGEX);
	string_replace(" brightblack", " dark_grey", REGEX);
	noteput("Working... 90%%");
	string_replace(" bright(blue|green|cyan|red|magenta|white)", " #1", REGEX);

	string_replace(" *= *", " = color_class ", REGEX);
	string_replace("%[(.* on .*)%]", "#1", REGEX);
	string_replace(" on 0x1000000;", ";", REGEX);

	stuff("// Automatically generated by import-colors.\n");
	stuff("// Save this file, then press Alt-F3 to import colors.\n\n");
	stuff("#include \"eel.h\"\n\n");
	point = 0;
	buf_delete(tmp);
	add_color_class_declarations();
	say("Done.");
}

export_colors_one_scheme(int i)
{
	char *nm;
	int j, k, fg, bg;

	nm = name_name(i);
	noteput("Working...%s", nm);
	bprintf("color_scheme \"%s\" {\n", nm);
	save_var selected_color_scheme = i;
	for (j = 0; j < number_of_color_classes(); j++) {
		k = get_foreground_color(j, 1);
		if (k & ECOLOR_UNKNOWN)
			continue;
		bprintf("\tcolor_class %s ", name_color_class(j));
		if (k & ECOLOR_COPY) {
			bprintf("= color_class %s;\n", name_color_class(COLOR_STRIP_ATTR(k)));
			continue;
		}
		fg = get_foreground_color(j, 0);
		bg = get_background_color(j, 0);
		bprintf("MAKE_RGB(%d, %d, %d)",
				GETRED(fg), GETGREEN(fg), GETBLUE(fg));
		if (fg & EFONT_BOLD)
			bprintf(" | EFONT_BOLD");
		if (fg & EFONT_ITALIC)
			bprintf(" | EFONT_ITALIC");
		if (fg & EFONT_UNDERLINED)
			bprintf(" | EFONT_UNDERLINED");
		if (!j || !(bg & ETRANSPARENT)
			  && bg != get_background_color(0))
			bprintf(" on MAKE_RGB(%d, %d, %d)",
					GETRED(bg), GETGREEN(bg), GETBLUE(bg));
		stuff(";\n");
	}
	stuff("};\n\n");
}

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1, skip = 0;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class") || !strcmp(func, "__interface"))
		return tag_c_struct(1);		// or similar class or managed c++ def
	else if (!strcmp(func, "interface") || !strcmp(func, "package"))
		return tag_c_struct(1);		// for java or managed c++
	else if ((c_extra_keywords & OBJC_KEYWORDS) && character(start - 1) == '@'
			 && word_in_list(func, "|interface|implementation|protocol|", 0))
		return tag_objc_block();
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_modifier_kwd |= 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "typedef")) {
		tag_modifier_kwd |= 2;	// next vars are typedefs
		return 1;
	} else if ((c_extra_keywords & CPP_KEYWORDS) && !strcmp(func, "using")) {
		tag_modifier_kwd |= 4;	// just a declaration 
		return 1;
	} else if (is_java_buffer() && (!strcmp(func, "import")
									|| !strcmp(func, "package"))) {
		tag_modifier_kwd |= 4;
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (!strcmp(func, "__attribute__") && parse_string(1, "<wspace>*%(%(")
			   && !skip_c_delims('(', ')', 1)) {
			return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	sprintf(tmp, "(%s)</word>", c_prototype_macro_names);
	if (parse_string(1, tmp)) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '<')
		tag_check_template(func);
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
		if (!(tag_which_items & 1)
			  || (c_extra_keywords & JS_KEYWORDS) && !strcmp(func, "function"))
			skip = 1;
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
			"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (c_extra_keywords & JS_KEYWORDS) {
		i = tag_jscript_ident(func, start);
	} else if (tag_c_variable() && ((tag_declarations & 2) || !(tag_modifier_kwd & 1)
									|| (tag_modifier_kwd & 2))) {
		i = 2;
		if (!(tag_which_items & 2))
			skip = 1;
	}
	if ((i == 2 || i == 3) && !isdigit(*func) && !(tag_modifier_kwd & 4)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		if (!skip)
			add_tag(func, start);
	}
	return i;
}

// If we have template syntax, add it.
tag_check_template(char *func)
{
	int pos = point;
	if (!find_template_end() || !parse_string(1, "[ \t\n]*(::<ident>+)?%("))
		point = pos;
	else
		grab(pos, matchstart, func + strlen(func));
}

tag_c_skip_before_body(int show)
{
	if (parse_string(1, "#.*(\\\n.*)*")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(try|const|volatile|NOTHROW|override|"
					 "final|import|module)</ident>")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(AVAILABLE|DEPRECATED_IN)_MAC_OS_X_VERSION_<word>+")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "throws[ \t]")) {	// for java
		point = matchend;
		skip_identifier_list();
		return 1;
	}
	if (parse_string(1, "(throw|noexcept|__attribute__|__DARWIN_ALIAS)[ \t]*!%(")) {
		point = matchend;	// throw for C++, attribute for gcc
		if (skip_c_delims('(', ')', 1)) {
			if (show)
				buffer_printf("-tag-errors", "Couldn't parse %s at line %d\n",
							  filename, lines_between(0, point, 0) + 1);
			return 0;
		}
		return 1;
	}
	if (parse_string(1, "(noexcept)</word>")) {
		point = matchend;
		return 1;
	}
	if ((c_extra_keywords & JS_KEYWORDS) && parse_string(1, ":<wspace>*<alnum>+")) {
		point = matchend;
		return 1;
	}
	return 0;
}

// Tag the following structure/class/enum if we're defining a tag for it.
// Flavor: 0=struct/union, 1=class/interface, 2=enum, 3=namespace.
tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN], *s;
	int start, i;

	if (flavor == 1 && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(-1, "@class"))
		return tag_objc_class_list();
	c_fwd_space();
	start = i = point;
	if (parse_string(1, "__declspec</word>")) {
		point = matchend;
		c_fwd_space();
		if (curchar() == '(')
			c_skip_level('(', ')');
		c_fwd_space();
	}

	if (!parse_string(1, tag_c_ident_pattern(), func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	if (flavor == 0 && (c_extra_keywords & CPP_KEYWORDS))
		flavor = 1;	// In C++, treat structs just like classes.
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (parse_string(1, "(final)</word>")) {
		point = matchend;
		c_fwd_space();
	}
	if (flavor == 1 && !is_java_buffer() && !(c_extra_keywords & JS_KEYWORDS)
		  && parse_string(1, tag_c_ident_pattern())) {
		i = point;
		parse_string(1, tag_c_ident_pattern(), func);
		c_fwd_space();
	}
	while (strchr(func, '.')) {
		strcpy(fullname, func);
		s = strchr(fullname, '.');
		sprintf(s, "::%s", func + 1 + (s - fullname));
		strcpy(func, fullname);
	}
	if (strchr(";[,", curchar()) && i > start) {
		point = i;	// A plain definition.
		return 1;
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		tag_c_add_class(func);
		i = ++point;	// skip over { so we do stuff within
		c_fwd_space();
		if (parse_string(1, "#(else|elif)"))
			if (locate_by_ifdef(1) != 2)
				point = i;
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

list_definitions_when_displaying(from, to)
{
	char tagname[FNAMELEN], msg[FNAMELEN];
	int dlg = window_handle;

	from = from;		// not used
	if (1) {
		set_wattrib(window_handle, PVERTBORDCOLOR, color_class text);
		set_wattrib(window_handle, PHORIZBORDCOLOR, color_class text);
//		set_wattrib(window_handle, PTITLECOLOR, color_class text);
		text_color = color_class text;
	}
	to = tag_line_to_position(lines_between(0, point, 0) + 1, tagname);
	save_var window_handle = _bufed_orig_window;
	if (!(list_definitions_live_update & 1))
		save_var point;
	point = to;
	if (*tagname) {
		strcpy(msg, tag_declarations ? " Select a declaration or definition - "
									 : " Select a definition - ");
		if (list_definitions_live_update & 2)
			sprintf(msg + strlen(msg), "Line %d ",
					buf_position_to_line_number(bufnum, to));
		sprintf(msg + strlen(msg), "%s", tagname);
		window_title(dlg, BTOP, TITLELEFT(2), msg);
		set_window_caption(dlg, msg);
	}
	if (list_definitions_live_update & 1)
		position_to_tag_line();
}

spell_configure_helper(int w)
{
	char file[FNAMELEN], ok[FNAMELEN], pr[FNAMELEN];

	sprintf(ok, "NAI%sU", (is_unix_flavor == IS_UNIX_MACOS || is_mswindows) ? "M" : "");
	sprintf(pr, "External spell helper? (None, Aspell, Ispell%s, Unchanged) [U] ",
			is_mswindows ? ", MicroSpell" :
			(is_unix_flavor == IS_UNIX_MACOS) ? ", Mac OS X" : "");
	show_popup_help(w, "The speller can optionally use an external helper "
					"program to provide better, faster suggestions when "
					"correcting misspellings.  If you have one of the supported "
					"programs installed, select it now.\n\nThe aspell and "
					"earlier ispell programs are commonly installed on Unix "
					"systems, and may also be used on Windows systems via the "
					"Cygwin system.\n\nMicroSpell is a commercial spelling "
					"checker that provides hotkey spell checking under "
					"Windows.  See www.microspell.com.  You must have version "
					"9.1f or later.\n\nThe OS X speller is available under Mac "
					"OS X systems (but not over ssh).\n\nThis option sets "
					"the spell-helper-program variable.  If your aspell/ispell "
					"program is not on your path, you'll have to modify this "
					"variable to include its full path.");
	switch (get_key_response(pr, ok, 'U', "spell-configure")) {
		case 'A':	strcpy(spell_helper_program, "aspell -a");
					break;
		case 'I':	strcpy(spell_helper_program, "ispell -a");
					break;
		case 'M':	if (is_unix_flavor == IS_UNIX_MACOS) {
						strcpy(spell_helper_program, "\"%xmacspell\"");
						break;
					}
					get_executable_file(file, "mspellcmd.exe", 0);
					if (!check_file(MICROSPELL_CMDLINE))
						say("Please install %s to %s as admin", file, MICROSPELL_CMDLINE);
					sprintf(spell_helper_program, "\"%s\"", MICROSPELL_CMDLINE);
					break;
		case 'N':	*spell_helper_program = 0;
					break;
	}
}

when_loading()
{
	right_pointer = CURSOR_RIGHTARR;
}
