/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 b3 to 14 b4.

#include "eel.h"
#include "epsnet.h"
#include "tags.h"
#include "chelp.h"
#include "colcode.h"
#include "c.h"
#include "proc.h"
#include "grep.h"
#include "info.h"
#include "menu.h"
#include "misclang.h"
#include "perl.h"
#include "tags.h"
#include "bufed.h"
#include "vspell.h"
#include "epsnet.h"

do_shift_selects()
{
	if (!shift_selects || IS_ASCII_KEY(key)
		  || (NUMPLAIN(key) < KEYEND || NUMPLAIN(key) > KEYPGUP)
		  && (NUMPLAIN(key) < NUMDIGIT(0) || NUMPLAIN(key) > NUMDIGIT(9))
	   ) {	// only affects cursor keys or numpad keys
		if (shift_selecting == 2)		// except at prompts
			highlight_off();
		return;
	}
	if (was_key_shifted()) {
		if (!is_highlight_on())		// if shift key is down, begin/extend
			do_set_mark(point);		// selection
		set_region_type(REGNORM);
		highlight_on();
		shift_selecting = 1;
	} else// if (shift_selecting)	// if shift was just removed, remove highlight
		highlight_off();
}

int c_mute_unmatched_delimiter(int dir);

command c_mode()
{
	mode_default_settings();
	mode_keys = c_tab;		/* use these keys */
	in_perl_buffer = 0;
	indenter = c_indenter;
	auto_indent = 1;
	major_mode = _c_mode_name;
	strcpy(comment_start, "(/</|*>)[ \t\f]*");
	strcpy(comment_pattern, "//.*$|/<*>(.|<newline>)*<*>/<FirstEnd>");
	when_setting_new_c_comments();
	mode_move_level = c_move_level;
	mute_unmatched_delimiter = c_mute_unmatched_delimiter;
	recolor_range = color_c_range;	// set up coloring rules
	recolor_from_here = color_c_from_here;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if (auto_show_c_delimiters)
		auto_show_matching_characters = c_auto_show_delim_chars;
	if (c_tab_override > 0)
		tab_size = c_tab_override;
	buffer_maybe_break_line = c_maybe_break_line;
	if (c_fill_column > 0)
		margin_right = c_fill_column;
	mode_auto_show_delimiter = ifdef_auto_show_delimiter;
	try_calling("c-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

// In C mode, which context should we use?
context_help_override_c(char *mode)
{
	if (c_extra_keywords == EEL_KEYWORDS)	// .h files have this bit too,
		strcpy(mode, "eel");		// but don't assume they're for EEL.
	else if (c_extra_keywords & JAVA_KEYWORDS)
		strcpy(mode, "java");
	else if (c_extra_keywords & JS_KEYWORDS)
		strcpy(mode, "jscript");
	else if (c_extra_keywords & OBJC_KEYWORDS)
		strcpy(mode, "objc");
	else
		strcpy(mode, "c");
}

// Copy src to dest, replacing the first instance of c with repl and
// return 1. If none, copy src to dest and return 0.
do_context_help_replace_in_string(char *dest, char *src, char c, char *repl)
{
	char *s;
	if (s = strchr(src, c)) {
		strncpy(dest, src, s - src);
		strcat(dest, repl);
		strcat(dest, s + 1);
		return 1;
	} 
	strcpy(dest, src);
	return 0;
}

// Display this url in a web browser.
do_context_help_url(char *url)
{
	char item[FNAMELEN], full_url[FNAMELEN];

	context_help_get_item(item, word_pattern);
	if (!do_context_help_replace_in_string(full_url, url, '*', item))
		strcat(full_url, item);
	show_url(full_url); 
}

do_context_help_site_search(char *site, char *site_alt)
{
	char replacement[FNAMELEN];
	do_context_help_replace_in_string(replacement, site_alt, '@', site);
	do_context_help_url(replacement);
}

// Interpret this rule for context help.
use_context_help_rule(char *rule, char *var)
{
	switch (*rule) {
		case '>':	do_context_help_info(rule + 1); break;
		case '!':	do_context_help_run(rule + 1); break;
		case '=':	do_context_help_url(rule + 1); break;
		case '&':	do_context_help_site_search(rule + 1, context_help_site_search); break;
		case '@':	do_context_help_site_search(rule + 1, context_help_site_search_lucky); break;
		case '$':	if (!(has_feature & FEAT_WINHELP))
						error("In variable %s, Winhelp rule unavailable.", var);
					do_context_help_winhelp(rule + 1); break;
		case '+':	try_calling(rule + 1); break;
		default:	if (*rule)
						error("Unknown rule code %c in variable %s.",
							  *rule, var);
					else
						error("Set the %s variable to locate help sources.",
							  var);
	}
}

comp_read(trystr, prmpt, finder, flags, def)
char *trystr, *prmpt;
char *(*finder)();
char *def;		/* default value */
{
	int oldbuf, editbuf, ocmd_len = cmd_len, olen_def_mac = _len_def_mac;
	int skipped_redisp = char_avail();
	char *s, subst[FNAMELEN];

	save_var postpone_asynchronous_notify = 1;
	error_if_input(1);
	*trystr = 0;
	if (run_by_mouse && ungot_key == -1 && finder
		&& ((flags & MUST_MATCH) || show_mouse_choices))
		ungot_key = '?';
	else if (insert_default_response)
		strcpy(trystr, def);
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var return_raw_buttons = 0;
	save_var permit_window_keys |= PERMIT_WHEEL_KEY;
	if ((flags & COMP_FILE_BIT) && !completion_literal_char)
		save_var completion_literal_char = literal_if_quoted;
	_grab_word_loc = point;
	_grab_word_start = 0;
	oldbuf = bufnum;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	editbuf = bufnum = tmp_buf();
	if (flags & COMP_FILE_BIT)
		double_quotes(trystr);
	stuff(trystr);
	full_comp_read(trystr, prmpt, finder, flags, editbuf, def);
	bufnum = oldbuf;
	buf_delete(editbuf);
	if (_len_def_mac >= 0)	// now treat as one big cmd
		cmd_len = ocmd_len + _len_def_mac - olen_def_mac;
	check_abort();
	hide_auto_show_delimiter();
	if (!skipped_redisp)		// skipped before, so skip now
		maybe_refresh();
	if (completion_column_marker) {
		s = strchr(trystr, completion_column_marker);
		if (s)
			*s = 0;
	}
	if (!(flags & PASSWORD_PROMPT))
		save_completion(trystr, finder);
	if (flags & COMP_FILE_BIT) {
		drop_quotes(trystr);
		if (!was_quoted && filename_subst_envvars(trystr, subst))
			strcpy(trystr, subst);
	}
	say("");
}

full_comp_read(trystr, prmpt, finder, flags, edit, def)
char *trystr, *prmpt, *def;
char *(*finder)();
{
	int i, j, num_match = 0, maxlen = ptrlen(trystr) - 1, pbuf, res;
	int pwin, editwin, owin = window_handle, recall_cnt = 0;
	int olen = 0, in_menu = 0, menu = -1, mbuf = 0;
	jmp_buf this_level;
	int *(*lister)();
	char password_display_class[256];

	save_var last_index;
	save_var _help_on_cmd = last_index;
	save_var _doing_input = DI_LINEINPUT;
	save_var _doing_completion = (finder != 0);
	save_var top_level = &this_level;

	while (!in_macro() && !char_avail() && unseen_msgs())
		check_abort();	// Wait for unseen msgs to display so we don't cover them.
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
		pbuf = tmp_buf());	/* make system window with prompt */
	stuff(prmpt);
	i = current_column();
	system_window = 1;	/* don't let user switch to it */
	point = display_column = 0;
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;

	editwin = add_popup(i, echo_line, screen_cols - i, 1,
							0, edit);
	display_column = 0;	/* horizontally scroll this window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	if (flags & PASSWORD_PROMPT) {
		set_chars(password_display_class, '*', 256);
		_display_class = password_display_class;
		build_first = 1;
		save_var disable_recording_last_keys = 1;
	}
	mouse_handler = allow_mouse_switching;
	if (insert_default_response == 1) {
		highlight_on();
		shift_selecting = 2;
	}
	if ((flags & POP_UP_PROMPT) && (has_feature & FEAT_GUI_DIALOGS)) {
		remove_window(pwin);
		do_display_dialog_box("OneLineBox", prmpt, editwin, 0, 0,
							  "OK", "Cancel", "Help", NULL, NULL);
	}
	maybe_refresh();
	if (setjmp(top_level)) {
		if (in_menu == 2) {	/* menu is half-created */
			remove_window(menu);
			buf_delete(mbuf);
			in_menu = mbuf = 0;
			if (is_gui && want_gui_prompts) {
				res = INP_ABORT;
				goto done;
			}
		}
	}
	for (;;) {
		if (mbuf) {
			window_handle = menu;
			if (in_menu && line_gets_highlight())
				make_line_highlight();
			else
				remove_line_highlight();
		}
		window_handle = in_menu ? menu : editwin;
		mode_keys = in_menu ? menu_tab : comp_tab;
		_doing_input = (response_tracks_list && mbuf)
					   ? DI_LISTBOX : DI_LINEINPUT;
		if (_doing_input == DI_LISTBOX) {
			maybe_refresh();
			getkey();
			cmd_result = 0;
			run_topkey();
			res = cmd_result;
		} else
			res = do_recursion();	/* let user edit input */
		if (!is_window(editwin) || in_menu && !is_window(menu)) {
			res = INP_ABORT;
			break;
		} else if (window_handle == (in_menu ? editwin : menu))
			in_menu = !in_menu;
		window_handle = editwin;
		if (res == INP_ADD) {
			window_handle = editwin;
			run_topkey();
			if (search_in_menu && mbuf) {
				window_handle = editwin;
				grab(0, maxlen, trystr);
				if (strlen(trystr) >= olen) {
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			} else
				in_menu = 0;
		} else if (res == INP_SEEN)
			in_menu = 0;
		else if (res == INP_TO_LIST && mbuf) {
			window_handle = menu;
			in_menu = 1;
		} else if ((res == INP_GET || res == INP_GET_EXIT) && mbuf) {
			window_handle = menu;
			get_menu_line(edit);
			window_handle = editwin;
			in_menu = 0;
		} else if (!res) {
			if (_doing_input == DI_LISTBOX)
				if (in_menu) {
					window_handle = menu;
					get_menu_line(edit);
				} else {
					window_handle = editwin;
					grab(0, maxlen, trystr);
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			continue;
		}
		restart_file_paths();
		if (res == INP_ABORT || res == INP_GET_EXIT || res == INP_EXIT)
			break;
		else if (res == INP_GRAB_WORD) {
			grab_buffer_word(owin, 0);
			continue;
		} else if (res == INP_GRAB_LINE) {
			grab_buffer_word(owin, 1);
			continue;
		} else if (res == INP_DEFAULT) {
			if (!size())	/* type in default text */
				stuff(def);
			in_menu = 0;
			continue;
		} else if (res == INP_LAST_RECALL || res == INP_NEXT_RECALL
				   || res == INP_PREV_RECALL) {
			if (res == INP_LAST_RECALL)
				get_last_recall(finder, trystr);
			else
				recall_cnt = get_last_recall_from_end(finder, trystr,
					recall_cnt + (res == INP_NEXT_RECALL ? -1 : 1));
			if ((*trystr || res != INP_LAST_RECALL) && !mbuf) {
				delete(0, size());
				stuff(trystr);
				if (recall_response_selected && *trystr) {
					highlight_on();
					shift_selecting = 2;
				}
			}
			in_menu = 0;
		} else if (res == INP_GET_OTHER) {
			if (inp_get_other_name && *inp_get_other_name) {
				delete(0, size());
				stuff(inp_get_other_name);
			}
			in_menu = 0;
		} else if (res == INP_SHOW_MATCHES || res == INP_RECALL) {
show_matches:		grab(0, maxlen, trystr);	/* typed '?' */
			if (is_highlight_on() || res == INP_RECALL) {
				if (flags & COMP_FILE_BIT)
					*get_tail(trystr, 0) = 0;
				else
					*trystr = 0;
				highlight_off();
			}
			olen = strlen(trystr);
			if (!mbuf) {
				menu = new_menu(mbuf = tmp_buf());
				if (is_gui && want_gui_prompts) {
					remove_window(pwin);
					say("");
					make_prompt_box(res == INP_RECALL, prmpt, editwin, menu, finder);
				}
			}
			in_menu = 2;		/* initializing */
			if (res == INP_RECALL)
				lister = list_recalls;
			else if (completion_lister)
				lister = completion_lister;
			else
				lister = list_matches;
			make_menu(menu, trystr, finder, flags, editwin, mbuf, lister);
			in_menu = 1;
			mode_keys = menu_tab;
			cursor_to_column = -1;
		} else if (res == INP_TRY || res == INP_COMPLETE) {
			if (res == INP_TRY && cursor_to_column
						== current_column() + 1)
				break;	/* exactly 1 match last time */
			grab(0, maxlen, trystr);
			i = strlen(trystr);
			num_match = complete(trystr, finder, flags);
			j = strlen(trystr);
			if ((flags & SPACE_VALID) && key == ' ' && trystr[i] == ' '
				  && could_add_space(trystr, finder, flags) == 1)
				i++, num_match++;
			else if (num_match && i != j)
				i = j;
			else if (num_match != 1) {
				if ((flags & SPACE_VALID) && key == ' ' && is_window(menu)
//					  && trystr[i] == ' '
// Given dirs c:\program files and c:\programs, this commented-out code
// made selecting the former fail.
					  && could_add_space(trystr, finder, flags)) {
					trystr[i++] = ' ';
				} else {
					maybe_ding(bell_on_completion);
					if (completion_pops_up)
						ungot_key = '?';
				}
			}
			if ((flags & COMP_FILE_BIT) && num_match == 1 && i > 0
			    && is_path_separator(trystr[i - 1]))
				num_match++;	/* might be directory */
			trystr[i] = 0;
			buf_zap(edit);
			stuff(trystr);
			if (num_match == 1) {
				if (res == INP_COMPLETE)
					break;	/* <Esc> worked */
				maybe_refresh();
				cursor_to_column = current_column() + 1;
			}
		}
	}
done:
	if (res != INP_ABORT) {
		grab(0, maxlen, trystr);
		if (size() > maxlen)
			say("Response exceeds internal limit of %d characters; truncated.",
				  maxlen);
	} else if (!user_abort)		/* user deleted edit window */
		user_abort = 1;
	if (res == INP_EXIT && must_match(trystr, finder, flags))
		goto show_matches;
	if (is_window(owin))
		window_handle = owin;
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(pbuf);
	if (mbuf) {
		remove_window(menu);
		buf_delete(mbuf);
	}
}

// Write to buffer b a list of previous responses to any prompt that
// match the given partial response.
previous_response_matches(int b, char *resp, int flags)
{
	char oldresp[FNAMELEN], pat[FNAMELEN];
	buf_zap(b);
	if (exist(OLDRESP)) {
		save_var bufname = OLDRESP;
		save_var point = 0;
		sprintf(pat, ": (%q.*)", resp);
		while (re_search(1, pat)) {
			grab(find_group(1, 1), find_group(1, 0), oldresp);
			if ((flags & EXACTONLY)
				  ? !strcmp(oldresp, resp)
				  : !strnfcmp(oldresp, resp, strlen(resp)))
				buf_printf(b, "%s\n", oldresp);
		}
		buf_sort_and_uniq(b);
	}
	set_buf_point(b, 0);
}

#include <rmode.e>

// Write to buffer b a list of URLs based on url, that match the
// partial host name.
remote_host_match(int b, char *url, char *parthost, int flags)
{
	char host[FNAMELEN], newurl[FNAMELEN], pat[FNAMELEN];
	int hostpos = strlen(url) - strlen(parthost);
	buf_zap(b);
	strcpy(newurl, url);
	if (exist(OLDRESP)) {
		save_var bufname = OLDRESP;
		save_var point = 0;
		sprintf(pat, "(%s):[/\\][/\\](%q[^/\\\n]+)", url_services, parthost);
		while (re_search(1, pat)) {
			grab(find_group(2, 1), find_group(2, 0), host);
			strcpy(newurl + hostpos, host);
			if ((flags & EXACTONLY)
				  ? !strcmp(parthost, host)
				  : !strnfcmp(parthost, host, strlen(parthost)))
				buf_printf(b, "%s/\n", newurl);
		}
		buf_sort_and_uniq(b);
	}
	set_buf_point(b, 0);
}

char *remote_file_match(char *s, int flags)
{
	struct url_parts parts;
	int b = create(_MATCH_BUF);

	if (!parse_url(s, &parts))
		return NULL;
	if (flags & STARTMATCH) {
		if (!*parts.fname && (!*parts.host || !is_path_separator(s[strlen(s) - 1])))
			remote_host_match(b, s, parts.host, flags);	// Complete on partial hostname.
		else if (!strcmp(parts.service, "scp"))
			scp_get_response((flags & EXACTONLY) ? 3 : 0, s, &parts, b);
		else if (!strcmp(parts.service, "ftp"))
			ftp_file_match(s, flags);
		else
			previous_response_matches(b, s, flags);
		if (flags & FM_NO_DIRS) {
			save_var bufnum = b;
			simple_re_replace(0, ".*[/\\]\n", "");
			restore_vars();
		}
	}
	if (!grab_line(b, _this_match))
		return NULL;
	return _this_match;
}

struct scp_job *do_scp(struct url_parts *parts, int op);

scp_misc_op(char *s, char *cmds)
{
	struct url_parts parts;
	struct scp_job *p;

	if (!parse_url(s, &parts) || strcmp(parts.service, "scp") || !*parts.host)
		return 0;
	p  = do_scp(&parts, FTP_MISC);
	if (p->flags & SCP_CLIENT_MODE)
		p->cmd = p->cmdlist = strsave(cmds);
	else
		scp_rewrite_commands(p, cmds);
	if (p->busy != 2)
		scp_next_cmd(p);
	return 1;
}

int parse_url(char *url, struct url_parts *p)
{
	memset(p, 0, sizeof(*p));
	strcpy(p->service, url);			// Parse service:// part.
	if (!split_string(p->service, ":", p->usr) || !*p->service
		|| is_path_separator(p->usr[0]) != 1
		|| is_path_separator(p->usr[1]) != 1)	// It's not valid.
		return 0;
	strcpy(p->usr, p->usr + 2);		// Split off the service part.
	split_string(p->usr, "/\\", p->fname);		// Remove the file name.

	if (!reverse_split_string(p->usr, "@", p->host)) {
										// Do we have a user part?
		strcpy(p->host, p->usr);		// No, so this must be the
		*p->usr = 0;					// host part.
	} else
		p->have_usr = 1;

	if (split_string(p->usr, ":", p->pwd))	// User part could have a pwd.
		p->have_password = 1;		// Pwd could be specified but empty.
	url_unescape(p->usr);
	url_unescape(p->pwd);
	if (split_string(p->host, ":", p->port))
		if (!*p->port)	// Host part could have a port (must be nonempty).
			return 0;
	return 1; // *p->host != 0;		// (Host name must be nonempty.) No, let caller check that.
}

remote_file_type(char *file)
{
	struct url_parts parts;
	char *s;

	save_var case_fold = 1;
	if (!is_remote_file(file))
		return 1;
	if (!parse_url(file, &parts) || !*parts.host)
		return 2;
	if (!strstr(" ftp scp ", parts.service))
		return 3;
	if (!*parts.fname)
		return 4;
	if (is_pattern(parts.fname) & 2)
		return 5;		// Use dired if it has * or ? chars.
	if (is_path_separator(parts.fname[strlen(parts.fname) - 1]))
		return 6;
	s = index(parts.fname, '~');	// a user's home dir
	if (s && !index(s, '/') && !index(s, '\\'))
		return 7;
	if (is_ftp_url_a_dir(file))
		return 8;
	return 0;
}

char *get_dired_buffer_pattern();

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN], dir[FNAMELEN];
	int show_rel = 1;		// Show the relative path of the default.

	if (!inbufs) {
		save_var recall_id = "f-match grep";	// Use same cmd history for all greplike cmds.
		if (is_dired_buf() && !inp_get_other_name)
			save_var inp_get_other_name = get_dired_buffer_pattern();
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory
			  || grep_default_directory == 3)
			show_rel = 0;
		if (grep_prompt_with_buffer_directory == 2 && !grep_default_directory)
			show_rel = 0;	// Ensure we show which files we'll search by default.
		build_prompt(pr, "In file(s)", _grep_files, 0, show_rel);
		save_var prompt_with_buffer_directory = grep_prompt_with_buffer_directory;
		get_file_dir(filepat, pr);
		if (!*filepat && grep_default_directory) {
			save_var _help_on_cmd = last_index;
			if (grep_default_directory == 3)
				strcpy(filepat, _grep_files);
			else
				get_last_recall(f_match, filepat);
			if (grep_default_directory >= 2 && *filepat
				  && get_buffer_directory(dir)) {
				if (grep_default_directory != 3)
					absolute(filepat, dir);
				return 1;
			}
		}
		if (*filepat) {
			if (is_directory(filepat))
				sprintf(filepat + strlen(filepat), "%c*", path_sep);
			strcpy(_grep_files, filepat);
		}
		if (grep_default_directory == 3)
			strcpy(filepat, _grep_files);
		else
			relative(_grep_files, filepat);
		if (grep_default_directory == 1 && !grep_prompt_with_buffer_directory)
			strcpy(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else {
		sprintf(pr, "In buffers matching pattern (default: this buffer): ");
		comp_read(filepat, pr, b_match,
				  ((preserve_filename_case & 4) ? FM_FOLD : 0), bufname);
		if (!*filepat)
			strcpy(filepat, bufname);
	}
	return 1;
}

// Display topic in a browser and return 1.  Return 0 if running a web
// browser is unsupported, or -1 if we couldn't for some other reason.
int try_html_help_string(char *topic)
{
	char url[FNAMELEN], prog[FNAMELEN], *p, *q, cmdline[FNAMELEN], dir[FNAMELEN];

	if (epsilon_manual_port > 0)
		sprintf(url, "http://localhost:%d/man/", epsilon_manual_port);
	else if (epsilon_manual_port == -1)
		strcpy(url, "https://www.lugaru.com/cgi-bin/lhelp/man/");
	else
		sprintf(url, "https://www.lugaru.com/xman/v%s/", versioned_file_string);
	for (p = topic, q = url + strlen(url); *p; p++, q++)
		if (*p == ' ') {
			strcpy(q, "%20");
			q += 2;
		} else
			*q = *p;
	*q = 0;
	if (!*topic)
		strcat(url, "0");
	p = lookpath("lhelp/epsilon.lhl");
	if (p) {		// Find name of directory containing our lhelp files.
		strcpy(dir, p);
		*get_tail(dir, 0) = 0;
	} else
		get_executable_directory(dir);
	convert_to_8_3_filename(dir);
	sayput("Starting a browser, please wait...");
	expire_message = 1;
	if (opsys == OS_UNIX) {
		get_executable_file(prog, "lhelpd", 1);
		sprintf(cmdline, "%s -p %d -d \"%s\" &", prog, epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			shell("", cmdline, "");
	} else {
		get_executable_file(prog, "lhelp.exe", 0);
		sprintf(cmdline, "lhelp -p %d -d \"%s\"", epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			winexec(prog, cmdline, SW_SHOWMINNOACTIVE, 0);
	}
	if (!try_show_url(url))
		return -1;
	note("");
	return 1;
}

char *man_use_program(int type);

make_man_page(char *topic, int type)
{
	char cmdline[FNAMELEN], buf[FNAMELEN];
	int color;

	save_var case_fold = 1;
	sprintf(cmdline, "%s %s", man_use_program(type), topic);
	sprintf(buf, "man-%s", topic);
	quiet_set_bookmark();
	to_buffer(buf);
	man_mode();
	putenv(strkeep("MAN_KEEP_FORMATTING=Y"));
	if (!size() || type != _man_program) {
		set_read_only(0);
		zap(buf);
		translation_type = FILETYPE_AUTO;
		pipe_text(NULL, buf, cmdline, NULL,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		point = 0;
		if (parse_string(1, "Reformatting.*please wait.*\n"))
			delete(0, matchend);
		simple_re_replace(1, "\n\n\n+", "\n\n");
		set_character_color(0, size(), -1);
		point = 0;
		while (re_search(1, ".\b.")) {
			if (character(point - 3) == '_')
				color = color_class man_underlined;
			else
				color = color_class man_bold;
			delete(point - 3, point - 1);
			set_character_color(point - 1, point, color);
			point--;
		}
		color_ansi_buffer();
		filename = "fake-man-file.man";	// So find-linked-files works.
		force_save_as = FSA_NEWFILE;
		discardable_buffer = buffer_not_saveable = 1;
	}
	modified = point = 0;
	_man_program = type;
	set_read_only(3);
}

// Display a text-based popup menu.
text_context_menu()
{
	char cmd[FNAMELEN], widthb[30];
	save_var mouse_mask |= MASK_MOVE | MASK_UP;
	load_menu();
	save_var mouse_cursor = left_pointer, window_handle;
	int popup = add_popup(mouse_x + 4, mouse_y, 30, screen_lines, ALL_BORD(BSINGLE), MENU_BNUM);
	prep_menu_window(color_class submenu);
	on_exit remove_window(popup);
	int isdir;
	display_scroll_bar = 0;
	point = 0;
	if (!re_search(1, "^ _popup[ \t]* ")	// look for popup menu
		  || !parse_string(1, "[0-9]+", widthb))	// get width of menu
		return 1;
	window_width = numtoi(widthb) + 2;
	nl_forward();
	window_start = point;
	save_var narrow_start = point;
	search(1, "\n\n");		// go to blank line after menu items
	save_var narrow_end = size() - matchstart;
	window_height = lines_between(window_start, point) + 3;
	point = window_start;
	while ((mouse_pressed() || char_avail()) && !user_abort)
		getkey();		// Wait for release of right button
	do {
		submenu_highlight(1);
		maybe_refresh();
		getkey();
		isdir = 0;
		switch (get_direction()) {
			case BTOP:
				point--;
				isdir = 1;
				break;
			case BBOTTOM:
				nl_forward();
				isdir = 1;
				break;
		}
		if (isalpha(key)) {
			go_line(do_menu_search(key) + 1);
			isdir = 1;
		}
		to_begin_line();
		if (IS_MOUSE_KEY(key) && window_at_coords(mouse_y, mouse_x, mouse_screen) == popup) {
			to_mouse_position(mouse_x, mouse_y);
			to_begin_line();
		}
	} while (IS_MOUSE_KEY(key) && !IS_MOUSE_UP(key) || isdir);
	*cmd = 0;
	if (IS_MOUSE_UP(key) && window_at_coords(mouse_y, mouse_x, mouse_screen) == popup
		  || exit_menu_doit())
		get_menu_cmd(cmd, popup);
	restore_vars();
	if (!*cmd || *cmd == '#')
		return 2;
	flash_menu_item();
	run_menu_selection(cmd, (shift_pressed()
							 & (KB_ALT_DN | KB_CTRL_DN | KB_SHIFT_DN)));
	return 0;
}

int color_nsis_range(int from, int to);

command nsis_mode()
{
	mode_default_settings();
	major_mode = strkeep("NSIS");
	mode_keys = nsis_tab;		/* Use these keys. */
	strcpy(comment_start, "[;#][ \t]*");
	strcpy(comment_pattern, "[;#].*$");
	strcpy(comment_begin, "; ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_nsis_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 0x200) != 0;
	auto_show_matching_characters = nsis_auto_show_delim_chars;
	mode_auto_show_delimiter = ifdef_auto_show_delimiter;
	try_calling("nsis-mode-hook");
	make_mode();
}

command mouse_to_tag()
{
	int k = key;

	if (mouse_other_handler())
		return;
	else if (!in_text_area(mouse_x, mouse_y)) // clicked in border
		mouse_move_border(1);
	else if (!IS_MOUSE_DOUBLE(key))
		mouse_selecting();
	else if (mouse_dbl_selects)
		run_enter_key();
	else {
		mouse_display = 0;
		point_to_pos(mouse_x, mouse_y, mouse_screen);
		pluck_tag();
	}
	mouse_call_hook(k);	// call correct hook for this key
}

command context_menu() on reg_tab[FSHIFT(10)], reg_tab[MOUSE_RIGHT_DN], reg_tab[MOUSE_DBL_RIGHT]
{
	if (has_feature & FEAT_OS_MENUS)
		win_menu_popup("_popup");
	else
		text_context_menu();
	invisible_cmd();
}

// If the file doesn't exist, substitute a file with the same basename.
proc_fix_filename(char *fname)
{
	if (!(process_next_error_options & 1) || is_remote_file(fname)
		  || check_file(fname))
		return;
	int orig = bufnum;
	int i = buf_list(0, 2);	/* start with most recently created buf */
	do {
		bufnum = i;
		if (*filename
			  && !fnamecmp(get_tail(filename, 0), get_tail(fname, 0))
			  && force_save_as != FSA_NEWFILE
			  && check_file(filename) == CHECK_FILE) {
			show_text((process_next_error_options & 2) ? 0 : -2, see_delay,
					  "No file %f found, picking another %s",
					  fname, get_tail(fname, 0));
			strcpy(fname, filename);
			return;
		}
	} while (i = buf_list(-1, 1));
	bufnum = orig;
}


// First, define some regex pieces that could be helpful in multiple
// definitions.

// These file name matchers say, optionally a drive spec c:, then a
// basename, a dot, and an extension.  The OK_SPACE version permits
// spaces in the basename, starting with the 2nd character.  Both permit
// tags like <3> at the end of the extension.
#define FNAME_NO_SPACE	"(([a-z]:)?[+-9A-z~]+<.>[+-9A-z]+[0-9<>]*)"
#define FNAME_OK_SPACE	"(([a-z]:)?[+-9A-z~][+-9A-z~ ()]*<.>[+-9A-z]+[0-9<>]*)"

// Some things we want to ignore in an error message.

// 3> or E " or W " or ***, for example.
#define START_STUFF1	"[1-9]>|e \"|w \"|<*>*"
// One of these three keywords, or [someother].
#define START_STUFF2	"error|warning|fatal| *%[[a-z]+%]"
// Optionally, a code like E123 with a space before it, then perhaps
// some punc chars * : after.  Or just punc chars " or (.
#define START_STUFF3	"( [a-z][0-9]+)?[*:]*|[\"(]?"

// Between a file name and line number, we must have space, open paren,
// comma or colon, then perhaps L or the word "line" or a [, then
// perhaps space.  Also skip over any " at the start, which may have
// surrounded file name.
#define BEFORE_NUM		"\"?([ (:,]( *line|L|%[)?) *"
// The line number.
#define LINE_NUM		"([0-9]+)"
// Stuff to require after line number.  Either ) or colon, or else some
// stuff with colon or semicolon after it, or else a column number and ].
#define END_STUFF		"(<rparen|:>|[^0-9\\\n].*[:;]|,[0-9]+%][ \n]*)"

/* If you change ERROR_PATTERN, you may have to change the next 3 macros
 * as well. */
#define ERROR_PATTERN	"^(" START_STUFF1 \
						"(" START_STUFF2 "|[a-z0-9]+ )" \
						START_STUFF3 "|[ \t]*from) *" \
						FNAME_NO_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN	4	/* filename is in 4th () pair above */
#define LINE_PAREN	8	/* line# is in 8th pair of parens above */

#define SIMP_PATTERN	"[:;)]| line "	/* search with simple pattern first */

// This version permits spaces in file names, but with more
// restrictions on ancillary text before it.
#define ERROR_PATTERN2	"^(" START_STUFF1 \
						"(" START_STUFF2 ")" \
						"( [a-z][0-9]+)?[*:]*|[\"(]?) *" \
						FNAME_OK_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN2	4	/* filename is in 4th () pair above */
#define LINE_PAREN2	8	/* line# is in 8th pair of parens above */

#define ERROR_PATTERN3	".* at (.+) line ([0-9]+)[.,].*"
#define FILE_PAREN3	1	/* filename in () pair above */
#define LINE_PAREN3	2	/* line# in pair of parens above */

// An error pattern that permits file names with no . character, but
// with more restrictions on the format.
#define ERROR_PATTERN4	"^([+-9A-z~:]+[0-9<>]*)" BEFORE_NUM \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN4	1	/* filename is in 1st () pair above */
#define LINE_PAREN4	2	/* line# is in 2nd pair of parens above */

// An error pattern for Java runtime errors.

#define ERROR_PATTERN5	"^.*%(([+-9A-z~:]+):([0-9]+)%)"
#define FILE_PAREN5	1	/* filename is in 1st () pair above */
#define LINE_PAREN5	2	/* line# is in 2nd pair of parens above */

// An error pattern that permits file names with no . character,
// allowing spaces, but with even more restrictions on the rest of the
// message.
#define ERROR_PATTERN6	"^([a-z]:\\[ +-9A-z~]+)[(]" \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN6	1	/* filename is in 1st () pair above */
#define LINE_PAREN6	2	/* line# is in 2nd pair of parens above */

/* Find next error in direction, return 1 and set vars to find_group() paren count */
find_error(int dir, int *file_paren, int *line_paren)
{
	int old = point;

	if (point > 0 && dir > 0)
		to_end_line();
	else
		to_begin_line();
	while (re_search(dir, SIMP_PATTERN)) {
		to_begin_line();
		if ((!*ignore_error || !parse_string(1, ignore_error, NULL))
			  && !parse_string(1, "In file included from ", NULL)) {
			if (parse_string(1, ERROR_PATTERN, (char *)0)) {
				*file_paren = FILE_PAREN;
				*line_paren = LINE_PAREN;
				return 1;
			}
#ifdef ERROR_PATTERN2
			if (parse_string(1, ERROR_PATTERN2, (char *)0)) {
				*file_paren = FILE_PAREN2;
				*line_paren = LINE_PAREN2;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN3
			if (parse_string(1, ERROR_PATTERN3, (char *)0)) {
				*file_paren = FILE_PAREN3;
				*line_paren = LINE_PAREN3;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN4
			if (parse_string(1, ERROR_PATTERN4, (char *)0)) {
				*file_paren = FILE_PAREN4;
				*line_paren = LINE_PAREN4;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN5
			if (parse_string(1, ERROR_PATTERN5, (char *)0)) {
				*file_paren = FILE_PAREN5;
				*line_paren = LINE_PAREN5;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN6
			if (parse_string(1, ERROR_PATTERN6, (char *)0)) {
				*file_paren = FILE_PAREN6;
				*line_paren = LINE_PAREN6;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN7
			if (parse_string(1, ERROR_PATTERN7, (char *)0)) {
				*file_paren = FILE_PAREN7;
				*line_paren = LINE_PAREN7;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN8
			if (parse_string(1, ERROR_PATTERN8, (char *)0)) {
				*file_paren = FILE_PAREN8;
				*line_paren = LINE_PAREN8;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN9
			if (parse_string(1, ERROR_PATTERN9, (char *)0)) {
				*file_paren = FILE_PAREN9;
				*line_paren = LINE_PAREN9;
				return 1;
			}
#endif
		}
		if (dir > 0)
			nl_forward();
	}
	*error_spot = old;
	return 0;
}

// At a $(, starting at p, so find matching ), and color through it,
// moving past.
shell_parse_substitution(int p)
{
	shell_fwd_quote('(');
	if (character(p + 2) == '(' && curchar() == ')')
		point++;			// Found $((..))
	set_character_color(p, point, color_class perl_constant);
}

load_color()
{
	int i, j, fore, back;
	char scheme[FNAMELEN], class[FNAMELEN], num[FNAMELEN], num2[FNAMELEN];

	point++;			/* skip initial & */
	parse_string(1, "[^ \t\n:]+", scheme);
	i = find_index(scheme);
	if (!parse_string(1, " color class for ", class)
			|| !parse_string(1, "[^ \t\n:]+", class)
			|| !parse_string(1, ": ", num)
			|| !parse_string(1, "[[=]?[-a-zA-Z0-9]+", num)) {
		init_warning("Bad color definition for %s.", scheme);
		nl_forward();
		return;
	}
	if (i <= 0)
		i = new_variable(scheme, NT_COLSCHEME, 0);
	for (j = 0; strcmp(class, name_color_class(j)); )
		if (++j >= number_of_color_classes()) {
			init_warning("Undefined color class %s.", class);
			nl_forward();
			return;
		}
	if (*num == '=') {		// copy definition
		for (fore = 0; strcmp(num + 1, name_color_class(fore)); )
			if (++fore >= number_of_color_classes()) {
				init_warning("Undefined color class %s.", num + 1);
				nl_forward();
				return;
			}
		fore |= ECOLOR_COPY;
		back = 0;
	} else if (*num == '[') {		// rgb-style definition
		fore = numtoi(num + 1);
		if (!parse_string(1, " on ", num)
			  || !parse_string(1, "[0-9a-fx]+]", num)) {
			init_warning("Bad rgb color definition for %s.", scheme);
			nl_forward();
			return;
		}
		back = numtoi(num);
		fore = SWAP_OLDCOLOR(fore);		// If that's all, then it uses pre-v14 coding.
		back = SWAP_OLDCOLOR(back);
		if (parse_string(1, "#RGB %[", num)) {	// Found new-style coding.
			if (!parse_string(1, "[0-9a-fx]+", num)
				  || !parse_string(1, " on ", num2)
				  || !parse_string(1, "[0-9a-fx]+]", num2)) {
				init_warning("Bad new-rgb color definition for %s.", scheme);
				nl_forward();
				return;
			}
			fore = numtoi(num);
			back = numtoi(num2);
		}
	} else {
		fore = find_color_code(num);
		if (fore < 0) {
			init_warning("Bad color name %s.", num);
			nl_forward();
			return;
		}
		if (parse_string(1, " on ", num)
			&& parse_string(1, "[-a-zA-Z0-9]+", num)) {
			back = find_color_code(num);
			if (back < 0) {
				init_warning("Bad color name %s.", num);
				nl_forward();
				return;
			}
		}
		fore = attr_to_rgb(fore);
		back = attr_to_rgb(back);
	}
	set_color_pair(j, fore, back, i);
}

init_warning(msg, s)		/* display error and return */
char *msg, *s;
{
	char full_msg[FNAMELEN];
	_errors_in_cmd_file++;
	if (_abort_load_on_error == 1) {
		sprintf(full_msg, msg, s);
		say("%s (line %d)", full_msg, lines_between(0, point, 0) + 1);
	} else if (_abort_load_on_error == 2) {
		buffer_printf(MSG_BUFFER, "%s %d: ", einit_file_name,
					  lines_between(0, point, 0) + 1);
		buffer_printf(MSG_BUFFER, msg, s);
		buffer_printf(MSG_BUFFER, "\n");
	} else {
		if (!nl_forward())
			insert('\n');
		stuff("# COMMAND FILE ERROR: ");
		bprintf(msg, s);
		stuff("\n");
	}
}

// Output color info on this class in current scheme to buf.
list_one_color_setting(int buf, int class, int raw)
{
	char colorpair[FNAMELEN];
	int fg, bg;

	fg = get_foreground_color(class, raw);
	if ((fg & ETRANSPARENT) && !raw)
		fg = get_foreground_color(0);
	bg = get_background_color(class, raw);
	if ((bg & ETRANSPARENT) && !raw)
		bg = get_background_color(0);
	if (fg & ECOLOR_COPY)
		sprintf(colorpair, "=%s", name_color_class(COLOR_STRIP_ATTR(fg)));
	else if (!(fg & ~0xffffff) && fg == attr_to_rgb(rgb_to_attr(fg))
			 && !(bg & ~0xffffff) && bg == attr_to_rgb(rgb_to_attr(bg)))
		give_color_combo(colorpair, alter_color(class, -1));
	else
		sprintf(colorpair, "[0x%x on 0x%x]#RGB [0x%x on 0x%x]",
				SWAP_OLDCOLOR(fg), SWAP_OLDCOLOR(bg), fg, bg);
	buf_printf(buf, "&%s color class for %s: %s\n",
			   name_name(selected_color_scheme),
			   name_color_class(class), colorpair);
}

spell_configure_helper(int w)
{
	char file[FNAMELEN], ok[FNAMELEN], pr[FNAMELEN];

	sprintf(ok, "NAI%sU", (is_unix_flavor == IS_UNIX_MACOS || is_mswindows) ? "M" : "");
	sprintf(pr, "External spell helper? (None, Aspell, Ispell%s, Unchanged) [U] ",
			is_mswindows ? ", MicroSpell" :
			(is_unix_flavor == IS_UNIX_MACOS) ? ", macOS" : "");
	show_popup_help(w, "The speller can optionally use an external helper "
					"program to provide better, faster suggestions when "
					"correcting misspellings.  If you have one of the supported "
					"programs installed, select it now.\n\nThe aspell and "
					"earlier ispell programs are commonly installed on Unix "
					"systems, and may also be used on Windows systems via the "
					"Cygwin system.\n\nMicroSpell is a commercial spelling "
					"checker that provides hotkey spell checking under "
					"Windows.  See www.microspell.com.  You must have version "
					"9.1f or later.\n\nThe macOS speller is available under "
					"macOS systems (but not over ssh).\n\nThis option sets "
					"the spell-helper-program variable.  If your aspell/ispell "
					"program is not on your path, you'll have to modify this "
					"variable to include its full path.");
	switch (get_key_response(pr, ok, 'U', "spell-configure")) {
		case 'A':	strcpy(spell_helper_program, "aspell -a");
					break;
		case 'I':	strcpy(spell_helper_program, "ispell -a");
					break;
		case 'M':	if (is_unix_flavor == IS_UNIX_MACOS) {
						strcpy(spell_helper_program, "\"%xmacspell\"");
						break;
					}
					get_executable_file(file, "mspellcmd.exe", 0);
					if (!check_file(MICROSPELL_CMDLINE))
						say("Please install %s to %s as admin", file, MICROSPELL_CMDLINE);
					sprintf(spell_helper_program, "\"%s\"", MICROSPELL_CMDLINE);
					break;
		case 'N':	*spell_helper_program = 0;
					break;
	}
}

when_loading()
{
	drop_name("mouse_goes_to_tag");
	drop_name("c_mode_mouse_to_tag");

	if (!strcmp(context_help_default_rule, "=http://www.google.com/search?q="))
		strcpy(context_help_default_rule, "=https://www.google.com/search?q=");

	if (!strcmp(context_help_rule_c_windows,
				"=http://www.google.com/search?q=*+site:docs.microsoft.com"))
		strcpy(context_help_rule_c_windows,
			   "@docs.microsoft.com");

	if (!strcmp(context_help_rule_php,
				"=http://www.google.com/search?q=*+site:www.php.net&l=en&btnI=I"))
		strcpy(context_help_rule_php, "@www.php.net");

	if (!strcmp(context_help_rule_java,
				"=http://www.google.com/search?q=*+site:java.sun.com&l=en&btnI=I"))
		strcpy(context_help_rule_java, "@www.w3schools.com");

	if (!strcmp(context_help_rule_html,
				"=http://www.google.com/search?q=*+site:www.htmlhelp.com&l=en&btnI=I"))
		strcpy(context_help_rule_html, "@www.w3schools.com");

	if (!strcmp(context_help_rule_jscript,
				"=http://www.google.com/search?q=*+site:developer.mozilla.org&l=en&btnI=I"))
		strcpy(context_help_rule_jscript, "@developer.mozilla.org");

	if (!strcmp(context_help_rule_vbscript,
				"=http://www.google.com/search?q=*+site:www.w3schools.com&l=en&btnI=I"))
		strcpy(context_help_rule_vbscript, "@www.w3schools.com");

	if (!strcmp(context_help_rule_vbasic_windows,
				"=http://www.google.com/search?q=*+site:docs.microsoft.com"))
		strcpy(context_help_rule_vbasic_windows, "@docs.microsoft.com");

	if (!strcmp(context_help_rule_python,
				"=http://www.google.com/search?q=*+site:docs.python.org&l=en&btnI=I"))
		strcpy(context_help_rule_python, "@docs.python.org");

	if (!strcmp(context_help_rule_latex,
				"=http://www.google.com/search?q=%22hypertext+help%22+latex+*&l=en&btnI=I"))
		strcpy(context_help_rule_latex,
			   "=https://www.google.com/search?q=%22hypertext+help%22+latex+*&l=en");

	if (!strcmp(context_help_rule_vhdl,
				"=http://www.google.com/search?q=*+site:www.acc-eda.com&l=en&btnI=I"))
		strcpy(context_help_rule_vhdl, "@www.vhdl.renerta.com");

	if (!strcmp(context_help_rule_tcl,
				"=http://www.google.com/search?q=*+-%22keyword+index%22+site:tmml.sourceforge.net&l=en&btnI=I"))
		strcpy(context_help_rule_tcl,
			   "=https://duckduckgo.com/?q=\\*+-%22keyword+index%22+site:tmml.sourceforge.net");

	if (!strcmp(context_help_rule_asm_windows,
				"=http://www.google.com/search?q=*+site:docs.microsoft.com"))
		strcpy(context_help_rule_asm_windows, "@docs.microsoft.com");
}
