/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 b4 to 14 b5.

#include "eel.h"
#include "colcode.h"
#include "html.h"
#include "info.h"

do_shift_selects()
{
	int arrow = (NUMPLAIN(key) >= KEYEND && NUMPLAIN(key) <= KEYPGUP)
				|| (NUMPLAIN(key) >= NUMDIGIT(0) && NUMPLAIN(key) <= NUMDIGIT(9));
	if (!shift_selects || IS_ASCII_KEY(key) || !arrow) {	// only affects cursor keys or numpad keys
		if (shift_selecting == 2)		// except at prompts
			highlight_off();
		return;
	}
	if (was_key_shifted()) {
		if (!is_highlight_on())		// if shift key is down, begin/extend
			do_set_mark(point);		// selection
		set_region_type(REGNORM);
		highlight_on();
		shift_selecting = 1;
	} else if (arrow && shift_selecting)	// if shift was just removed, remove highlight
		highlight_off();
}

// On each refresh, recolor_partial_code() will be called with the
// buffer range that was modified since the last redisplay. It handles
// changes in syntax highlighting rules, and sees that the visible
// buffer region is syntax highlighted.
recolor_partial_code(from, to)
{
	init_color_code();
	if (!(coloring_flags & COLOR_IN_PROGRESS)) {
		coloring_flags |= COLOR_IN_PROGRESS; // ensure errors don't restart us
		if (!(coloring_flags & COLOR_MINIMAL) != !minimal_coloring) {
			drop_all_colored_regions();		// If setting's changed.
			coloring_flags &= ~COLOR_MINIMAL;
			if (minimal_coloring)
				coloring_flags |= COLOR_MINIMAL;
		}
		save_var point, matchstart, matchend, abort_searching = 0;
		if (!(coloring_flags & COLOR_RETAIN_NARROWING))
			save_var narrow_start = 0, narrow_end = 0;
		shrink_colored_area();		// Buffer modifications shrink list of regions.
		if (size() < color_whole_buffer) // Now expand list of regions.
			color_for_area(0, size());
		else {
			int end = MIN(window_end + 1, size());
			if (display_column >= 0)
				hscroll_color_for_area(window_start, end);
			else
				color_for_area(window_start, end);
		}
		coloring_flags &= ~COLOR_IN_PROGRESS;	// no errors aborted us,
	}					// let us run next time
	default_when_displaying(from, to);
}

// We're displaying the buffer region start...end, so syntax-highlight
// it if needed. But since we're horizontally scrolling, and lines might
// extend far to the right, we'll avoid wasting time coloring those
// parts, by coloring line by line.
void hscroll_color_for_area(int start, int end)
{
	int i, from, to, next, max = text_height();
	int last_column = text_width() + display_column;
	save_var point;
	for (i = 0; i < max; i++) {		// For each screen line of this window
		start = window_line_to_position(i);
		if (start >= end)	// If its start is past the region to color, done.
			break;
		// To avoid an expensive move_to_column(), we first check if
		// this line is already entirely colored.
		if (!get_tagged_region("needs-color", start, &from, &to)) {
			// The start of this line is already colored and it's part
			// of the already-colored region from...to.
			next = window_line_to_position(i + 1);	// Start of next screen line.
			if (next == -1 && i == max - 1)	// Last screen line, so we can't get
				next = end;		// start of the next one -- substitute.
			if (next != -1 && next <= to)
				continue;		// Entire line is already colored.
		}
		point = start;		// Find buffer pos of the last column we need to color.
		move_to_column(last_column);
		color_for_area(start, point);
	}
}

// Fill paragraph between a and b.  Make the first line up to
// first_width long.  Make remaining lines up to rest_width.
region_fill_first_line(int a, int b, int first_width, int rest_width)
{
	int start, atend, len;

	if (a > b)			/* switch so a < b */
		start = a, a = b, b = start;
	save_spot point = a;
	if (character(b - 1) == '\n')	// Don't merge into line following this region.
		b--;
	save_var narrow_start = a, narrow_end = size() - b;
	save_var abort_searching = 0;
	point = 0;		/* remove previous line breaks */
	while (re_search(1, "( [ \t\n]+)|([\t\n][ \t\n]*)")) {
		len = point - matchstart;
		atend = character(matchstart) == '\n';
		delete(matchstart, point);
		if (sentence_end_double_space && (len >= 2 || atend)
			  && parse_string(-1, "[.?!][])'\"]*", NULL))
			stuff("  ");
		else
			insert(' ');
	}
	point = 0;
	widened_move_to_column(first_width);
	on_exit build_first = 1;		/* redisplay hint */
	for (;;) {		/* put in new line breaks */
		if (point >= size() - narrow_end)
			break;
		while (curchar() == ' ')
			point++;
		while (re_search(-1, " +"))
			if (!mode_restrict_break || mode_restrict_break(point))
				break;
		do {
			if (!re_search(1, " +"))
				return;
		} while (mode_restrict_break && !mode_restrict_break(matchstart));
		delete(matchstart, point);
		insert('\n');
		narrow_start = point;	/* never look before here */
		widened_move_to_column(rest_width);
	}
}

int html_restrict_break(int pos)
{
	save_var matchstart, matchend, point;
	if (get_character_syntax_color(pos) == color_class html_value) {
		int start;
		get_character_color(pos, &start);
		return !strchr("\"'", character(start));
	}
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_fill_combine
												: html_auto_fill_combine;
	if (var && parse_string(-1, "%<[ \t]*(" HTML_NAME ")[ \t]*")
		  && find_group(1, 0) - find_group(1, 1) <= var)
		return 0;
	return 1;
}

// Return a file pattern matching places where a dir file might be.
char *info_get_std_dirs()
{
	if (is_mswindows)
		return "{/cygwin*,}/usr/*/info/dir";
	else
		return "/{usr,opt}/*/{share/,}info/dir";
}

read_info_dir(int transl, char *path)
{
	int i, cnt = 0;
	char *s, file[FNAMELEN];

	buf_zap(bufnum);
	s = lookpath("info/dir", 0);
	if (s)
		bprintf("%s\n", s);
	for (i = 0; s = look_on_path("dir", 0, path, i); i++)
		bprintf("%s\n", s);		// Use the provided path variable.
	path = getenv("INFOPATH");
	if (path)	// And the environment setting.
		for (i = 0; s = look_on_path("dir", 0, path, i); i++)
			bprintf("%s\n", s);
	s = do_file_match(info_get_std_dirs(), STARTMATCH | EXACTONLY);
	if (s)		// And search in other places too.
		do {
			bprintf("%s\n", s);
		} while (s = do_file_match(info_get_std_dirs(), EXACTONLY));
	point = 0;
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);		// Discard duplicates.
	point = 0;		// We now have a buffer of dir file names.
	for (;;) {
		i = point;
		if (!grab_line(bufnum, file))
			break;
		delete(i, point);
		do_insert_file(file, transl);
		save_spot point = i;
		if (++cnt > 1 && re_search(1, "^%* Menu:.*\n"))
			delete(i, point);	// Discard the extra headings.
		restore_vars();
	}
	return 0;
}

info_to_epsilon_node(char *node, int record)
{
	char full[FNAMELEN], *s;

	s = lookpath("info/epsilon.inf", 0);
	if (!s)
		s = lookpath("info/epsilon.info", 0);
	if (s) {			/* Find it even if it's not installed properly. */
		sprintf(full, "(%s", s);
		*get_extension(full) = 0;
		strcat(full, ")");
	} else
		strcpy(full, "(epsilon)");
	strcat(full, node);
	return info_go_node(full, record);
}

when_loading()
{
	char buf[FNAMELEN];
	sprintf(buf, "%s:%s", "%x../info", info_path_unix);
	if (!strstr(info_path_unix, "%x../info"))
		strcpy(info_path_unix, buf);
}
