/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 b5 to 14 b6.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "dired.h"
#include "grep.h"
#include "html.h"
#include "hex.h"
#include "info.h"
#include "makemode.h"
#include "rmode.h"
#include "perl.h"
#include "proc.h"
#include "session.h"
#include "tags.h"

int find_template_end()
{
	int opos = point, level = 1, parens = 0, bad = 0;

	if (curchar() != '<')
		return 0;
	point++;
	for (;; point++) {
		if (parse_string(1, "[ \n\t,a-zA-Z0-9_:*=.]+|//.*"))
			point = matchend - 1;
		else
			switch (curchar()) {
				case '/':
					if (character(point + 1) != '*')
						bad = 1;
					point += 2;
					if (!search(1, "*/"))
						bad = 1;
					break;
				case '<':
					if (parens == 0)
						level++;
					break;
				case '>':
					if (--level <= 0) {
						point++;
						return 1;
					}
					break;
				case '(':
					parens++;
					break;
				case ')':
					if (--parens <= 0)
						bad = 1;
					break;
				default:
					bad = 1;
					break;
			}
		if (bad) {
			point = opos;
			return 0;
		}
	}
}

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN];
	int class_start = point, i;

	strcpy(last, "");	// See if "class foo" is before this.
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (c_ignore_declspec())
			continue;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, (c_extra_keywords & (CSHARP_KEYWORDS
				| JAVA_KEYWORDS | JS_KEYWORDS))
							  ? "[a-zA-Z0-9_.]+" : "[a-zA-Z0-9_:]+", this))
				return 0;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")
				  || !strcmp(this, "enum") || !strcmp(this, "__interface")
				  || !strcmp(this, "interface")) {
				if (class) {
					if (*class && ((display_c_options & 1)
								   || strcmp(type, "namespace"))) {
						strcat(class, c_class_separator());
						class += strlen(class);
					}
					strcpy(class, last);	// Is place where we started within func?
				}
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

int find_c_func_info(char *type, char *class, char *func, int stop_on_key)
{
	char funcname[FNAMELEN], classname[FNAMELEN], pat[FNAMELEN], pat2[FNAMELEN];
	int start, funcstart, j, got_class = 0;

	start_of_function = -1;
	*classname = *funcname = 0;
	if (func)
		*func = 0;
	if (class)
		*class = 0;
	if (type)
		*type = 0;
	save_var point, matchstart, matchend, narrow_start = 0, narrow_end = 0;
	save_var tag_declarations = 0;	// We use tagging sometimes.
	if (c_extra_keywords & OBJC_KEYWORDS)
		got_class |= find_objc_class(type, class);
	to_begin_line();
	start = point;
	if (c_quick_scan) {
		point--;		// Begin searching at the last } char on the left margin...
		while (search(-1, "\n}")) {
			point++;		// ...that's not in a comment.
			if (!in_c_comment(c_look_back))
				break;
			point--;
		}
	} else			// Required if inner braces might be at the left margin.
		point = 0;

	strcpy(pat, "{|/<*|/>|(~?[a-zA-Z0-9_]+|" CPP_OPSPEC ")[ \t\n]*%(");
	if (c_extra_keywords & OBJC_KEYWORDS)
		strcat(pat, "|\n[-+]");
	while (re_search(1, pat)) {
		point = matchstart;
		if (stop_on_key && (char_avail() || is_key_repeating()))
			return -1;		// Give up search now?
		if (parse_string(-1, "^[ \t]*#[ \t]*(define|if|pragma).*(\\\n.*)*")) {
			to_end_line();	// Ignore everything in a macro definition.
			continue;
		}
		switch (curchar()) {
			case '{':		// Could be start of class def.
				j = point;
				if (c_inside_block(point, start)) {
					got_class |= find_c_class_info(start, type, class, func);
					point = j + 1;
					if (got_class)
						continue;
				}
				point = j;
				c_skip_level('{', '}');
				break;
			case '/':
				if (character(point + 1) == '*') {		// Skip over comment.
					point += 2;
					search(1, "*/");
				} else
					to_end_line();	/* c++ comment */
				break;
			case '\n':		// Objective-C method
				point++;
				if (give_begin_line() > start)
					return got_class;		// Gone too far, so give up.
				if (find_objc_method(func) == 1)
					return CF_INFO_FUNC | got_class;
				break;
			default:		// Found an identifier; is it a function def?
				if (give_begin_line() > start)
						return got_class;		// Gone too far, so give up.
				funcstart = point;
				if (!parse_string(1, "~?[a-zA-Z0-9_]+|" CPP_OPSPEC, funcname))
					return got_class;
				sprintf(pat2, "([a-zA-Z0-9_]+)[ \t\n]+(%s)",
						c_prototype_macro_names);
				if (parse_string(-1, pat2))
					grab(find_group(1, 0), find_group(1, 1), funcname);
				if (parse_string(1, "[ \t\n]*%("))
					narrow_end = size() - (point + c_biggest_declarator);
				if (tag_c_func(0) == 2 && !find_c_func_oneliner(start)) {
					point = funcstart;	// Checks out as a function def.
					start_of_function = give_begin_line();
					*classname = 0;
					if ((c_extra_keywords & JS_KEYWORDS) && !strcmp(funcname, "function"))
						find_js_func_info(funcname);
					if (parse_string(-1, (c_extra_keywords & (CSHARP_KEYWORDS | JS_KEYWORDS | JAVA_KEYWORDS))
									 ? "[ \t\n]*%.[ \t\n]*"
									 : "[ \t\n]*::[ \t\n]*")) {
						point = matchend;
						if (!parse_string(-1, "[a-zA-Z0-9_]+", classname) && parse_string(-1, "%>")) {
							int template_end = point;
							if (find_template_start() && parse_string(-1, "[a-zA-Z0-9_]+", classname))
								grab(matchstart, template_end, classname + strlen(classname));
						}
					}
					if (func)
						strcpy(func, funcname);
					if (class && *classname)
						strcpy(class, classname);
					return CF_INFO_FUNC | (*classname ? CF_INFO_CLASS : 0)
							| got_class;
				}
				*funcname = 0;
				narrow_end = 0;
				break;
		}
	}
	return got_class;
}

char *c_class_separator()
{
	return (c_extra_keywords & (CSHARP_KEYWORDS | JS_KEYWORDS | JAVA_KEYWORDS)) ? "." : "::";
}

c_statement_start()
{
	save_var point;
	if ((c_extra_keywords & CPP_KEYWORDS) && parse_string(1, "<hspace>*%[%["))
		search(1, "]]");
	return parse_string(1, "[{} \t\f]*(if|else|while|do|for|switch|try|"
						"catch|finally)[^a-zA-Z0-9]");
}

is_contin_statement(p, fd, orig) // would line after p be a continuation line?
{			// assumes p is just after last interesting char
	// if fd, pretend func param decls are a "continuation" of func decl
	// orig is later position in buffer, for unclosed "for" scan
	switch (character(p - 1)) {
		case ';':		// check for multi-line "for" statemnt
			if (unmatched_rparen(p, orig))
				return 1;
		case '{':
		case '}':
			return 0;
		case ')':
			save_var point = p;
			save_var narrow_start = p - 500;
			if (!c_rev_skip_level('(', ')'))
				return 1;
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && parse_string(-1, "@(synchronized|catch)"))
				return 0;
			if (sublanguage_code == SUBLANG_PHP)
				return !parse_string(-1, "[^a-z0-9_](if|while|for"
									 "|switch|foreach)[ \t\f]*")
						&& (fd || !Topindent || get_indentation(point));
			if (after_c_macro_line())
				return 0;
			if (c_extra_keywords & CPP_KEYWORDS)
				point -= parse_string(-1, "constexpr<wspace>*");
			return !parse_string(-1, "[^a-z0-9_](if|while|for"
								 "|switch)[ \t\f]*")
					&& (fd || !Topindent || get_indentation(point))
					|| in_perl_buffer;
		case 'e':
			return !is_else(p - 4);
		case 'o':
			save_var point = p;
			return !parse_string(-1, "[^a-z0-9_]do");
		case 'y':
			if ((c_extra_keywords & OBJC_KEYWORDS)
				  && matches_at(p, -1, "@(try|finally)"))
				return 0;
			return 1;
		case ':':
			if (is_php_alt_open(p))
				return 0;
			save_var point = p;
			to_indentation();
			return !is_label(point);
		default:
			return 1;
	}
}

int do_color_c_range(int from, int to)
{
	int t = -1, talk, s, talk_at = 0;
	char pat[200];
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
			   "|-?%.?[0-9]([A-Za-z0-9._']|[Ee]-)*");
	if (c_extra_keywords & JS_KEYWORDS)
		strcat(pat, "|`");	// JavaScript: `${template} ${literals}` syntax
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '/':			// found // one-line comment
				while (nl_forward() && character(point - 2) == '\\')
					;		// Continued // comment.
				set_character_color(t, point, color_class c_comment);
				break;
			case '*':			// found /* starting comment
				search(1, "*/");
				set_character_color(t, point, color_class c_comment);
				if (get_character_color(point, (int *) 0, &s)
					  == color_class c_comment && s > to) {
					c_init_color(point, s);
					to = s;	// Comment extended further before; rescan it.
				}
				break;
			case '#':			// found preproc line
				if (matches_at(t, 1, c_color_preproc_if_pattern)) {
					if (locate_by_ifdef(1))
						nl_forward();
					else
						point = size();
					set_character_color(t, point, color_class c_comment);
				} else {
					c_preproc_color();	// move past correct part
					set_character_color(t, point, color_class c_preprocessor);
				}
				break;
			case '"':		// found a string literal
				point = t;
				if ((c_extra_keywords & OBJC_KEYWORDS) && character(t - 1) == '@')
					t--;
				if ((c_extra_keywords & CSHARP_KEYWORDS)
					  && character(t - 1) == '@')
					t--, re_search(1, "\"([^\"]|\"\")*\"![^\"]");
				else
					re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class c_string);
				if (get_character_color(point, (int *) 0, &s) == 
					  color_class c_string && s > to)  // fix up after
					c_init_color(point, to = s); // quoted "'s
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class c_charconst);
				break;
			case '`':		// found JavaScript template literal
				color_js_template_literal(t);
				break;
			default:		// found identifier, kywd, or number
				if ((c_extra_keywords & (OBJC_KEYWORDS | JAVA_KEYWORDS))
					&& character(t - 1) == '@')
						t--;
				set_character_color(t, point, c_keyword_color(t));
				if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
					  && (point - t == 4)			// uuid's in IDL files.
					  && parse_string(-1, "uuid")
					  && parse_string(1, "%([0-9A-Fa-f]"))
					maybe_color_uuid();
				break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

replace_file_with_output(char *file, char *cmdpat)
{
	char cmdline[FNAMELEN];
	int outbuf = tmp_buf();
	on_exit buf_delete(outbuf);
	save_var bufnum = outbuf;
	sprintf(cmdline, cmdpat, file);
	if (pipe_text(bufname, bufname, cmdline, NULL,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0)
	   && lines_between(0, size(), 0) == 1 && character(size() - 1) == '\n')
		grab(0, size() - 1, file);
}

copy_matches(int tobuf, int flags, char *pat, char *rel, char *extra)
{
	int sbuf = bufnum, cnt = 0, i, more, srcherr, lines;
	int before, file, do_color;

	do_color = want_code_coloring && recolor_range;
	save_var point = 0;
	while ((srcherr = do_color_searching(flags, pat)) == 1) {
		i = give_position_at_column(matchstart, 0);
		more = nl_forward();
		lines = lines_between(i, point, 0);
		before = get_buf_point(tobuf);
		if (lines > 1)
			buf_printf(tobuf, "%s%s [%d]: ", extra, rel, lines);
		else
			buf_printf(tobuf, "%s%s: ", extra, rel);
		file = get_buf_point(tobuf);
		if (do_color) {			// Color in place, then copy.
			if (do_color == 1 && idle_coloring_size > 1000
				  && (recolor_from_here != recolor_by_lines)) {
				recolor_buffer_range(0, size());
				do_color = 2;		// Color all at once if coloring pieces is
			}						// slow.
			if (do_color == 1)
				recolor_buffer_range(i, point - more);
		}
		buf_xfer_colors(tobuf, i, point - more);
		bufnum = tobuf;
		set_character_color(before, file, color_class grep_filename);
		bprintf("\n");
		set_character_color(point - 1, point, color_class grep_filename);
		point = file;
		// Set a tagged region to mark any lines after the first for a
		// given match, to handle multi-line matches.
		set_tagged_region("grep-continuation", before, size(), -1);	// Clear markings.
		if (size() - 1 > give_end_line() + 1)
			set_tagged_region("grep-continuation", give_end_line() + 1, size() - 1, 1);
		color_internal_matches(flags, pat);	// Now color matches within line.
		point = size();
		bufnum = sbuf;
		cnt++;
		if (!more)
			return cnt;
	}
	return srcherr ? -srcherr : cnt;
}

grep_move_by_file(int dir)
{
	char start[FNAMELEN], thisfile[FNAMELEN];

	if (dir < 0)
		nl_reverse();
	grep_move_by_matches(0);
	parse_string(1, "<Min>.*!: ", start);
	int atbreak = parse_string(1, "\n|Found [0-9]+");
	while (grep_move_by_matches(dir)) {
		if (atbreak ? !parse_string(1, "\n|Found [0-9]+")
					: (!parse_string(1, "<Min>.*!: ", thisfile)
					   || strcmp(start, thisfile))) {
			grep_move_by_matches(dir < 0 ? 1 : 0);
			return;
		}
	}
}

do_grep_examine(char *cmd)
{
	int start, end, colon, old, instance, i, toolong = 0, extra;
	int lines = 1, precolon;
	char file[FNAMELEN], string[FNAMELEN], line[FNAMELEN], curdir[FNAMELEN], *p;
	char prefix[FNAMELEN];

	save_var synchronous_file_io = 1;
	_last_was_grep = 1;
	grep_move_by_matches(0);
	save_var case_fold = 0;		/* turn off for grep buffer */
	if (!search(1, ": "))
		error("Can't read file name");
	colon = point;
	precolon = colon - 2;
	if (parse_string(-1, " %[[0-9]+%]: ")) {
		precolon = matchend;
		grab(precolon + 2, point, line);
		lines = numtoi(line);
	}
	to_begin_line();
	extra = parse_string(1, "%[[0-9][0-9]+-[^]]*[0-9]%] ");
	start = point;
	for (i = lines; i > 1; i--)
		nl_forward();
	to_end_line();
	end = point;
	point = start;
	grab(start, precolon, prefix);
	grab(start + extra, precolon, file);
	if (start + FNAMELEN - 3 < end)
		end = toolong = start + FNAMELEN - 3;
	grab(colon, end, string);
	line[0] = '\n';		// Put \n at start and end of line.
	grab(start, end + 1, line + 1);
	instance = count_instances(line) + 1;
	grep_get_curdir(curdir);
	if (cmd && *cmd)
		try_calling(cmd);
	if (strstr(file, MANPREFIX) == file) {
		grep_examine_man_page(file + sizeof(MANPREFIX) / sizeof(char) - 1);
		return;
	} else if (strstr(file, BUFPREFIX) != file) {
		absolute(file, curdir);
		if (ask_find_it(file))
			return;
	} else if (exist(p = file + sizeof(BUFPREFIX) / sizeof(char) - 1))
		to_buffer(p);
	else
		error("Can't find buffer %s.", p);
	if (strlen(line) != end - start + 2)
		error("Can't find matching lines with null characters.");
	old = point;
	save_var case_fold = 0;		/* turn off for user buffer */
	point = 0;
	for (i = 0; i < instance; i++)
		if (toolong ? !search(1, string) : !line_search(1, string)) {
			point = old;
			if (i == 0 || instance == 1)
				error("Can't find \"%s\"", string);
			else
				error("Can't find this copy of repeated line");
		} else {
			old = point;
			if (i + 1 < instance)
				nl_forward();
		}
	to_begin_line();
	if (toolong)
		say("Line too long; this may not be the correct match.");
}

// Handle continuation lines (where a single match in the grep buffer
// spans multiple lines). If dir > 0 go to start of next match line
// group, if dir < 0 to previous match line group, if dir == 0 to start
// of current match line group. Return 1 if found.
int grep_move_by_matches(int dir)
{
	int start, end;
	if (!dir) {
		to_begin_line();
		if (get_tagged_region("grep-continuation", point, &start, &end) == 1) {
			point = start - 1;
			to_begin_line();
		}
	} else if (dir > 0) {
		if (!nl_forward())
			return 0;
		if (get_tagged_region("grep-continuation", point, &start, &end) == 1)
			point = end + 1;
		if (point >= size() - narrow_end)
			return 0;
		grep_move_by_matches(0);
	} else {
		grep_move_by_matches(0);
		if (!nl_reverse())
			return 0;
		if (get_tagged_region("grep-continuation", point - 1, &start, &end) == 1)
			point = start - 1;
		to_begin_line();
	}
	return 1;
}

by_matches(val)
{
	int owin, win, pt, dir = RE_FORWARD, buf = name_to_bufnum("grep");

	tiled_only();
	_last_was_grep = 1;
	if (!buf)
		error("No grep buffer.");
	win = buf_in_window(buf);
	if (win >= 0) {
		owin = window_handle;
		window_handle = win;
	} else
		bufnum = buf;
	grep_move_by_matches(0);
	if (val < 0)
		val = -val, dir = RE_REVERSE;
	else
		to_end_line();
	while (val-- > 0)
		do {
			if (!grep_move_by_matches(dir == RE_REVERSE ? -1 : 1))
				error("No more matches.");
		} while (!parse_string(1, "^(<(>buffer<)> )?[^\n\t]+: "));
	if (win >= 0) {
		pt = point;
		window_handle = owin;
		bufnum = buf;
		point = pt;
	}
	has_arg = iter = 0;
	grep_examine();
}

hex_save_key()
{
	char orig[FNAMELEN];
	int hexbuf = bufnum, pos, answer = 2;
	struct file_info saved = file_info;

	if (!is_hex_buffer(bufnum))
		error("Not a hex buffer");
	if (hex_maybe_overtype())
		return;
	pos = hex_get_offset();
	strcpy(orig, ORIG_BUF(bufname));
	if (modified)
		answer = hex_ask_apply();
	else
		say("Leaving hex mode.");
	if (answer == 3 || user_abort)
		return;
	if (answer == 1)
		hex_apply_all(orig);
	check_abort();
	to_buffer(orig);
	file_info = saved;		// Restore old file date checking info.
	buf_delete(hexbuf);
	point = pos;
	buffer_on_modify = 0;
	check_dates(0);
}

void hex_old_on_modify();
void hex_on_modify();
int color_hex_range(int from, int to);
char *hex_prefix();

command hex_mode()
{
	char bname[FNAMELEN];
	int pos = point, trans = GET_LINE_TRANSLATE(translation_type);
	struct file_info saved = file_info;

	if (is_hex_buffer(bufnum)) {	// "Turn off mode" -- save back to real buf.
		hex_save_key();
		return;
	}
	if (!size())
		error("Buffer is empty.");
	sprintf(bname, "%s%s", HEX_BUFNAME_PREFIX, bufname);
	if (exist(bname)) {
		say("This buffer is in hex mode.  Press %s? for help or %sq to quit.",
			  hex_prefix(), hex_prefix());
		to_buffer(bname);
		return;
	}
	if (trans == FILETYPE_MSDOS || trans == FILETYPE_MAC)
		hex_reload_file();
	buffer_to_hex(bname);
	if (user_abort) {
		delete_buffer(bname);
		error("Cancelled.");
	}
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_old_on_modify;	// special funcs can modify it.
	file_info.check_type = 0;	// Disable file date checking during hex mode.
	to_buffer(bname);
	file_info = saved;		// Remember old file date checking info.
	call_on_modify = 1;		// Make the buffer read-only so only our
	buffer_on_modify = hex_on_modify;	// special funcs can modify it.
	mode_keys = hex_tab;		/* Use these keys. */
	major_mode = hex_overtype_mode ? _hex_overtype_mode_name : _hex_mode_name;
	recolor_range = color_hex_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_delay = -1; // No idle coloring, it slows down apply-changes.
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("hex-mode-hook");
	drop_all_colored_regions();
	make_mode();
	say("Entering hex mode, press %s? for help or %sq to quit.",
		  hex_prefix(), hex_prefix());
	hex_goto_place(pos, hex_overtype_mode);
}

char *info_filename_end(char *s);

// Look in node for the first of several patterns that indicate s
// is being defined there.  Return 1 if we found one, else 0.
info_locate_match(char *s)
{
	char pat[FNAMELEN], pat2[FNAMELEN], *p;

	point = 0;
	sprintf(pat, "[a-z]+: %q!( |$)", s);
	if (re_search(1, pat))		// sometext: ourmatch
		return 1;
	point = 0;
	sprintf(pat, "`%q'", s);	// `ourmatch'
	if (re_search(1, pat))
		return 1;
	point = 0;
	sprintf(pat, "^%q$", s);	// ourmatch on its own line
	if (re_search(1, pat))
		return 1;
	point = 0;
	strcpy(pat, s);			// If we're searching for ourmatch (sometext),
	p = strstr(pat, " (");		// try looking for ourmatch.
	if (p && info_filename_end(p)) {
		*p = 0;
		sprintf(pat2, "`%q'", s);	// `ourmatch'
		if (re_search(1, pat2))
			return 1;
		point = 0;
		sprintf(pat2, "^%q$", pat);	// ourmatch on its own line
		if (re_search(1, pat2))
			return 1;
		point = 0;
		if (search(1, pat))
			return 1;
	}
	p = strstr(pat, "(");		// if it's a function, find first reference
	if (p && p > pat + 3) {
		*p = 0;
		point = 0;
		if (search(1, pat))
			return 1;
	}
	point = 0;
	if (search(1, s))		// Look for the plain text.
		return 1;
	point = 0;
	return 0;
}

color_makefile_range(from, to)
{			// last colored region may go past to
	int t = -1, s;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;		// Never stop coloring on a continued line.
	re_search(1, "[^\\]\n");
	to = point;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();
	for (; point < to; ) {
		t = point;
		if (!re_search(1, ":([ \t]|$)|[#=]|^[ \t]*[.!ie]"))
			break;
		set_character_color(t, matchstart, -1);
		switch (character(point - 1)) {
			case '#':		// Found a comment.
				if (parse_string(-1, "^\t.*"))
					break;	// Not in a command.
				set_character_color(point - 1, give_end_line() + 1,
									color_class makefile_comment);
				nl_forward();
				break;
			case 'i':
			case 'e':
				if (matches_at(point - 1, 1,
							   "^ *(ifn?(eq|def)|else|endif)")) {
					s = point - 1;
					while (nl_forward() && character(point - 2) == '\\')
						;			// include contin lines
					set_character_color(s, point, color_class makefile_preproc);
				}
				break;
			case '.':
				if (!parse_string(1, "(if|else|elsif|end)"))
					break;		// Fall through.
			case '!':		// A preprocessor directive starting with !.
				s = point - 1;
				while (nl_forward() && character(point - 2) == '\\')
					;			// include contin lines
				set_character_color(s, point, color_class makefile_preproc);
				break;
			case '=':		// A macro definition.
				if (parse_string(-1, "^[^\t\n ]([^\n=:]|\\\n)*="))
					set_character_color(matchstart - 1, matchend,
										color_class makefile_macro);
				break;
			default:		// Perhaps a target specification?
// A target specification ends with : followed by eol or whitespace,
// and includes previous lines that end with a \.
				if (parse_string(-1, "^[^\t\n ](.|\\\n)*(:)[ \t]*")) {
					s = matchend;	// Don't color : or :: after it.
					set_character_color(s, point - parse_string(-1,
						"[ :]*"), color_class makefile_target);
				}
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

#define XWIN_XFLD_COUNT	13	/* # of fields in an X11 font logical descriptor */
int xwin_font_name_when_idle(int cnt);
int xwin_font_size_when_idle(int cnt);
int xwin_pixel_size_when_idle(int cnt);

// Make a list of fonts by running xlsfonts, and have the user narrow
// down the choices until just one is left. Return 0 if we can't get the
// list.
int set_font_for_unix()
{
	char fontname[1000], orig[1000], fontpat[1000], fontfamily[1000];
	char fields[XWIN_XFLD_COUNT][1000], cust_point_sz[1000];
	jmp_buf *old_level = top_level, this_level;
	int i, mbuf = zap(_MATCH_BUF);
	struct screen_info retain_windows;

	retain_windows.wins = 0;
	save_screen(&retain_windows);
	strcpy(orig, font_fixed);
	*cust_point_sz = 0;
	for (i = 0; i < XWIN_XFLD_COUNT; i++)
		strcpy(fields[i], "*");
	save_var show_status = 0;
	note("Listing fonts...");
	int font_list = zap("-font-list");

	// First, extract the foundry+family of cur font.
	buf_printf(font_list, "%s\n", font_fixed);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	grab_full_line(mbuf, fontfamily);

	buf_zap(font_list);
	pipe_text(NULL, "-font-list", "if [ -x /opt/x11/bin/xlsfonts ] ; "
			  "then /opt/x11/bin/xlsfonts ; else xlsfonts ; fi", NULL,
			  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
	save_var bufnum = font_list;	// Got a list of all available fonts.
	if (lines_between(0, size(), 0) < 5) {
		say("Couldn't run xlsfonts to list the fonts.");
		return 0;
	}
	top_level = &this_level;
	if (setjmp(top_level)) {	// On abort, come here.
		font_fixed = orig;
		restore_screen(&retain_windows);	// Restore windows, since size changes
		free(retain_windows.wins);			// might have deleted some.
		top_level = old_level;
		quick_abort();
	}
	xwin_set_font_first_prune();	// Drop fonts we know we can't use.
	note("");

	// Extract the foundry+family names from font_list.
	xwin_make_font_name_list(_MATCH_BUF, font_list, 0, NULL);
	xwin_prune_to_usable_fonts(_MATCH_BUF);
	xwin_prompt_for_field(fields[0], "Select a font:", " Select a font ",
						 fontfamily, xwin_font_name_when_idle);

	// Now make a list of allowed point sizes for the selected font
	// family. Zero means any size is OK.
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 7, fontpat);
	xwin_font_size_prettify(_MATCH_BUF);

	i = xwin_prompt_for_field(fields[6], "Select a point size:",
							 " Point Size ", "12",
							 xwin_font_size_when_idle);

	xwin_font_size_unprettify(fields[6]);
	// If point size is 0 (custom), prompt for it and remember it.
	if (!strcmp(fields[6], "0")) {
		if (i)
			get_string(cust_point_sz, "Point size, such as 12 or 14: ");
		else
			get_strdef(cust_point_sz, "Font lacks recommended point size. "
					   "Please pick one", "14");
		xwin_font_size_fix_user(cust_point_sz);
	}

	// Ask the user to choose among other options the font family might
	// include.
	xwin_query_one_field(11, fields, "Select a char set (iso8859-1 "
						"recommended, or iso10646-1 for Unicode):",
						" Character Set ", "iso8859-1");
	xwin_query_one_field(1, fields, "Make all characters bold?",
						" Force Bold? ", "medium");
	xwin_query_one_field(2, fields, "Make all characters italic (o) not regular (r)?",
						" Force Italic? ", "r");

	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, 6, fontpat);
	xwin_prompt_for_field(fields[5], "Select a pixel size:", " Pixel size ", "",
						  xwin_pixel_size_when_idle);

// any other fields?

	save_var resize_menu_list = 1;
	xwin_font_array_to_pattern(fontpat, fields);
	xwin_make_font_name_list(_MATCH_BUF, font_list, -1, fontpat);

	if (*cust_point_sz)
		xwin_add_point_size(_MATCH_BUF, cust_point_sz);
	xwin_prompt_for_field(fontname, "These fonts match. Please pick one:",
						 " Font names ", "", xwin_font_name_when_idle);

	note("");
	aside("Switch to %s if %d", fontname, !user_abort);
	font_fixed = user_abort ? orig : fontname;
	if (!user_abort && strcmp(font_fixed, fontname))
		say("Sorry, can't use that font.");

	restore_screen(&retain_windows);
	free(retain_windows.wins);
	top_level = old_level;
	return 1;
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;
	TIMER talk_now;

	save_var point, matchstart, matchend, case_fold = 0;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// But wait a bit before chatter.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
											   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				if (in_shell_buffer && !parse_string(-1, "([ \n\t|&;()<>]|^)#"))
					break;
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && strchr("\"'", curchar())) {
					set_character_color(t, point, color_class perl_string);
					break;			// Translated string.
				}
				if (in_shell_buffer) {
					if (curchar() == '(') {
						shell_parse_substitution(point - 1);
						break;
					}
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									 "|[^ \t\n`(]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+((::|-%>)<word>+)*"
									 "|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "[#$@]*(((<word>+('|::))*|'|::)<word>+|[-+])")) {
					set_character_color(t, point = matchend,
										color_class perl_variable);
				} else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer) {
					if (!(matches_at_length(t, -1, "\\+") & 1)) {
						search(1, "'");	// No \-quoting in '' for shell mode.
						set_character_color(t, point, color_class perl_string);
					}
				} else if (!strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && (matches_at_length(t, -1, "\\+") & 1))
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				if (curchar() == '/') {  // Found defined-or operator //
					point++;
					break;
				}
				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if (perl_after_expression(t, 0) || character(t - 1) == '<')
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_><word>*)"
									 : "~?(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "~?[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "~?[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "~?[ \t]*\\(<alpha|_><word>*)"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "~?[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
					break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

tag_mode_r()
{
	char func[TAGLEN];
	int start, color;

	r_mode();
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	while (re_search(1, "([a-zA-Z._]<word>*)[ \t]*%<-[ \t]*(function)%(")) {
		color = get_character_color(matchstart);
		if (color == color_class rmode_comment
			  || color == color_class rmode_string)
			continue;
		grab(start = find_group(1, 1), find_group(1, 0), func);
		add_tag(func, start);
	}
}

remember_this_buffer()	/* should we record this buffer in session file? */
{
	char *fname = filename;
	int limit;

	if (*bufname == '-')
		return 0;
	if (is_process_buffer(bufnum) || (buffer_not_saveable & (NOSAVE_ALL | NOSAVE_SESSION))
		  || _read_aborted || (force_save_as & FSA_NEWFILE))
		return 0;
	if (file_info.year && file_info.check_type != CHECK_FILE)
		return 0;
	if (is_dired_buf()) {
		fname = get_dired_buffer_pattern();
		if (!(session_restore_directory_buffers
			  & (is_remote_file(fname) ? 2 : strstr(fname, "**") ? 4 : 1)))
			return 0;
	} else {
		if (!*filename)
			return 0;
		limit = is_remote_file(filename) ? session_restore_biggest_remote_file
										 : session_restore_biggest_file;
		if (limit >= 0 && size() > limit)
			return 0;		// too big to restore
		if (!size() && check_file(filename) != CHECK_FILE)
			return 0;
	}
	if (session_restore_exclude_file_types) {
		int flags = filename_rules(fname);
		if ((flags & FSYS_NETWORK) && (session_restore_exclude_file_types & 1))
			return 0;
		if ((flags & FSYS_REMOVABLE) && (session_restore_exclude_file_types & 2))
			return 0;
		if ((flags & FSYS_CDROM) && (session_restore_exclude_file_types & 4))
			return 0;
	}
	return 1;
}

save_screen(p)		/* save current window info in p */
struct screen_info *p;	/* on first call with p, p->wins must be 0 */
{
	int i, oldroom = 0, nwins = number_of_windows() - number_of_popups();

	p->lines = screen_lines, p->cols = screen_cols;
	p->cur_window = window_number;
	p->echo_line = echo_line;
	p->avoid_top = avoid_top_lines;
	p->avoid_bottom = avoid_bottom_lines;
	p->permanent_menu = permanent_menu;
	if (p->wins)
		oldroom = ptrlen(p->wins, 1) / sizeof(struct window_info);
	if (oldroom < nwins) {
		if (p->wins)
			free(p->wins);
		p->wins = (struct window_info *) malloc(nwins
												* sizeof(struct window_info));
	}
	for (i = 0; i < nwins; i++)
		get_window_info(i, p->wins + i);
	p->num = i;
}

restore_screen(p)	/* restore windows from data in p */
struct screen_info *p;
{
	int i, origbuf, origpt;

	bufnum = origbuf = window_bufnum;
	origpt = point;
//	if (screen_lines != p->lines || screen_cols != p->cols)
//		error("Screen is in wrong mode for restore.");
	for (i = 0; i < p->num; i++)
		window_create(!i, p->wins + i);
	echo_line = p->echo_line;
	select_low_window(p->cur_window, p->avoid_top, p->avoid_bottom,
					  p->lines, p->cols);
	restore_menu(p->permanent_menu);
	if (bufnum == origbuf)
		point = origpt;
}

color_scheme "standard-gui" {
	color_class c_keyword dark_brown;		// new default
};

when_loading()
{
	strcpy(push_cmd_unix_macos_interactive, "(xterm || /opt/X11/bin/xterm) &");
}
