/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 B8 to 14 B9.

#include "eel.h"
#include "c.h"
#include "chelp.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "hex.h"
#include "info.h"
#include "makemode.h"
#include "misclang.h"
#include "rmode.h"
#include "perl.h"
#include "proc.h"
#include "session.h"
#include "tags.h"
#include "tex.h"

suffix_json()
{
	javascript_mode();
}

suffix_tsx()
{
	suffix_ts();
}

after_many_comments()
{
	char pat[FNAMELEN];
	int cnt = 0;

	if (comment_repeat_indentation_lines < 0)
		return 0;
	save_var point;
	sprintf(pat, "[ \t]*%s", get_comment_start());
	while (nl_reverse()) {
		to_begin_line();
		if (!parse_string(1, pat))
			break;
		if (++cnt >= comment_repeat_indentation_lines)
			return 1;
	}
	return 0;
}

// Return nonzero if the ' character at pos should be treated as part
// of a C++ numeric constant like 123'456'789.
int have_cpp_quote_in_number(int pos)
{
	save_var matchstart;
	return (c_extra_keywords & CPP_KEYWORDS)
			&& matches_at(pos - 1, 1, "[0-9a-f]'[0-9a-f]");
}

// Mark comments and strings in region from...to in the current C mode
// buffer.  Return 0 if last comment or string we found never ended.
// The opening delimiter character(s) of the comment or string isn't included.
int c_lex_region(int from, int to)
{
	// If from is the in middle of an already-lexed region, go to its start.
	get_tagged_region("c-lex", from, &from);
	save_var point = from, narrow_start = 0, narrow_end = 0;
	int start, c, val, init = from, p, found_end = 1;
	while (re_search(1, "<squote|dquote>|//|/<*>")) {
		start = matchstart;
		c = character(start);
		p = point;	// Mark that we're not in comment/string before the march.
		set_tagged_region("c-lex", init, p, 0);
		if (c == '/' && character(start + 1) == '/') {
			val = IN_NEW_COMMENT;	// Handle // comment (c_skip_item() won't).
			found_end = nl_forward();
		} else if (c == '\'' && have_cpp_quote_in_number(point - 1)) {
			continue;		// A number like 123'456'789.
		} else {
			point = start;
			found_end = c_skip_item(c, 0);
			switch (c) {
				case '/':	val = IN_OLD_COMMENT; break;
				case '"':	val = IN_STR_CONST; break;
				case '\'':	val = IN_CHAR_CONST; break;
				default:	say("Internal error in c_lex_region in %s at %d (%d)",
								  bufname, point, c);
							val = 0;
							break;
			}
		}
		set_tagged_region("c-lex", p, point, val);
		if (point > to)
			return found_end;
		init = point;
	}
	set_tagged_region("c-lex", init, point, 0);	// Mark final region.
	return found_end;
}

// Force this line to match the last line?
int c_next_same_indent(int orig, int pstart)
{
	if ((c_extra_keywords & (JAVA_KEYWORDS)) && matches_at(pstart, 1, "@[A-Z0-9]"))
		return 1;	// Java annotation on prev line of this definition.
	if ((c_extra_keywords & (JS_KEYWORDS)) && matches_at(orig, 1, "%.[A-Z0-9]"))
	   return 1;	// Javascript chaining.
	return 0;
}

/* Indent continuation line. */
indent_c_contin(ind, orig, lastl, in_initializer)
{
	int c, plevel = 0, qlevel = 0, findc, alevel = 0, slevel = 0;
	int objc_sel = 0, objc_indent = -1;

 	if (c_align_inherit >= 0 && matches_at(orig, -1, ":\n[ \t]*")
 		  && !c_matches_in(give_begin_line(), lastl, "%?"))
 		return get_indentation(point) + c_align_inherit;
	if (!c_align_contin_lines)	// never line up
		if (!in_initializer)
			return c_default_indent(ind, orig, 1);
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	if ((c_extra_keywords & OBJC_KEYWORDS) && c_align_selectors
		  && matches_at(give_begin_line(), 1, "^[ \t]*(<word>+):"))
		objc_sel = find_group(1, 0) - find_group(1, 1);
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[][()?:\"'\n<>]|[=!<>]?=")) {
		c = curchar();
		if (in_perl_buffer && (in_perl_comment(point) || strchr("\"'\n", c)))
			continue;	// Skip over comments differently.
		switch (c) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0 || objc_indent >= 0)
					break;
				if (c_align_open_paren && findc != ')') {
					if (c_align_open_paren == 1)
						return ind + get_c_indent();
					else if (strchr(",(", character(lastl - 1)))
						return get_indentation(point) + get_c_indent();
				}
indhere:		if (objc_indent >= 0)
					break;
				if (findc != ')')
					point++;
				else if (Closeback & 2)
					return get_indentation(point);
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind, orig);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				if (objc_sel && !plevel && !slevel
					  && parse_string(-1, "(^|<word>|[]])[ \t]*<word>+")
					  && current_column() <= c_align_selectors)
					objc_indent = current_column() - objc_sel;
				else
					qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
				    || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				if (objc_indent >= 0)
					return objc_indent;
				cind_try_space(); // at ends of lines
				break;
			case '\'':
				if (have_cpp_quote_in_number(point))
					continue;		// A number like 123'456'789.
			case '"':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
			case '[':
				if (slevel-- > 0 || objc_indent >= 0)
					break;
				if (findc != ']')
					point++;
				save_var c_align_extra_space = 0;
				return limit_c_contin_indent(ind, orig);
			case ']':
				slevel++;
				break;
			case '<':
				if (character(point + 1) == '=')
					break;
				if (alevel-- > 0 || !(c_extra_keywords & (CPP_KEYWORDS
					| JAVA_KEYWORDS | CSHARP_KEYWORDS)) || objc_indent >= 0)
					break;
				// Someday, test if there's an identifier before the <
				// that's the name of a template, not just if it's a
				// template keyword.
				if (parse_string(-1, "</word>template[ \t\n]*")) {
					point++;
					point += parse_string(1, "[ \t]*!<word>");
					return limit_c_contin_indent(ind, orig);
				}
				break;
			case '>':
				if (character(point + 1) != '=')
					alevel++;
				break;
		}
	}
	if (objc_indent >= 0)
		return objc_indent;
	if (c_align_inherit >= 0 && c_matches_in(point, orig, ":[ \t]*($|//|/%*)")
		  && !c_matches_in(point, orig, "%?"))
		return get_indentation(point) + c_align_inherit;

	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		if (curchar() == '#')
			past_initial_preprocessor();
		return limit_c_contin_indent(ind, orig);
	}
	return c_default_indent(ind, orig, 2);
}

to_statement_start()
{
	int lastcont, count = 100;	/* give up after this many lines */
	int orig = point;

	to_begin_line();
	if (parse_string(1, "[ \t\f]*}")) {
		search(1, "}");
		c_rev_skip_level('{', '}');
	}
	if (!is_js_object_literal(point)) {
		do {
			to_begin_line();
			lastcont = point;
			if (parse_string(1, "[ \t\f]*},")	// initializer
				  || (c_extra_keywords & (JS_KEYWORDS) && parse_string(1, "<hspace>*}"))) {
				search(1, "}");
				c_rev_skip_level('{', '}');
			}
		} while (!cind_try_space() && is_contin_statement(point, 0, orig)
				 && count-- > 0);
		point = lastcont;
	}
	to_indentation();
	if (curchar() == '#')
		past_initial_preprocessor();
}

cind_close()	/* indent a line that starts with a } character */
{		/* find matching { and indent based on that one */
	int ind, solitary, chaining = 0;

	point++;		/* go past } char */
	if (curchar() == ')' && (c_extra_keywords & JS_KEYWORDS)) {
		point++;
		c_rev_skip_level('(', ')');	/* find matching open paren */
		if (parse_string(-1, "^<hspace>*%.<ident>*"))	// Chaining?
			chaining = 1;
	} else
		c_rev_skip_level('{', '}');	/* find matching open brace */
	solitary = parse_string(-1, "\n[ \t\f]*");	/* stuff before? */
	if (solitary || chaining)
		to_indentation();
	else
		to_statement_start();
	past_labels();
	ind = current_column();
	if (!Closeback && !solitary)
		ind += get_c_indent() + c_brace_offset;
	return ind;
}

int c_rev_skip_level(left_d, right_d)
char left_d, right_d;
{
	char pattern[40], ch;
	int level = 0;

	if (in_perl_buffer)
		return perl_skip_level(-1, left_d, right_d);
	sprintf(pattern, "<*>/|[%c%c'\"]|//.*\n", right_d, left_d);

bgn:
	do {
		if (!re_search(RE_REVERSE, pattern))
			return 0;
		switch (ch = curchar()) {
			case '/':
				if (at_c_comment(point + 2, 100) & IN_STRING)
					to_end_line();
				goto bgn;
			case '\'':
				if (have_cpp_quote_in_number(point))
					continue;		// A number like 123'456'789.
			case '"': case '*':
				point = matchstart;
				if (ch != '*' && (in_c_comment(100) & IN_COMMENT)) {
					  point--;
					  goto bgn;
				}
				c_skip_item(character(point - 1), RE_REVERSE);
				goto bgn;
			default:
				if (ch == right_d)
					level++;
				else if (ch == left_d)
					level--;
		}
	} while (level > 0);
	return 1;
}

int c_skip_level(left_d, right_d)
char left_d, right_d;
{
	char pattern[40], ch;
	int level = 0;

	if (in_perl_buffer)
		return perl_skip_level(1, left_d, right_d);
	sprintf(pattern, "/[/*]|[%c%c'\"]", right_d, left_d);

bgn:	do {
		if (!re_search(RE_FORWARD, pattern))
			return 0;
		switch (ch = character(matchstart)) {
			case '/':
				search(1, character(matchstart + 1) == '/'
				       ? "\n" : "*/");
				break;
			case '\'':
				if (have_cpp_quote_in_number(matchstart))
					continue;		// A number like 123'456'789.
			case '"':
				point = matchstart;
				c_skip_item(ch, RE_FORWARD);
				goto bgn;
			default:
				if (ch == right_d)
					level++;
				else if (ch == left_d)
					level--;
		}
	} while (level < 0);
	return 1;
}

/*  fix indentation if necessary when # is typed */
command c_hash_mark() on c_tab['#']
{
	normal_character();
	if (!in_perl_buffer && parse_string(-1, "^[ \t\f]*#"))
		fix_c_indentation();
}

early_restoring()
{
	char fname[FNAMELEN];
	int i;

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_WINDOWS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	char *s = getenv("PERLDB_OPTS"), *t = malloc((s ? strlen(s) : 0) + 30);
	sprintf(t, "PERLDB_OPTS=ReadLine=0 %s", s ? s : "");
	putenv(t);	// Make Perl debugger run in process buffer.
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	else
		get_standard_color_scheme();
	use_default = 1;		/* default is fundamental mode */
	mode_default_settings();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */
	do_set_show_graphic(get_show_graphic());

	force_save_as = FSA_NEWFILE;	// make startup buffer prompt
	strcpy(fname, "startup.txt");	// for filename
	absolute(fname);
	filename = fname;
	strcpy(_docname, "edoc");
	if (load_from_state)
		show_text(-2, 0, "Loaded from state file %s", state_file);
	else if (*state_file)
		show_text(-2, 0, "Loaded from state file %s (quickup)", state_file);
	else
		show_text(-2, 0, "Loaded from bytecode files");
	for (i = 0; original_argv(i); i++)
		show_text(-2, 0, "Argv[%d]=%s", i, original_argv(i));
	get_customization_directory(fname);
	show_text(-2, 0, "Customization directory is %s", fname);
	char *path = verenv("EPSPATH");
	show_text(-2, 0, "EPSPATH is %s", path ? path : "undefined");
}

int do_color_c_range(int from, int to)
{
	int t = -1, talk, s, talk_at = 0;
	char pat[200];
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
		       "|-?%.?[0-9]([A-Za-z0-9._']|[Ee]-)*");
	if (c_extra_keywords & JS_KEYWORDS)
		strcat(pat, "|`");	// JavaScript: `${template} ${literals}` syntax
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
		case '/':			// found // one-line comment
			while (nl_forward() && character(point - 2) == '\\')
				;		// Continued // comment.
			set_character_color(t, point, color_class c_comment);
			break;
		case '*':			// found /* starting comment
			search(1, "*/");
			set_character_color(t, point, color_class c_comment);
			if (get_character_color(point, (int *) 0, &s)
				  == color_class c_comment && s > to) {
				c_init_color(point, s);
				to = s;	// Comment extended further before; rescan it.
			}
			break;
		case '#':			// found preproc line
			if (matches_at(t, 1, c_color_preproc_if_pattern)) {
				if (locate_by_ifdef(1))
					nl_forward();
				else
					point = size();
				set_character_color(t, point, color_class c_comment);
			} else {
				c_preproc_color();	// move past correct part
				set_character_color(t, point, color_class c_preprocessor);
			}
			break;
		case '"':		// found a string literal
			point = t;
			if ((c_extra_keywords & OBJC_KEYWORDS) && character(t - 1) == '@')
				t--;
			if ((c_extra_keywords & CPP_KEYWORDS) && character(t - 1) == 'R'
				  && move_past_cpp_raw_literal(0, 0))
				t--;	// Color R too.
			else if ((c_extra_keywords & CSHARP_KEYWORDS)
				  && character(t - 1) == '@')
				t--, re_search(1, "\"([^\"]|\"\")*\"![^\"]");
			else
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
			set_character_color(t, point, color_class c_string);
			if (get_character_color(point, (int *) 0, &s) == 
			    color_class c_string && s > to)  // fix up after
				c_init_color(point, to = s); // quoted "'s
			break;
		case '\'':		// found a char const
			point = t;
			re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
			set_character_color(t, point, color_class c_charconst);
			break;
		case '`':		// found JavaScript template literal
			color_js_template_literal(t);
			break;
		default:		// found identifier, kywd, or number
			if ((c_extra_keywords & (OBJC_KEYWORDS | JAVA_KEYWORDS))
				  && character(t - 1) == '@')
				t--;
			set_character_color(t, point, c_keyword_color(t));
			if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
				&& (point - t == 4)			// uuid's in IDL files.
				&& parse_string(-1, "uuid")
				&& parse_string(1, "%([0-9A-Fa-f]"))
				maybe_color_uuid();
			break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

// Parse a C++ raw string literal like R"delim(text)delim". For forward
// (reverse is 0), point+offset is just before the opening ". Move to
// just after the final " and return 1. If it wasn't a valid raw string
// literal, just return 0 without moving. If unterminated, move to EOF
// and return 2.

// If reverse is nonzero, we try instead to parse it backwards, and
// return 0 if it wasn't a raw string literal after all. We expect
// point+offset to be just after the final ", and move to just before
// the starting R on success.
int move_past_cpp_raw_literal(int offset, int reverse)
{
	char delim[FNAMELEN], pat[FNAMELEN];

	save_var case_fold = 0;
	if (reverse) {
		int orig = point;
		if (!matches_at(point + offset, -1, "%)([^\n()\\]*)\""))
			return 0;
		grab(find_group(1, 1), find_group(1, 0), delim);
		sprintf(pat, "R\"%q%%(", delim);
		if (re_search(-1, pat))
			return 1;
		point = orig;
		return 0;
	}		
	if (character(point + offset - 1) != 'R'
		  || !matches_at(point + offset, 1, "\"([^\n()\\]*)%("))
		return 0;
	grab(find_group(1, 1), find_group(1, 0), delim);
	sprintf(pat, "%%)%q\"", delim);
	return re_search(1, pat) ? 1 : 2;
}

// If there's a directory name before an env variable substitution,
// remove the directory name.
delete_file_path_before_subst()
{
	char pat[FNAMELEN], var[FNAMELEN];
	if (opsys == OS_WINDOWS)
		sprintf(pat, "(((%s|[a-z]):)?"		// optional a: or ftp:
			"[\\/].*[\\/]|[a-z]:([\\/]))!"	// slash, stuff, slash
			"[%%](<word|@|(|)>+)[%%]",		// env var subst
			url_services);
	else
		sprintf(pat, "((((%s):)?"			// optional ftp:
			"[\\/].*[\\/]))!"				// slash, stuff, slash
			"[$](<word|@|(|)>+|{<word|@|(|)>+})", // env var subst
			url_services);
	int res = parse_string(1, pat);
	if (res) {		// If it looks like it has an env var, grab it.
		grab(find_group(5, 1), find_group(5, 0), var);
		if (var[0] == '{') {		// Found {varname}, use varname.
			var[strlen(var) - 1] = 0;
			if (filename_subst_var(var + 1))
				delete(0, res);
		} else if (filename_subst_var(var))
			delete(0, res);
	}
}

get_show_graphic()	// what's mode code for current graphic representation?
{
	switch (_display_class[0]) { /* check value for null char */
		case BNORMAL: return 1;
		case BHEX: return 2;
	}
	return _display_class[128 + 'A'] == BNORMAL ? 3 : 0;	// int'l vs multi-char
}

do_set_show_graphic(mode)	// set graphic rep by mode code
{				// 0: multi-char, 1: graphic, 2: hexadecimal, 3: int'l
	int i, std, stdctrl;

	if (show_spaces && _display_class == _show_disp_class)
		save_var _display_class = _std_disp_class;
	std = (mode == 0) ? BM : (mode == 2) ? BHEX : BNORMAL;
	stdctrl = (mode == 3) ? BC : std;
	for (i = 0; i < 128; i++)
		if (i < 32 || i == 127) {
			if (i != '\n' && i != '\t')
				_display_class[i] = mode ? stdctrl : BC;
			int dispwidth = set_character_property(i + 128, CPROP_DISP_WIDTH, -1);
			if (mode == 0 || dispwidth == 0 && mode != 2)
				_display_class[i + 128] = BMC;
			else
				_display_class[i + 128] = std;
		} else {
			_display_class[i] = BNORMAL;
			_display_class[i + 128] = std;
		}
	for (i = 256; i  < 65536; i++)
		_display_class[i] = (mode == 2) ? BHEX : BNORMAL;
	fix_display_class();
	make_show_spaces();
	full_redraw = 1;
}

int get_password(char *res, char *host, char *usr)
{
	char msg[FNAMELEN];
	struct password_data *p;
	int use_old;

	for (p = password_data; p; p = p->next)
		if (!strcmp(p->host, host) && !strcmp(p->usr, usr)) {
			if (p->invalid)
				break;
			if (res)
				strcpy(res, p->pwd);
			else				// Mark it invalid so we prompt next time.
				p->invalid = 1;
			return 0;
		}
	if (!res)		// Told to discard a pwd, but it's not there.
		return 0;
	if (_doing_input) {
		say("Can't prompt for a password within another prompt.");
		return 1;
	}
	use_old = (p != 0);		// Replace existing pwd if it was refused.
	sprintf(msg, "Password for %s@%s: ", usr, host);
	comp_read(res, msg, (int (*)()) 0, POP_UP_PROMPT | PASSWORD_PROMPT, "");
	if (!use_old)
		p = (struct password_data *) malloc(sizeof(struct password_data));
	strcpy(p->host, host);
	strcpy(p->usr, usr);
	strcpy(p->pwd, res);
	p->invalid = 0;
	if (!use_old) {		// If we're not replacing an invalid pwd,
		p->next = password_data;	// add to the list.
		password_data = p;
	}
	return 0;
}

int prepare_url_operation(char *file, int op, struct url_parts *parts)
{
	char *p;

	save_var case_fold = 1;
	if (!parse_url(file, parts))
		error("Bad URL %s", file);
	if (!strstr(" ftp scp ", parts->service) && (op & FTP_OP_MASK) != FTP_RECV)
		error("Can't do that with %s: buffers, only ftp: or scp: buffers.",
			  parts->service);
	if (!strstr(" ftp scp http https telnet ssh ", parts->service))
		error("Unknown service %s: in URL.", parts->service);

	if (!strcmp(parts->service, "ftp")) {	// An FTP URL?
		if (!parts->have_usr)		// No user specified.
			strcpy(parts->usr, "anonymous");
		if (!parts->have_password)	// If no password yet,
			if (!strfcmp(parts->usr, "anonymous")	// Use anon pwd?
				|| !strfcmp(parts->usr, "guest"))
				strcpy(parts->pwd, anon_ftp_password);
			else				// Or prompt for one/retrieve old.
				if (get_password(parts->pwd, parts->host, parts->usr))
					return 1;
	} else if (!strcmp(parts->service, "http") || !strcmp(parts->service, "https")) {
		if (parts->have_usr && !parts->have_password)
			get_password(parts->pwd, parts->host, parts->usr);
	} else if (!strcmp(parts->service, "scp") || !strcmp(parts->service, "ssh")) {
	} else if (parts->have_usr)
		error("User name not supported in %s URL's", parts->service);

	while (p = index(parts->fname, '\\'))
		*p = '/';			// Convert all \'s to /'s.
	return 0;
}

int https_activity(int activity, int buf, int from, int to);

do_https(struct url_parts *parts, int port)
{
	char cmdline[FNAMELEN], dir[FNAMELEN];

	when_activity = https_activity;
	set_mode_message("[HTTPS: Connecting] ");
	get_executable_directory(dir);
	sprintf(cmdline, ".%cgethttps ", path_sep);
	if (parts->have_usr)
		sprintf(cmdline + strlen(cmdline), "--password=\"%s\" ", parts->pwd);
	sprintf(cmdline + strlen(cmdline), "\"https://");
	if (parts->have_usr)
		sprintf(cmdline + strlen(cmdline), "%s@", parts->usr);
	sprintf(cmdline + strlen(cmdline), "%s", parts->host);
	if (port)
		sprintf(cmdline + strlen(cmdline), ":%d", port);
	sprintf(cmdline + strlen(cmdline), "/%s\"", parts->fname);
	if (!buf_pipe_text(0, bufnum, cmdline, dir, PIPE_CLEAR_BUF))
		file_error(errno, "Running gethttps helper", "Couldn't start gethttps helper");
}

int url_operation(char *file, int op)
{
	struct url_parts parts;
	int port, i;

	if (prepare_url_operation(file, op, &parts))
		return 1;
	port = numtoi(parts.port);
	if (*parts.port && got_bad_number)
		say("Ignoring invalid port number %s", parts.port);
	if (i = find_variable_by_host("fallback_remote_translation_type", parts.host))
		fallback_translation_type = get_num_var(i);
	if (!strfcmp(parts.service, "http")) {
		if (got_bad_number)
			port = 80;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with HTTP files.");
		do_http(&parts, port);
	} else if (!strfcmp(parts.service, "https")) {
		if (got_bad_number)
			port = 443;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with HTTPS files.");
		do_https(&parts, port);
	} else if (!strfcmp(parts.service, "telnet")) {
		if (got_bad_number)
			port = 23;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with telnet buffers.");
		do_telnet(parts.host, port, bufname);
	} else if (!strfcmp(parts.service, "ssh")) {
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with ssh buffers.");
		do_ssh(parts.host, bufname, parts.usr);
	} else if (!strfcmp(parts.service, "scp")) {
		do_scp(&parts, op);
	} else
		do_ftp_op(bufname, parts.host, parts.port, parts.usr, parts.pwd,
				  parts.fname, op);
	return 0;		// No error possible yet (asynchronous).
}

remote_dirname_absolute(char *s)
{
	char *t1, *t2, cmd[FNAMELEN], *name;
	struct url_parts parts;
	int port;

	if (prepare_url_operation(s, FTP_MISC | FTP_WAIT, &parts))
		return;
	if (!strcmp(parts.service, "scp")) {
		scp_remote_dirname_absolute(s, &parts);
		return;
	}
	noteput("Expanding home directory, please wait...");
	t1 = temp_buf();
	t2 = temp_buf();
	name = get_url_file_part(s, 0);
	if (*parts.fname)
		sprintf(cmd, "CWD %s\nPWD\n", parts.fname);
	else
		sprintf(cmd, "CWD\nPWD\n");
	port = numtoi(parts.port);
	if (got_bad_number)
		port = 21;
	ftp_op(t1, t2, parts.host, port, parts.usr, parts.pwd, cmd,
		   FTP_MISC | FTP_WAIT);
	save_var bufname = t2;
	point = 0;
	if (re_search(1, "^257 \"(.*)\"")) {
		grab(find_group(1, 1), find_group(1, 0), name);
		restore_vars();
		delete_buffer(t1);
		delete_buffer(t2);
		note("");
	} else {
		say("Can't expand pathname %s, log %s is %d",
			s, t2, size());
		buffer_printf("Epsnet-Debug", "Can't expand pathname %s,"
					  " log %s is %d\n", s, t2, size());
	}
}

int is_ftp_url_a_dir(char *s)
{
	char cmd[FNAMELEN];
	struct url_parts parts;
	int port, res;

	if (prepare_url_operation(s, FTP_MISC | FTP_WAIT, &parts))
		return 0;
	if (!strcmp(parts.service, "scp"))
		return scp_is_url_a_dir(s, &parts);
	if (ftp_compatible_dirs & 4)
		return 0;
	if (*parts.fname)
		sprintf(cmd, "CWD %s", parts.fname);
	else
		return 1;
	port = numtoi(parts.port);
	if (got_bad_number)
		port = 21;
	char *t1 = temp_buf(), *t2 = temp_buf();
	set_mode_message("[FTP: Resolving link] ");
	ftp_op(t1, t2, parts.host, port, parts.usr, parts.pwd, cmd,
		   FTP_MISC | FTP_WAIT);
	set_mode_message(NULL);
	save_var bufname = t2;
	point = 0;
	res = !re_search(1, "^(550|530) ");
	restore_vars();
	delete_buffer(t1);
	delete_buffer(t2);
	return res;
}

ftp_misc_operation(char *url, char *cmd)
{
	char *t1;
	struct url_parts parts;

	if (scp_misc_op(url, cmd))
		return;
	t1 = temp_buf();
	on_exit delete_buffer(t1);
	if (!prepare_url_operation(url, FTP_MISC, &parts))
		do_ftp_op(t1, parts.host, parts.port, parts.usr, parts.pwd, cmd,
				  FTP_MISC);
}

when_loading()
{
}
