/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2011 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file shows how to add a new mode.  Make a copy of it and replace
// every use of "sample" with the name of your mode.

#include "eel.h"
#include "colcode.h"

keytable sample_tab;		// The mode will define some keys.

// We'll only syntax-highlight comments.
color_class sample_comment = color_class c_comment;

// When the cursor is at a delimiter, highlight the matching one.
user char auto_show_sample_delimiters = 1;
user char sample_auto_show_delim_chars[20] = "{[()]}";

// Uncomment to tell compile-buffer which external program to run on the file.
// user char compile_sample_cmd[128] = "samplecomp \"%r\"";


// Color lines that start with a # as comments.
color_sample_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = size() - to;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		// Change the search pattern to find whatever syntax you want
		// to color.  See existing modes for examples.
		if (!re_search(1, "^[ \t]*#")) {
			t = size();
			break;
		}
		t = matchstart;
		// Check last char. (Not needed for this simple example.)
		switch (character(point - 1)) {
			case '#':
				nl_forward();
				set_character_color(t, point, color_class sample_comment);
				break;
			default:
				// Add coloring logic for other syntax here.
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

command sample_mode()
{
	mode_default_settings();
	major_mode = strkeep("Sample");
	mode_keys = sample_tab;		// Use these keys.

	strcpy(comment_start, "#[ \t]*");	// Have the commenting commands use #.
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	comment_column = 0;

	// Set up coloring rules.
	recolor_range = color_sample_range;
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// Maybe turn on coloring.
		when_setting_want_code_coloring();

	fill_mode = 1;		// Assume we want comment-filling.
	buffer_maybe_break_line = generic_maybe_break_line;

	if (auto_show_sample_delimiters)
		auto_show_matching_characters = sample_auto_show_delim_chars;

	try_calling("sample-mode-hook");
	make_mode();
}

// Set up some keys.
when_loading()
{
	sample_tab[ALT('q')] = (short) fill_comment;
	sample_tab[ALT('Q')] = (short) prefix_fill_paragraph;

	// If ) should jump back to its matching (, not merely highlight
	// it, uncomment this line.
//	sample_tab[')'] = (short) show_matching_delimiter;

}

// This sample function name finder assumes that the files use the word
// SECTION at the start of a line to start a section, whose name follows
// on that line. This code then displays the current section name in the
// mode line. Modify as needed, or comment out this subroutine if you
// don't need this functionality.
sample_func_name_finder()
{
	save_var point;
	save_var case_fold = 0;
	*display_func_name = 0;
	if (re_search(-1, "^SECTION</word> *")) {
		int to = give_end_line();
		if (to - matchstart < ptrlen(display_func_name) - 2)
			grab(matchstart, to, display_func_name);
	}
	return 1;
}

// This makes .samp files use the new mode.
suffix_samp()
{
	sample_mode();
}

