/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has the searching and replacing commands.

#include "eel.h"
#include "colcode.h"

// Remember this position.
remember_old_place(old_place *p, int pos)
{
	if (search_continuation)
		(*search_continuation)(SCON_RECORD, pos, p->item);
	p->pos = pos;
}

// Go to this remembered position (including its item).
int to_old_place(old_place *p)
{
	save_var window_start;
	if (search_continuation)
		(*search_continuation)(SCON_RESTORE, 0, p->item);
	point = p->pos;
}

// Return 1 if point/current item is bigger than old position.
int past_old_place(old_place *p)
{
	int c;

	save_var window_start;
	if (search_continuation)
		if (c = (*search_continuation)(SCON_COMPARE, 0, p->item))
			return c > 0;
	return point > p->pos;
}

save_old_place_to_bookmark(old_place *p)
{
	if (point == p->pos)		// Don't save it unless we moved away.
		return;
	save_var window_start, point = p->pos;
	quiet_set_bookmark();
}

// Help set up to restrict searching to region; return handle for
// highlighting.
int make_restriction(int flags)
{
	int h;

	set_spot(&restrict_search.start, MIN(point, mark));
	set_spot(&restrict_search.end, MAX(point, mark));
	restrict_search.colstart = virtual_column();
	restrict_search.colend = virtual_mark_column();
	if (restrict_search.colstart > restrict_search.colend) {
		int t = restrict_search.colstart;
		restrict_search.colstart = restrict_search.colend;
		restrict_search.colend = t;
	}
	restrict_search_type = region_type();
	highlight_off();
	if ((point < mark) != !(flags & REVERSE))	// Forward-search region from top.
		exchange_point_and_mark();				// Backward-search region from bottom.
	h = add_region(restrict_search.start, restrict_search.end,
				   color_class pull_highlight, region_type());
#if defined(MRSTARTCOL)
	if (region_type() == REGRECT) {
		modify_region(h, MRSTARTCOL, restrict_search.colstart);
		modify_region(h, MRENDCOL, restrict_search.colend);
	}
#endif
	return h;
}

// Grab the entire current prompt buffer and put in the string *str,
// expanding it as needed.  Take the current flag settings and return
// new ones (usually unaltered).  But if the buffer contains nulls,
// maybe translate it into a regex pattern so we can search for it.

int grab_search_string(int flags, char **str)
{
	save_var abort_searching = 0;
	grab_expanding(0, size(), str, 256);
	if (size() != strlen(*str)) {		// Has nulls.
		if ((flags & SEARCH_REGEX) && !(flags & SEARCH_REGEX_HIDDEN)) {
			// If it's a regex pattern, replace literal nulls with
			// regex coding that finds them.
			spot sp = alloc_spot(1);	// Preserve point w/left-inserting spot.
			on_exit point = *sp, free_spot(sp);
		} else {
			int tmp = tmp_buf();		// Convert literal text to regex.
			on_exit buf_delete(tmp);
			buf_xfer(tmp, 0, size());	// Make a copy of buffer with nulls.
			save_var bufnum = tmp;
			point = size();
			while (re_search(-1, "[][|()+*?.$%<>^]")) {
				insert('%');		// Quote these literal characters for regex.
				point--;
			}
			flags |= SEARCH_REGEX | SEARCH_REGEX_HIDDEN;
		}
		simple_re_replace(1, "<#0>", "<#0>");	// Nulls => 4 chars "<#0>".
		grab_expanding(0, size(), str, 256);	// Get revised pattern.
	}
	return flags;
}

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, ostart, cnt;
	int oreg_type, old_mark = -1, search_contin = search_continuation != 0;
	int restrict_region = 0, recall_cnt = 0;
	char tmpbuf[FNAMELEN], search_other_buf_name[FNAMELEN];
	old_place orig, mstart, okpoint;

	if (_doing_input == DI_SEARCH || _doing_input == DI_SEARCH_NOINCR)
		error_if_input(0);
	if (search_in_region && is_highlight_on()) {
		save_var _highlight_control;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	} else if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight && typing_deletes_highlight != 3
		  && !in_macro(1) && _len_def_mac < 1 && search_defaults_from == 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var _highlight_control;
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
	save_var prompt_orig_buf = bufnum;
	save_var prompt_orig_win = window_handle;
	save_var last_index;
	save_var mark_to_column = -1, mark = point;
	save_var cursor_to_column = -1;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY | SEARCH_CONTIN;
	remember_old_place(&orig, point);
	remember_old_place(&mstart, point);
	remember_old_place(&okpoint, point);
	ostart = window_start;
	if (!search_other_name) {
		sprintf(search_other_buf_name, "search-%s", bufname);
		save_var search_other_name = search_other_buf_name;
	}
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	} else if (search_defaults_from == 2 && !size()) {
		if (flags & UNKNOWN)
			flags = known_flags(flags);
		recall_cnt++;
		stuff(default_search_string(flags));
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
		shift_selecting = 2;
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = full_key;
				break;
			}
		cnt = 1;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			to_old_place(&okpoint);
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_contin) {
					flags |= SEARCH_CONTIN;
					s_code = FHERE;			/* Try other records. */
				} else if (search_wraps & 1) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(full_key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(full_key, reg_tab)
				   == (short) backward_delete_character && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (is_recall_key(flags)) {
			if ((recall_prior_response_options & 2) && !IS_ALT_KEY(key))
				recall_cnt = pick_prev_search_from_end(&flags,
					recall_cnt + ((is_recall_key(flags))), "search");
			else
				flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (binding_of(full_key, comp_tab) == (short) inp_get_other) {
			flags = pick_search_recall(flags);
		} else if (binding_of(full_key, reg_tab) == (short) pull_word
				  || binding_of(full_key, reg_tab) == (short) pull_word_fwd) {
			flags &= ~UNKNOWN;
			window_handle = origwin;
			_highlight_control = 0;
			window_handle = editwin;
			pull_word();
		} else if (binding_of(full_key, reg_tab) == (short) delete_character
				   && (search_delete_match & 1) && (flags & INCR)) {
			window_handle = origwin;
			if (!(oldflags & FAILING))
				delete(point, mark);
			full_key = -1;
			break;
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(full_key, comp_tab) == (short) inp_grab_word
				 || binding_of(full_key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
						(binding_of(full_key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						remember_old_place(&orig, point);
						remember_old_place(&mstart, point);	// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = (flags & INCR) ? DI_SEARCH : DI_SEARCH_NOINCR;
		if (!is_window(origwin) || !is_window(editwin)
		    || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		flags = grab_search_string(flags, &trystr);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			to_old_place(&orig);
		else if (s_code >= FMATCH) {
			to_old_place(&mstart);
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
search_again:
		switch (do_color_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 0:
				if ((flags & SEARCH_CONTIN) && search_contin
					  && (*search_continuation)(cnt++, flags, trystr))
					goto search_again;
				flags |= FAILING;
				to_old_place(&okpoint);
				if (in_macro(1))
					goto end;
				else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:
				remember_old_place(&okpoint, point);
				remember_old_place(&mstart, matchstart);
				mark = matchstart;
				highlight_on();
				if ((flags & SEARCH_WRAPPED)
					  && past_old_place(&orig) == !(flags & REVERSE))
					flags |= SEARCH_OVERWRAPPED;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				to_old_place(&okpoint);
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				while (ungot_key == -1 && char_avail()
					   && getkey() == abort_key || !IS_TRUE_KEY(key))
					;
				if (key != abort_key && IS_TRUE_KEY(key))
					ungot_key = full_key;
				drop_pending_says();
				break;
		}
	}
	window_handle = editwin;
	if (size()) {
		flags = grab_search_string(flags, &trystr);
		save_search_recall(flags, trystr);
	}
	if (key == abort_key || user_abort) {
		window_handle = origwin;
		to_old_place(&orig);
		window_start = ostart;
	} else {
		window_handle = origwin;
		if (set_bookmark_automatically & 1)
			save_old_place_to_bookmark(&orig);
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
					  : &_default_flags) = flags;
			if (_last_was_regex)
				grab_search_string(flags, &_default_regex_search);
			else
				grab_search_string(flags, &_default_search);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = full_key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if (restrict_region)
		remove_region(restrict_region);
	if ((flags & FAILING) && in_macro(1))
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

char *help_on_search(flags)	/* give help on correct command */
{
	if (flags & GREP)
		return (flags & QUERY) ? "file-query-replace" : "grep";
	else if (flags & REGEX)
		return (flags & REPLACE) ? "regex-replace"
				: "regex-search";
	else
		return !(flags & REPLACE) ? "incremental-search"
				: (flags & QUERY) ? "query-replace"
				: "replace-string";
}

exit_incr()	/* does this key exit an incremental search? */
{
	int b = get_binding(key, mode_keys);
	if (name_type(b) == NT_MACRO)
		return 0;
	return key != '\t' && key != CTRL('Q') && (key < ' ' || !IS_ASCII_KEY(key))
			&& ((binding_of(full_key, comp_tab) != (short) inp_grab_word
				 && binding_of(full_key, comp_tab) != (short) inp_grab_line)
				|| !IS_ALT_KEY(key)) && !norm_binding();
}

// Should this key list previous responses, or perhaps select one?
is_recall_key(int flags)
{
	if (IS_ALT_KEY(key) || !(flags & INCR))
		if (binding_of(full_key, comp_tab) == (short) inp_recall)
			return 1;
		else if (binding_of(full_key, comp_tab) == (short) inp_grab_word
				 && (recall_prior_response_options & 2) && !IS_ALT_KEY(key))
			return -1;
	return 0;
}

norm_binding()	/* does this key run normal-character? */
{
	if (!mode_keys)
		return 0;
	return ((key & CMD_INDEX_KEY) ? (key & ~CMD_INDEX_KEY)
			: (mode_keys[key] ? mode_keys[key] : root_keys[key]))
			== (short) normal_character;
}

set_flag_modes(fp)	/* maybe set *fp from key, return 1 if we could */
int *fp;
{
	if (key == CTRL('C')) {
		if ((*fp & REPLACE) && replace_by_case == 2) {
			if (*fp & CASE_REPLACE)
				*fp &= ~(CASE_REPLACE | FOLD);
			else if (*fp & FOLD)
				*fp |= CASE_REPLACE;
			else
				*fp |= FOLD;
		} else
			*fp ^= FOLD;
		*fp |= MODFOLD;
		return 1;
	} else if (key == CTRL('W'))
		*fp ^= WORD;
	else if (key == CTRL('T')) {
		if (*fp & SEARCH_REGEX_HIDDEN)
			*fp &= ~SEARCH_REGEX_HIDDEN;
		else
			*fp ^= REGEX;
		if (*fp | REGEX) /* if entering regex, not incremental */
			*fp &= ~INCR;
	} else
		return 0;
	if (*fp & UNKNOWN)	/* setting regex or word makes type known */
		*fp &= ~(INCR | UNKNOWN);
	return 1;
}

is_search_dir()	/* return 1 if it's a direction key inside search */
{		/* also replace with standard key value */
	if (key == CTRL('R') || key == ALT(CTRL('R'))
					|| full_key == rev_search_key || key == rev_search_key)
		key = CTRL('R');
	else if (key == CTRL('S') || key == ALT(CTRL('S'))
					|| full_key == fwd_search_key || key == fwd_search_key)
		key = CTRL('S');	/* substitute standard one */
	else
		return 0;
	return 1;
}

known_flags(flags)	/* set flags when user pulls in default string */
{
	int def, new;

	flags &= ~(SMODES | UNKNOWN);
	def = _last_was_regex ? _default_regex_flags : _default_flags;
	def &= (SMODES | INCR);
	new = flags | (def & ~INCR);	/* no incr if grep, e.g., calls */
	if ((flags & REGEX) && !(def & INCR))
		new &= ~INCR;	/* only allow explicit incr regex search */
	if (((new & FOLD) != 0) != case_fold.default)
		new |= MODFOLD;
	return new;
}

make_search_prompt(s, flags)
char *s;
{
	sprintf(s, "%s%s%s%s%s%s%s",
		(flags & SEARCH_CONTIN) ? "Continued " :
		(flags & SEARCH_OVERWRAPPED) ? "Overwrapped " :
			(flags & SEARCH_WRAPPED) ? "Wrapped " : "",
		(flags & BAD) ? "Bad " : (flags & FAILING) ? "Failing " : "",
		(flags & REVERSE) ? "Reverse " : "",
		!(flags & MODFOLD) ? "" : (flags & CASE_REPLACE) ? "CaseRepl "
			: (flags & FOLD) ? "Folding " : "Exact ",
		(flags & REGEX) && !(flags & SEARCH_REGEX_HIDDEN) ? "R-E " : "",
		(flags & WORD) ? "Word " : "",
		(flags & INCR) ? "I-" : "");
}

show_search_prompt(prmpt, flags, origwin, pwin, editwin)
char *prmpt;
{
	char tmp[80];

	window_handle = pwin;
	buf_zap(bufnum);
	make_search_prompt(tmp, flags);
	stuff(tmp);
	stuff(prmpt);
	point = 0;
	set_wattrib(editwin, BLEFT, size());	/* reset window edges */
	window_handle = origwin;
	if (display_column > 0
		  && get_column((flags & REVERSE) ? mark : point) < text_width())
		display_column = 0;	// Position to left edge if whole match fits.
	window_handle = (flags & (INCR | SONCE)) ? origwin : editwin;
	maybe_refresh();
}

save_search_recall(int flags, char *str)
{
	char *srch;
	int buf = tmp_buf();

	save_var bufnum = buf, abort_searching = 0;
	bprintf("%s%s%s",
		!(flags & MODFOLD) ? "" : (flags & FOLD) ? "Fold " : "Nofold ",
		(flags & WORD) ? "Word " : "",
		(flags & REGEX) ? "Regex " : "");
	if (character(point - 1) == ' ')
		delete(point - 1, point);
	bprintf(": %s", str);
	simple_re_replace(1, "%<NL>", "%<NL>");	// Put a % before actual <NL> strings.
	simple_re_replace(1, "\n", "<NL>");	// Convert newlines to <NL>.
	srch = malloc(size() + 1);
	grab(0, size(), srch);
	restore_vars();
	buf_delete(buf);
	save_recall(srch, "search");
	if (search_other_name)
		save_recall(srch, search_other_name);
	free(srch);
}

pick_prev_search(flags)	// select old search string from list, insert in curbuf
{
	int w = window_handle, menu, mbuf, obuf = bufnum, res;

	buf_zap(bufnum);
	save_var recall_id = "search";
	menu = new_menu(mbuf = tmp_buf());
	make_menu(menu, "", (char *(*)()) 0, 0, w, mbuf, list_recalls);
	if (is_gui && want_gui_prompts)
		one_window_to_dialog("Search for", menu, "OK", "Cancel", "");
	res = select_menu_item(obuf, menu, w, -1);
	buf_delete(mbuf);
	point = 0;			// parse and remove flag info
	switch (res) {
		case INP_ABORT:
			user_abort = 1;
		default:
			return flags;
		case INP_GET_EXIT:
		case INP_GET:
			break;
	}
	flags = extract_search_flags(flags);
	if ((flags & GREP) && res == INP_GET_EXIT)
		ungot_key = '\n';
	return flags;
}

// The buffer contains saved search info.  Remove flags and return
// them, leaving the actual search text in the buffer.

int extract_search_flags(int flags)
{
	point = 0;
	if (!search(1, ":")) {
		user_abort = 1;
		return flags;
	}
	narrow_end = size() - point - 1;
	point = 0;
	if (curchar() == 'N') {		// explicit Nofold flag
		flags &= ~FOLD;
		search(1, " ");
	}
	flags = recall_search_flag('F', flags, FOLD, 0);
	flags = recall_search_flag('W', flags, WORD, 1);
	flags = recall_search_flag('R', flags, REGEX, 1);
	if (((flags & FOLD) != 0) != case_fold.default)
		flags |= MODFOLD;
	point = size();		// get to end of flag info
	narrow_end = 0;
	delete(0, point);	// remove flag info
	point = 0;
	while (search(1, "<NL>")) {
		if (character(matchstart - 1) == '%')	// Unquote actual <NL> we searched for.
			delete(matchstart, matchstart - 1);
		else {						// Convert <NL> back to \n.
			delete(matchstart, matchend);
			stuff("\n");
		}
	}
	point = size();
	if (recall_response_selected && typing_deletes_highlight)
		highlight_on();
	return flags;
}

// Insert pos'th search entry from the end in cur buffer, counting last
// as 0'th.  Wrap suitably.  Return the number of the entry we actually
// return.

int pick_prev_search_from_end(int *flags, int pos, char *tag)
{
	char res[FNAMELEN];
	int ret;

	save_var recall_id = tag;
	ret = get_last_recall_from_end((char *(*)())0, res, pos);
	buf_zap(bufnum);
	stuff(res);
	*flags &= ~(MODFOLD | FOLD);
	if (case_fold)
		*flags |= FOLD;
	*flags = extract_search_flags(*flags);
	if (recall_response_selected) {
		highlight_on();
		shift_selecting = 2;
	}
	return ret;
}

recall_search_flag(letter, flags, thisflag, def)
{			// set flag based on keyword, skip past it
	if (curchar() == letter) {
		flags |= thisflag;
		search(1, " ");
	} else if (def)		// if letter's missing, maybe turn off flag
		flags &= ~thisflag;
	return flags;
}

// Copy word or line from origwin to current window.
// Return nonzero if we moved the origin.

int grab_search_word(flags, origwin, line)
{
	char text[FNAMELEN];
	int orig, len = size(), ret = 0;

	save_var window_handle = origwin;
			// First time, if we're inside a word, go to its start.
	if (!len && !line && !(flags & FAILING) && parse_string(1, word_pattern)
		&& parse_string(-1, word_pattern)) {
		point -= parse_string(-1, word_pattern);
		ret = 1;
	}
	if (!(flags & REVERSE))
		save_var point;
	else if (len && !(flags & FAILING))
		point = matchstart;
	orig = point;
	if (!(flags & FAILING))
		if (line) {
			if (curchar() == '\n')
				point++;
			to_end_line();
		} else
			re_search(1, word_pattern);		// Get the next word.
	int max = orig + sizeof(text) / sizeof(*text) - 1;
	if (point >= max)
		point = max;
	grab(orig, point, text);
	restore_vars();
	if (search_defaults_from && shift_selecting == 2 && is_highlight_on()) {
		delete(0, size());
		shift_selecting = 0;
	}
	point = size();
	bprintf((flags & REGEX) ? "%q" : "%s", text);
	return ret;
}

int pick_search_recall(int flags)
{
	if (search_other_name)
		pick_prev_search_from_end(&flags, 1, search_other_name);
	return flags;
}

/* string-search command */
command regex_search() on reg_tab[ALT(CTRL('S'))]
{
	conv_search(REGEX);
}

/* reverse-string-search backward */
command reverse_regex_search() on reg_tab[ALT(CTRL('R'))]
{
	conv_search(REGEX | REVERSE);
}

command incremental_search() on reg_tab[CTRL('S')]
{
	conv_search(UNKNOWN | INCR);
}

command reverse_incremental_search() on reg_tab[CTRL('R')]
{
	conv_search(UNKNOWN | INCR | REVERSE);
}

command string_search()
{
	conv_search(0);
}

command reverse_string_search()
{
	conv_search(REVERSE);
}

command toggle_case_fold()
{
	case_fold = has_arg ? (iter != 0) : !case_fold;
	say("Case fold %sabled for this buffer.",
		  case_fold ? "en" : "dis");

}

get_search_string(pr, flags)	/* prompt for search string & return flags */
char *pr;
{
	return search_read("", pr, default_fold(flags));
}

search_flags_with_direction(flags)
{			// return default search flags including direction
	int def = _last_was_regex ? _default_regex_flags : _default_flags;

	return known_flags(flags) | (def & REVERSE);
}

command search_again()
{	// search for last thing without prompting, in same direction
	do_search_again(search_flags_with_direction(UNKNOWN));
}

command forward_search_again()
{
	do_search_again(known_flags(UNKNOWN) & ~REVERSE);
}

command reverse_search_again()
{
	do_search_again(known_flags(REVERSE));
}

do_search_again(flags)
{
	char tmp[FNAMELEN];

	make_search_prompt(tmp, flags);
	say("%sSearch: %s", tmp, default_search_string(flags));
	if (search_positions_at_start && !(flags & REVERSE))
		point++;
	repeat_search(iter, flags);
	iter = 0;
	say("");
}

command search_region()
{
	save_var search_in_region = 1;
	on_exit highlight_off();
	highlight_on();
	incremental_search();
}

command replace_again()
{	// right now, we can only replace forward
	int flags = search_flags_with_direction(UNKNOWN) | QUERY;
	char tmp[FNAMELEN], *s, *t;

	if (_buf_readonly)
		check_modify(bufnum);
	make_search_prompt(tmp, flags);
	say("%sReplace string: %s", tmp, default_search_string(flags));
	s = default_search_string(flags);
	t = *default_replace_string(flags);
	last_index = find_index(help_on_search(flags));
	show_replace(s, t, flags);
	string_replace(s, t, flags | INCR);
	iter = 0;
	say("");
}

conv_search(flags) /* prompt for search string, then search, but see flags */
{
	flags = get_search_string("Search: ", flags);
	if (!(flags & (INCR | SONCE)))
		repeat_search(iter, flags);
	iter = 0;
}

ask_wrap(int flags)
{
	char msg[FNAMELEN];
	sprintf(msg, "Not found.  Wrap to %s of buffer and continue searching?",
			(flags & SEARCH_REVERSE) ? "end" : "start");
	return ask_yn("Wrap and continue searching?", msg, "Yes", "No", 1);
}

repeat_search(cnt, flags)	 // search & report errors
{
	int orig, ret = 0, wrap = (search_wraps & 2) ? 1 : 0;
	char *s;

	s = default_search_string(flags);
	while (cnt-- > 0) {
        if (wrap == 2)      // Just wrapped, no more wrapping.
            wrap = 0;
        else
            orig = point;
		ret = do_color_searching(flags, s);
		if (ret == 1)
			continue;
		if (ret == DSABORT)
			say("Search cancelled.");
        else if (wrap == 1 && ask_wrap(flags)) {
            wrap = 2;
            point = (flags & REVERSE) ? size() : 0;
            say("Search wrapped.");
            cnt++;
            continue;
        } else {
			if (ret == DSBAD)
				say("%s: bad regex pattern.", s);
			else
				say("%s: not found.", s);
			point = orig;
			if (in_macro(1))
				quick_abort();
		}
		break;
	}
	if (search_positions_at_start && ret == 1)
		point = matchstart;
}

default_fold(flags)		/* initialize flag from buffer variable */
{
	if (case_fold)
		flags |= FOLD;
	if (case_fold != case_fold.default)
		flags |= MODFOLD;
	return flags;
}

// Return the most recent search string.  For functions that may wish
// to write short amounts to these, we ensure there's some room.
char *default_search_string(int flags)
{
	if (flags & REGEX) {
		if (!_default_regex_search)
			_default_regex_search = malloc(FNAMELEN);
	} else {
		if (!_default_search)
			_default_search = malloc(FNAMELEN);
	}
	return (flags & REGEX) ? _default_regex_search : _default_search;
}

// Return a pointer to the most recent replace string's location.  For
// functions that may wish to write short amounts to these, we ensure
// there's some room.
char **default_replace_string(int flags)
{
	if (flags & REGEX) {
		if (!_default_regex_replace)
			_default_regex_replace = malloc(FNAMELEN);
	} else {
		if (!_default_replace)
			_default_replace = malloc(FNAMELEN);
	}
	return (flags & REGEX) ? &_default_regex_replace : &_default_replace;
}

command replace_string() on reg_tab[ALT('&')]
{
	gen_replace(0);
}

command query_replace() on reg_tab[ALT('%')], reg_tab[ALT('r')]
{
	gen_replace(QUERY);
}

command regex_replace() on reg_tab[ALT('*')], reg_tab[NUMALT(GREYSTAR)]
{
	gen_replace(QUERY | REGEX);
}

command reverse_replace()
{
	gen_replace(QUERY | REVERSE);
}

gen_replace(flags)	/* get input, then do replace */
{
	char *str, with[10000];
	int restrict_region = 0, want_restrict = 0;

	save_var mark, mark_to_column, point, cursor_to_column;
	if (replace_in_region && is_highlight_on()) {
		want_restrict = 1;
		save_var search_in_region = 1, typing_deletes_highlight = 3;
	}
	if (_buf_readonly)
		check_modify(bufnum);
	if (has_arg)
		flags |= WORD;
	restore_vars();
	save_var mark, mark_to_column, point, cursor_to_column;
	if (want_restrict) {
		save_var search_in_region = 1, typing_deletes_highlight = 3;
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
		on_exit remove_region(restrict_region), restrict_region = 0;
	}
	flags = get_replace_text(flags, with);
	restore_vars();
	if (want_restrict) {
		if (restrict_region)
			remove_region(restrict_region);
		save_var restrict_search_type;
		restrict_region = make_restriction(flags);
	}
	str = default_search_string(flags);
	if (flags != -1) {
		if (!want_restrict && point > mark && is_highlight_on() && !(flags & REVERSE))
			exchange_point_and_mark();
		show_replace(str, with, flags);
		string_replace(str, with, flags | INCR);
	}
	if (restrict_region)
		remove_region(restrict_region);
	if (want_restrict && replace_in_region == 2)
		highlight_on();
}

get_replace_text(int flags, char *with)
{
	char prompt[80], **str;

	iter = 0;
	flags = default_fold(flags);
	sprintf(prompt, "%s%sreplace string: ", (flags & FILEGREP) ? "file " :
		(flags & BUFGREP) ? "buffer " : "",
		(flags & QUERY) ? "query " : "");
	*prompt = toupper(*prompt);
	flags = search_read("", prompt, flags | REPLACE);
	flags &= ~(INCR | FAILING | BAD);
	if (flags & SEARCH_EMPTY)
		return -1;
	str = default_replace_string(flags);
	last_index = find_index(help_on_search(flags));
	save_var recall_id = "replace";
	save_var inp_get_other_name = default_search_string(flags);
	get_strnone(with, "with: ", *str);
	copy_expanding(with, str, FNAMELEN);
	return flags;
}

show_replace(str, with, flags)	/* print in echo area */
char *str, *with;
{
	char msg[FNAMELEN];

	make_search_prompt(msg, flags);
	if (flags & BAD) {
		say("%sReplace %s", msg, str);
		return;
	}
	strcat(msg, "Replacing ");
	if (strlen(str) > screen_cols / 2) {
		strncat(msg, str, screen_cols / 2);
		strcat(msg, "...");
	} else
		strcat(msg, str);
	strcat(msg, " with ");
	if (strlen(with) > screen_cols / 2) {
		strncat(msg, with, screen_cols / 2);
		strcat(msg, "...");
	} else
		strcat(msg, with);
	if (!in_macro(1))
		if (flags & QUERY)
			note("%s", msg);
		else
			noteput("%s", msg);
}

any_uppercase(char *p)	// return nonzero if p has any uppercase letters
{
	for (; *p; p++)
		if (isupper(*p))
			return 1;
	return 0;
}

// Return 1 if we should do old-style case-based replace, 2 for smart
// case replace, or 0 for none.
want_case_replace(char *str, char *with, int flags)
{
	if (flags & MODFOLD) {
		if (!(flags & FOLD))
			return 0;
	} else if (!case_fold)
		return 0;
	if (flags & CASE_REPLACE)
		return 2;
	return !any_uppercase(str) && !any_uppercase(with);
}

// Are words in the range i..j entirely uppercase, all capitalized
// (each word starts with an uppercase letter, optionally followed
// by nothing but lowercase letters until spaces), or something else?
// Return 1 for uppercase, 2 for capitalized, 0 otherwise.

check_for_case_replace(i, j)
{
	int k;
	int sofar = 0, oneupper = -1;

	if (i > j)
		k = i, i = j, j = k;
	while (i < j) {
		for (; !isalpha(character(i)); )
			if (++i >= j)
				return (oneupper == 1) ? 1 : sofar;
		if (isupper(character(i)) && oneupper == -1)
			oneupper = 1;
		if (islower(character(i)))		// Is 1st alpha character lowercase?
			return 0;			// Otherwise, must look further.
		else if (++i >= j)
			return sofar ? sofar : 1;	// Only found 1 ltr words, assume UC.
		while (isalpha(character(i))) {
			if (islower(character(i))) {	// Are other alpha chars lowercase?
				if (sofar == 1)		// Last word was upper, this is mixed.
					return 0;
				sofar = 2;
				oneupper = 0;
			} else {
				if (sofar == 2)		// Last word was mixed, this is upper.
					return 0;
				sofar = 1;
			}
			if (isupper(character(i)))
				oneupper = 0;
			if (++i >= j)
				return sofar;
		}
	}
	return sofar;
}

// Is text in the range i..j entirely uppercase (return 1), entirely
// lowercase (return 2), or mixed case?  If mixed, is the first
// character in the region uppercase (return 3) or not (return 4)?
// Return 5 if there are no letters in the range at all.

check_alt_case_replace(i, j)
{
	int k, start;
	int upper = 0, lower = 0;

	if (i > j)
		k = i, i = j, j = k;
	for (start = character(i); i < j; i++) {
		if (isupper(character(i)))
			upper = 1;
		else if (islower(character(i)))
			lower = 1;
	}
	if (upper && !lower)
		return 1;
	if (!upper && lower)
		return 2;
	if (upper && lower)
		return isupper(start) ? 3 : 4;
	return 5;
}

/* General replace routine. */

/* Interprets flag INCR to mean "interactive, show results". */

/* Returns DSABORT if the user aborted, DSBAD if illegal search string, */
/* 1 if user said to exit early, or 0 if we got to the last match. */

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, i;
	int null_detect = -1, new_null_detect;
	int perform_case_replace = 0, this_case_replace = 0;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	if (replace_by_case)
		perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_color_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end)) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(full_key, root_keys) == (short) help) {
			help_on_command(help_on_search(flags));
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		} else if (IS_MOUSE_SINGLE(key)) {
			ret = 1;
			ungot_key = full_key;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				null_detect = -1;
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
					/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
						show = 1;
						goto ask;
					}
					if (!IS_ENTER_KEY(key))
						ungot_key = full_key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = (perform_case_replace == 1)
									? check_for_case_replace(matchstart, point)
									: check_alt_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(str, with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace == 1)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			else if (perform_case_replace == 2) {
				i = MIN(matchstart, point);
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_lowercase_region(matchstart, point); break;
					case 3: do_uppercase_region(i, i + 1); break;
					case 4: do_lowercase_region(i, i + 1); break;
				}
			}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
			case ',': case '^':	skip = 1; goto ask;
			case '.':		ret = 1; goto done;
			case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement cancelled.");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro(1))
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
			    replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

get_replace_key(flags, str, skip)	// show match, then get a key
char *str;
{
	int oreg_type = region_type();

	if (!skip) {	// make sure we're in the right place given current flags
		save_var abort_searching = 0;
		point = matchstart;
		if (!do_color_searching(flags, str))
			return 0;
	}
	save_var mark = matchstart, _highlight_control;
	set_region_type(REGNORM);
	highlight_on();
	save_var matchstart, matchend;	// refresh could clobber this
	if (display_column > 0
		  && get_column((flags & REVERSE) ? mark : point) < text_width())
		display_column = 0;	// Position to left edge if whole match fits.
	do {
		maybe_refresh();		/* show the selection */
		getkey();
	} while (!IS_TRUE_KEY(key) && !IS_MOUSE_SINGLE(key));
	if (key == CTRL('^'))
		full_getkey("", CTRLIFY_KEY);
	user_abort = 0;
	restore_vars();
	set_region_type(oreg_type);
	if (!skip) {	// refresh could've clobbered regex state info
		save_var abort_searching = 0;	// so repeat srch
		point = matchstart;
		if (!do_color_searching(flags, str))
			return 0;
	}
	return 1;
}

simple_search(int flags, char *str)
{
	int dir = (flags & REVERSE) ? -1 : 1;
	if (flags & REGEX)
		return do_regex_searching(flags, str);
	else if (flags & WORD)
		return word_search(dir, str);
	return search(dir, str);
}

// Is the match inside the specified region?
int restricted_search_ok(int flags)
{
	int ret = 1, start, end;	// Boundaries of possible matches.

	if (spot_to_buffer(restrict_search.start) != bufnum
		  || spot_to_buffer(restrict_search.end) != bufnum)
		return 1;
	start = *restrict_search.start;
	end = *restrict_search.end;
	if (restrict_search_type == REGRECT) {
		int col1 = get_column(matchstart);
		int col2 = get_column(matchend);

		if (col1 < restrict_search.colstart || col2 < restrict_search.colstart
			  || col1 > restrict_search.colend || col2 > restrict_search.colend)
			ret = 0;
		start = give_position_at_column(start, 0);
		end = give_position_at_column(end, -1);
	}
	if (matchstart < start || matchend < start
		  || matchstart > end || matchend > end)
		ret = 0;
	if (point < start)		// Advance to first possible match.
		point = (flags & REVERSE) ? 0 : start;
	else if (point > end)
		point = (flags & REVERSE) ? end : size();
	return ret;
}

// Does this search string rely on accurate coloring of the buffer?
has_color_check(int flags, char *str)
{
	return ((flags & REGEX) && (fpatmatch(str, "*<c:*", 0, 0)
								|| fpatmatch(str, "*<?c:*", 0, 0)));
}

// Search according to flags, like do_searching, and return 0=not found,
// 1=found, 2=aborted, 3=bad pattern.  But if str might use color
// assertions, color the buffer as needed to ensure correct matching.
// Specifically, we search once while ignoring coloring.  Then we color
// that section of the buffer and search again.  After rejecting a few
// matches for incorrect coloring, color everything and search again.

do_color_searching(int flags, char *str)
{
	int res, mstart, mend, cnt = 0;

	// Do simple if it's not color-based searching, or buffer's all colored
	// already.
	if (!(flags & REGEX) || !has_color_check(flags, str) || !chars_need_coloring())
		return do_searching(flags, str);

	if ((res = do_searching(flags | SEARCH_IGNORE_COLOR, str)) != 1)
		return res;
	do {		// Check that the match still holds after applying color.
		mstart = matchstart;
		mend = matchend;
		recolor_buffer_range(MIN(mstart, mend), MAX(mstart, mend));
		point = mstart;		// Now search using color assertions.
		res = do_searching(flags, str);
		if (!res) {		// That match was bogus, but see if there's another
			point = mend;	// that's not yet colored right.
			res = do_searching(flags | SEARCH_IGNORE_COLOR, str);
			if (point == mend)
				break;
		}
		if (cnt++ > 5) {	// That's enough tries for slow incremental coloring.
			recolor_buffer_range(0, size());	// Color whole buffer and retry.
			return do_searching(flags, str);
		}
	} while (res == 1 && point != mend);
	return res;
}

do_searching(flags, str)	/* search according to flags */
char *str;	/* return 0=not found, 1=found, 2=aborted, 3=bad pattern */
{
	int ret;
	int null_detect = -1;

	if (flags & MODFOLD)
		save_var case_fold = (flags & FOLD);
	save_var abort_searching = ABORT_ERROR;
	for (;;) {
		ret = simple_search(flags, str);
		if (ret != 1 || !restrict_search_type || restricted_search_ok(flags))
			break;

		if (null_detect == point)	// Don't get stuck on nullable pat.
			if ((flags & REVERSE) ? (point-- <= narrow_start)
								  : (point++ >= size() - narrow_end))
				return 0;
		null_detect = point;
	}
	if (ret == ABORT_ERROR)
		ret = DSABORT;
	return ret;
}

do_regex_searching(flags, str)
char *str;
{
	int rflags = (flags & REVERSE) ? RE_REVERSE : 0;

	if (flags & WORD) {
		char *pat = malloc(strlen(str) + 20);
		on_exit free(pat);
		sprintf(pat, "</word>(?:%s)</word>", str);
		str = pat;
	}

	if (flags & SEARCH_IGNORE_COLOR)
		rflags |= RE_IGNORE_COLOR;
	if (regex_first_end)		/* construct flags for */
		rflags |= RE_FIRST_END;	/* regex search from vars */
	if (regex_shortest)
		rflags |= RE_SHORTEST;
	return re_compile(rflags, str) ? DSBAD : re_search(rflags, str);
}

word_search(dir, str)		/* search for whole words only */
char *str;
{
	char *pat, *s, *t;
	int ret, a = abort_searching;

	if (a == ABORT_JUMP)
		save_var abort_searching = ABORT_ERROR;
	pat = malloc(strlen(str) * 8 + 20);
	*pat = 0;
	if (isident(str[0]))
		strcpy(pat, "</word>");
	for (s = str, t = pat + strlen(pat); *s; s++)
		if (isalnum(*s))
			*t++ = *s;
		else if (*s == ' ') {
			strcpy(t, "<wspace>+");
			t += strlen(t);
		} else {
			*t++ = '%';
			*t++ = *s;
		}
	if (strlen(str) >= 1 && isident(str[strlen(str) - 1]))
		strcpy(t, "</word>");
	ret = re_search(dir, pat);
	free(pat);
	if (ret == ABORT_ERROR && a == ABORT_JUMP)
		quick_abort();
	return ret;
}

// Fold characters in range according to rule.
fold_replacement_text(int rule, char *from, char *to)
{
	int first = 1;

	for (; from < to; from++)
		switch (rule) {
			case 'L': *from = tolower(*from); break;
			case 'U': *from = toupper(*from); break;
			case 'C':
				*from = first ? toupper(*from) : tolower(*from);
				first = !isalpha(*from);
				break;
		}
}


// Fold characters in range according to rule.
fold_replacement_buf(int rule, int from, int to)
{
	switch (rule) {
		case 'L': do_lowercase_region(from, to); break;
		case 'U': do_uppercase_region(from, to); break;
		case 'C': do_capitalize_region(from, to); break;
	}
}

// Assuming pat contains a series of |-separated things, set res to the
// num'th one (counting from 0) and return 1.  Return 0 if there aren't
// that many.
get_pattern_alternative(char *pat, int num, char *res)
{
	*res = 0;
	while (num-- > 0)
		if (pat = strchr(pat, '|'))
			pat++;
		else
			return 0;
	strcpy(res, pat);
	if (pat = strchr(res, '|'))
		*pat = 0;
	return 1;
}

// See which alternative in pat matches text in orig.
pick_pattern_alternative(int orig, char *pat)
{
	char part[1000];
	int i, out = bufnum;

	save_var bufnum = orig;
	for (i = 0; get_pattern_alternative(pat, i, part); i++) {
		if (parse_string(-1, part)) {
			if (!get_pattern_alternative(pat, i + 1, part))
				get_pattern_alternative(pat, 0, part);
			buf_printf(out, "%s", part);
			point -= parse_string(-1, pat);
			point += parse_string(1, pat);
			return;
		}
	}
}

fromorig_bigrepl(char *str, char *with)	/* do replace with #0, #1 */
{
	int group;
	int orig = bufnum, tmp = tmp_buf();
	int fold_as = 0, fold_from = -1;

	bufnum = tmp;
	for (; *with; with++) {
		if (*with != '#')
			insert(*with);
		else if (isdigit(*++with)) {
			bufnum = orig;
			group = *with - '0';
			buf_xfer(tmp, find_group(group, 1),
				find_group(group, 0));
			bufnum = tmp;
		} else
			switch (*with) {
				case 'U':
				case 'L':
				case 'C':
				case 'E':
					if (fold_from >= 0)
						fold_replacement_buf(fold_as, fold_from, point);
					fold_as = *with;
					fold_from = point;
					break;
				case 'S':
					pick_pattern_alternative(orig, str);
					break;
				default:
					insert(*with);
					break;
				case 0:
					with--;
					break;
			}
	}
	if (fold_from >= 0)
		fold_replacement_buf(fold_as, fold_from, point);
	bufnum = orig;
	delete(matchstart, point);
	grab_buffer(tmp);
	buf_delete(tmp);
}

// Parse a character value like <newline> or <#123> from the string
// *s, advancing *s past it.  Return -1 if the string was invalid
// somehow.
int get_angled_value_from_string(char **s)
{
	char *end;
	if (**s != '<' || strlen(*s) <= 2)
		return -1;
	int i = key_value(*s, &end);
	if (i < 0)
		return i;
	*s = end;
	return i;
}

// Replace match with new text, interpreting #1 and such.  Fast version
// that can't handle long replacement text, #S, nulls.  Return 1
// normally, 0 if we detected something we couldn't handle (in which
// case, we've done nothing).
int fromorig_quickrepl(char *pat, char *with)
{
	int group, fold_as = 0, from, to;
	char *p, str[1000], *q = str, *fold_from = 0;

	UNREFERENCED_PARAMETER(pat);
	if (strlen(with) > 900 || strstr(with, "#S"))
		return 0;
	while (*with) {
		p = with;
		if (p)
			p = index(p, '#');
		if (!p) break;
		strncpy(q, with, p - with);	/* add till just before # */

		q += strlen(q);
		if (fold_from)
			fold_replacement_text(fold_as, fold_from, q);
		if (isdigit(*++p)) {
			group = *p - '0';
			from = find_group(group, 1);
			to = find_group(group, 0);
			if (q - str + (to - from) > 900)
				return 0;
			grab(from, to, q);
			if (to - from != strlen(q))
				return 0;
			q += strlen(q);		/* add variable part */
		} else
			switch (*p) {
				case 'U':
				case 'L':
				case 'C':
				case 'E':
					fold_as = *p;
					fold_from = q;
					break;
				case '<':
					int val = get_angled_value_from_string(&p);
					if (val >= 0 && IS_ASCII_KEY(val))
						*q++ = val;
					p--;
					break;
				default:
					*q++ = *p;		/* #x gives x */
					break;
			}
		with = p + 1;
	}
	strcpy(q, with);			/* add trailing part */
	if (fold_from)
		fold_replacement_text(fold_as, fold_from, q + strlen(q));
	delete(matchstart, point);		/* do replacement */
	stuff(str);
	return 1;
}

// Replace match with new text, interpreting #1 and such.  Try quick
// method first; if it fails, use slower method.
fromorig(char *pat, char *with)
{
	if (!fromorig_quickrepl(pat, with))
		fromorig_bigrepl(pat, with);
}

col_search(pat, col)		/* go to start of next line w/ string pat */
char *pat;			/* starting in column col */
{				/* return 1 if found, else 0 & go to start */
	int i;

	while (search(1, pat)) {
		i = point;
		point = matchstart;
		if (current_column() == col) {
			if (nl_reverse())
				point++;
			return 1;
		}
		point = i;
	}
	point = 0;
	return 0;
}

// Search like do_color_searching, but skip matches that don't start or
// end in given columns.
int column_color_searching(int flags, char *pat, int startcol, int endcol)
{
	while (do_color_searching(flags, pat) == 1)
		if (startcol < 0 || get_column(matchstart) == startcol)
			if (endcol < 0 || get_column(matchend) == endcol)
				return 1;
	return 0;
}

narrowed_search(int flags, char *str, int limit)
{
	save_var narrow_start = point - limit;
	save_var narrow_end = size() - (point + limit);

	return do_searching(flags, str);
}

// A quick and simple function for replacing one regex string with
// another.  Returns number of replacements done.  Use string_replace
// for replacement text with #1, etc.
simple_re_replace(int flags, char *str, char *repl)
{
	int cnt = 0;
	save_spot point = (flags & RE_REVERSE) ? size() : 0;
	while (re_search(flags, str)) {
		delete(matchstart, matchend);
		stuff(repl);
		cnt++;
	}
	return cnt;
}

int regex_replace_in_string(char *dest, char *src, char *pat, char *repl)
{
	int tmp = tmp_buf();
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	stuff(src);
	point = 0;
	save_var abort_searching = 0;
	save_var replace_num_found, replace_num_changed;
	int ret = string_replace(pat, repl, REGEX);
	grab(0, size(), dest);
	return ret ? -1 : replace_num_changed;
}

// Search and replace from a dialog box, for the Windows version.

struct find {
	char *find_text;
	char *replace_text;
	int flags;
} find_data;		// The find dialog looks at this variable directly.

// The search and replace strings.
char sbuf[FNAMELEN], rbuf[FNAMELEN];

setup_find_struct(replace)
{
	if (_last_was_regex)
		find_data.flags = _default_regex_flags;
	else
		find_data.flags = _default_flags;
	strncpy(sbuf, default_search_string(find_data.flags),
			sizeof(sbuf) / sizeof(*sbuf));
	find_data.find_text = sbuf;
	if (replace) {
		strncpy(rbuf, *default_replace_string(find_data.flags),
				sizeof(rbuf) / sizeof(*rbuf));
		find_data.replace_text = rbuf;
	} else
		find_data.replace_text = 0;
}


// Do a redisplay, but make sure that point is displayed
// within the top n lines.

redisplay_in(n)
int n;
{
	if (n > text_height())
		n = text_height();
	if (point < window_start
		|| point > move_by_screen_lines(window_start, n))
		window_start = move_by_screen_lines(point, -n/2);
	redisplay();
}

// Set the default search string & mode, and record in list of strings.
record_search(flags, s)
char *s;
{
	_last_was_regex = ((flags & REGEX) != 0);
	*(_last_was_regex ? &_default_regex_flags : &_default_flags) = flags;
	strcpy(default_search_string(flags), s);
	save_search_recall(flags, default_search_string(flags));
	if (find_data.replace_text) {
		save_recall(find_data.replace_text, "replace");
		strcpy(*default_replace_string(flags), find_data.replace_text);
	}
}

int our_matchstart;	// make sure matchstart no longer changes unexpectedly

// Before replacing, we position to the start of the match and search
// for it again.  We need to do this for regex matches since code
// coloring might have occurred, preventing #1, #2 from working right.
// Also, flags might have changed since we searched, and the next
// match might have moved.  We don't replace when this happens.

perform_replace()
{
	point = our_matchstart;
	if (find_from(find_data.flags, find_data.find_text, our_matchstart,
				  our_matchstart) != 1 || our_matchstart != matchstart)
		return;		// Match has moved (flags changed?)
	if (find_data.flags & REGEX)
		fromorig(find_data.find_text, find_data.replace_text);
	else {
		delete(matchstart, point);
		stuff(find_data.replace_text);
	}
}

find_from(flags, text, start_at, on_failure)
char *text;
{
	int r;
	point = start_at;
	r = do_color_searching(flags, text);
	if ( r != 1)
		point = on_failure;
	return r;
}

int search_and_show(flags, find_text, start_here, on_failure)
char *find_text;
{
	char msg[256];
	int r = find_from(flags, find_text, start_here, on_failure);
	switch (r){
		case 0:
			sprintf(msg, "Not found: %s", find_data.find_text);
			redisplay_in(window_lines_visible(window_handle));
			find_dialog_say(msg);
			if (!(search_wraps & 2) || !ask_wrap(flags)
				  || !find_from(flags, find_text,
								(flags & SEARCH_REVERSE) ? size() : 0, on_failure))
				break;
			// Fall through.
		case 1:
			mark = matchstart;
			highlight_on();
			redisplay_in(window_lines_visible(window_handle));
			our_matchstart = matchstart;
			find_dialog_say("");
			break;
		case 3:
			sprintf(msg, "Bad regex pattern: %s", find_data.find_text);
			find_dialog_say(msg);
			break;
	}
	return r;
}

int start_here;

int _basic_search()
{
	int r = search_and_show(find_data.flags, find_data.find_text, point, point);
	if (r){
			start_here = our_matchstart;
			if (matchstart != our_matchstart)
				say("matchstart trouble");
	}
	return r;
}

// If a flag is present in flags, turn it on.
// If a bit is present in mask but not in flags, turn it off.
// The bits we know about are MODFOLD, FOLD, WORD, REGEX, and REVERSE.
// We also accept REPLACE in flags to display a replace dialog instead
// of a search dialog.

_find(flags, mask)
{
	int oreg_type;

	save_var mark = point, _highlight_control = 0;
	oreg_type = region_type();
	set_region_type(REGNORM);

	setup_find_struct(flags & REPLACE);
	find_data.flags &= ~mask;
	find_data.flags |= flags;
	find_dialog(1);
	start_here = point;
	while(1){
		int k = getkey();
		switch (k){
			case 'a':  // Replace All
				if (is_highlight_on())
					point = matchstart;		// include current match
				record_search(find_data.flags, find_data.find_text);
				set_region_type(oreg_type);
				find_dialog(0);
				string_replace(find_data.find_text, find_data.replace_text,
							   ((SMODES | REVERSE) & find_data.flags)
							   | REPLACE | INCR);
				return;
			case 'p':	// Replace
				if (is_highlight_on())		// Is there a current match?
					perform_replace();
//				else
//					note("No current match, doing find...");
				if (_basic_search() == 0){
					note("Replacement finished.");
					record_search(find_data.flags, find_data.find_text);
					find_dialog(0);
					set_region_type(oreg_type);
					return;
				}
				break;
			case CTRL('r'):
				find_data.flags |= REVERSE;
				find_dialog(1);
				_basic_search();
				break;
			case CTRL('s'):
				find_data.flags &= ~REVERSE;
				find_dialog(1);
				_basic_search();
				break;
			case CTRL('m'):
				_basic_search();
				break;
			case 's': // Incremental character.
				if (!(find_data.flags & REGEX) && !(find_data.flags & WORD))
					search_and_show(find_data.flags, find_data.find_text,
									start_here, point);
				break;
			case 'g':
				record_search(find_data.flags, find_data.find_text);
				find_dialog(0);
				set_region_type(oreg_type);
				return;
			case CTRL('t'):
				find_data.flags ^= REGEX;
				find_dialog(1);
				break;
			case CTRL('w'):
				find_data.flags ^= WORD;
				find_dialog(1);
				break;
			case CTRL('c'):
				find_data.flags ^= FOLD;
				find_data.flags |= MODFOLD;
				find_dialog(1);
				break;
			case 'h':
				help_on_command(find_data.replace_text ? "dialog-replace"
					: "dialog-search");
				break;
		}
	}
}

command dialog_search()
{
	if (has_feature & FEAT_SEARCH_DIALOG)
		_find(0, REVERSE);
	else
		incremental_search();
}

command dialog_reverse_search()
{
	if (has_feature & FEAT_SEARCH_DIALOG)
		_find(REVERSE, REVERSE);
	else
		reverse_incremental_search();
}


command dialog_replace()
{
	if (has_feature & FEAT_SEARCH_DIALOG)
		_find(REPLACE, REVERSE | REGEX);
	else
		query_replace();
}

command dialog_regex_replace()
{
	if (has_feature & FEAT_SEARCH_DIALOG)
		_find(REPLACE | REGEX, REVERSE | REGEX);
	else
		regex_replace();
}
