/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// Some Unicode-related commands.

#include "eel.h"

do_unicode_convert(int flags)
{
	char msg[FNAMELEN], *enc;
	int err = 0, from = 0, to = size(), high, scol, ecol;

	high = is_highlight_on();
	check_modify(bufnum);
	sprintf(msg, "Convert %s %s", high ? "region" : "buffer",
			(flags & CONV_TO_16) ? "to Unicode from encoding"
								 : "from Unicode to encoding");
	if (!(enc = encoding_to_name(ask_encoding(msg, 1))))
		return;
	if (high) {
		save_spot point, mark;
		fix_region();
		from = point;
		to = mark;
	} else
		save_spot point;
	if (high == REGRECT) {
		scol = get_column(point);
		ecol = get_column(mark);
		for (; point < mark && !err; nl_forward())
			err = perform_unicode_conversion(bufnum, column_to_pos(scol),
											 column_to_pos(ecol), flags, enc);
	} else
		err = perform_unicode_conversion(bufnum, from, to, flags, enc);
	if (err == ETOOBIG)
		error("This is not an 8-bit encoding; a character is already wide.");
	else if (err)
		file_error(err, filename ? filename : bufname, "conversion error");
	if (!high && (flags & CONV_TO_16)) {
		int trans = GET_LINE_TRANSLATE(translation_type);
		int conv = encoding_from_name(enc);
		translation_type = MAKE_TRANSLATE(conv, trans,
										  (trans != FILETYPE_AUTO || conv > 0));
		make_mode();
	}
}

command unicode_convert_from_encoding()
{
	do_unicode_convert(CONV_TO_16);
}

command unicode_convert_to_encoding()
{
	do_unicode_convert(0);
}

// Convert some punctuation and a few symbols from Unicode and from
// Windows-1252 to ASCII.  This doesn't try to strip accent marks from
// letters or do full transliteration.
command unicode_convert_to_ascii()
{
	save_spot point;
	if (is_highlight_on()) {
		save_spot mark, narrow_start, narrow_end;
		fix_region();
		narrow_start = point;
		narrow_end = size() - mark;
	}

	// First we'll do Unicode.
	simple_re_replace(1, "<#0x00A0>", " ");	   // NO-BREAK SPACE
	simple_re_replace(1, "<#0x00A8>", "''");   // DIAERESIS
	simple_re_replace(1, "<#0x00AB>", "<<");   // LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
	simple_re_replace(1, "<#0x00AD>", "-");	   // SOFT HYPHEN
	simple_re_replace(1, "<#0x00B4>", "'");	   // ACUTE ACCENT
	simple_re_replace(1, "<#0x00B8>", ",");	   // CEDILLA
	simple_re_replace(1, "<#0x00BB>", ">>");   // RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
	simple_re_replace(1, "<#0x00D7>", "x");	   // MULTIPLICATION SIGN
	simple_re_replace(1, "<#0x2000>", " ");	   // EN QUAD
	simple_re_replace(1, "<#0x2001>", " ");	   // EM QUAD
	simple_re_replace(1, "<#0x2002>", " ");	   // EN SPACE
	simple_re_replace(1, "<#0x2003>", " ");	   // EM SPACE
	simple_re_replace(1, "<#0x2004>", " ");	   // THREE-PER-EM SPACE
	simple_re_replace(1, "<#0x2005>", " ");	   // FOUR-PER-EM SPACE
	simple_re_replace(1, "<#0x2006>", " ");	   // SIX-PER-EM SPACE
	simple_re_replace(1, "<#0x2010>", "-");	   // HYPHEN
	simple_re_replace(1, "<#0x2011>", "-");	   // NON-BREAKING HYPHEN
	simple_re_replace(1, "<#0x2013>", "-");	   // EN DASH
	simple_re_replace(1, "<#0x2014>", "-");	   // EM DASH
	simple_re_replace(1, "<#0x2018>", "'");	   // LEFT SINGLE QUOTATION MARK
	simple_re_replace(1, "<#0x2019>", "'");	   // RIGHT SINGLE QUOTATION MARK
	simple_re_replace(1, "<#0x201A>", "'");	   // SINGLE LOW-9 QUOTATION MARK
	simple_re_replace(1, "<#0x201C>", "\"");   // LEFT DOUBLE QUOTATION MARK
	simple_re_replace(1, "<#0x201D>", "\"");   // RIGHT DOUBLE QUOTATION MARK
	simple_re_replace(1, "<#0x201E>", "\"");   // DOUBLE LOW-9 QUOTATION MARK
	simple_re_replace(1, "<#0x2022>", "*");	   // BULLET
	simple_re_replace(1, "<#0x2026>", "...");  // HORIZONTAL ELLIPSIS
	simple_re_replace(1, "<#0x2032>", "'");	   // PRIME
	simple_re_replace(1, "<#0x2035>", "`");	   // REVERSED PRIME
	simple_re_replace(1, "<#0x2039>", "<");	   // SINGLE LEFT-POINTING ANGLE QUOTATION MARK
	simple_re_replace(1, "<#0x203A>", ">");	   // SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
	simple_re_replace(1, "<#0x20AC>", "E");	   // EURO SIGN
	simple_re_replace(1, "<#0x2122>", "TM");   // TRADE MARK SIGN
	simple_re_replace(1, "<#0x02DC>", "~");	   // SMALL TILDE
	simple_re_replace(1, "<#0x02C6>", "^");	   // MODIFIER LETTER CIRCUMFLEX ACCENT
	simple_re_replace(1, "<#0x02CA>", "'");	   // MODIFIER LETTER ACUTE ACCENT
	simple_re_replace(1, "<#0x02CB>", "`");	   // MODIFIER LETTER GRAVE ACCENT
	simple_re_replace(1, "<#0x02DD>", "''");   // DOUBLE ACUTE ACCENT
	simple_re_replace(1, "<#0x0152>", "OE");   // LATIN CAPITAL LIGATURE OE

// Also convert some non-Unicode Windows-1252 characters to ASCII
	simple_re_replace(1, "<#0x0082>", ",");	   // SINGLE LOW-9 QUOTATION MARK
	simple_re_replace(1, "<#0x0084>", ",,");   // DOUBLE LOW-9 QUOTATION MARK
	simple_re_replace(1, "<#0x0085>", "...");  // HORIZONTAL ELLIPSIS
	simple_re_replace(1, "<#0x008B>", "<");	   // SINGLE LEFT-POINTING ANGLE QUOTATION MARK
	simple_re_replace(1, "<#0x008C>", "OE");   // LATIN CAPITAL LIGATURE OE
	simple_re_replace(1, "<#0x0091>", "`");	   // LEFT SINGLE QUOTATION MARK
	simple_re_replace(1, "<#0x0092>", "'");	   // RIGHT SINGLE QUOTATION MARK
	simple_re_replace(1, "<#0x0093>", "\"");   // LEFT DOUBLE QUOTATION MARK
	simple_re_replace(1, "<#0x0094>", "\"");   // DOUBLE ACUTE ACCENT
	simple_re_replace(1, "<#0x0095>", "*");	   // BULLET
	simple_re_replace(1, "<#0x0096>", "-");	   // EN DASH
	simple_re_replace(1, "<#0x0097>", "--");   // EM DASH
	simple_re_replace(1, "<#0x0099>", "TM");   // TRADE MARK SIGN
	simple_re_replace(1, "<#0x009B>", ">");	   // SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
}
