/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*	Copyright (C) 2003, 2009 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "python.h"
#include "vhdl.h"

color_vhdl_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	save_var case_fold = 1;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	while (point < to) {
		if (!re_search(1, "[a-z_][a-z0-9_]*('[a-z0-9_]+)*"
					   "|[0-9.]([a-z0-9._]|e[-+][0-9])*l?|--|[\"'\\]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '-':			// found comment
				nl_forward();
				set_character_color(t, point, color_class vhdl_comment);
				break;
			case '"':		// found a string literal
				t = --point;
				if (strchr("BOX", toupper(character(point - 1)))
					  && parse_string(-1, "</word>[box]"))
					t = --point;
				if (re_search(1, "[box]?\"([^\"\\\n]|\\.)*[\"\n]"))
					set_character_color(t, point, color_class vhdl_string);
				if (get_character_color(point, (int *) 0, &s) ==
					  color_class vhdl_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '\'':
//				if (re_search(1, "([^'\\\n]|\\.)*['\n]"))
				if (re_search(1, "([^'\n])*['\n]"))
					set_character_color(t, point, color_class vhdl_string);
				if (get_character_color(point, (int *) 0, &s) ==
					  color_class vhdl_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '\\':		// \arbitrary name\ is an identifier.
				if (re_search(1, "[\\\n]"))
					set_character_color(t, point, color_class vhdl_identifier);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, vhdl_keyword_color(t));
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring VHDL program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

vhdl_keyword_color(from)	// return color for "identifier" from here to point
{			// (something with alpha or digits)
	char buf[500];

	if (point - from > sizeof(buf) / sizeof(*buf) - 10)
		save_var point = from + sizeof(buf) / sizeof(*buf) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-", buf[1]) || buf[1] == '.' && isdigit(buf[2]))
		return c_number_color(buf + 1);
	strcpy(buf + point - from + 1, "|");
	if (is_vhdl_keyword(buf))
		return color_class vhdl_keyword;
	return color_class vhdl_identifier;
}

is_vhdl_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr("|abs|access|after|alias|all|and|architecture|array|"
			   "assert|attribute|begin|block|body|buffer|bus|case|"
			   "component|configuration|constant|disconnent|downto|"
			   "else|elsif|end|entity|exit|file|for|function|generate|"
			   "generic|group|guarded|if|impure|in|inertial|inout|is|"
			   "label|library|linkage|literal|loop|map|mod|nand|new|"
			   "next|nor|null|of|on|open|or|others|out|package|port|"
			   "postponed|procedure|process|pure|range|record|register|"
			   "reject|rem|report|return|rol|ror|select|severity|"
			   "signal|sla|sll|sra|srl|subtype|then|to|transport|type|"
			   "unaffected|units|until|use|variable|wait|when|while|"
			   "with|xnor|xor|", p))
		return 1;
	return 0;
}

// The indenter just indents like the previous non-comment line for now.
// Some rules it might use someday:

// If the previous line has an open parenthesis, align under that.
// If the previous line's last word is "loop", "is", "then", or "begin",
// indent more.  But don't if the current line starts with "begin".
// If the current line starts with "end", indent less, or maybe search
// for its matching line and indent like that.

do_vhdl_indent()
{
	int ind;

	save_var point;
	do {		// Find prev non-comment line.
		if (!nl_reverse())		// First line in the file, no indent.
			return 0;
		to_begin_line();
	} while (parse_string(1, "[ \t]*(--|\n)"));
	to_indentation();
	ind = current_column();
	restore_vars();
	to_column(ind);
}

vhdl_indenter() on vhdl_tab['\t']
{
	int orig = point;
	int orig_column = current_column();

	if (maybe_indent_rigidly(0))
		return;
	to_indentation();
	if (orig_column > current_column()) {
					/* if not in indentation */
		point = orig;
		to_column(orig_column + vhdl_indent
				  - orig_column % vhdl_indent);
	} else if (prev_cmd == C_INDENT)	/* repeated, make bigger */
		to_column(orig_column + vhdl_indent);
	else {
		do_vhdl_indent();
		if (!get_indentation(point) && !orig_column
			  && this_cmd != CMD_INDENT_REG)	// 0 indent=>0 indent,
			to_column(vhdl_indent);		// but indent anyway
	}
	if (this_cmd != CMD_INDENT_REG)
		this_cmd = C_INDENT;
}


command vhdl_mode()
{
	mode_default_settings();
	mode_keys = vhdl_tab;		/* Use these keys. */
	major_mode = _vhdl_mode_name;
	compile_buffer_cmd = compile_vhdl_cmd;	// can compile this
	indenter = do_vhdl_indent;
	strcpy(comment_start, "--[ \t]*");
	strcpy(comment_pattern, "--.*$");
	strcpy(comment_begin, "-- ");
	strcpy(comment_end, "");
	recolor_range = color_vhdl_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if (auto_show_vhdl_delimiters)
		auto_show_matching_characters = vhdl_auto_show_delim_chars;
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 64) != 0;
	try_calling("vhdl-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

when_loading()
{
	vhdl_tab[ALT('q')] = (short) fill_comment;
	vhdl_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

suffix_vhd()
{
	vhdl_mode();
}

suffix_vhdl()
{
	vhdl_mode();
}

// Set display_func_name to the name of the function we're editing, and
// return 1.  If not in a function, set display_func_name to "" and
// return 1.  If user pressed a key and we gave up for now, return 0.

vhdl_func_name_finder()
{
	int from, to, pos, orig = point;
	char pat[FNAMELEN * 2];

	save_var point, case_fold = 1;
	to_end_line();
	while (do_color_searching(REVERSE | REGEX, "^[ \t]*<^c:vhdl-comment"
							  "|vhdl-string>(procedure|function)</word>")) {
		if (!must_find_func_name && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		pos = point;
		if (!parse_string(1, "^[ \t]*(procedure|function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		if (!parse_string(1, "[ \t]*(return[ \t\n]+[a-z0-9_]+[ \t]*)?is</word>")) {
			point = pos;
			continue;
		}
		if (matchend - matchstart < FNAMELEN) {
			sprintf(pat, "^[ \t]*end[ \t]+%s</word>",
					display_func_name);
			if (re_search(1, pat) && point < orig)
				*display_func_name = 0;
		}
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

tag_mode_vhdl()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (do_color_searching(REGEX, "^[ \t]*<^c:vhdl-comment"
							  "|vhdl-string>(procedure|function|entity)"
							  "[ \t]+([a-zA-Z_]<word>*)")) {
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}
