/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has commands for changing Epsilon's window size.

#include "eel.h"
#include "lowlevel.h"
#include "video.h"

term_init()		/* record outside size */
{
	_find_outside();
	_inside_size = _outside_size;
}

term_cmd_line()		/* go to mode indicated by cmd line switches */
{
	int use_lines, use_cols, newsize;

	if (want_lines || want_cols) {
		use_lines = want_lines ? want_lines : screen_lines;
		use_cols = want_cols ? want_cols : screen_cols;
		newsize = MAKE_SIZE(use_cols, use_lines);
		if (newsize != _inside_size) {
			change_video(_inside_size, 0, "");
			change_video(newsize, 1, "");
		}
		resize_screen(use_lines, use_cols);
		_inside_size = MAKE_SIZE(screen_cols, screen_lines);
	}
}

term_mode(active)	/* set screen mode in(out) of epsilon */
{
	prev_screen_lines = 0;
	prev_screen_cols = 0;
	if (_inside_size != _outside_size) {
		change_video(active ? _outside_size : _inside_size, 0, "");
		change_video(active ? _inside_size : _outside_size, 1, "");
	}
	return _inside_size != _outside_size;
}

show_screen_size()
{
	if (prev_screen_lines && prev_screen_cols
		&& (prev_screen_lines != screen_lines
			|| prev_screen_cols != screen_cols))
		note("Now %d x %d", screen_cols, screen_lines);
	prev_screen_lines = screen_lines;
	prev_screen_cols = screen_cols;
}

try_set_screen_size(lines, cols, mode)		// lines & cols required
char *mode;			// mode name can be "" if we don't care
{
	int ok;
	int nsize, osize = MAKE_SIZE(screen_cols, screen_lines);

	nsize = MAKE_SIZE(cols, lines);
	if (nsize == osize)
		return 1;
	if (!change_video(nsize, 2, mode))	/* check if legal */
		return 0;
	change_video(osize, 0, mode);		/* turn off old */
	ok = change_video(nsize, 1, mode);
	if (!ok)		/* if error, restore old */
		change_video(osize, 1, mode);
	_inside_size = MAKE_SIZE(screen_cols, screen_lines);
	return ok;
}

change_video(size, active, mode)	/* set to size code, return 1 if ok */
char *mode;
{
	char name[20];
	int ret = 0, lines, cols;

	lines = GET_LINES(size);
	cols = GET_COLS(size);
	save_var mouse_auto_on = 0, mouse_display = 0;
	if (!mode || !*mode)	// make up appropriate mode name
		sprintf(mode = name, "%dx%d", cols, lines);
	ret = 1;
	if (active == 1) {
		resize_screen(lines, cols);
		if (screen_lines != lines || screen_cols != cols)
			ret = 0;
	}
	return ret;
}

_find_outside()	/* set variables for screen size outside of epsilon */
{
	_outside_size = MAKE_SIZE(screen_cols, screen_lines);
	if (opsys == OS_WINDOWS && restore_color_on_exit)
		alter_color(color_class after_exiting, orig_screen_color());
	selectable_colors = 256;
}
